/*                                                                     
** COMPONENT_NAME: MET / Aria
**                                                                      
** Licensed Materials - Property of IBM                                  
**                                                                     
** IBM Confidential (IBM Confidential-Restricted when combined          
**                   with aggregated modules for this product)          
**                                                                     
**                                                                     
** OCO Source Materials                                                
**                                                                     
**                                                                     
** GA-0100 (C) Copyright IBM Corp. 1997                                 
**                                                                     
** US Government Users Restricted RIghts - Use, duplication or         
** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.   
**                                                                     
** The source code for this program is not published or otherwise      
** divested of its trade secrets, irrespective of what has been        
** deposited within the U.S. Copyright Office.                         
**                                                                     
*/


 
/*******************************************************************
  PROGRAM NAME..... analyze_branch
  FUNCTION ........ Analyze the user-code basic block ending branch and
     determine how the Aria code should simulate the effect of that branch.
     Also generate the code to set the next address (following this block)
     and the next not-taken address. 

  NOTES............
  ASSUMPTIONS... None
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... the next available translation address
  CALLED-ROUTINES.. None
  CALLED-BY........ do_one_translation (see do_translation.c)
  INCLUDED-BY...... None
  MACROS-USED...... several: some are defined here as code-writing 
     shortcuts (where sequences of duplicated code are pulled out 
     into a macro).  Also the opc_asm macros (see opc_asm.gen).
  INCLUDES......... <see below>

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/
#include <stdio.h>

#include "aria_xlate_macros.h"
#include "aria_debug.h"
#include "globtypes.h"

#include "io_msgs.h"

#include "proc_state.h"
#include "aria_state.h"
#include "global_variables.h"
#include "opc.h"

#include "dump_description.h"

/** Store the corrected CTR value in the processor state **/
#define GET_CTR_VALUE \
   *next_xlate_address = LOAD_REG     (scratch, CTR_SAVE_OFFSET, base);       next_xlate_address++;  \
   *next_xlate_address = OPC_ASM_MTSPR( 9, scratch);                          next_xlate_address++;  \

#define SET_NEW_CTR_VALUE \
   *next_xlate_address = OPC_ASM_MFSPR(scratch,  9);                          next_xlate_address++;  \
   *next_xlate_address = STORE_REG    (scratch, CTR_SAVE_OFFSET, base);       next_xlate_address++;  \

/**************************************************************/
/** THE FOLLOWING ARE THE 32-BIT MACROS FOR BRANCH ANALYSIS  **/
/**************************************************************/
#ifndef DO_64_BIT_COMPILE

#define ADDRESS_STORAGE_DECLARES \
   unsigned lo_addr; \
   unsigned hi_addr

/** Code to handle the setting (and storing) of the target address  **/
#define BRANCH_TARGET_CODE \
   *next_xlate_address = OPC_ASM_ADDIS(scratch,  0, hi_addr);                 next_xlate_address++; \
   *next_xlate_address = OPC_ASM_ORI  (scratch, scratch, lo_addr);            next_xlate_address++; \
   *next_xlate_address = STORE_REG    (scratch, NEXT_TARGET_OFFSET, base);    next_xlate_address++

/** Code to handle the setting (and storing) of the target address  **/
#define BRANCH_NOT_TAKEN_TARGET_CODE \
   *next_xlate_address = OPC_ASM_ADDIS(scratch,  0, hi_addr);                 next_xlate_address++; \
   *next_xlate_address = OPC_ASM_ORI  (scratch, scratch, lo_addr);            next_xlate_address++; \
   *next_xlate_address = STORE_REG    (scratch, NOT_TAKEN_OFFSET, base);      next_xlate_address++

/* Code to handle when Branch is taken **/ 
#define TAKEN_BRANCH_CODE \
   hi_addr = taken_target >> 16; \
   lo_addr = taken_target & 0x0000ffff; \
   BRANCH_TARGET_CODE; \
   if (not_taken_address_required_flag == 1) { \
     hi_addr = fall_thru_path >> 16; \
     lo_addr = fall_thru_path & 0x0000ffff; \
     BRANCH_NOT_TAKEN_TARGET_CODE; \
   }

/* Code to handle when Branch is not taken **/ 
#define NOT_TAKEN_BRANCH_CODE \
   hi_addr = fall_thru_path >> 16; \
   lo_addr = fall_thru_path & 0x0000ffff; \
   BRANCH_TARGET_CODE; \
   if (not_taken_address_required_flag == 1) { \
     hi_addr = taken_target >> 16; \
     lo_addr = taken_target & 0x0000ffff; \
     BRANCH_NOT_TAKEN_TARGET_CODE; \
   }

/* Code to save the link register target address */
#define SAVE_LINK_TARGET \
    hi_addr = link_target >> 16; \
    lo_addr = link_target & 0x0000ffff; \
    *next_xlate_address = OPC_ASM_ADDIS(scratch,  0, hi_addr);           next_xlate_address++;  \
    *next_xlate_address = OPC_ASM_ORI  (scratch, scratch, lo_addr);      next_xlate_address++;  \
    *next_xlate_address = STORE_REG    (scratch, LR_SAVE_OFFSET, base);  next_xlate_address++;  

/* Code to get the program counter value */
#define SET_SPECIAL_BRANCH \
  hi_addr = (((unsigned) next_xlate_address) & 0xffff0000) >> 16; \
  lo_addr = (((unsigned) next_xlate_address) & 0x0000ffff); \
  *temp_save_address = OPC_ASM_ADDIS (scratch_reg, 0, hi_addr);           temp_save_address++; \
  *temp_save_address = OPC_ASM_ORI  (scratch_reg, scratch_reg, lo_addr);  temp_save_address++; \

/**************************************************************/
/** THE FOLLOWING ARE THE 64-BIT MACROS FOR BRANCH ANALYSIS  **/
/**************************************************************/
#else

#define ADDRESS_STORAGE_DECLARES \
  unsigned addr_0_15;  \
  unsigned addr_16_31; \
  unsigned addr_32_47; \
  unsigned addr_48_63

/** Code to handle the setting (and storing) of the target address for 64 bit machines **/
#define BRANCH_TARGET_CODE \
   *next_xlate_address = OPC_ASM_ADDI  (scratch, 0, addr_0_15);               next_xlate_address++; \
   *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);             next_xlate_address++; \
   *next_xlate_address = OPC_ASM_ORI   (scratch, scratch, addr_16_31);        next_xlate_address++; \
   *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);             next_xlate_address++; \
   *next_xlate_address = OPC_ASM_ORI   (scratch, scratch, addr_32_47);        next_xlate_address++; \
   *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);             next_xlate_address++; \
   *next_xlate_address = OPC_ASM_ORI   (scratch, scratch, addr_48_63);        next_xlate_address++; \
   *next_xlate_address = OPC_ASM_STD   (scratch, NEXT_TARGET_OFFSET, base);   next_xlate_address++

/** Code to handle the setting (and storing) of the target address for 64 bit machines **/
#define BRANCH_NOT_TAKEN_TARGET_CODE \
   *next_xlate_address = OPC_ASM_ADDI  (scratch, 0, addr_0_15);               next_xlate_address++; \
   *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);             next_xlate_address++; \
   *next_xlate_address = OPC_ASM_ORI   (scratch, scratch, addr_16_31);        next_xlate_address++; \
   *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);             next_xlate_address++; \
   *next_xlate_address = OPC_ASM_ORI   (scratch, scratch, addr_32_47);        next_xlate_address++; \
   *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);             next_xlate_address++; \
   *next_xlate_address = OPC_ASM_ORI   (scratch, scratch, addr_48_63);        next_xlate_address++; \
   *next_xlate_address = OPC_ASM_STD   (scratch, NOT_TAKEN_OFFSET, base);     next_xlate_address++

/* Code to handle when Branch is taken for 64 bit machines **/ 
#define TAKEN_BRANCH_CODE \
   addr_0_15  = (taken_target >> 48) & 0xffffLL; \
   addr_16_31 = (taken_target >> 32) & 0xffffLL; \
   addr_32_47 = (taken_target >> 16) & 0xffffLL; \
   addr_48_63 = (taken_target      ) & 0xffffLL; \
   BRANCH_TARGET_CODE; \
   if (not_taken_address_required_flag == 1) { \
     addr_0_15  = (fall_thru_path >> 48) & 0xffffLL; \
     addr_16_31 = (fall_thru_path >> 32) & 0xffffLL; \
     addr_32_47 = (fall_thru_path >> 16) & 0xffffLL; \
     addr_48_63 = (fall_thru_path      ) & 0xffffLL; \
     BRANCH_NOT_TAKEN_TARGET_CODE; \
   }

/* Code to handle when Branch is not taken for 64 bit machines **/ 
#define NOT_TAKEN_BRANCH_CODE \
   addr_0_15  = (fall_thru_path >> 48) & 0xffffLL; \
   addr_16_31 = (fall_thru_path >> 32) & 0xffffLL; \
   addr_32_47 = (fall_thru_path >> 16) & 0xffffLL; \
   addr_48_63 = (fall_thru_path      ) & 0xffffLL; \
   BRANCH_TARGET_CODE; \
   if (not_taken_address_required_flag == 1) { \
     addr_0_15  = (taken_target >> 48) & 0xffffLL; \
     addr_16_31 = (taken_target >> 32) & 0xffffLL; \
     addr_32_47 = (taken_target >> 16) & 0xffffLL; \
     addr_48_63 = (taken_target      ) & 0xffffLL; \
     BRANCH_NOT_TAKEN_TARGET_CODE; \
   }

/* Code to save the link register target address */
#define SAVE_LINK_TARGET \
    addr_0_15  = (link_target >> 48) & 0xffffLL;  \
    addr_16_31 = (link_target >> 32) & 0xffffLL;  \
    addr_32_47 = (link_target >> 16) & 0xffffLL;  \
    addr_48_63 = (link_target      ) & 0xffffLL;  \
    *next_xlate_address = OPC_ASM_ADDI  (scratch, 0, addr_0_15);          next_xlate_address++;  \
    *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);        next_xlate_address++;  \
    *next_xlate_address = OPC_ASM_ORI   (scratch,  scratch, addr_16_31);  next_xlate_address++;  \
    *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);        next_xlate_address++;  \
    *next_xlate_address = OPC_ASM_ORI   (scratch,  scratch, addr_32_47);  next_xlate_address++;  \
    *next_xlate_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);        next_xlate_address++;  \
    *next_xlate_address = OPC_ASM_ORI   (scratch,  scratch, addr_48_63);  next_xlate_address++;  \
    *next_xlate_address = OPC_ASM_STD (scratch, LR_SAVE_OFFSET, base);    next_xlate_address++;

/* Code to get the program counter value */
#define SET_SPECIAL_BRANCH \
  addr_0_15  = (((BASE_TYPE) next_xlate_address) >> 48) & 0xffffLL; \
  addr_16_31 = (((BASE_TYPE) next_xlate_address) >> 32) & 0xffffLL; \
  addr_32_47 = (((BASE_TYPE) next_xlate_address) >> 16) & 0xffffLL; \
  addr_48_63 = (((BASE_TYPE) next_xlate_address)      ) & 0xffffLL; \
  *temp_save_address = OPC_ASM_ADDI  (scratch, 0, addr_0_15);          temp_save_address++; \
  *temp_save_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);        temp_save_address++; \
  *temp_save_address = OPC_ASM_ORI   (scratch, scratch, addr_16_31);   temp_save_address++; \
  *temp_save_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);        temp_save_address++; \
  *temp_save_address = OPC_ASM_ORI   (scratch, scratch, addr_32_47);   temp_save_address++; \
  *temp_save_address = OPC_ASM_RLDICL(scratch, scratch, 16, 0);        temp_save_address++; \
  *temp_save_address = OPC_ASM_ORI   (scratch, scratch, addr_48_63);   temp_save_address++;

#endif

/**************************************************************/
/** THE FOLLOWING ARE THE COMMON MACROS FOR BRANCH ANALYSIS  **/
/**************************************************************/

/** Code to handle a taken branch through register **/
#define TAKEN_BRANCH_THRU_REGISTER_CODE \
   if ((index == OPC_BCCTR) || (index == OPC_BCCTRL)) {                         \
     *next_xlate_address = LOAD_REG(scratch, CTR_SAVE_OFFSET, base); next_xlate_address++; \
   }                                                                            \
   else {                                                                       \
     *next_xlate_address = LOAD_REG(scratch, LR_SAVE_OFFSET, base); next_xlate_address++; \
   }                                                                            \
   *next_xlate_address = STORE_REG(scratch, NEXT_TARGET_OFFSET, base); next_xlate_address++; 

/* General code to handle the conditional branches for 64 bit machines */
#define CONDITIONAL_BRANCH_CODE(crbit, cnd) \
  if (br_to_register) {                                                                     \
    branch_posn_addr = next_xlate_address;                                    next_xlate_address++;  \
    NOT_TAKEN_BRANCH_CODE;                                                                  \
    *branch_posn_addr = OPC_ASM_BC((crbit), (cnd), (next_xlate_address + 1 - branch_posn_addr)); \
    branch_posn_addr = next_xlate_address;                                    next_xlate_address++;  \
    TAKEN_BRANCH_THRU_REGISTER_CODE;                                                        \
    *branch_posn_addr = OPC_ASM_B ((next_xlate_address - branch_posn_addr));          \
  } else {                                                                                  \
    branch_posn_addr = next_xlate_address;                                    next_xlate_address++;  \
    NOT_TAKEN_BRANCH_CODE;                                                                  \
    *branch_posn_addr = OPC_ASM_BC((crbit), (cnd), (next_xlate_address + 1 - branch_posn_addr)); \
    branch_posn_addr = next_xlate_address;                                    next_xlate_address++;  \
    TAKEN_BRANCH_CODE;                                                                      \
    *branch_posn_addr = OPC_ASM_B ((next_xlate_address - branch_posn_addr));          \
  }

/**************************************************************/
/** THE FOLLOWING IS THE MAIN ROUTINE FOR ANALYZING A BRANCH **/
/**************************************************************/

unsigned * 
analyze_branch(unsigned   branch_iword,
	       unsigned * prog_address,
	       unsigned * base_reg_ptr,
	       unsigned   dtrbf_ptr,
	       unsigned   dyntr_posn
	       )
{
  unsigned * branch_posn_addr;
  unsigned base  = *base_reg_ptr;

  unsigned iword = branch_iword;
  opc_idx  index = opc_index(iword);

  BASE_TYPE orig_pc = (BASE_TYPE) prog_address;
  
  BASE_TYPE fall_thru_path  = (BASE_TYPE) (prog_address + 1); /** The fall-through path **/
  BASE_TYPE taken_target    = 0;
  BASE_TYPE link_target     = 0;	  

  int br_and_link    = OPC_DEC_LK(iword);
  
  int br_to_register = OPC_BRANCH_THROUGH_REGISTER(index);
  
  int offset;

  ADDRESS_STORAGE_DECLARES;

  unsigned scratch  = 12;
  unsigned scratch2 = 11;

  unsigned * temp_ptr = next_xlate_address;

  IFDEBUG(warn_exec("In analyze_branch...\n"); );

  if (do_no_output_during_runs == 0) {
    DUMP_DESCRIPTION(temp_ptr, next_xlate_address);    
    warn_xlate("* (analyze branch) --------------------------------------------------------\n");
    temp_ptr = next_xlate_address;
  }

  /* Find a scratch register (unique from the base and dynamic trace buffer pointer registers) */
  while ( (scratch == base) || (scratch == dtrbf_ptr) ) {
    scratch--;
  }
  
  /* Find a second scratch register (unique from the scratch, base and dynamic trace buffer pointer registers) */
  while ( (scratch2 == base) || (scratch2 == dtrbf_ptr) || (scratch2 == scratch) ) {
    scratch2--;
  }
  if (scratch2 == 3) { 
    warn("PROBLEM in analyze_branch: scratch2 == 3!\n");
  }

  /** Now set the taken path target address **/
  switch (index)
    {
    case OPC_B:
      offset = OPC_DEC_LI(iword) << 2;
      taken_target = orig_pc + offset;
      break;
    case OPC_BA:
      offset = OPC_DEC_LI(iword) << 2;
      taken_target = offset;
      break;
    case OPC_BL:
      offset = OPC_DEC_LI(iword) << 2;
      taken_target = orig_pc + offset;
      break;
    case OPC_BLA:
      offset = OPC_DEC_LI(iword) << 2;
      taken_target = offset;
      break;
    case OPC_BC:
      offset = OPC_DEC_BD(iword) << 2;
      taken_target = orig_pc + offset;
      break;
    case OPC_BCA:
      offset = OPC_DEC_BD(iword) << 2;
      taken_target = offset;
      break;
    case OPC_BCL:
      offset = OPC_DEC_BD(iword) << 2;
      taken_target = orig_pc + offset;
      break;
    case OPC_BCLA:
      offset = OPC_DEC_BD(iword) << 2;
      taken_target = offset;
      break;
    case OPC_BCCTR:
    case OPC_BCCTRL:
    case OPC_BCLR:
    case OPC_BCLRL:
      break;
      
    default:
      fprintf(mystderr, "Unknown branch instruction: oaddr = 0x%016LLx, xaddr = 0x%016LLx, iw = 0x%08x\n",
	      (BASE_TYPE)prog_address, (BASE_TYPE)next_xlate_address, iword);
      break;
    }

  if (OPC_IS_A_CONDITIONAL_BRANCH(index)) {
    unsigned bo_val = OPC_DEC_BO(iword);
    unsigned bi_val = OPC_DEC_BI(iword);

    /** Dynamically determine if the branch is taken or not  **/
    switch(bo_val & 0x1e) {
    case 0x00: /* Dec CTR, BR if !0 and FALSE */
    case 0x02: /* Dec CTR, BR if =0 and FALSE */
    case 0x08: /* Dec CTR, BR if !0 and TRUE  */
    case 0x0a: /* Dec CTR, BR if =0 and TRUE  */
    case 0x10: /* Dec CTR, BR if !0           */
    case 0x18: /* Dec CTR, BR if !0           */
    case 0x12: /* Dec CTR, BR if =0           */
    case 0x1a: /* Dec CTR, BR if =0           */
      GET_CTR_VALUE;
      CONDITIONAL_BRANCH_CODE(bo_val, bi_val);
      SET_NEW_CTR_VALUE;
      break;

    case 0x04: /* BR if condition is FALSE */
    case 0x06: /* BR if condition is FALSE */
    case 0x0c: /* BR if condition is TRUE  */
    case 0x0e: /* BR if condition is TRUE  */
      CONDITIONAL_BRANCH_CODE(bo_val, bi_val);
      break;

    case 0x14: /* BR always */
    case 0x15: /* BR always */
    case 0x16: /* BR always */
    case 0x17: /* BR always */
    case 0x1c: /* BR always */
    case 0x1d: /* BR always */
    case 0x1e: /* BR always */
    case 0x1f: /* BR always */
      if ((bo_val & 0x14) == 0x14)	{ /* BR always */
	fall_thru_path = taken_target;
      
	if (br_to_register) {
	  TAKEN_BRANCH_THRU_REGISTER_CODE;
	}
	else {
	  TAKEN_BRANCH_CODE;
	}
      }
      break;

    default:
      fprintf(mystderr, "[an_br] ERROR: conditional branch bo_val = 0x%02x\n", bo_val);
      fprintf(mystderr, "[an_br]  Don't have a case for this...\n");
    }
  }
  else {
    fall_thru_path = taken_target;
    TAKEN_BRANCH_CODE;
  }
  
  if (br_and_link) {
    /** Put the link target into the saved state **/
    link_target = (BASE_TYPE)(prog_address + 1); \
    SAVE_LINK_TARGET;
  }
  
  if (do_dynamic_trace_generation) {
    if (do_no_output_during_runs == 0) {
      DUMP_DESCRIPTION(temp_ptr, next_xlate_address);    
      warn_xlate("*   dyn trce       --------------------------------------------------------\n");
      temp_ptr = next_xlate_address;
    }

    /* Set the dynamic trace data for this branch */
    *next_xlate_address = OPC_ASM_ADDIS(dtrbf_ptr, 0, ARIA_STATE_SAVE_HI_16);                         next_xlate_address++;
    *next_xlate_address = LOAD_REG    (dtrbf_ptr, PROC_DYN_TRBUF_OFFSET, dtrbf_ptr);  next_xlate_address++;

    /* Get the return address */
    *next_xlate_address = LOAD_REG    ( scratch,  NEXT_TARGET_OFFSET, base);          next_xlate_address++;
    
    if (do_iaddr_fixup) {
      /** If either the taken or fall-through path is into user space,  **/
      /**  OR this is a branch through a register, check whether the    **/
      /**  iaddress (target) is in segment 0x2 and needs to be adjusted **/
      if ( ((taken_target & 0xf0000000) == global_user_program_iaddr_mask) || 
	   ((fall_thru_path & 0xf0000000) == global_user_program_iaddr_mask) ||
	   br_to_register ) {
	/* Modify it to the true user space address */
	*next_xlate_address = OPC_ASM_RLWINM(scratch2, scratch, 4, 28, 31);        next_xlate_address++;
	*next_xlate_address = OPC_ASM_CMPI ( 7, 0, scratch2, (global_user_program_iaddr_mask>>28)&0xf);                 next_xlate_address++;
	*next_xlate_address = OPC_ASM_BC   ( 0x4, 30, 3);                          next_xlate_address++;
	/* Load the process iaddr_fixup offset */
	*next_xlate_address = LOAD_REG    ( scratch2, IADDR_OFFSET, base);        next_xlate_address++;
	*next_xlate_address = OPC_ASM_SUBFC( scratch, scratch2, scratch);          next_xlate_address++;
      }
    }
    
    /* Store the user-view target address in the dynamic trace */
    *next_xlate_address = STORE_REG( scratch, ARIA_TB_PER_INSTR_SIZE*ARIA_TB_ITEM_SIZE*dyntr_posn+ARIA_TB_BRTARGET_OFFSET, dtrbf_ptr);   next_xlate_address++;

    if (do_no_output_during_runs == 0) {
      DUMP_DESCRIPTION(temp_ptr, next_xlate_address);    
      warn_xlate("*   sim code       --------------------------------------------------------\n");
      temp_ptr = next_xlate_address;
    }

  }

  if (not_taken_address_required_flag == 1) { 
    /* Get the not-taken path address */
    /*  NOTE: do this before setting GPR 3 since GPR 3 may be the base register */
    *next_xlate_address = LOAD_REG    (scratch2, NOT_TAKEN_OFFSET, base);         next_xlate_address++; 
  }

  /**************************************************************************/
  /** Set up the subroutine return value (the taken path address) in GPR 3 **/
  /**************************************************************************/
  /* Set the function return value (GPR3) to be the next BB address */  
  *next_xlate_address = LOAD_REG    ( 3,  NEXT_TARGET_OFFSET, base);              next_xlate_address++;

  /** If the base WAS GPR 3, find a new base register */
  if (base == 3) {
    base++;
    while ( (base == scratch) || (base == scratch2) || (base == dtrbf_ptr) ) {
      base++;
    }
  }

  /**********************************************/
  /** Set up the not-taken path return address **/
  /**********************************************/

  /* Load GPRbase_register with the address of our ARIA state saving space    */               
  *next_xlate_address = OPC_ASM_ADDIS(base,  0, ARIA_STATE_SAVE_HI_16);        next_xlate_address++; 

  if (not_taken_address_required_flag == 1) { 
    /* Load the address to which the not-taken value should be stored */
    *next_xlate_address = LOAD_REG (scratch, NOT_TAKEN_RVAL_OFFSET, base);   next_xlate_address++; 
    /* And store the not-taken address there */
    *next_xlate_address = STORE_REG(scratch2, 0, scratch);                   next_xlate_address++; 
  }

  *base_reg_ptr = base;

  if (do_no_output_during_runs == 0) {
    DUMP_DESCRIPTION(temp_ptr, next_xlate_address);    
    warn_xlate("* (rest of exit) ----------------------------------------------------------\n");
    temp_ptr = next_xlate_address;
  }

  return next_xlate_address;
}




/*******************************************************************
  PROGRAM NAME..... bb_fallthrough
  FUNCTION ........ Generate the code to set the next address 
     (following this block) and the next not-taken address, 
     when a basic block does not end in a branch (e.g. 
     when it is broken because of too many instructions, etc.).
  NOTES............
  ASSUMPTIONS... None
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... the next available translation address
  CALLED-ROUTINES.. None
  CALLED-BY........ do_one_translation (see do_translation.c)
  INCLUDED-BY...... None
  MACROS-USED...... several: some are defined above as code-writing 
     shortcuts (where sequences of duplicated code are pulled out 
     into a macro).  Also the opc_asm macros (see opc_asm.gen).
  INCLUDES......... <see above>

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

unsigned * 
bb_fallthrough(unsigned  * prog_address,
	       unsigned  * base_reg_ptr
	       )
{
  unsigned base = *base_reg_ptr;

  BASE_TYPE fall_thru_path = (BASE_TYPE) (prog_address + 1); /** The fall-through path **/
  BASE_TYPE taken_target   = fall_thru_path;

  ADDRESS_STORAGE_DECLARES;

  unsigned scratch  = 12;
  unsigned scratch2 = 11;

  unsigned * temp_ptr = next_xlate_address;

  IFDEBUG(warn_exec("[bb_fallthrough]  Entered the bb_fallthrough subroutine\n"); );

  /* Find a scratch register (unique from the base and dynamic trace buffer pointer registers) */
  while ( (scratch == base) ) {
    scratch--;
  }
  
  /* Find a second scratch register (unique from the scratch, base and dynamic trace buffer pointer registers) */
  while ( (scratch2 == base) || (scratch2 == scratch) ) {
    scratch2--;
  }

  IFDEBUG(warn_xlate("[bb_fallthrough]  Inserting the \"fall-through\" code\n");
	  warn_xlate("[bb_fallthrough]     fall-through target = 0x%016LLx\n", fall_thru_path);
	  warn_xlate("[bb_fallthrough]     (taken target)      = 0x%016LLx\n", taken_target); 
	  );

  /* insert the not-taken branch code (i.e. the fall-through code) */
  NOT_TAKEN_BRANCH_CODE;

  if (do_no_output_during_runs == 0) {
    warn_xlate("*   fall-thru code --------------------------------------------------------\n");
    DUMP_DESCRIPTION(temp_ptr, next_xlate_address);    
    temp_ptr = next_xlate_address;
  }

  if (not_taken_address_required_flag == 1) { 
    /* Get the not-taken path address */
    /*  NOTE: do this before setting GPR 3 since GPR 3 may be the base register */
    *next_xlate_address = LOAD_REG    (scratch2, NOT_TAKEN_OFFSET, base);       next_xlate_address++; 
  }

  /**************************************************************************/
  /** Set up the subroutine return value (the taken path address) in GPR 3 **/
  /**************************************************************************/
  /* Set the function return value (GPR3) to the next BB address */
  *next_xlate_address = LOAD_REG  ( 3,  NEXT_TARGET_OFFSET, base);    next_xlate_address++;

  /** If the base WAS GPR 3, find a new base register */
  if (base == 3) {
    base++;
    while ( (base == scratch) || (base == scratch2) ) {
      warn("base = %u scratch = %u scratch2 = %u dtrb = %u\n", base, scratch, scratch2);
      base++;
    }
  }

  /**********************************************/
  /** Set up the not-taken path return address **/
  /**********************************************/

  /* Load GPRbase_register with the address of our ARIA state saving space    */               
  *next_xlate_address = OPC_ASM_ADDIS(base,  0, ARIA_STATE_SAVE_HI_16);        next_xlate_address++; 

  if (not_taken_address_required_flag == 1) { 
    /* Load the address to which the not-taken value should be stored */
    *next_xlate_address = LOAD_REG (scratch, NOT_TAKEN_RVAL_OFFSET, base);   next_xlate_address++; 
    *next_xlate_address = STORE_REG(scratch2, 0, scratch);                   next_xlate_address++; 
  }

  if (do_no_output_during_runs == 0) {
    DUMP_DESCRIPTION(temp_ptr, next_xlate_address);    
    warn_xlate("* (rest of exit) ----------------------------------------------------------\n");
    temp_ptr = next_xlate_address;
  }

  *base_reg_ptr = base;

  return next_xlate_address;
}








/*******************************************************************
  PROGRAM NAME..... find_relative_branch_target
  FUNCTION ........ Analyze the user-code branch and, if it is a 
     branch with a displacement, determine the (taken path) branch
     target address (in the original user code).  The fallthrough
     address is also computed.

  NOTES............
  ASSUMPTIONS... None
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... None
  CALLED-ROUTINES.. None
  CALLED-BY........ analyze_lwarx_stwcx.c
  INCLUDED-BY...... None
  MACROS-USED...... several: some are defined here as code-writing 
     shortcuts (where sequences of duplicated code are pulled out 
     into a macro).  Also the opc_asm macros (see opc_asm.gen).
  INCLUDES......... <see below>

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

unsigned * 
find_relative_branch_target(unsigned   branch_iword,
			    unsigned * prog_address,
			    unsigned * target_addr,
			    unsigned * fallthru_addr
			    )
{
  unsigned iword = branch_iword;
  opc_idx  index = opc_index(iword);

  BASE_TYPE orig_pc  = (BASE_TYPE) prog_address;
  
  BASE_TYPE fall_thru_path  = (BASE_TYPE) (prog_address + 1); /** The fall-through path **/
  BASE_TYPE taken_target    = 0;

  int offset;

  /** Now set the taken path target address **/
  switch (index)
    {
    case OPC_B:
      offset = OPC_DEC_LI(iword) << 2;
      taken_target = orig_pc + offset;
      break;
    case OPC_BA:
      offset = OPC_DEC_LI(iword) << 2;
      taken_target = offset;
      break;
    case OPC_BL:
      offset = OPC_DEC_LI(iword) << 2;
      taken_target = orig_pc + offset;
      break;
    case OPC_BLA:
      offset = OPC_DEC_LI(iword) << 2;
      taken_target = offset;
      break;
    case OPC_BC:
      offset = OPC_DEC_BD(iword) << 2;
      taken_target = orig_pc + offset;
      break;
    case OPC_BCA:
      offset = OPC_DEC_BD(iword) << 2;
      taken_target = offset;
      break;
    case OPC_BCL:
      offset = OPC_DEC_BD(iword) << 2;
      taken_target = orig_pc + offset;
      break;
    case OPC_BCLA:
      offset = OPC_DEC_BD(iword) << 2;
      taken_target = offset;
      break;
    case OPC_BCCTR:
    case OPC_BCCTRL:
    case OPC_BCLR:
    case OPC_BCLRL:
      break;
      
    default:
      fprintf(mystderr, "Unknown branch instruction: oaddr = 0x%016LLx, xaddr = 0x%016LLx, iw = 0x%08x\n",
	      (BASE_TYPE)prog_address, (BASE_TYPE)next_xlate_address, iword);
      break;
    }

  /* For conditional branch, see if the condition is "take always" */
  if (OPC_IS_A_CONDITIONAL_BRANCH(index)) {
    unsigned bo_val = OPC_DEC_BO(iword);
    if ((bo_val & 0x1e) == 0x14) {  
      fall_thru_path = taken_target;
    }
  }

  *target_addr   = taken_target;
  *fallthru_addr = fall_thru_path;

  return next_xlate_address;
}





/*******************************************************************
  PROGRAM NAME..... insert_embedded_branch_code
  FUNCTION ........ Analyze the user-code basic block ending branch and
     determine how the Aria code should simulate the effect of that branch.
     Also generate the code to set the next address (following this block)
     and the next not-taken address. 

  NOTES............
  ASSUMPTIONS... None
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... the next available translation address
  CALLED-ROUTINES.. None
  CALLED-BY........ do_one_translation (see do_translation.c)
  INCLUDED-BY...... None
  MACROS-USED...... several: some are defined here as code-writing 
     shortcuts (where sequences of duplicated code are pulled out 
     into a macro).  Also the opc_asm macros (see opc_asm.gen).
  INCLUDES......... <see below>

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/


unsigned * 
insert_embedded_branch_code(unsigned   branch_iword,
			    unsigned   branch_number,
			    branch_list_type * the_branches_list,
			    unsigned   base_reg,
			    unsigned   scratch_reg,
			    unsigned   dtrbf_ptr,
			    unsigned   br_num_reg,
			    unsigned   br_target_reg,
			    unsigned   br_fallthru_reg
	       )
{
  unsigned base  = base_reg;

  unsigned iword = branch_iword;
  opc_idx  index = opc_index(iword);

  int br_to_register = OPC_BRANCH_THROUGH_REGISTER(index);

  if ( (the_branches_list[branch_number].rel_target != 0) && (the_branches_list[branch_number].fall_thru != 0) ) {
    if (OPC_IS_A_CONDITIONAL_BRANCH(index)) {
      unsigned bo_val = OPC_DEC_BO(iword);
      unsigned bi_val = OPC_DEC_BI(iword);
      if ((bo_val & 0x1c) != 0x0c)	{ /* NOT a BR always */
	*next_xlate_address = OPC_ASM_BC(bo_val, bi_val, 2);  next_xlate_address++;
	*next_xlate_address = OPC_ASM_B(3);  next_xlate_address++;
      }
    }
    *next_xlate_address = OPC_ASM_ADDI(br_num_reg, 0, branch_number+1); next_xlate_address++;
    
    the_branches_list[branch_number].skip_posn = next_xlate_address;
    *next_xlate_address = 0x00000000; next_xlate_address++;
  }
  else {
    if (br_to_register) { 
      unsigned bo_val = OPC_DEC_BO(iword);
      unsigned bi_val = OPC_DEC_BI(iword);
      if ((bo_val & 0x1c) != 0x0c)	{ /* NOT a BR always */
	*next_xlate_address = OPC_ASM_BC(bo_val, bi_val, 2);  next_xlate_address++;
	*next_xlate_address = OPC_ASM_B(4);  next_xlate_address++;
      }
      *next_xlate_address = OPC_ASM_ADDI(br_num_reg, 0, branch_number+1); next_xlate_address++;
      if ((index == OPC_BCCTR) || (index == OPC_BCCTRL)) {
	*next_xlate_address = LOAD_REG   (br_target_reg, CTR_SAVE_OFFSET, base); next_xlate_address++;
      }
      else {
	*next_xlate_address = LOAD_REG   (br_target_reg, LR_SAVE_OFFSET, base); next_xlate_address++;
      }

      the_branches_list[branch_number].skip_posn = next_xlate_address;  
      *next_xlate_address = 0x00000000; next_xlate_address++;
    } 
    else {
      warn("ARIA: Error -- unset target or fallthrough for NON br_to_register...\n");
      warn("\t target    = 0x%08x\n", the_branches_list[branch_number].rel_target);
      warn("\t fall_thru = 0x%08x\n", the_branches_list[branch_number].fall_thru);
      abort();
    }
  }
  return next_xlate_address;
}




/*******************************************************************
  PROGRAM NAME..... insert_embedded_branch_cleanup_code
  FUNCTION ........ Analyze the branches in the branch list and 
     decide (dynamically) what the effect of the taken branch is on 
     the user process state and next addresses.

  NOTES............
  ASSUMPTIONS... None
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... the next available translation address
  CALLED-ROUTINES.. None
  CALLED-BY........ analyze_lwarx_stwcx.c
  INCLUDED-BY...... None
  MACROS-USED...... several: some are defined here as code-writing 
     shortcuts (where sequences of duplicated code are pulled out 
     into a macro).  Also the opc_asm macros (see opc_asm.gen).
  INCLUDES......... <see below>

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/


unsigned * 
insert_embedded_branch_cleanup_code(unsigned   num_branches,
				    branch_list_type * the_branches_list,
				    unsigned  *last_user_instr_addr,
				    unsigned  *base_reg_ptr,
				    unsigned   scratch_reg,
				    unsigned   dtrbf_ptr,
				    unsigned   br_num_reg,
				    unsigned   br_target_reg,
				    unsigned   br_fallthru_reg
				    )
{
  unsigned * temp_save_address;
  unsigned * no_branch_taken_address;
  unsigned base  = *base_reg_ptr;

  BASE_TYPE fall_thru_path  = 0;
  BASE_TYPE taken_target    = 0;
  BASE_TYPE link_target     = 0;	  

  ADDRESS_STORAGE_DECLARES;

  unsigned scratch  = scratch_reg;

  unsigned ii;

  /** Implement a "jump_table" based on the "taken" branch value **/

  /** Skeleton code to get the offset into a jump table **/
  *next_xlate_address = OPC_ASM_RLWIMI(br_num_reg, br_num_reg, 2, 0, 31);     next_xlate_address++;

  /* keep a pointer to the positions to load the curr PC address */
  temp_save_address = next_xlate_address;
  next_xlate_address++;		/* Load the high half-word */
  next_xlate_address++;		/* Load the low  half-word */

  /* Add in the offset into the table */
  *next_xlate_address = OPC_ASM_ADD   (scratch_reg, scratch_reg, br_num_reg);  next_xlate_address++;
  /* Move the address to the count register */
  *next_xlate_address = OPC_ASM_MTSPR (9, scratch_reg);                        next_xlate_address++;
  /* branch through the count register */
  *next_xlate_address = OPC_ASM_BCCTR (0x14, 0x0);                             next_xlate_address++;

  /* Now that we know the address of the first jump table entry, set up the PC value **/
  SET_SPECIAL_BRANCH;
  
  /* Save a pointer to the first entry, and advance xlation pointer to after the table */
  temp_save_address = next_xlate_address;
  for (ii = 0; ii <= num_branches; ii++) {
    next_xlate_address++;
  }
  
  /** Set up the complete fall-thru case **/
  *temp_save_address = OPC_ASM_B((next_xlate_address - temp_save_address));  temp_save_address++;
  taken_target   = (BASE_TYPE) (last_user_instr_addr + 1);
  fall_thru_path = taken_target;
  TAKEN_BRANCH_CODE;
  no_branch_taken_address = next_xlate_address; 
  next_xlate_address++;

  /* Now set up the code to handle each original user-code branch */
  for (ii = 0; ii < num_branches; ii++) {
    unsigned * br_addr     = (unsigned *) the_branches_list[ii].address;
    unsigned   iword       = *br_addr;
    int        br_and_link = OPC_DEC_LK(iword);

    /* Set the jump-table entry */
    *temp_save_address = OPC_ASM_B((next_xlate_address - temp_save_address));  temp_save_address++;
   
    /** Insert the branch-handling code **/
    if (br_and_link) {
      /** Put the link target into the saved state **/
      link_target = the_branches_list[ii].address + 4;
      SAVE_LINK_TARGET;
    }
    
    fall_thru_path = the_branches_list[ii].fall_thru;

    if (the_branches_list[ii].rel_target != 0x0) {
      taken_target = the_branches_list[ii].rel_target;
      TAKEN_BRANCH_CODE;
    }
    else { /** Assumed to be a branch through register **/
      *next_xlate_address = STORE_REG(br_target_reg, NEXT_TARGET_OFFSET, base); next_xlate_address++; 
      if (not_taken_address_required_flag == 1) { 
	*next_xlate_address = STORE_REG(br_fallthru_reg, NEXT_TARGET_OFFSET, base); next_xlate_address++; 
      }
    }
    /* Now save a place (and pointer to where) for a branch over rest of branch handlers */
    the_branches_list[ii].skip_posn = next_xlate_address;
    next_xlate_address++;
  }

  /* Set the skip out addresses */
  *no_branch_taken_address = OPC_ASM_B((next_xlate_address - no_branch_taken_address));

  for (ii = 0; ii < num_branches; ii++) {
    if (the_branches_list[ii].skip_posn != 0) {
      if ((next_xlate_address - the_branches_list[ii].skip_posn) == 1) {
	*the_branches_list[ii].skip_posn = OPC_ASM_OR(0,0,0);  /* NOP */
      }
      else {
	*the_branches_list[ii].skip_posn = OPC_ASM_B((next_xlate_address - the_branches_list[ii].skip_posn));
      }
    }
    else {
      warn("The jump table has entries with no skips?\n");
    }
  }
  
  /** Set the dynamic trace information if it is being gathered **/
  if (do_dynamic_trace_generation) {
    /* Set the dynamic trace data for this branch */
    *next_xlate_address = OPC_ASM_ADDIS(dtrbf_ptr, 0, ARIA_STATE_SAVE_HI_16);                         next_xlate_address++;
    *next_xlate_address = LOAD_REG    (dtrbf_ptr, PROC_DYN_TRBUF_OFFSET, dtrbf_ptr);  next_xlate_address++;

    /* Get the return address */
    *next_xlate_address = LOAD_REG  (br_target_reg, NEXT_TARGET_OFFSET, base);  next_xlate_address++; 

    if (do_iaddr_fixup) {
      /* Modify it to the true user space address */
      *next_xlate_address = OPC_ASM_RLWINM(scratch, br_target_reg, 4, 28, 31);      next_xlate_address++;
      *next_xlate_address = OPC_ASM_CMPI  ( 7, 0, scratch, (global_user_program_iaddr_mask>>28)&0xf);                    next_xlate_address++;
      *next_xlate_address = OPC_ASM_BC    ( 0x4, 30, 3);                            next_xlate_address++;
      *next_xlate_address = LOAD_REG     (scratch, IADDR_OFFSET, base);            next_xlate_address++;
      *next_xlate_address = OPC_ASM_SUBFC (br_target_reg, scratch, br_target_reg);  next_xlate_address++;
    }
    
    /* Store the user-view target address in the dynamic trace */
    *next_xlate_address = OPC_ASM_MULLI( br_target_reg, br_num_reg, ARIA_TB_PER_INSTR_SIZE);       next_xlate_address++;
    *next_xlate_address = OPC_ASM_ADDI ( br_target_reg, br_target_reg, ARIA_TB_BRTARGET_OFFSET);   next_xlate_address++;
    *next_xlate_address = STORE_REG_X  ( scratch, br_target_reg, dtrbf_ptr);  next_xlate_address++;
  }

  if (not_taken_address_required_flag == 1) { 
    /* Get the not-taken path address */
    /*  NOTE: do this before setting GPR 3 since GPR 3 may be the base register */
    if (br_fallthru_reg == 3) {
      *next_xlate_address = LOAD_REG    (br_target_reg, NOT_TAKEN_OFFSET, base);       next_xlate_address++; 
    }
    else {
      *next_xlate_address = LOAD_REG    (br_fallthru_reg, NOT_TAKEN_OFFSET, base);       next_xlate_address++; 
    }
  }

  /**************************************************************************/
  /** Set up the subroutine return value (the taken path address) in GPR 3 **/
  /**************************************************************************/
  /* Set the function return value (GPR3) to be the next BB address */  
  *next_xlate_address = LOAD_REG    ( 3,  NEXT_TARGET_OFFSET, base);                next_xlate_address++;

  /** If the base WAS GPR 3, find a new base register */
  if (base == 3) {
    base++;
    while ( (base == scratch) || (base == br_fallthru_reg) || (base == br_target_reg) || (base == dtrbf_ptr) ) {
      base++;
    }
  }

  /**********************************************/
  /** Set up the not-taken path return address **/
  /**********************************************/

  /* Load GPRbase_register with the address of our ARIA state saving space    */               
  *next_xlate_address = OPC_ASM_ADDIS(base,  0, ARIA_STATE_SAVE_HI_16);        next_xlate_address++; 

  if (not_taken_address_required_flag == 1) { 
    /* Load the address to which the not-taken value should be stored */
    *next_xlate_address = LOAD_REG(scratch, NOT_TAKEN_RVAL_OFFSET, base);   next_xlate_address++; 
    if (br_fallthru_reg == 3) {
      /* And store the not-taken address there */
      *next_xlate_address = STORE_REG(br_target_reg, 0, scratch);                   next_xlate_address++; 
    }
    else {
      *next_xlate_address = STORE_REG(br_fallthru_reg, 0, scratch);                   next_xlate_address++; 
    }
  }

  *base_reg_ptr = base;


  return next_xlate_address;
}


