/*                                                                     
** COMPONENT_NAME: MET / Aria
**                                                                      
** Licensed Materials - Property of IBM                                  
**                                                                     
** IBM Confidential (IBM Confidential-Restricted when combined          
**                   with aggregated modules for this product)          
**                                                                     
**                                                                     
** OCO Source Materials                                                
**                                                                     
**                                                                     
** GA-0100 (C) Copyright IBM Corp. 1997                                 
**                                                                     
** US Government Users Restricted RIghts - Use, duplication or         
** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.   
**                                                                     
** The source code for this program is not published or otherwise      
** divested of its trade secrets, irrespective of what has been        
** deposited within the U.S. Copyright Office.                         
**                                                                     
*/


 
/*******************************************************************
  PROGRAM NAME..... aria_copy_process_state_from_to
  FUNCTION ........ Copy a given processor state space into a 
     second, given processor state.  Useful for duplicating the 
     user process state when a second copy (e.g. for checkpointing
     or before speculative execution) is needed.

  NOTES............
  ASSUMPTIONS... None
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... None
  CALLED-ROUTINES.. None
  CALLED-BY........ the model that drives aria
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... <see below>

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

#include <stdio.h>

#include "proc_state.h"

#include "global_variables.h"
#include "io_msgs.h"


void local_compare_processor_states(proc_state_type * from_ptr, proc_state_type * to_ptr, int check_FP);


void 
aria_copy_process_state_from_to(proc_state_type * from_ptr,
				proc_state_type * to_ptr
				)
{
  int i;
  int j;
  /* Copy the segment register validity flags */
  for (j = 0; j < 16; j++) {
    for (i = 0; i < 16; i++) {
      proc_state_valid_segment(to_ptr, j, i) = proc_state_valid_segment(from_ptr, j, i);
    }
  }

  /* Copy the processor special state registers */
  proc_state_LR(to_ptr)         = proc_state_LR(from_ptr);
  proc_state_CTR(to_ptr)        = proc_state_CTR(from_ptr);
  proc_state_CR(to_ptr)         = proc_state_CR(from_ptr);
  proc_state_XER(to_ptr)        = proc_state_XER(from_ptr);
  proc_state_MQ(to_ptr)         = proc_state_MQ(from_ptr);

  /* Copy the next target and next not0taken targets */
  proc_state_TARGET(to_ptr)     = proc_state_TARGET(from_ptr);
  proc_state_NT_TARGET(to_ptr)  = proc_state_NT_TARGET(from_ptr);  

  /* Copy the FPSCR space */
  to_ptr->_aria_fpscr_dummy     = from_ptr->_aria_fpscr_dummy;
  proc_state_FPSCR(to_ptr)      = proc_state_FPSCR(from_ptr);

  /* Copy the GPRs and FPRs */
  for (i = 0; i < 32; i++) {
    proc_state_GPR(to_ptr,i)            = proc_state_GPR(from_ptr,i);
    proc_state_FPR(to_ptr,i)            = proc_state_FPR(from_ptr,i);
  }

#ifdef DO_FULL_DEBUGGING
  warn("--------------------------------------------------------------------------------\n");
  warn("The copied processor state is:\n");
  local_compare_processor_states(from_ptr, to_ptr, 0);
  warn("--------------------------------------------------------------------------------\n");
#endif
}


#ifdef DO_FULL_DEBUGGING
void local_compare_processor_states(proc_state_type * from_ptr, 
				    proc_state_type * to_ptr, 
				    int check_FP
				    )
{
  int same = 1;
  int i;

  /* Check the segment register validity flags */
  for (i = 0; i < 16; i++) {
    if (proc_state_valid_segment(to_ptr, i) != proc_state_valid_segment(from_ptr, i)) {
      warn("proc_state_valid_segment(%d) differs : %u vs %u\n", i,
	   proc_state_valid_segment(to_ptr, i), proc_state_valid_segment(from_ptr, i));
      same = 0;
    }
  }

  /* Check the processor special state registers */
  if (proc_state_LR(to_ptr) != proc_state_LR(from_ptr) ) {
    warn("proc_state_LR differs : %08x vs %08x\n",
	 proc_state_LR(to_ptr), proc_state_LR(from_ptr));
    same = 0;
  }

  if (proc_state_CTR(to_ptr) != proc_state_CTR(from_ptr) ) {
    warn("proc_state_CTR differs : %08x vs %08x\n",
	 proc_state_CTR(to_ptr), proc_state_CTR(from_ptr));
    same = 0;
  }

  if (proc_state_CR(to_ptr) != proc_state_CR(from_ptr) ) {
    warn("proc_state_CR differs : %08x vs %08x\n",
	 proc_state_CR(to_ptr), proc_state_CR(from_ptr));
    same = 0;
  }

  if (proc_state_XER(to_ptr) != proc_state_XER(from_ptr) ) {
    warn("proc_state_XER differs : %08x vs %08x\n",
	 proc_state_XER(to_ptr), proc_state_XER(from_ptr));
    same = 0;
  }

  if (proc_state_MQ(to_ptr) != proc_state_MQ(from_ptr) ) {
    warn("proc_state_MQ differs : %08x vs %08x\n",
	 proc_state_MQ(to_ptr), proc_state_MQ(from_ptr));
    same = 0;
  }

  /* Check the next target and next not0taken targets */
  if (proc_state_TARGET(to_ptr) != proc_state_TARGET(from_ptr) ) {
    warn("proc_state_TARGET differs : %08x vs %08x\n",
	 proc_state_TARGET(to_ptr), proc_state_TARGET(from_ptr));
    same = 0;
  }

  if (proc_state_NT_TARGET(to_ptr) != proc_state_NT_TARGET(from_ptr) ) {
    warn("proc_state_NT_TARGET differs : %08x vs %08x\n",
	 proc_state_NT_TARGET(to_ptr), proc_state_NT_TARGET(from_ptr));
    same = 0;
  }

  /* Check the FPSCR space */
  if (to_ptr->_aria_fpscr_dummy != from_ptr->_aria_fpscr_dummy) {
    warn("_aria_fpscr_dummy differs : %08x vs %08x\n",
	 to_ptr->_aria_fpscr_dummy, from_ptr->_aria_fpscr_dummy);
    same = 0;
  }
  if (proc_state_FPSCR(to_ptr) != proc_state_FPSCR(from_ptr) ) {
    warn("proc_state_FPSCR differs : %08x vs %08x\n",
	 proc_state_FPSCR(to_ptr), proc_state_FPSCR(from_ptr));
    same = 0;
  }
  
  /* Check the GPRs and FPRs */
  for (i = 0; i < 32; i++) {
    if (proc_state_GPR(to_ptr,i) != proc_state_GPR(from_ptr,i)) { 
      warn("proc_state_GPR(%d) differs : %08x vs %08x\n", i,
	   proc_state_GPR(to_ptr,i), proc_state_GPR(from_ptr,i));
      same = 0;
    }
  }
  for (i = 0; i < 32; i++) {
    if (proc_state_FPR(to_ptr,i) != proc_state_FPR(from_ptr,i)) {
      unsigned * pt1a = (unsigned *) (&proc_state_FPR(to_ptr,i));
      unsigned * pt2a = (unsigned *) ((unsigned *)(&proc_state_FPR(to_ptr,i))) + 1;
      unsigned * pt1b = (unsigned *) (&proc_state_FPR(to_ptr,i));
      unsigned * pt2b = (unsigned *) ((unsigned *)(&proc_state_FPR(to_ptr,i))) + 1;
      warn("proc_state_FPR(%d) differs : %08x%08x vs %08x%08x\n", i,
	   *pt1a, *pt2a, *pt1b, *pt2b);
      same = 0;
    }
  }

  if (same) { 
    warn("proc state copied identically\n");
  }
}
#endif
