/*                                                                     
** COMPONENT_NAME: MET / Aria
**                                                                      
** Licensed Materials - Property of IBM                                  
**                                                                     
** IBM Confidential (IBM Confidential-Restricted when combined          
**                   with aggregated modules for this product)          
**                                                                     
**                                                                     
** OCO Source Materials                                                
**                                                                     
**                                                                     
** GA-0100 (C) Copyright IBM Corp. 1997                                 
**                                                                     
** US Government Users Restricted RIghts - Use, duplication or         
** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.   
**                                                                     
** The source code for this program is not published or otherwise      
** divested of its trade secrets, irrespective of what has been        
** deposited within the U.S. Copyright Office.                         
**                                                                     
*/


 
/*******************************************************************
  PROGRAM NAME..... aria_expanded_interface.c

  FUNCTION ........ Provides an expanded interface to Aria that 
     includes functionality similar to DEC's ATOM tool and many 
     others, that allow the user to specify actions to take on a 
     per-instruction (type) basis.

  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 
  CALLED-ROUTINES.. 
  CALLED-BY........ None
  INCLUDED-BY...... 
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

#include <limits.h>

#include "aix_a.out.h"
#include "aix_errno.h"
#include "aix_ldr.h"
#include "aix_mman.h"
#include "posix_fcntl.h"
#include "ansi_stdio.h"
#include "ansi_stdlib.h"

#include "ansi_assert.h"

#include "aria_state.h"

#include "io_msgs.h"
#include "opc.h"
#include "proc_state.h"
#include "hashing.h"
#include "do_translation.h"
#include "do_svc_translation.h"
#include "save_user_state.h"

#include "globdefs.h"
#include "globtypes.h"

#include "global_variables.h"

#include "interface_globals.h"

#include "aria_interface.h"

/** definitions of the per-instruction data types, access macros, etc. **/
#include "aria_per_instr_interface.h"

/*******************************************************************
  PROGRAM NAME..... aria_register_function

  FUNCTION ........ Adds an entry to the set of per-instruction 
     actions defined for the current run of Aria on a user program,
     indicating that whenever an instruction of a given index
     (where the indices are defined in opc_decl.h) is encountered,
     the current basic block should be broken and a user function 
     is called (the_function is a pointer to the function to call).

  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... returns a success or failure value
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

int 
aria_register_function(unsigned instr_index,
		       short act_at_time,
		       aria_user_function_type the_function
		       )
{
  if (instr_index > OPC_NUM) {
    fprintf(mystderr, "ARIA: ERROR: Cannot register instruction index %u, \n", instr_index);
    fprintf(mystderr, "        There are only %u valid instruction indices\n", OPC_NUM);
    return -1;
  }

  switch (act_at_time) 
    {
    case DO_ARIA_ACTION_BEFORE:
      if ( (per_instruction_action_time(instr_index) & DO_ARIA_ACTION_BEFORE) == 0 ) {
	per_instruction_action_time(instr_index)    |= act_at_time;
	per_instruction_action_type(instr_index)     = PER_INSTR_FUNCTION_CALL_ACTION;
	per_instruction_before_function_ptr(instr_index)  = the_function;
	return 0;
      }
      else {
	fprintf(mystderr, "ARIA: ERROR: Cannot register instruction index %u, \n", instr_index);
	fprintf(mystderr, "        There is already a BEFORE registration for that index.\n");
	return -1;
      }
      break;
      
    case DO_ARIA_ACTION_AFTER:
      if ( (per_instruction_action_time(instr_index) & DO_ARIA_ACTION_AFTER) == 0 ) {
	per_instruction_action_time(instr_index)    |= act_at_time;
	per_instruction_action_type(instr_index)     = PER_INSTR_FUNCTION_CALL_ACTION;
	per_instruction_after_function_ptr(instr_index)  = the_function;
      }
      else {
	fprintf(mystderr, "ARIA: ERROR: Cannot register instruction index %u, \n", instr_index);
	fprintf(mystderr, "        There is already an AFTER registration for that index.\n");
	return -1;
      }
      break;
      
    case DO_NO_ARIA_ACTION:
    default:
      break;
    }
  
  return 0;
}


/*******************************************************************
  PROGRAM NAME..... aria_register_itype_function

  FUNCTION ........ Adds an entry to the set of per-instruction 
     actions defined for the current run of Aria on a user program,
     indicating that whenever an instruction of a given itype
     (where the itypes are defined in opc_itype_defns.h) is encountered,
     the current basic block should be broken and a user function 
     is called (the_function is a pointer to the function to call).
     This simply adds the same action for all instruction indices of 
     the given instruction itype.

  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... returns a success or failure value
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

int 
aria_register_itype_function(unsigned instr_itype,
			     short act_at_time,
			     aria_user_function_type the_function
			     )
{
  unsigned instr_index;
  for (instr_index = 0; instr_index < OPC_NUM; instr_index++) {

    if (opc_itype(instr_index) == instr_itype) {
      int rval = aria_register_function(instr_index, act_at_time, the_function);
      if (rval != 0) { 
	return rval;
      }
    }
  }
  return 0;
}

