/*                                                                     
** COMPONENT_NAME: MET / Aria
**                                                                      
** Licensed Materials - Property of IBM                                  
**                                                                     
** IBM Confidential (IBM Confidential-Restricted when combined          
**                   with aggregated modules for this product)          
**                                                                     
**                                                                     
** OCO Source Materials                                                
**                                                                     
**                                                                     
** GA-0100 (C) Copyright IBM Corp. 1997                                 
**                                                                     
** US Government Users Restricted RIghts - Use, duplication or         
** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.   
**                                                                     
** The source code for this program is not published or otherwise      
** divested of its trade secrets, irrespective of what has been        
** deposited within the U.S. Copyright Office.                         
**                                                                     
*/


 
/*******************************************************************
  PROGRAM NAME..... aria_skip_interface

  FUNCTION ........ Provides a set of interface routines to aria, 
     expanding the user interface, to provide faster, skip-ahead
     style execution calls

  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 
  CALLED-ROUTINES.. 
  CALLED-BY........ None
  INCLUDED-BY...... 
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

#include <limits.h>

#include "aix_a.out.h"
#include "aix_errno.h"
#include "aix_ldr.h"
#include "aix_mman.h"
#include "posix_fcntl.h"
#include "ansi_stdio.h"
#include "ansi_stdlib.h"

#include "ansi_assert.h"

#include "aria_state.h"

#include "io_msgs.h"
#include "opc.h"
#include "proc_state.h"
#include "hashing.h"
#include "do_translation.h"
#include "do_svc_translation.h"
#include "save_user_state.h"

#include "globdefs.h"
#include "globtypes.h"

#include "global_variables.h"

#include "interface_globals.h"

/** Forward function declarations **/

#include "aria_interface.h"
#include "aria_per_instr_interface.h"

#include "analyze_branch.h"

/*******************************************************************
  PROGRAM NAME..... do_aria_skip_bbs_ll

  FUNCTION ........ Executes a number (specified in the call) of 
     user basic blocks in a faster, skip-ahead mode, which allows
     the user to quickly move some distance into a user program 
     execution before beginning tracing.
 
  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

#define TRANSLATE_TAKEN_PATH_CODE  0

#define  SKIP_MODE_FLAG  2


unsigned * 
do_aria_skip_bbs_ll(unsigned *  next_bb_address, 
		    proc_state_type * saved_proc_state_ptr, 
		    unsigned long long num_bbs_to_skip
		    )
{
  unsigned long long num_bbs_skipped = 0;

  unsigned ubb_size;
  unsigned taken_xbb_size;
  unsigned not_taken_xbb_size;

  unsigned * next_xlated_bb_address;
  unsigned * next_not_taken_xl_bb_address;
  unsigned * microtrace_buffer_ptr;

  unsigned * taken_path;
  unsigned * not_taken_path;
  unsigned * temp_ptr;

  unsigned long long num_instrs_skipped = 0;

  hash_entry * the_hash_entry_ptr = NULL;

  unsigned saved_do_dynamic_trace_generation = do_dynamic_trace_generation;
  unsigned saved_do_iaddr_fixup = do_iaddr_fixup;
  unsigned saved_do_not_taken_path_translations = do_not_taken_path_translations;
  unsigned saved_no_FP_registers_used_in_model_flag = no_FP_registers_used_in_model_flag;  

  /* Set the "translations mode" flag to indicate       */
  /*  we are doing skip-ahead mode translations (so the */
  /*  existing non skip-ahead xlations are "invalid"    */

  /* Check for an execution of the exit flag address */
  if ( (((unsigned)next_bb_address) == ARIA_EXIT_FLAG_ADDRESS) ) {
    return ARIA_EXIT_FLAG_ADDRESS;    
  }

  /** If iaddresses are being adjusted, the user could be sending an   **/
  /**  adjusted address, so Aria checks it and corrects the address    **/
  /**  to recover the appropriate address in Aria's view of the loaded **/
  /**  user program.                                                   **/
  if ( (do_iaddr_fixup) && (((unsigned)next_bb_address & 0xf0000000) == 0x10000000) ) {
    next_bb_address = (unsigned *) ((unsigned)next_bb_address + global_user_iaddr_offset);
  }

  /*************************************************/
  /** Now set up to do skip-ahead mode executions **/
  /*************************************************/

  /* Turn off any unnecessary translation mode flags that */
  /*  would slow down skip-ahead mode (saved above)       */
  do_dynamic_trace_generation = 0;
  do_iaddr_fixup = 0;
  do_not_taken_path_translations = 0;
  no_FP_registers_used_in_model_flag = 1;  

  /***********************************************************/
  /** Now we need to start executing from the user-provided **/
  /** address in skip-ahead mode.                           **/
  /***********************************************************/
  #ifdef DEBUG_MODE
    warn("In Skip-Mode: skipping %llu basic blocks\n", num_bbs_to_skip);
  #endif

  while ( (num_bbs_skipped < num_bbs_to_skip) && ((unsigned)next_bb_address != 0x00000000) ) {

    #ifdef DEBUG_MODE
      warn("\nLooking up hash table entry for Oaddr = 0x%08x\n", next_bb_address);
    #endif

    /** Check for a translation of this basic block **/
    the_hash_entry_ptr = find_hash_table_entry_for(next_bb_address, SKIP_MODE_FLAG);

    next_xlated_bb_address = HashTable_xlation_address(*the_hash_entry_ptr);

    #ifdef DEBUG_MODE
      warn("  ... found translation entry with Xaddr = 0x%08x\n", next_xlated_bb_address);
    #endif

    /** If no translation already exists, then create one **/
    #ifdef DEBUG_MODE	
      temp_ptr = next_bb_address;
    #endif

    if (next_xlated_bb_address == NULL) {
      /* If the execution is a system call, then translate it to a canned sys call block **/
      if ( (global_use_safe_svc_code && (next_bb_address <= (unsigned *)0x4000)) ||
	   (OPC_IS_A_SYSTEM_CALL(opc_iword_itype(*next_bb_address))) ) {
	next_xlate_address = do_one_svc_translation(next_bb_address, the_hash_entry_ptr, 
						    TRANSLATE_TAKEN_PATH_CODE, SKIP_MODE_FLAG);
      }
      /* ELSE this is a normal (user code) basic block, so translate its contents */
      else {
	next_xlate_address = do_one_translation(next_bb_address, the_hash_entry_ptr, 
						TRANSLATE_TAKEN_PATH_CODE, SKIP_MODE_FLAG);
      }

      next_xlated_bb_address = HashTable_xlation_address(*the_hash_entry_ptr);
      ubb_size = HashTable_user_bb_size(*the_hash_entry_ptr);
      taken_xbb_size = HashTable_xlated_bb_size(*the_hash_entry_ptr);
      not_taken_xbb_size = HashTable_NT_xlated_bb_size(*the_hash_entry_ptr);

      global_user_line_index   += ubb_size;
      global_xlated_line_index += taken_xbb_size + not_taken_xbb_size;
      global_bb_index++;

      #ifdef DEBUG_MODE	
        next_not_taken_xl_bb_address = HashTable_NT_xlation_address(*the_hash_entry_ptr);
	microtrace_buffer_ptr = HashTable_trace_buffer(*the_hash_entry_ptr);
        warn("Skip-Mode: translated BB: Oa = 0x%08x Xa = 0x%08x SXa = 0x%08x Sbf = 0x%08x Us = %u Ts = %u NTs = %u\n", 
	     temp_ptr, next_xlated_bb_address, next_not_taken_xl_bb_address, *microtrace_buffer_ptr, 
	     ubb_size, taken_xbb_size, not_taken_xbb_size); 
      #endif
    }

    /****************************************/
    /** Execute the translated basic block **/
    pc_func_desc(prog_desc) = next_xlated_bb_address;

    next_bb_address = (unsigned *)prog_ptr(saved_proc_state_ptr, &not_taken_path);

    #ifdef DEBUG_MODE
      next_not_taken_xl_bb_address = HashTable_xlation_address(*the_hash_entry_ptr);
      microtrace_buffer_ptr = HashTable_trace_buffer(*the_hash_entry_ptr);
      ubb_size = HashTable_user_bb_size(*the_hash_entry_ptr);
      taken_xbb_size = HashTable_xlated_bb_size(*the_hash_entry_ptr);
      not_taken_xbb_size = HashTable_NT_xlated_bb_size(*the_hash_entry_ptr);
      warn("Skip-Mode: Executed BB %lld, the code for Oa = %08x Xa = %08x SXa = 0x%08x Sbf = 0x%08x Us = %u Ts = %u\n", 
		 num_bbs_skipped, temp_ptr, pc_func_desc(prog_desc),
		 next_not_taken_xl_bb_address, *microtrace_buffer_ptr, ubb_size, taken_xbb_size);
    #endif 

    ubb_size = HashTable_user_bb_size(*the_hash_entry_ptr);
    taken_xbb_size = HashTable_xlated_bb_size(*the_hash_entry_ptr);

    num_instrs_skipped += ubb_size;
    num_bbs_skipped++;

    global_user_instructions   += ubb_size;
    global_xlated_instructions += taken_xbb_size;
    global_bb_executions++;
    
    if ( (do_safe_exit_flag) && (((unsigned)next_bb_address) == the_exit_routine_address) ) {
     #ifdef FULL_DEBUG
      fprintf(mystderr, "<END> Hit the exit routine address ...\n");
     #endif
      return ARIA_EXIT_FLAG_ADDRESS;
    }
  }

  #ifdef DEBUG
    warn("In Skip-Mode: skipped %u basic blocks, %u instructions\n", num_bbs_to_skip, num_instrs_skipped);
    fflush(mystderr);
  #endif

  /* Restore the global translation control flags */
  do_dynamic_trace_generation = saved_do_dynamic_trace_generation;
  do_iaddr_fixup = saved_do_iaddr_fixup;
  do_not_taken_path_translations = saved_do_not_taken_path_translations;
  no_FP_registers_used_in_model_flag = saved_no_FP_registers_used_in_model_flag;  
  /* Return the proper next_bb_address to the user */
  if ( (do_iaddr_fixup) && (((unsigned)next_bb_address & 0xf0000000) == global_user_program_iaddr_mask) ) {
    return (unsigned *) ((unsigned)next_bb_address - global_user_iaddr_offset);
  }
  else {
    return  next_bb_address;
  }
}


/*******************************************************************
  PROGRAM NAME..... do_aria_skip_bbs

  FUNCTION ........ Calls the do_aria_skip_bbs_ll
     routine using the old interface with unsigned
     rather than unsigned long long number of BBs.

  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

unsigned * 
do_aria_skip_bbs(unsigned *  next_bb_address, 
		 proc_state_type * saved_proc_state_ptr, 
		 unsigned num_bbs_to_skip
		 )
{
  unsigned * ret_addr;
  unsigned long long ll_bbs_to_skip = num_bbs_to_skip;
  
  ret_addr = do_aria_skip_bbs_ll(next_bb_address, saved_proc_state_ptr, ll_bbs_to_skip);

  return ret_addr;
}

/*******************************************************************
  PROGRAM NAME..... do_aria_skip_instructions_ll

  FUNCTION ........ Executes a number (specified in the call) of 
     user basic blocks until the number of user program instructions
     executed exceeds the number of instructions that were to be
     skipped in the user program.  This skipping forward takes place
     in a faster, skip-ahead mode, which allows the user to quickly 
     move some distance into a user program execution before 
     beginning tracing.
 
  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

unsigned * 
do_aria_skip_instructions_ll(unsigned *  next_bb_address, 
			     proc_state_type * saved_proc_state_ptr, 
			     unsigned long long   num_instrs_to_skip,
			     unsigned long long * actual_num_instrs_skipped
			     )
{
  unsigned long long num_instrs_skipped = 0;

  unsigned ubb_size;
  unsigned taken_xbb_size;
  unsigned not_taken_xbb_size;

  unsigned * next_xlated_bb_address;
  unsigned * next_not_taken_xl_bb_address;
  unsigned * microtrace_buffer_ptr;

  unsigned * taken_path;
  unsigned * not_taken_path;
  unsigned * temp_ptr;

  hash_entry * the_hash_entry_ptr = NULL;

  unsigned saved_do_dynamic_trace_generation = do_dynamic_trace_generation;
  unsigned saved_do_iaddr_fixup = do_iaddr_fixup;
  unsigned saved_do_not_taken_path_translations = do_not_taken_path_translations;
  unsigned saved_no_FP_registers_used_in_model_flag = no_FP_registers_used_in_model_flag;  
  
  /* Set the "translations mode" flag to indicate       */
  /*  we are doing skip-ahead mode translations (so the */
  /*  existing non skip-ahead xlations are "invalid"    */

  /* Check for an execution of the exit flag address */
  if ( (((unsigned)next_bb_address) == ARIA_EXIT_FLAG_ADDRESS) ) {
    return ARIA_EXIT_FLAG_ADDRESS;    
  }
  /** If iaddresses are being adjusted, the user could be sending an   **/
  /**  adjusted address, so Aria checks it and corrects the address    **/
  /**  to recover the appropriate address in Aria's view of the loaded **/
  /**  user program.                                                   **/
  if ( (do_iaddr_fixup) && (((unsigned)next_bb_address & 0xf0000000) == 0x10000000) ) {
    next_bb_address = (unsigned *) ((unsigned)next_bb_address + global_user_iaddr_offset);
  }

  /*************************************************/
  /** Now set up to do skip-ahead mode executions **/
  /*************************************************/

  /* Turn off any unnecessary translation mode flags that */
  /*  would slow down skip-ahead mode (saved above)       */
  do_dynamic_trace_generation = 0;
  do_iaddr_fixup = 0;
  do_not_taken_path_translations = 0;
  no_FP_registers_used_in_model_flag = 1;

  /***********************************************************/
  /** Now we need to start executing from the user-provided **/
  /** address in skip-ahead mode.                           **/
  /***********************************************************/
  #ifdef DEBUG_MODE
    warn("In Skip-Mode: skipping %u instructions\n", num_instrs_to_skip);
  #endif

  while ( (num_instrs_skipped < num_instrs_to_skip) && ((unsigned)next_bb_address != 0x00000000) ) {

    #ifdef DEBUG_MODE
      warn("\nLooking up hash table entry for Oaddr = 0x%08x\n", next_bb_address);
    #endif

    /** Check for a translation of this basic block **/
    the_hash_entry_ptr = find_hash_table_entry_for(next_bb_address, SKIP_MODE_FLAG);

    next_xlated_bb_address = HashTable_xlation_address(*the_hash_entry_ptr);

    #ifdef DEBUG_MODE
      warn("  ... found translation entry with Xaddr = 0x%08x\n", next_xlated_bb_address);
    #endif

    /** If no translation already exists, then create one **/
    #ifdef DEBUG_MODE	
      temp_ptr = next_bb_address;
    #endif

    if (next_xlated_bb_address == NULL) {
      /* If the execution is a system call, then translate it to a canned sys call block **/
      if ( (global_use_safe_svc_code && (next_bb_address <= (unsigned *)0x4000)) ||
	   (OPC_IS_A_SYSTEM_CALL(opc_iword_itype(*next_bb_address))) ) {
	next_xlate_address = do_one_svc_translation(next_bb_address, the_hash_entry_ptr, 
						    TRANSLATE_TAKEN_PATH_CODE, SKIP_MODE_FLAG);
      }
      /* ELSE this is a normal (user code) basic block, so translate its contents */
      else {
	next_xlate_address = do_one_translation(next_bb_address, the_hash_entry_ptr, 
						TRANSLATE_TAKEN_PATH_CODE, SKIP_MODE_FLAG);
      }

      next_xlated_bb_address = HashTable_xlation_address(*the_hash_entry_ptr);
      ubb_size = HashTable_user_bb_size(*the_hash_entry_ptr);
      taken_xbb_size = HashTable_xlated_bb_size(*the_hash_entry_ptr);
      not_taken_xbb_size = HashTable_NT_xlated_bb_size(*the_hash_entry_ptr);

      global_user_line_index   += ubb_size;
      global_xlated_line_index += taken_xbb_size + not_taken_xbb_size;
      global_bb_index++;

      #ifdef DEBUG_MODE	
        next_not_taken_xl_bb_address = HashTable_NT_xlation_address(*the_hash_entry_ptr);
	microtrace_buffer_ptr = HashTable_trace_buffer(*the_hash_entry_ptr);
	not_taken_xbb_size = HashTable_NT_xlated_bb_size(*the_hash_entry_ptr);
        warn("Skip-Mode: translated BB: Oa = 0x%08x Xa = 0x%08x SXa = 0x%08x Sbf = 0x%08x Us = %u Ts = %u NTs = %u\n", 
	     temp_ptr, next_xlated_bb_address, next_not_taken_xl_bb_address, *microtrace_buffer_ptr, 
	     ubb_size, taken_xbb_size, not_taken_xbb_size); 
      #endif
    }

    /****************************************/
    /** Execute the translated basic block **/
    pc_func_desc(prog_desc) = next_xlated_bb_address;

    next_bb_address = (unsigned *)prog_ptr(saved_proc_state_ptr, &not_taken_path);

    #ifdef DEBUG_MODE
      next_not_taken_xl_bb_address = HashTable_xlation_address(*the_hash_entry_ptr);
      microtrace_buffer_ptr = HashTable_trace_buffer(*the_hash_entry_ptr);
      ubb_size = HashTable_user_bb_size(*the_hash_entry_ptr);
      taken_xbb_size = HashTable_xlated_bb_size(*the_hash_entry_ptr);
      not_taken_xbb_size = HashTable_NT_xlated_bb_size(*the_hash_entry_ptr);
      warn("Skip-Mode: Executed BB %lld, the code for Oa = %08x Xa = %08x SXa = 0x%08x Sbf = 0x%08x Us = %u Ts = %u\n", 
		 num_instrs_skipped, temp_ptr, pc_func_desc(prog_desc),
		 next_not_taken_xl_bb_address, *microtrace_buffer_ptr, ubb_size, taken_xbb_size);
    #endif 

    ubb_size = HashTable_user_bb_size(*the_hash_entry_ptr);
    taken_xbb_size = HashTable_xlated_bb_size(*the_hash_entry_ptr);

    num_instrs_skipped += ubb_size;

    global_user_instructions   += ubb_size;
    global_xlated_instructions += taken_xbb_size;
    global_bb_executions++;
    
    if ( (do_safe_exit_flag) && (((unsigned)next_bb_address) == the_exit_routine_address) ) {
     #ifdef FULL_DEBUG
      fprintf(mystderr, "<END> Hit the exit routine address ...\n");
     #endif
      return ARIA_EXIT_FLAG_ADDRESS;
    }
  }

  #ifdef DEBUG
    fflush(mystderr);
  #endif

  /* Restore the global translation control flags */
  do_dynamic_trace_generation        = saved_do_dynamic_trace_generation;
  do_iaddr_fixup                     = saved_do_iaddr_fixup;
  do_not_taken_path_translations     = saved_do_not_taken_path_translations;
  no_FP_registers_used_in_model_flag = saved_no_FP_registers_used_in_model_flag;
  
  /* Indicate how many actual user instructions were skipped */
  *actual_num_instrs_skipped = num_instrs_skipped;

  /* Return the proper next_bb_address to the user */
  if ( (do_iaddr_fixup) && (((unsigned)next_bb_address & 0xf0000000) == global_user_program_iaddr_mask) ) {
    return (unsigned *) ((unsigned)next_bb_address - global_user_iaddr_offset);
  }
  else {
    return  next_bb_address;
  }
}

/*******************************************************************
  PROGRAM NAME..... do_aria_skip_instructions

  FUNCTION ........ Calls the do_aria_skip_instructions_ll
     routine using the old interface with unsigned
     rather than unsigned long long number of instructions.

  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

unsigned * 
do_aria_skip_instructions(unsigned *  bb_address, 
			  proc_state_type *     proc_state_ptr, 
			  unsigned   num_instrs_to_skip,
			  unsigned * actual_num_instrs_skipped
			  )
{
  unsigned * ret_addr;
  unsigned long long ll_instrs_to_skip = num_instrs_to_skip;
  unsigned long long ll_act_num_instrs_skipped = 0;
  unsigned temp_unsigned;

  ret_addr = do_aria_skip_instructions_ll(bb_address, proc_state_ptr, 
					  ll_instrs_to_skip, &ll_act_num_instrs_skipped);

  temp_unsigned = ll_act_num_instrs_skipped;
  *actual_num_instrs_skipped = temp_unsigned;

  return ret_addr;
}


/*******************************************************************
  PROGRAM NAME..... do_aria_full_execution

  FUNCTION ........ Executes the entire user program under Aria,
     returning to the driver model only when the user program 
     exits.
 
  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... May, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/


unsigned * 
do_aria_full_execution(unsigned *  next_bb_address, 
		       proc_state_type * saved_proc_state_ptr,
		       void * user_ptr
		       )
{
  unsigned ubb_size;
  unsigned taken_xbb_size;
  unsigned not_taken_xbb_size;
  unsigned * next_xlated_bb_address;
  unsigned * next_not_taken_xl_bb_address;
  unsigned * temp_ptr;
  unsigned * microtrace_buffer_ptr;
  unsigned * last_bb_printed = 0;

  unsigned * taken_path;
  unsigned * not_taken_path;
  unsigned * temp_xl_addr;

  hash_entry * the_hash_entry_ptr = NULL;
  
  unsigned saved_do_iaddr_fixup = do_iaddr_fixup;
  unsigned * user_stop_iaddr;

  /** If iaddresses are being adjusted, the user could be sending an   **/
  /**  adjusted address, so Aria checks it and corrects the address    **/
  /**  to recover the appropriate address in Aria's view of the loaded **/
  /**  user program.                                                   **/
  if ( (do_iaddr_fixup) && (((unsigned)next_bb_address & 0xf0000000) == 0x10000000) ) {
    next_bb_address = (unsigned *) ((unsigned)next_bb_address + global_user_iaddr_offset);
  }

  do_iaddr_fixup = 0;

  while (next_bb_address != (unsigned *)0x0) {
    unsigned iword = *next_bb_address;
    unsigned iword_index = opc_index(iword);
    unsigned iword_ikind = opc_itype(iword_index);
    unsigned iword_daddr = 0;
    unsigned iword_dlen  = 0;

    #ifdef DEBUG_MODE
      warn("Executing at user address 0x%08x\n", next_bb_address);
    #endif

    /************************************************************************/
    /** Check for user-specified stop-before condition at first instruction */
    /************************************************************************/
    if ( (per_instruction_action_time(iword_index) & DO_ARIA_ACTION_BEFORE) != 0 ){      
      if (per_instruction_action_type(iword_index) == PER_INSTR_FUNCTION_CALL_ACTION) {
	aria_user_function_type user_func = per_instruction_before_function_ptr(iword_index);

	/* Determine the daddr and dlen from the iword and processor state. */
	/*  We cannot use the microtrace since the instruction has yet to run */
        #include "opc_idx_set_daddr_dlen_code.gen"

	if (OPC_IS_A_BRANCH(iword_ikind)) {
	  iword_daddr = (unsigned)next_bb_address+1;
	}
	if ( (*user_func)(next_bb_address, iword, (unsigned *)iword_daddr, iword_dlen, saved_proc_state_ptr, user_ptr) != 0 ) {
	  fprintf(stderr, "Error detected in user function for instruction: (time = BEFORE)\n");
	  fprintf(stderr, "    iaddr = 0x%08x  iword = 0x%08x  index = %u\n", 
		  next_bb_address, iword, iword_index);
	  abort();
	}
      }
    }
    
    /** Check for a translation of this basic block **/
    the_hash_entry_ptr = find_hash_table_entry_for(next_bb_address, 1);

    next_xlated_bb_address = HashTable_xlation_address(*the_hash_entry_ptr);

    /** If no translation already exists, then create one **/
    temp_ptr = next_bb_address;
    
    if (next_xlated_bb_address == NULL) {

      /** Determine the kind of tranlsation to do... **/
      /* If the execution is a system call, then translate it to a canned sys call block **/
      if ( (global_use_safe_svc_code && (next_bb_address <= (unsigned *)0x4000)) ||
	   (OPC_IS_A_SYSTEM_CALL(opc_iword_itype(*next_bb_address))) ) {
	next_xlate_address = do_one_svc_translation(next_bb_address, the_hash_entry_ptr, 
						    TRANSLATE_TAKEN_PATH_CODE, 1);
      }
      /** ELSE this is a normal (user code) basic block, so translate its contents **/
      else {
	next_xlate_address = do_one_translation(next_bb_address, the_hash_entry_ptr, 
						TRANSLATE_TAKEN_PATH_CODE, 1);
      }
      
      next_xlated_bb_address = HashTable_xlation_address(*the_hash_entry_ptr);

      ubb_size            = HashTable_user_bb_size(*the_hash_entry_ptr);
      taken_xbb_size      = HashTable_xlated_bb_size(*the_hash_entry_ptr);
      not_taken_xbb_size  = HashTable_NT_xlated_bb_size(*the_hash_entry_ptr);	  

      if (do_no_output_during_runs == 0) {
	if ( (output_after_dynamic_bb_execution <= global_bb_executions) 
	     && (no_output_after_dynamic_bb_execution  >= global_bb_executions) 
	     && (output_after_static_bb_translation <= global_bb_index) 
	     && (no_output_after_static_bb_translation >= global_bb_index)) {
	  next_not_taken_xl_bb_address = HashTable_NT_xlation_address(*the_hash_entry_ptr);
	  microtrace_buffer_ptr       = HashTable_trace_buffer(*the_hash_entry_ptr);
	  warn_trace("  ...translated BB: Oa = 0x%08x Xa = 0x%08x SXa = 0x%08x Sbf = 0x%08x Us = %u Ts = %u NTs = %u\n", 
		     temp_ptr, next_xlated_bb_address, next_not_taken_xl_bb_address, microtrace_buffer_ptr, 
		     ubb_size, taken_xbb_size, not_taken_xbb_size); 
	}
      }
      
      global_user_line_index   += ubb_size;
      global_xlated_line_index += taken_xbb_size + not_taken_xbb_size;
      global_bb_index++;
    }
    
    /*************************************************/
    /** Get the rest of the basic block information **/
    /*************************************************/
    next_not_taken_xl_bb_address = HashTable_NT_xlation_address(*the_hash_entry_ptr);
    microtrace_buffer_ptr        = HashTable_trace_buffer(*the_hash_entry_ptr);

    ubb_size           = HashTable_user_bb_size(*the_hash_entry_ptr);
    taken_xbb_size     = HashTable_xlated_bb_size(*the_hash_entry_ptr);
    not_taken_xbb_size = HashTable_NT_xlated_bb_size(*the_hash_entry_ptr);

    user_stop_iaddr    = HashTable_user_stop_iaddr(*the_hash_entry_ptr);
    
    /****************************************/
    /** Execute the translated basic block **/
    /****************************************/
    #ifdef DEBUG_MODE
     warn("NEXT_XLATED_BB_ADDRESS = 0x%08x\n", next_xlated_bb_address);
    #endif
    pc_func_desc(prog_desc) = next_xlated_bb_address;
    
    /** The copy to run when debugging output desired **/
    if (do_no_output_during_runs == 0) {
      if ( (output_after_dynamic_bb_execution <= global_bb_executions) 
	   && (no_output_after_dynamic_bb_execution  >= global_bb_executions) 
	   && (output_after_static_bb_translation <= global_bb_index) 
	   && (no_output_after_static_bb_translation >= global_bb_index) ) {
	warn_exec("\nExecuting the taken-path translated code:\n");
	warn_exec("     Starting address %08x:\n",   pc_func_desc(prog_desc));
	warn_exec("     TOC address      %08x:\n",   toc_func_desc(prog_desc));
	warn_exec("     Other data       %08x:\n", other_func_desc(prog_desc));
	warn_processor_state(saved_proc_state_ptr, show_proc_float_state);
      }

      taken_path = (unsigned *)prog_ptr(saved_proc_state_ptr, &not_taken_path);
      next_bb_address = taken_path;
      
      if ( (output_after_dynamic_bb_execution <= global_bb_executions) && 
	   (no_output_after_dynamic_bb_execution  >= global_bb_executions) && 
	   (output_after_static_bb_translation <= global_bb_index) &&
	   (no_output_after_static_bb_translation >= global_bb_index) ) {
	warn_exec("  ... Done. Next address = %08x (not_taken = %08x)\n", next_bb_address, not_taken_path);
	warn_trace("  ...Executed.. BB %d, the code for Oa = %08x Xa = %08x Next = %08x\n", 
		   global_bb_executions, temp_ptr, pc_func_desc(prog_desc), next_bb_address);
	warn_processor_state(saved_proc_state_ptr, show_proc_float_state);
      }
    }
    /** The faster copy to run when no debugging output desired **/
    else {
      taken_path = (unsigned *)prog_ptr(saved_proc_state_ptr, &not_taken_path);
      next_bb_address = taken_path;
    }
  
    global_bb_executions++;
    global_user_instructions   += ubb_size;
    global_xlated_instructions += taken_xbb_size;

    /*******************************************************************************/
    /** Check for user-specified stop-after conditions in the block just executed **/
    /*******************************************************************************/
    if (user_stop_iaddr != (unsigned *)0x0) {
      iword       = *user_stop_iaddr;
      iword_index = opc_index(iword);
      if ((per_instruction_action_time(iword_index) & DO_ARIA_ACTION_AFTER) != 0) {
	if (per_instruction_action_type(iword_index) == PER_INSTR_FUNCTION_CALL_ACTION) {
	  aria_user_function_type user_func = per_instruction_after_function_ptr(iword_index);

	  /* Get the daddr and dlen from the trace buffer. */
	  /*  The stop-after MUST be the last instruction... */
	  iword_daddr = aria_Trace_Daddr(microtrace_buffer_ptr, ubb_size-1);
	  iword_dlen  = aria_Trace_Dlen(microtrace_buffer_ptr, ubb_size-1);

	  if ( (*user_func)(user_stop_iaddr, iword, (unsigned *)iword_daddr, iword_dlen, saved_proc_state_ptr, (void *)user_ptr) != 0 ) {
	    fprintf(stderr, "Error detected in user function for instruction (time = AFTER):\n");
	    fprintf(stderr, "    iaddr = 0x%08x  iword = 0x%08x  index = %u\n", 
		    user_stop_iaddr, *user_stop_iaddr, iword_index);
	    abort();
	  }
	}
      }
    }
  }

  /******************************************************/
  /** Done executing the program (found the exit call) **/
  /******************************************************/
  
  /** Output the run flag check **/
  if (global_bb_executions > show_run_update_after > 0) {
    if ((global_bb_executions % show_run_update_every) == 0) {
      int i = 0;
      unsigned curr_orig_address = (unsigned)temp_ptr;
      unsigned next_orig_address = (unsigned)next_bb_address;
      if ( (do_iaddr_fixup) && (((unsigned)next_bb_address & 0xf0000000) == global_user_program_iaddr_mask) ) {
	curr_orig_address -= global_user_iaddr_offset;
	next_orig_address -= global_user_iaddr_offset;
      }
      
      warn("  ...Executed BB %d: Oa= 0x%08x Xa= 0x%08x Nxt= 0x%08x\n", 
	   global_bb_executions, curr_orig_address, pc_func_desc(prog_desc), next_orig_address);
      
      if (last_bb_printed != temp_ptr) {
	for (i = 0; i < ubb_size; i++) {
	  unsigned * trbuf = microtrace_buffer_ptr;
	  fprintf(mystderr, "\t0x%08x  0x%08x  %u  0x%08x  %u\n", 
		  trbuf[ARIA_TB_PER_INSTR_SIZE*i+0], trbuf[ARIA_TB_PER_INSTR_SIZE*i+1], 
		  trbuf[ARIA_TB_PER_INSTR_SIZE*i+2], trbuf[ARIA_TB_PER_INSTR_SIZE*i+3], 
		  trbuf[ARIA_TB_PER_INSTR_SIZE*i+4]);
	}
      }
      last_bb_printed = temp_ptr;
    }
  }
  
  /** Dump final statistics check **/
  if ( (next_bb_address == 0x00000000) && (show_final_aria_statistics == 1) ) {
    unsigned save_print = do_proc_state_dump;
    unsigned long long save_dstart = output_after_dynamic_bb_execution;
    unsigned long long save_dend   = no_output_after_dynamic_bb_execution;
    unsigned long long save_sstart = output_after_static_bb_translation;
    unsigned long long save_send   = no_output_after_static_bb_translation;
    do_proc_state_dump = 1;
    output_after_dynamic_bb_execution = 0;
    no_output_after_dynamic_bb_execution = INT_MAX;
    output_after_static_bb_translation = 0;
    no_output_after_static_bb_translation = INT_MAX;
    {
      unsigned curr_orig_address = (unsigned)temp_ptr;
      unsigned next_orig_address = (unsigned)next_bb_address;
      if ( (do_iaddr_fixup) && (((unsigned)next_bb_address & 0xf0000000) == global_user_program_iaddr_mask) ) {
	curr_orig_address -= global_user_iaddr_offset;
	next_orig_address -= global_user_iaddr_offset;
      }
      warn("Exit from program detected...\n");
      warn("  ** Executed BB %d: Oa= 0x%08x (0x%08x) Xa= 0x%08x Next = 0x%08x (0x%08x)\n", 
	   global_bb_executions, temp_ptr, curr_orig_address, pc_func_desc(prog_desc), 
	   next_bb_address, next_orig_address);
      warn_processor_state(saved_proc_state_ptr, show_proc_float_state);
      
      warn("\n===========================================================================\n");
      warn("And the final statistics (within Aria) are:\n\n");
      dump_aria_run_statistics(saved_proc_state_ptr);
      warn("Returning to timer model...\n");
    }
    
    do_proc_state_dump = save_print;
    output_after_dynamic_bb_execution = save_dstart;
    no_output_after_dynamic_bb_execution = save_dend;
    output_after_static_bb_translation = save_sstart;
    no_output_after_static_bb_translation = save_send;
  }

  fflush(mystderr);
  
  do_iaddr_fixup = saved_do_iaddr_fixup;

  if ( (do_iaddr_fixup) && (((unsigned)next_bb_address & 0xf0000000) == global_user_program_iaddr_mask) ) {
    return (unsigned *) ((unsigned)next_bb_address - global_user_iaddr_offset);
  }
  else {
    return  next_bb_address;
  }
}

