/*                                                                     
** COMPONENT_NAME: MET / Aria
**                                                                      
** Licensed Materials - Property of IBM                                  
**                                                                     
** IBM Confidential (IBM Confidential-Restricted when combined          
**                   with aggregated modules for this product)          
**                                                                     
**                                                                     
** OCO Source Materials                                                
**                                                                     
**                                                                     
** GA-0100 (C) Copyright IBM Corp. 1997                                 
**                                                                     
** US Government Users Restricted Rights - Use, duplication or         
** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.   
**                                                                     
** The source code for this program is not published or otherwise      
** divested of its trade secrets, irrespective of what has been        
** deposited within the U.S. Copyright Office.                         
**                                                                     
*/

 
/*******************************************************************
  PROGRAM NAME..... do_svc_translation
  FUNCTION ........ Creates a translation wrapper for a user-program
     supervisor call.  This wrapper has a predefined form for both 
     the taken-path and not-taken path codes.  The taken-path code
     will check whether this is a call to exit from the program, and
     otherwise set up the user state, and invoke the underlying 
     supervisor function.  When control returns to this translated
     code, the Aria and user states are swapped and control is 
     handed back to Aria.  To be safe, ALL user process state is 
     restored prior to the supervisor call, and ALL state is 
     saved after the supervisor call returns.  

  NOTES............
  ASSUMPTIONS... The area at TEMP_SPOT off the user stack (in this 
     case -216 bytes down from the current user stack pointer) is 
     NOT in use by the user program; it is overwritten during the 
     swapping of the user and aria states.
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... the next available translation address
  CALLED-ROUTINES.. 
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... <see below> plus 
     xlation_alignment_code.c, 
     save_aria_state_code.c, bb_prolog_code.c, bb_epilog_code.c, 
     restore_aria_state_code.c

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/
#include <stdio.h>
#include <sys/time.h>

#include "aria_xlate_macros.h"

#include "opc.h"

#include "proc_state.h"
#include "aria_state.h"
#include "hashing.h"

#include "io_msgs.h"

#include "dump_description.h"

#include "globtypes.h"
#include "global_variables.h"

#include "do_svc_translation.h"

#include "display_reg_usedef.h"

#define TEMP_SPOT  -216

/**************************************************************/
/** THE FOLLOWING ARE THE 32-BIT MACROS FOR SVC TRANSLATIONS **/
/**************************************************************/
#ifndef DO_64_BIT_COMPILE

#define INSERT_DYNAMIC_TRACE_CODE(x)\
      unsigned static_hi = ((unsigned)microtrace_buffer_address & 0xffff0000) >> 16; \
      unsigned static_lo = ((unsigned)microtrace_buffer_address & 0x0000ffff);       \
      *next_xlate_address = OPC_ASM_ADDIS((x), 0, static_hi);  next_xlate_address++;  \
      *next_xlate_address = OPC_ASM_ORI  ((x), (x), static_lo);  next_xlate_address++;  \
      *next_xlate_address = STORE_REG ((x), PROC_DYN_TRBUF_OFFSET, base_register);  next_xlate_address++;

/**************************************************************/
/** THE FOLLOWING ARE THE 64-BIT MACROS FOR SVC TRANSLATIONS **/
/**************************************************************/
#else

#define INSERT_DYNAMIC_TRACE_CODE(x)\
      unsigned addr_0_15  = ((BASE_TYPE)microtrace_buffer_address >> 48) & 0xffffLL; \
      unsigned addr_16_31 = ((BASE_TYPE)microtrace_buffer_address >> 32) & 0xffffLL; \
      unsigned addr_32_47 = ((BASE_TYPE)microtrace_buffer_address >> 16) & 0xffffLL; \
      unsigned addr_48_63 = ((BASE_TYPE)microtrace_buffer_address      ) & 0xffffLL; \
      *next_xlate_address = OPC_ASM_ADDI  ((x), 0, addr_0_15);                           next_xlate_address++; \
      *next_xlate_address = OPC_ASM_RLDICL((x), (x), 16, 0);                               next_xlate_address++; \
      *next_xlate_address = OPC_ASM_ORI   ((x), (x), addr_16_31);                         next_xlate_address++; \
      *next_xlate_address = OPC_ASM_RLDICL((x), (x), 16, 0);                               next_xlate_address++; \
      *next_xlate_address = OPC_ASM_ORI   ((x), (x), addr_32_47);                         next_xlate_address++; \
      *next_xlate_address = OPC_ASM_RLDICL((x), (x), 16, 0);                               next_xlate_address++; \
      *next_xlate_address = OPC_ASM_ORI   ((x), (x), addr_48_63);                         next_xlate_address++; \
      *next_xlate_address = OPC_ASM_STD   ((x), PROC_DYN_TRBUF_OFFSET, base_register);   next_xlate_address++;

#endif


/***************************************************************/
/** THE SUBROUTINE TO INSERT THE SVC SYSTEM CALL TRANSLATIONS **/
/***************************************************************/

unsigned * 
do_one_svc_translation(unsigned   * next_prog_address,
		       hash_entry * the_hash_entry_ptr,
		       unsigned     do_speculative_path_xlation,
		       unsigned     generation
		       )
{
  unsigned * microtrace_buffer_address = NULL;

  unsigned * xlate_ptr;
  unsigned total_instr_in_xl_bb = 0;
  unsigned total_instr_in_bb = 1;

  unsigned * original_bb_start_address   = next_prog_address;
  unsigned * translated_bb_start_address;

  unsigned timebase_struct_address;

  unsigned base_register;
  unsigned scratch_reg;
  unsigned scratch_reg_2;

  /** Check whether we already have a translation for the   **/
  /**  other "flavor" of this BB (e.g. whether we have      **/
  /**  already done a taken-path when we are about to do a  **/
  /**  not-taken path translation).                         **/
  unsigned first_xlation = (do_speculative_path_xlation) ?
    (HashTable_xlation_address(*the_hash_entry_ptr) == NULL) : 
    (HashTable_NT_xlation_address(*the_hash_entry_ptr) == NULL);

  /***********************************************************/
  /** Now set up the static space for the user trace buffer **/
  /**  if necessary (i.e. if this is the first translation) **/
  /***********************************************************/  
  /**
    warn("turn_svc_into_a_bclr_instruction = %u\n", turn_svc_into_a_bclr_instruction);
    **/

/**
warn("global_user_iaddr_offset = %08x\n", global_user_iaddr_offset);
**/

  if (first_xlation) {
    unsigned br_addr = (unsigned) original_bb_start_address;
    unsigned iword   = (turn_svc_into_a_bclr_instruction) ? OPC_ASM_BCLR(0x14, 0) : *original_bb_start_address;
    unsigned iword_index = opc_index(iword);
    unsigned iword_kind  = opc_itype(iword_index);
    unsigned iword_daddr    = 0xdeadbeef;
    unsigned iword_dlength  = 0;

    *next_xlate_address  = 0x00000000;        next_xlate_address++;
    
    microtrace_buffer_address = next_xlate_address;    
    
    *next_xlate_address  = br_addr;           next_xlate_address++;
    *next_xlate_address  = iword;             next_xlate_address++;
    *next_xlate_address  = iword_kind;        next_xlate_address++;
    *next_xlate_address  = iword_daddr;       next_xlate_address++;
    *next_xlate_address  = iword_dlength;     next_xlate_address++;
    *next_xlate_address  = 0x00000000;        next_xlate_address++;

    /** Update the addresses for the program and translation space **/
    #include "xlation_alignment_code.c"

    /***************************************************************/
    /* adjust the hash table entry for the newly translated block **/
    /***************************************************************/
    HashTable_original_address(*the_hash_entry_ptr)   = original_bb_start_address;
    HashTable_trace_buffer(*the_hash_entry_ptr)       = microtrace_buffer_address;
    HashTable_user_bb_size(*the_hash_entry_ptr)       = 1;
    HashTable_generation(*the_hash_entry_ptr)         = generation;
    HashTable_overflow(*the_hash_entry_ptr)           = NULL;
  }
  else {
    /** Get the values we need from the previous translation **/
    original_bb_start_address = HashTable_original_address(*the_hash_entry_ptr);
    microtrace_buffer_address = HashTable_trace_buffer(*the_hash_entry_ptr);
  }
    
  /** Enter a header for this BB translation **/
  #include "add_xlation_header_code.c"
  
  /* Create a space for the timebasestruct information on calls to read_real_time **/
  if (global_get_system_call_execution_times == 1) {
    unsigned * ta_ptr = next_xlate_address;

    /* define the time base structure space (with appropriate markers) */
    *next_xlate_address = 0x00ffff00; next_xlate_address++;
    timebase_struct_address = (unsigned) next_xlate_address;
    next_xlate_address += (TIMEBASE_SZ>>2);
    *next_xlate_address = 0x00ffff00; next_xlate_address++;

    if (do_no_output_during_runs == 0) {
      int iii;
      warn_xlate("\nThe space for the timebase structure is:\n");
      warn_xlate("\t0x%08x  %08x\n", ta_ptr, *ta_ptr);  ta_ptr++;
      for (iii = 0; iii < (TIMEBASE_SZ>>2); iii++) {
	warn_xlate("\t0x%08x  %08x\n", ta_ptr, *ta_ptr);  ta_ptr++;
      }
      warn_xlate("\t0x%08x  %08x\n", ta_ptr, *ta_ptr);  ta_ptr++;      
      warn_xlate("\n");
    }
  }

  /*************************************/
  /** The taken-path translation code **/
  /*************************************/
  if (do_speculative_path_xlation == 0) {
    translated_bb_start_address = next_xlate_address;

    /********************************************************/
    /** Save off all of Aria's important (clobbered) state **/
    /********************************************************/

    base_register = 12;
    scratch_reg   = 11;
    scratch_reg_2 = 10;

    /* Load GPR base_register with the address of our ARIA state saving space */          
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16); next_xlate_address++; 

    /* Save off the input parameters */
    *next_xlate_address = STORE_REG ( 3, PROC_STATE_OFFSET,  base_register);     next_xlate_address++;

    if (not_taken_address_required_flag == 1) { 
      *next_xlate_address = STORE_REG ( 4, NOT_TAKEN_RVAL_OFFSET, base_register);  next_xlate_address++;
    }

    /** Create a pointer to the trace buffer **/
    if ( (do_dynamic_trace_generation) ) {
      INSERT_DYNAMIC_TRACE_CODE(4);
    }

    /*************************************/
    /** Do the saving of the Aria state **/
    /*************************************/
    {
      unsigned gprs = 0xffffffff;
      unsigned fprs = 0xffffffff;
      unsigned crs  = 0xffffffff;

      #include "save_aria_state_code.c"
    }

    if (do_no_output_during_runs == 0) {
      warn_xlate(" (aria prolog) ------------------------------------------------------------\n");
      DUMP_DESCRIPTION(translated_bb_start_address, next_xlate_address);    
      xlate_ptr = next_xlate_address;
    }

    /** If the user wants the times for the execution, get them... **/
    if (global_get_system_call_execution_times == 1) {
      unsigned address;
      unsigned addr_hi;
      unsigned addr_lo;
      unsigned * Aaddr = (unsigned *) (ARIA_STATE_SAVE_SPACE + ARIA_READ_REAL_TIME);
      
      /* Set the CTR to branch through */
      address = *(Aaddr);
      addr_hi = (address & 0xffff0000) >> 16;
      addr_lo = (address & 0x0000ffff);
      *next_xlate_address = OPC_ASM_ADDIS(3, 0, addr_hi);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_ORI  (3, 3, addr_lo);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_MTSPR(9, 3);           next_xlate_address++;

      /* Set the TOC value */
      address = *(Aaddr+1);
      addr_hi = (address & 0xffff0000) >> 16;
      addr_lo = (address & 0x0000ffff);
      *next_xlate_address = OPC_ASM_ADDIS(2, 0, addr_hi);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_ORI  (2, 2, addr_lo);  next_xlate_address++;

      /* Set the first parameter */
      address = timebase_struct_address;
      addr_hi = (address & 0xffff0000) >> 16;
      addr_lo = (address & 0x0000ffff);
      *next_xlate_address = OPC_ASM_ADDIS(3, 0, addr_hi);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_ORI  (3, 3, addr_lo);  next_xlate_address++;
      
      /* Set the second parameter */
      *next_xlate_address = OPC_ASM_ADDI(4, 0, TIMEBASE_SZ);  next_xlate_address++;

      /* branch and link */
      *next_xlate_address = OPC_ASM_BCCTRL(20, 0);  next_xlate_address++;
      
      /* Save the returned data in the aria state */
      *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16); next_xlate_address++; 
      *next_xlate_address = LOAD_REG(5, 0, 4);  next_xlate_address++;
      *next_xlate_address = STORE_REG(5, ARIA_SAVE_START_TIME, base_register);  next_xlate_address++;
      *next_xlate_address = LOAD_REG(5, 4, 4);  next_xlate_address++;
      *next_xlate_address = STORE_REG(5, ARIA_SAVE_START_TIME+4, base_register);  next_xlate_address++;
      *next_xlate_address = LOAD_REG(5, 8, 4);  next_xlate_address++;
      *next_xlate_address = STORE_REG(5, ARIA_SAVE_START_TIME+8, base_register);  next_xlate_address++;

      if (do_no_output_during_runs == 0) {
	warn_xlate(" (time base prolog) -------------------------------------------------------\n");
	DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
	warn_xlate("read_real_time = 0x%08x TOC = %08x tbasestruct = 0x%08x\n",
		   *(Aaddr), *(Aaddr+1), timebase_struct_address);
	xlate_ptr = next_xlate_address;
      }
    }

    /* Load base_register with the processor state save space address from the reserved work area */
    *next_xlate_address = LOAD_REG ( base_register,  0, base_register);              next_xlate_address++;  
  
    /****************************************************************************/
    /** Insert the bb prolog, which inserts instructions to restore user state **/
    /****************************************************************************/
    {
      unsigned gpr_uses   = 0xffffffff;
      unsigned fpr_uses   = 0xffffffff;
      unsigned cr_touches = 0xffffffff;
      unsigned spr_uses   = 0xffffffff;
      unsigned spr_defs   = 0xffffffff;
      
      #include "bb_prolog_code.c"
    }
    
    if (do_no_output_during_runs == 0) {
      warn_xlate("* (prolog) ----------------------------------------------------------------\n");
      DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);    
      xlate_ptr = next_xlate_address;
      warn_xlate("* (svc prep translated body) ==============================================\n");
    }
    
    /*********************************************************/
    /** Now generate some code for the SVC call basic block **/
    /*********************************************************/

    /*******************************************************************/
    /** Check the value in GPR 2 (the system call routine identifier) **/
    /**  to see if the program is trying to exit; the exit routine    **/
    /**  flags differs by operating system version, so check for any. **/
    /*******************************************************************/
    /** Save GPR 12 on the stack (so we can use it for some case checking) **/
    *next_xlate_address = STORE_REG  ( 12,  TEMP_SPOT, STACK_PTR);     next_xlate_address++;
    *next_xlate_address = OPC_ASM_MFCR ( 12);                            next_xlate_address++;
    *next_xlate_address = OPC_ASM_CMPI (  7,0,  2, the_exit_code_value); next_xlate_address++;


    /********************************************************/
    /** If the GPR2 code is not EXIT, go to the BLA code   **/
    /**     (i.e. skip the exit code below)                **/
    /** The branch is inserted at the end of the exit code **/
    /**  (below) in space saved at branch_over_address     **/
    /********************************************************/
    {
      unsigned * branch_over_address = next_xlate_address;
      unsigned instrs_in_exit_code = 0;

      /** Skip where the conditional branch will be placed **/
      next_xlate_address++;

      /** Insert the bb epilog, which inserts instructions to save new user state **/
      {
	unsigned   gpr_defs   = 0xffffffff;
	unsigned   fpr_defs   = 0xffffffff;
	unsigned   cr_touches = 0xffffffff;
	unsigned   spr_uses   = 0xffffffff;
	unsigned   spr_defs   = 0xffffffff;
      
        #include "bb_epilog_code.c"
      }

      /** Load the scratch register with the link register value **/
      *next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, base_register);    next_xlate_address++;
      if ( (do_dynamic_trace_generation) && (do_iaddr_fixup) ) {
	*next_xlate_address = OPC_ASM_RLWINM(scratch_reg, scratch_reg, 4, 28, 31);       next_xlate_address++;
	*next_xlate_address = OPC_ASM_CMPI ( 7, 0, scratch_reg, ((global_user_program_iaddr_mask>>28)& 0xf));                    next_xlate_address++;
	*next_xlate_address = OPC_ASM_BC   ( 0x4, 30, 5);                                next_xlate_address++;
	*next_xlate_address = LOAD_REG  (scratch_reg, LR_SAVE_OFFSET, base_register); next_xlate_address++;
	*next_xlate_address = LOAD_REG  (scratch_reg_2, IADDR_OFFSET, base_register); next_xlate_address++;
	*next_xlate_address = OPC_ASM_SUBFC(scratch_reg, scratch_reg_2, scratch_reg);    next_xlate_address++;
	*next_xlate_address = OPC_ASM_B    ( 2);                                         next_xlate_address++;
	*next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, base_register);  next_xlate_address++;      
      }

      /********************************/
      /** Now restore the Aria state **/

      /* Load GPRbase with the address of our ARIA state saving space    */               
      *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);               next_xlate_address++; 

      /* Store the dynamic trace information (if user wants it) */
      if (do_dynamic_trace_generation) {
	*next_xlate_address = LOAD_REG  (base_register, PROC_DYN_TRBUF_OFFSET, base_register);  next_xlate_address++;
	*next_xlate_address = STORE_REG  (scratch_reg, ARIA_TB_BRTARGET_OFFSET, base_register);  next_xlate_address++;
	*next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);             next_xlate_address++; 
      }

      /******************************************/
      /** Do the restoration of the Aria state **/
      /******************************************/
      {
	unsigned gprs = 0xffffffff;
	unsigned fprs = 0xffffffff;
	unsigned crs  = 0xffffffff;
      
        #include "restore_aria_state_code.c"
      }

      /**********************************************************************************************/
      /* Set the next_address return value to be address 0x00000000 (indicates user program exits) **/ 
      *next_xlate_address = OPC_ASM_ADDIS( 3, 0, 0);           next_xlate_address++; 

      if (not_taken_address_required_flag == 1) { 
	/* And the not-taken next address */
	*next_xlate_address = LOAD_REG  (10, NOT_TAKEN_RVAL_OFFSET, 12);  next_xlate_address++; 
	*next_xlate_address = STORE_REG  ( 3, 0, 10);                      next_xlate_address++; 
      }
    
      /* Replace the branch in the original code with code to return through link register branch */
      *next_xlate_address = OPC_ASM_BCLR (0x14, 0);        next_xlate_address++;

      /** Determine the number of instructions to branch over this exit code **/
      instrs_in_exit_code = (next_xlate_address - branch_over_address);
      *branch_over_address = OPC_ASM_BC ( 0x05, 7*4+2, instrs_in_exit_code);

      if (do_no_output_during_runs == 0) {
	warn_xlate("* (exit branch code) ------------------------------------------------------\n");
	DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
	xlate_ptr = next_xlate_address;
      }
    }

    /***********************************************************/
    /** ELSE, not an EXIT call, so do the underlying SVC call **/
    /***********************************************************/

    *next_xlate_address = OPC_ASM_MTCRF(0xff, 12);                             next_xlate_address++; 
    *next_xlate_address = LOAD_REG  ( 12, TEMP_SPOT, STACK_PTR);            next_xlate_address++;

    /* Insert a BLA to the slot before the sc address            */ 
    /*  (POWER machines require a NOP immediately before the sc) */
    {
      unsigned br_addr = (unsigned) original_bb_start_address;
      *next_xlate_address = OPC_ASM_BLA( ((br_addr-4)>>2) );                   next_xlate_address++;
    }

    if (do_no_output_during_runs == 0) {
      warn_xlate("* (sc body) ---------------------------------------------------------------\n");
      DUMP_FAKE_DESCRIPTION(xlate_ptr, next_xlate_address);
      xlate_ptr = next_xlate_address;
    }

    *next_xlate_address = STORE_REG  ( 12,  TEMP_SPOT, STACK_PTR);        next_xlate_address++;  
    *next_xlate_address = OPC_ASM_ADDIS( 12,  0, ARIA_STATE_SAVE_HI_16);    next_xlate_address++;  
    *next_xlate_address = LOAD_REG  ( 12,  0, 12);                       next_xlate_address++;  
    *next_xlate_address = LOAD_REG  ( 12, LR_SAVE_OFFSET, 12);           next_xlate_address++;  
    *next_xlate_address = OPC_ASM_MTSPR(  8, 12);                           next_xlate_address++; 
    *next_xlate_address = LOAD_REG  ( 12,  TEMP_SPOT, STACK_PTR);        next_xlate_address++;  
  
    /*****************************************************************************/
    /** Insert the bb epilog, which inserts instructions to save new user state **/
    /*****************************************************************************/
    
    {
      unsigned   gpr_defs   = 0xffffffff;
      unsigned   fpr_defs   = 0xffffffff;
      unsigned   cr_touches = 0xffffffff;
      unsigned   spr_uses   = 0xffffffff;
      unsigned   spr_defs   = 0xffffffff;
    
      #include "bb_epilog_code.c"
    }

    if (do_no_output_during_runs == 0) {
      warn_xlate("* (BB epilog) -------------------------------------------------------------\n");
      DUMP_FAKE_DESCRIPTION(xlate_ptr, next_xlate_address);
      xlate_ptr = next_xlate_address;
    }

    /** If the user wants the times for the execution, get them... **/
    if (global_get_system_call_execution_times == 1) {
      unsigned address;
      unsigned addr_hi;
      unsigned addr_lo;
      unsigned * Aaddr = (unsigned *) (ARIA_STATE_SAVE_SPACE + ARIA_READ_REAL_TIME);
      
      /* Set the CTR to branch through */
      address = *(Aaddr);
      addr_hi = (address & 0xffff0000) >> 16;
      addr_lo = (address & 0x0000ffff);
      *next_xlate_address = OPC_ASM_ADDIS(3, 0, addr_hi);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_ORI  (3, 3, addr_lo);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_MTSPR(9, 3);           next_xlate_address++;

      /* Set the TOC value */
      address = *(Aaddr+1);
      addr_hi = (address & 0xffff0000) >> 16;
      addr_lo = (address & 0x0000ffff);
      *next_xlate_address = OPC_ASM_ADDIS(2, 0, addr_hi);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_ORI  (2, 2, addr_lo);  next_xlate_address++;

      /* Set the first parameter */
      address = timebase_struct_address;
      addr_hi = (address & 0xffff0000) >> 16;
      addr_lo = (address & 0x0000ffff);
      *next_xlate_address = OPC_ASM_ADDIS(3, 0, addr_hi);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_ORI  (3, 3, addr_lo);  next_xlate_address++;
      
      /* Set the second parameter */
      *next_xlate_address = OPC_ASM_ADDI(4, 0, TIMEBASE_SZ);  next_xlate_address++;

      /* branch and link */
      *next_xlate_address = OPC_ASM_BCCTRL(20, 0);  next_xlate_address++;
      
      /* Save the returned data in the aria state */
      *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16); next_xlate_address++; 
      *next_xlate_address = LOAD_REG(5, 0, 4);  next_xlate_address++;
      *next_xlate_address = STORE_REG(5, ARIA_SAVE_FINISH_TIME, base_register);  next_xlate_address++;
      *next_xlate_address = LOAD_REG(5, 4, 4);  next_xlate_address++;
      *next_xlate_address = STORE_REG(5, ARIA_SAVE_FINISH_TIME+4, base_register);  next_xlate_address++;
      *next_xlate_address = LOAD_REG(5, 8, 4);  next_xlate_address++;
      *next_xlate_address = STORE_REG(5, ARIA_SAVE_FINISH_TIME+8, base_register);  next_xlate_address++;

      /* Get back saved proc state addressibility */
      *next_xlate_address = LOAD_REG(base_register, PROC_STATE_OFFSET, base_register); next_xlate_address++; 

      if (do_no_output_during_runs == 0) {
	warn_xlate(" (time base prolog) -------------------------------------------------------\n");
	DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
	warn_xlate("read_real_time = 0x%08x TOC = %08x tbasestruct = 0x%08x\n",
		   *(Aaddr), *(Aaddr+1), timebase_struct_address);
	xlate_ptr = next_xlate_address;
      }
    }

    /** Load the scratch register with the link register value **/
    *next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, base_register);     next_xlate_address++;

    /** Output the dynamic trace information (if appropriate) **/
    if ( (do_dynamic_trace_generation) && (do_iaddr_fixup) ) {
      *next_xlate_address = OPC_ASM_RLWINM(scratch_reg, scratch_reg, 4, 28, 31);        next_xlate_address++;
      *next_xlate_address = OPC_ASM_CMPI ( 7, 0, scratch_reg, ((global_user_program_iaddr_mask >> 28) & 0xf));                     next_xlate_address++;
      *next_xlate_address = OPC_ASM_BC   ( 0x4, 30, 5);                                 next_xlate_address++;
      *next_xlate_address = LOAD_REG  (scratch_reg, LR_SAVE_OFFSET, base_register);  next_xlate_address++;
      *next_xlate_address = LOAD_REG  (scratch_reg_2, IADDR_OFFSET, base_register);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_SUBFC(scratch_reg, scratch_reg_2, scratch_reg);     next_xlate_address++;      
      *next_xlate_address = OPC_ASM_B    ( 2);                                          next_xlate_address++;
      *next_xlate_address = LOAD_REG  (scratch_reg, LR_SAVE_OFFSET, base_register);  next_xlate_address++;
    }

    /* Load GPRbase with the address of our ARIA state saving space    */               
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);      next_xlate_address++; 

    /* Store the dynamic trace information (if appropriate) */
    if (do_dynamic_trace_generation) {
      *next_xlate_address = LOAD_REG (base_register, PROC_DYN_TRBUF_OFFSET, base_register);  next_xlate_address++;
      *next_xlate_address = STORE_REG (scratch_reg, ARIA_TB_BRTARGET_OFFSET, base_register);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);            next_xlate_address++; 
    }
    
    /** Store the link register value as the taken and not-taken addresses **/
    *next_xlate_address = LOAD_REG  (base_register, PROC_STATE_OFFSET, base_register);  next_xlate_address++; 
    *next_xlate_address = LOAD_REG  (scratch_reg, LR_SAVE_OFFSET, base_register);       next_xlate_address++;
    *next_xlate_address = STORE_REG  (scratch_reg, NEXT_TARGET_OFFSET, base_register);   next_xlate_address++;
    if (not_taken_address_required_flag == 1) {   
      *next_xlate_address = STORE_REG  (scratch_reg, NOT_TAKEN_OFFSET, base_register);   next_xlate_address++;
    }

    if (do_no_output_during_runs == 0) {
      warn_xlate("* (exit) ------------------------------------------------------------------\n");
      DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
      xlate_ptr = next_xlate_address;
    }

    /********************************/
    /** Now restore the Aria state **/
    /********************************/
    
    /* Load GPRbase_register with the address of our ARIA state saving space    */               
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);         next_xlate_address++; 

    /******************************************/
    /** Do the restoration of the Aria state **/
    /******************************************/
    {
      unsigned gprs = 0xffffffff;
      unsigned fprs = 0xffffffff;
      unsigned crs  = 0xffffffff;
    
      #include "restore_aria_state_code.c"
    }

    /**********************************************************/
    /* Determine the next basic block address and the         */
    /*   next not taken basic block address.                  */
    /**********************************************************/
    base_register = 12;
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);               next_xlate_address++;

    if (not_taken_address_required_flag == 1) { 
      *next_xlate_address = LOAD_REG  (10, NOT_TAKEN_RVAL_OFFSET, base_register);             next_xlate_address++; 
      *next_xlate_address = LOAD_REG  (base_register,  PROC_STATE_OFFSET, base_register);     next_xlate_address++;
      *next_xlate_address = LOAD_REG  (11, NOT_TAKEN_OFFSET, base_register);                  next_xlate_address++;
      *next_xlate_address = STORE_REG  (11, 0, 10);                                            next_xlate_address++; 
    }
    else {
      *next_xlate_address = LOAD_REG  (base_register,  PROC_STATE_OFFSET, base_register);     next_xlate_address++;
    }

    *next_xlate_address = LOAD_REG  ( 3, NEXT_TARGET_OFFSET, base_register);                  next_xlate_address++;

    /********************************************************************************************/
    /* Replace the branch in the original code with code to return through link register branch */
    /********************************************************************************************/
    *next_xlate_address = OPC_ASM_BCLR (0x14, 0);        next_xlate_address++;

    /* Output the translated basic block exit code */
    if (do_no_output_during_runs == 0) {
      DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
      xlate_ptr = next_xlate_address;
    }

    if (do_no_output_during_runs == 0) {
      warn_xlate("* (Aria restore) ----------------------------------------------------------\n");
      DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
      warn_xlate("===============================================================================\n");
      warn_xlate("    end of BB %d [%u user, %u xlated instr] ; oa=%08x xa = %08x\n", 
		 global_bb_index, 0,
		 (next_xlate_address - translated_bb_start_address),
		 original_bb_start_address, translated_bb_start_address);
    }

    /** Determine the number of instructions in the xlated basic block **/
    total_instr_in_xl_bb = (next_xlate_address - translated_bb_start_address);

    /* Update the addresses for the program and translation space **/
    #include "xlation_alignment_code.c"

    /** Adjust the hash table entry to reflect the taken-path translation **/
    HashTable_xlation_address(*the_hash_entry_ptr) = translated_bb_start_address;
    HashTable_xlated_bb_size(*the_hash_entry_ptr)  = total_instr_in_xl_bb;

    /** Dump the translated code out of the data cache **/
    _sync_cache_range(translated_bb_start_address, 4*(next_xlate_address - translated_bb_start_address));

    /***************************************/
    /** END OF THE TAKEN-PATH TRANSLATION **/
    /***************************************/
  }
  else {
    /***************************************************************/
    /* Insert the tranlsation code for speculative execution       */
    /***************************************************************/
    unsigned proc_base_reg = 3;
    base_register = 12;
    scratch_reg   = 11;
    scratch_reg_2 = 10;

    /* Store the address for the speculation code */
    translated_bb_start_address = next_xlate_address;

    if (do_no_output_during_runs == 0) {
      warn_xlate("* (svc speculation prep translated body) -------------------------------\n");
    }
    
    /* Load base_register with the address of our ARIA state saving space    */               
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);   next_xlate_address++; 
    
    #define NOT_TAKEN_REG 4
    #define TRACE_POINTER 5
    /* Save off the input parameters */
    *next_xlate_address = STORE_REG ( 3, PROC_STATE_OFFSET,  base_register);       next_xlate_address++;

    if (not_taken_address_required_flag == 1) { 
      *next_xlate_address = STORE_REG ( 4, NOT_TAKEN_RVAL_OFFSET, base_register);    next_xlate_address++;
    }

    /* Set the dynamic trace buffer to the static trace buffer */
    if (do_dynamic_trace_generation) {
      INSERT_DYNAMIC_TRACE_CODE(TRACE_POINTER);
    }

    /*******************************************************************/
    /** Check the value in GPR 2 (the system call routine identifier) **/
    /**  to see if the program is trying to exit; the exit routine    **/
    /**  flags differs by operating system version, so check for any. **/
    /*******************************************************************/
    /* Get the user's GPR2 value */
    *next_xlate_address = LOAD_REG  ( 11, GPR_SAVE_OFFSET+ARIA_GPR_SIZE*2, proc_base_reg);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_CMPI (  7,0,  2, the_exit_code_value);             next_xlate_address++;

    /* If this speculates exit from program, return a zero next address */
    {
      unsigned * branch_over_address = next_xlate_address;
      unsigned instrs_in_exit_code = 0;
      
      /** Skip where the conditional branch will be placed **/
      next_xlate_address++;

      /* Get the user's link register value */
      *next_xlate_address = OPC_ASM_ADDI ( 3, 0, 0);                                 next_xlate_address++;

      /* Enter the link register target address in the dynamic trace if appropriate */
      if (do_dynamic_trace_generation) {
	*next_xlate_address = STORE_REG  ( 3, 12, TRACE_POINTER);                  next_xlate_address++; 
      }

      /* Set the function not_taken return value to the link register address (if desired) */
      if (not_taken_address_required_flag == 1) { 
	*next_xlate_address = STORE_REG  ( 3,  0, NOT_TAKEN_REG);                  next_xlate_address++; 
      }

      /* Return to the Aria code */
      *next_xlate_address = OPC_ASM_BCLR(0x14, 0);                                   next_xlate_address++;

      /** Determine the number of instructions to branch over this exit code **/
      instrs_in_exit_code = (next_xlate_address - branch_over_address);
      *branch_over_address = OPC_ASM_BC ( 0x05, 7*4+2, instrs_in_exit_code);
    }

    /* Else return the link register next address */
    /* Get the user's link register value */
    *next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, proc_base_reg);    next_xlate_address++;
      
    /* Store the link register value as the not_taken return value if desired */
    *next_xlate_address = STORE_REG(scratch_reg, NEXT_TARGET_OFFSET, proc_base_reg);   next_xlate_address++; 
    if (not_taken_address_required_flag == 1) { 
      *next_xlate_address = STORE_REG(scratch_reg,  0, NOT_TAKEN_REG);                 next_xlate_address++; 
    }
      
    if ( (do_dynamic_trace_generation) && (do_iaddr_fixup) ) {
      *next_xlate_address = OPC_ASM_RLWINM(scratch_reg, scratch_reg, 4, 28, 31);       next_xlate_address++;
      *next_xlate_address = OPC_ASM_CMPI ( 7, 0, scratch_reg, ((global_user_program_iaddr_mask >> 28) & 0xf));                    next_xlate_address++;
      *next_xlate_address = OPC_ASM_BC   ( 0x4, 30, 5);                                next_xlate_address++;
      *next_xlate_address = LOAD_REG  (scratch_reg, LR_SAVE_OFFSET, base_register); next_xlate_address++;
      *next_xlate_address = LOAD_REG  (scratch_reg_2, IADDR_OFFSET, base_register); next_xlate_address++;
      *next_xlate_address = OPC_ASM_SUBFC(scratch_reg, scratch_reg_2, scratch_reg);    next_xlate_address++;
      *next_xlate_address = OPC_ASM_B    ( 2);                                         next_xlate_address++;
      *next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, base_register);  next_xlate_address++;      
    }

    /* Store the link register value in the dynamic trace if appropriate */
    if (do_dynamic_trace_generation) {
      *next_xlate_address = STORE_REG(scratch_reg, 12, TRACE_POINTER);                 next_xlate_address++; 
    }
      
    /* Set the function return value (GPR3) to the next BB address */
    *next_xlate_address = LOAD_REG( 3, NEXT_TARGET_OFFSET, proc_base_reg);   next_xlate_address++; 
    
    /* Return to the Aria code */
    *next_xlate_address = OPC_ASM_BCLR(0x14, 0);                                         next_xlate_address++;
    
    if (do_no_output_during_runs == 0) {
      DUMP_FAKE_DESCRIPTION(translated_bb_start_address, next_xlate_address);
      warn_xlate("* (exit) ----------------------------------------------------------------\n");
    }
    
    xlate_ptr = next_xlate_address;

    /** Determine the total number of instructions in the speculation xlate bb code **/
    total_instr_in_xl_bb = (next_xlate_address - translated_bb_start_address);

    /* Update the addresses for the program and translation space **/
    #include "xlation_alignment_code.c"

    /** Adjust the hash table entry to reflect the taken-path translation **/
    HashTable_NT_xlated_bb_size(*the_hash_entry_ptr)  = total_instr_in_xl_bb;
    HashTable_NT_xlation_address(*the_hash_entry_ptr) = translated_bb_start_address;

    /** Dump the translated code out of the data cache **/
    _sync_cache_range(translated_bb_start_address, 4*(next_xlate_address - translated_bb_start_address));
  }

/**
warn("global_user_iaddr_offset = %08x\n", global_user_iaddr_offset);
**/

  return next_xlate_address;
}






