/*                                                                     
** COMPONENT_NAME: MET / Aria
**                                                                      
** Licensed Materials - Property of IBM                                  
**                                                                     
** IBM Confidential (IBM Confidential-Restricted when combined          
**                   with aggregated modules for this product)          
**                                                                     
**                                                                     
** OCO Source Materials                                                
**                                                                     
**                                                                     
** GA-0100 (C) Copyright IBM Corp. 1997                                 
**                                                                     
** US Government Users Restricted RIghts - Use, duplication or         
** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.   
**                                                                     
** The source code for this program is not published or otherwise      
** divested of its trade secrets, irrespective of what has been        
** deposited within the U.S. Copyright Office.                         
**                                                                     
*/


 
/*******************************************************************
  PROGRAM NAME..... do_one_translation
  FUNCTION ........ Translates one user basic block into an Aria 
     represetnation of that block, including instructions to 
     pull out relevant micro-trace information.
       The general process is to analyze the basic block to 
     find the start and end instrucitons, then to generate code
     that swaps Aria state into user process state (for as much 
     state as is used by the basic block).  The original basic block 
     code is then copied into the translation space (along with 
     additional instructions to fill in the trace buffer), except 
     for any final branch instruction.  Code is then generated to 
     swap back from user to Aria state, and any code needed to 
     analyze and simulate the final branch is added.

  NOTES............
  ASSUMPTIONS... None
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... the next available translation address
  CALLED-ROUTINES.. analyze_branch, bb_fallthrough
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... free_regs_in_8_bits_of_vector.init_dat,
     fill_static_buffer_code.c, xlation_alignment_code.c,
     save_aria_state_code.c, bb_prolog_code.c, opc_idx_dtrb_code.gen,
     bb_epilog_code.c, restore_aria_state_code.c, opc_idx_spec_xform.gen

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/
#include <stdio.h>
#include <stdlib.h>

#include "aria_xlate_macros.h"
#include "aria_debug.h"
#include "io_msgs.h"

#include "opc.h"

#include "proc_state.h"
#include "aria_state.h"
#include "hashing.h"
#include "analyze_branch.h"

#include "dump_description.h"

#include "globtypes.h"
#include "global_variables.h"

#include "do_translation.h"

#include "canned_code.h"

#include "analyze_lwarx_stwcx.h"

#include "display_reg_usedef.h"

#include "aria_interface.h"
#include "aria_per_instr_interface.h"

/************************************************************************
 * These are static (local) variable definitions used in analysis
 ***********************************************************************/
/** This table is used to speed up the calculation of free registers **/
static unsigned free_regs_in[257] = {
    #include "free_regs_in_8_bits_of_vector.init_dat"
    0
  };

/** MACROS used to shorthand some pieces of code **/

#include "translation_macros.h"

/** Helper Functions **/

unsigned * 
do_one_translation(unsigned * next_prog_address,
		   hash_entry * the_hash_entry_ptr,
		   unsigned   do_speculative_path_xlation,
		   unsigned   generation
		   )
{
  unsigned * last_orig_bb_instr_ptr;
  unsigned * orig_ptr;

  unsigned * xlate_save_ptr;
  unsigned * temp_ptr;
  unsigned total_instr_in_bb = 0;

  use_def_data_type the_instr_reg_data[ABSOLUTE_MAXIMUM_USER_INSTR_IN_A_BLOCK];
  use_def_data_type the_full_block;

  instruction_xlate_type instr_xlate_info[ABSOLUTE_MAXIMUM_USER_INSTR_IN_A_BLOCK];

  unsigned translation_type;
  
  unsigned * original_bb_start_address = next_prog_address;
  unsigned * translated_bb_start_address;
  unsigned * microtrace_buffer_address = 0;

  unsigned * prolog_start_address;
  unsigned * epilog_start_address;

  unsigned * br_posn;
  unsigned * br_posn_2;

  opc_idx last_iword_index;

  short instr_in_prolog = 0;
  short instr_in_epilog = 0;

  unsigned base_register;
  unsigned scratch_reg;
  unsigned trbf_base_reg;

  unsigned original_base_register;
  unsigned original_scratch_reg;
  unsigned original_trbf_base_reg;

  unsigned top_visible_reg_use_vectors[4] = {0, 0, 0, 0};

  unsigned total_instr_in_xl_bb = 0;

  /* flag to keep track of whether dynamic data will have to be written for this block */
  unsigned dynamic_data_in_block = 0; 
  unsigned block_ending_branch   = 0; 


  unsigned instr_num;
  unsigned CR_must_be_saved;

  /** Check whether we already have a translation for the   **/
  /**  other "flavor" of this BB (e.g. whether we have      **/
  /**  already done a taken-path when we are about to do a  **/
  /**  not-taken path translation).                         **/
  unsigned first_xlation = (HashTable_original_address(*the_hash_entry_ptr) == NULL);

  IFDEBUG(warn("\n[do_xlate] :: Entered do_one_translation with...\n");
	  warn("[do_xlate]        next_prog_address           = 0x%016LLx\n", next_prog_address);
	  warn("[do_xlate]        the_hash_entry_ptr          = 0x%016LLx\n", the_hash_entry_ptr);
	  warn("[do_xlate]        do_speculative_path_xlation = 0x%x\n", do_speculative_path_xlation);
	  warn("[do_xlate]   first_xlation = 0x%x\n", first_xlation);	  
	  );

  if (!first_xlation) {
    /*****************************************************************/
    /** We have an existing translation, so recover the previously  **/
    /** determined data for this user basic block (from hash table) **/
    /*****************************************************************/
    /* The microtrace buffer that holds the original basic block copy */
    microtrace_buffer_address = HashTable_trace_buffer(*the_hash_entry_ptr);
        
    /* The flag for whether the block has dynamic data in it */
    dynamic_data_in_block = HashTable_has_dynamic_data(*the_hash_entry_ptr);
    block_ending_branch   = HashTable_ends_in_a_branch(*the_hash_entry_ptr);

    /* The original register numbers used for the base, scratch and buffer pointer */
    original_base_register = HashTable_base_reg(*the_hash_entry_ptr);
    original_scratch_reg   = HashTable_scratch_reg(*the_hash_entry_ptr);
    original_trbf_base_reg      = HashTable_buffer_reg(*the_hash_entry_ptr);

    /* The register use and definition for the full block */
    the_full_block.use_vectors[0] = HashTable_block_gpr_use(*the_hash_entry_ptr);
    the_full_block.def_vectors[0] = HashTable_block_gpr_def(*the_hash_entry_ptr);
    the_full_block.use_vectors[1] = HashTable_block_fpr_use(*the_hash_entry_ptr);
    the_full_block.def_vectors[1] = HashTable_block_fpr_def(*the_hash_entry_ptr);
    the_full_block.use_vectors[2] = HashTable_block_cr_use(*the_hash_entry_ptr);
    the_full_block.def_vectors[2] = HashTable_block_cr_def(*the_hash_entry_ptr);
    the_full_block.use_vectors[3] = HashTable_block_spr_use(*the_hash_entry_ptr);
    the_full_block.def_vectors[3] = HashTable_block_spr_def(*the_hash_entry_ptr);

    /* Calculate the full block touch vectors (to save storing them) */
    the_full_block.touch_vectors[0] = the_full_block.use_vectors[0] | the_full_block.def_vectors[0];
    the_full_block.touch_vectors[1] = the_full_block.use_vectors[1] | the_full_block.def_vectors[1];
    the_full_block.touch_vectors[2] = the_full_block.use_vectors[2] | the_full_block.def_vectors[2];
    the_full_block.touch_vectors[3] = the_full_block.use_vectors[3] | the_full_block.def_vectors[3];

    /* The register use visible at the top of the basic block */
    top_visible_reg_use_vectors[0] = HashTable_top_vis_gpr_use(*the_hash_entry_ptr);
    top_visible_reg_use_vectors[1] = HashTable_top_vis_fpr_use(*the_hash_entry_ptr);
    top_visible_reg_use_vectors[2] = HashTable_top_vis_cr_use(*the_hash_entry_ptr);
    top_visible_reg_use_vectors[3] = HashTable_top_vis_spr_use(*the_hash_entry_ptr);

    /* Get the size (in instrs) of the prolog and epilog so they can be copied */
    instr_in_prolog  = HashTable_prolog_size(*the_hash_entry_ptr);
    instr_in_epilog  = HashTable_epilog_size(*the_hash_entry_ptr);

    /** Set the last_orig_bb_instr_ptr and the last_iword_index **/
    total_instr_in_bb      = HashTable_user_bb_size(*the_hash_entry_ptr);    
    last_orig_bb_instr_ptr = aria_Trace_Iword_ptr(microtrace_buffer_address, (total_instr_in_bb - 1));
    last_iword_index = opc_index(*last_orig_bb_instr_ptr);

    /** Set up the instruction xlation information structure **/
    {
      int iii;
      for (iii = 0; iii < total_instr_in_bb; iii++) {
	instr_xlate_info[iii].iaddr = aria_Trace_Iaddr(microtrace_buffer_address, iii);
	instr_xlate_info[iii].iword = aria_Trace_Iword(microtrace_buffer_address, iii);
	instr_xlate_info[iii].index = opc_index(aria_Trace_Iword(microtrace_buffer_address, iii));
	instr_xlate_info[iii].ikind = aria_Trace_Ikind(microtrace_buffer_address, iii);
	instr_xlate_info[iii].daddr = aria_Trace_Daddr(microtrace_buffer_address, iii);
	instr_xlate_info[iii].dlen  = aria_Trace_Dlen(microtrace_buffer_address, iii);
      }
    }

    /* Dump out the initial trace buffer contents */
    #include "show_trace_buffer_contents.c"
  }
  else {
    /* Now analyze the code for this basic block */
    orig_ptr  = next_prog_address;
    xlate_save_ptr = next_xlate_address;

    /*********************************/
    /** FIND THE END OF BASIC BLOCK **/
    /*************************************************************************/
    /** This code determines the extent of the user-code basic block that will
     *   be translated.  There are three conditions that can end a basic-block:
     *    1. The discovery of a branch instruction
     *    2. The user-code basic block becomes longer than 
     *          the maximum allowed size (in instructions)
     *    3. The register use/def for next instr would not leave enough 
     *          registers for dynamic trace info generation.
     **/

    CLEAR_FULL_BLOCK_USE_DEF_TOUCH_VECTORS;

    /** We create the trace buffer as we scan the original basic block **/
    /**  To prepare, set a 0x00000000 value as a buffer spot, then     **/
    /**   set teh start of the trace buffer                            **/
    *next_xlate_address = 0x00000000;           next_xlate_address++;
    microtrace_buffer_address = next_xlate_address;

    do { 
      unsigned test_vector;
      unsigned tv_0_7  ;
      unsigned tv_8_15 ;
      unsigned tv_16_23;
      unsigned tv_24_31;
      unsigned free_gpr_regs;

      /* Trace information */
      unsigned iword         = *orig_ptr; 
      unsigned iword_index   = opc_index(iword);
      unsigned iword_ikind   = opc_itype(iword_index);

      instr_xlate_info[total_instr_in_bb].iaddr = (BASE_TYPE)orig_ptr;
      instr_xlate_info[total_instr_in_bb].iword = *orig_ptr;
      instr_xlate_info[total_instr_in_bb].index = opc_index(iword);
      instr_xlate_info[total_instr_in_bb].ikind = opc_itype(iword_index);
      instr_xlate_info[total_instr_in_bb].daddr = 0xdeadbeef;
      instr_xlate_info[total_instr_in_bb].dlen  = 0;

      IFDEBUG(warn_xlate(" %3u  ::  0x%08x 0x%08x %4u %4u \n", total_instr_in_bb,
			 instr_xlate_info[total_instr_in_bb].iaddr,
			 instr_xlate_info[total_instr_in_bb].iword,
			 instr_xlate_info[total_instr_in_bb].index,
			 instr_xlate_info[total_instr_in_bb].ikind
			 );
	      );

      /** Determine if the user specified a BEFORE action for this iword **/
      if ( ((per_instruction_action_time(iword_index) & DO_ARIA_ACTION_BEFORE) != 0) &&
	   (total_instr_in_bb > 0)
	   ) {
	global_user_stop_condition_counter++; 
	goto stop_analysis_because_of_user_action;
      }

      /** Determine whether we need a special lwarx/stwcx execution path translation **/
      if ( ((iword_index == OPC_LWARX) || (iword_index == OPC_LDARX)) && (do_speculative_path_xlation == 0) ) {
	next_xlate_address = xlate_save_ptr;
	next_xlate_address = do_lwarx_stwcx_translation(original_bb_start_address,
							the_hash_entry_ptr,
							generation
							);
	return next_xlate_address;
      }
      /**/

      /** Determine if the instruction is a system call (and thus ends the BB) **/
      if (OPC_IS_A_SYSTEM_CALL(iword_ikind)) {
	goto block_ends_before_an_svc;
      }

      last_iword_index = iword_index;

      opc_full_use_def_vecs(iword, the_instr_reg_data[total_instr_in_bb].use_vectors, 
			    the_instr_reg_data[total_instr_in_bb].def_vectors, the_instr_reg_data[total_instr_in_bb].touch_vectors);

      /**
	warn_xlate("For instr 0x%08x the use/def vecs are:\n", iword);
	warn_xlate("  gpr 0x%08x   0x%08x   0x%08x\n",
	the_instr_reg_data[total_instr_in_bb].use_vectors[0],
	the_instr_reg_data[total_instr_in_bb].def_vectors[0],
	the_instr_reg_data[total_instr_in_bb].touch_vectors[0]
	);
	warn_xlate("  fpr 0x%08x   0x%08x   0x%08x\n",
	the_instr_reg_data[total_instr_in_bb].use_vectors[1],
	the_instr_reg_data[total_instr_in_bb].def_vectors[1],
	the_instr_reg_data[total_instr_in_bb].touch_vectors[1]
	);
	warn_xlate("  cr  0x%08x   0x%08x   0x%08x\n",
	the_instr_reg_data[total_instr_in_bb].use_vectors[2],
	the_instr_reg_data[total_instr_in_bb].def_vectors[2],
	the_instr_reg_data[total_instr_in_bb].touch_vectors[2]
	);
	warn_xlate("  spr 0x%08x   0x%08x   0x%08x\n",
	the_instr_reg_data[total_instr_in_bb].use_vectors[3],
	the_instr_reg_data[total_instr_in_bb].def_vectors[3],
	the_instr_reg_data[total_instr_in_bb].touch_vectors[3]
	);
	**/
      /** Check that we would still have one free register **/
      test_vector = ( the_instr_reg_data[total_instr_in_bb].use_vectors[0] | the_full_block.use_vectors[0] |
		      the_instr_reg_data[total_instr_in_bb].def_vectors[0] | the_full_block.def_vectors[0] );
      tv_0_7   = (test_vector & 0x000000f8);
      tv_8_15  = (test_vector & 0x0000ff00) >> 8;
      tv_16_23 = (test_vector & 0x00ff0000) >> 16;
      tv_24_31 = (test_vector & 0xff000000) >> 24;
      free_gpr_regs = (free_regs_in[tv_0_7] + free_regs_in[tv_8_15] 
		       + free_regs_in[tv_16_23] + free_regs_in[tv_24_31]);
    
      /** Determine if three registers are avaiable,   **/
      /**    NOT including registers 0, 1 or 2.        **/
      /**      THREE registers are required to:        **/
      /**        1. Address the dynamic trace buffer   **/
      /**        2. Address the proc state space       **/
      /**        3. use as a scratch register          **/
      /** OR if there are no instructions in the BB    **/
      /**  (this is necessary for lswx, lmw instrs)    **/
      /**  in which case we make this a one-instr BB   **/
    
      if ( free_gpr_regs < 6 ) {
	global_addressibility_counter++;

	if (total_instr_in_bb == 0) {
	  /** HAVE to add this instruction to make a one-instruction BB **/
	  UPDATE_TOTAL_BLOCK_USE_DEF_TOUCH_VECTORS;
	
	  if (do_canned_string_code) {
	    /* Set that the block has dynamic data */
	    dynamic_data_in_block = 1;
	    
	    /** Set up the trace buffer entry for this instruction **/
            #include "fill_trace_buffer_entry_code.c"

	    /* Update the addresses for the program and translation space **/
            #include "xlation_alignment_code.c"

	    /* Increment the number of instructions in the block */
	    total_instr_in_bb++; 	  

	    /* Dump out the initial trace buffer contents */
            #include "show_trace_buffer_contents.c"

	    /* Set the starting address for the translated basic block */
	    /*  NOTE: this is used in synchronizing the cache...       */
	    translated_bb_start_address = next_xlate_address;

	    /* Call canned routines to fill in the translated basic block */
	    switch (instr_xlate_info[total_instr_in_bb - 1].index) {
	    case OPC_LSCBX:
	    case OPC_LSCBX_:
	    case OPC_LSWX:
	    case OPC_STSWX:
	      next_xlate_address = insert_canned_string_code(&the_full_block, 
							     original_bb_start_address,
							     microtrace_buffer_address,
							     the_hash_entry_ptr
							     );
	      break;
	    case OPC_LMW:
	    case OPC_STMW:
	      next_xlate_address = insert_canned_multiple_code(&the_full_block, 
							       instr_xlate_info,
							       original_bb_start_address,
							       microtrace_buffer_address,
							       the_hash_entry_ptr
							       );
	      break;
	    default:
	      fprintf(mystderr, "ARIA: ERROR: Addressability problem for iword = 0x%08x , index = %u\n", 
		      iword, instr_xlate_info[total_instr_in_bb - 1].index);
	      fprintf(mystderr, "  assumed that this could not happen for this type of iword!\n");
	      abort();
	      break;
	    }
	  
	    orig_ptr++; 

	    /* Set the generation of this (canned-code) translation */
	    HashTable_generation(*the_hash_entry_ptr) = generation;

	    /* Exit the translation routine */
	    goto exit_from_do_one_translation;
	  }
	  /* Increment the number of instructions in the block */
	  total_instr_in_bb++; 	  

	  orig_ptr++; 
	}
	goto block_ends_on_register_addresibility;
      }
  
      /** We know now that this instruction will be added to this basic block **/
      UPDATE_TOTAL_BLOCK_USE_DEF_TOUCH_VECTORS;

      /********************************************************/
      /** Set up the trace buffer entry for this instruction **/
      #ifdef DO_64_BIT_COMPILE
        #include "fill_trace_buffer_entry_code_64.c"
      #else 
        #include "fill_trace_buffer_entry_code.c"
      #endif

      /** Increment the orig_ptr to point to the next user program word **/
      orig_ptr++; 

      /** Increment the number of instructions in the user basic block **/
      total_instr_in_bb++; 

      /** Check whether this instruction puts us over the max block size limit **/
      if ( total_instr_in_bb == global_max_block_size_limit ) { 
	global_max_size_ubb_counter++; 
      }
    
      /** Check whether there are any memory access instructions in the block **/
      if (OPC_IS_A_MEMORY_ACCESS_OP(iword_ikind)) {
	dynamic_data_in_block = 1;
      }
      
      /** Check for a user-specified stop-after condition **/
      if ( (per_instruction_action_time(instr_xlate_info[total_instr_in_bb - 1].index) > DO_NO_ARIA_ACTION) ){
	HashTable_user_stop_iaddr(*the_hash_entry_ptr) = (orig_ptr-1);
	global_user_stop_condition_counter++; 
	goto stop_analysis_because_of_user_action;
      }

    } while ( (!OPC_IS_A_BRANCH(instr_xlate_info[total_instr_in_bb - 1].ikind)) && 
	      (total_instr_in_bb < global_max_block_size_limit)
	      ); 

    if (OPC_IS_A_BRANCH(instr_xlate_info[total_instr_in_bb - 1].ikind)) {
      block_ending_branch = 1;
    }

  block_ends_on_register_addresibility:
  block_ends_before_an_svc:
  stop_analysis_because_of_user_action:
    /*********************************/
    /** END OF BASIC BLOCK DETECTED **/
    /*********************************/

    /** Mark the end of the trace buffer **/
    *next_xlate_address = 0x00000000;           next_xlate_address++;
  
    /* Update the xlation address alignment for the subsequent translation */
    #include "xlation_alignment_code.c"

    /* Dump out the initial trace buffer contents */
    #include "show_trace_buffer_contents.c"

    /* Set a pointer to the last instruction in the basic block */
    /*  (often this is the ending branch instruction)           */
    last_orig_bb_instr_ptr = (unsigned *)instr_xlate_info[total_instr_in_bb-1].iaddr;

    /*************************************************************************/
    /*  Now go bottom-up through the BB to determine the outward-visible     */
    /*   (here called top_visible) register uses (in a vector for each type) */ 
    /*************************************************************************/
    {
      int inum;
      for (inum = (total_instr_in_bb - 1); inum >= 0; inum--) {
	top_visible_reg_use_vectors[0] = ((top_visible_reg_use_vectors[0] & ~(the_instr_reg_data[inum].def_vectors[0]))
					  | the_instr_reg_data[inum].use_vectors[0]);
	top_visible_reg_use_vectors[1] = ((top_visible_reg_use_vectors[1] & ~(the_instr_reg_data[inum].def_vectors[1]))
					  | the_instr_reg_data[inum].use_vectors[1]);
	top_visible_reg_use_vectors[2] = ((top_visible_reg_use_vectors[2] & ~(the_instr_reg_data[inum].def_vectors[2]))
					  | the_instr_reg_data[inum].use_vectors[2]);
	top_visible_reg_use_vectors[3] = ((top_visible_reg_use_vectors[3] & ~(the_instr_reg_data[inum].def_vectors[3]))
					  | the_instr_reg_data[inum].use_vectors[3]);
      }
    }

    /**********************************************************************/
    /* Now select 3 registers to be our base, scratch and dynamic trace   */
    /*  buffer address registers, and include them in the touch vectors,  */
    /*  so that they will retain a consistent value in ARIA's view.       */
    /**********************************************************************/
    {
      int i;
      unsigned regs_to_use[3] = {3, 4, 5};
      unsigned index = 0;
      for (i = 3; i <= 31; i++) {
	if ((the_full_block.touch_vectors[0] & (1 << i)) == 0) {
	  regs_to_use[index] = i;
	  if (i > 12) {
	    the_full_block.touch_vectors[0] |= (1 << i);
	  }
	  index++;
	}
	if (index == 3) {
	  break;
	}
      }
      original_base_register = regs_to_use[0];
      original_scratch_reg   = regs_to_use[1];
      original_trbf_base_reg      = regs_to_use[2];
      /* 
	 warn("original_base_register = %d\n", regs_to_use[0]);
	 warn("original_scratch_reg   = %d\n", regs_to_use[1]);
	 warn("original_trbf_base_reg      = %d\n", regs_to_use[2]);
	 **/
    }
  
    /** Store the information common to all translations (i.g. info about original BB) **/
    HashTable_original_address(*the_hash_entry_ptr)   = original_bb_start_address;
    HashTable_trace_buffer(*the_hash_entry_ptr)       = microtrace_buffer_address;
    HashTable_user_bb_size(*the_hash_entry_ptr)       = total_instr_in_bb;
    HashTable_generation(*the_hash_entry_ptr)         = generation;
    HashTable_overflow(*the_hash_entry_ptr)           = NULL;

    /** Store the information we may need to do the other translation **/
    HashTable_has_dynamic_data(*the_hash_entry_ptr)  = dynamic_data_in_block;
    HashTable_ends_in_a_branch(*the_hash_entry_ptr)  = block_ending_branch;

    HashTable_base_reg(*the_hash_entry_ptr)    = original_base_register;
    HashTable_scratch_reg(*the_hash_entry_ptr) = original_scratch_reg;
    HashTable_buffer_reg(*the_hash_entry_ptr)  = original_trbf_base_reg;

    HashTable_block_gpr_use(*the_hash_entry_ptr) = the_full_block.use_vectors[0];
    HashTable_block_gpr_def(*the_hash_entry_ptr) = the_full_block.def_vectors[0];
    HashTable_block_fpr_use(*the_hash_entry_ptr) = the_full_block.use_vectors[1];
    HashTable_block_fpr_def(*the_hash_entry_ptr) = the_full_block.def_vectors[1];
    HashTable_block_cr_use(*the_hash_entry_ptr)  = the_full_block.use_vectors[2];
    HashTable_block_cr_def(*the_hash_entry_ptr)  = the_full_block.def_vectors[2];
    HashTable_block_spr_use(*the_hash_entry_ptr) = the_full_block.use_vectors[3];
    HashTable_block_spr_def(*the_hash_entry_ptr) = the_full_block.def_vectors[3];
    HashTable_top_vis_gpr_use(*the_hash_entry_ptr)   = top_visible_reg_use_vectors[0];
    HashTable_top_vis_fpr_use(*the_hash_entry_ptr)   = top_visible_reg_use_vectors[1];
    HashTable_top_vis_cr_use(*the_hash_entry_ptr)    = top_visible_reg_use_vectors[2];
    HashTable_top_vis_spr_use(*the_hash_entry_ptr)   = top_visible_reg_use_vectors[3];
  }

  /** Enter a header for this BB translation **/
  #include "add_xlation_header_code.c"
  
  /* Output the use and def of registers for this user block */
  if ( (do_no_output_during_runs == 0) ) { 
    /** warn_bb_regs("The total sets of registers touched are:\n"); **/
    display_register_uses_and_defs(the_full_block.use_vectors[0], the_full_block.use_vectors[1], 
				   the_full_block.use_vectors[2], the_full_block.use_vectors[3], 
				   the_full_block.def_vectors[0], the_full_block.def_vectors[1], 
				   the_full_block.def_vectors[2], the_full_block.def_vectors[3], 
				   top_visible_reg_use_vectors); 
    if (do_speculative_path_xlation == 0) {
      warn_xlate("\nThe taken-path-execution (non-speculative) code is:\n");
    }
    else {
      warn_xlate("\nThe speculative-execution (not-taken path) code is:\n");
    }

    /* Set a temp pointer (for printing output) to the start of the translated block */
    temp_ptr = next_xlate_address;
  }

  if (do_speculative_path_xlation == 0) {
    /** Do the taken path translations **/
    translation_type = TAKEN_XLATE;   
    translated_bb_start_address = next_xlate_address;
    
    /** In case this is not a first xlation, get the start of prolog and epilog addresses **/
    prolog_start_address = HashTable_NT_xlation_address(*the_hash_entry_ptr);
    epilog_start_address = prolog_start_address + (HashTable_NT_xlated_bb_size(*the_hash_entry_ptr) - instr_in_epilog);
  }
  else {
    /** Do the taken path translations **/
    translation_type = NOT_TAKEN_XLATE;   
    translated_bb_start_address = next_xlate_address;

    /** In case this is not a first xlation, get the start of prolog and epilog addresses **/
    prolog_start_address = HashTable_xlation_address(*the_hash_entry_ptr);
    epilog_start_address = prolog_start_address + (HashTable_xlated_bb_size(*the_hash_entry_ptr) - instr_in_epilog);
  }

  if (first_xlation) {
    /* Store (temporarily) the start of the translated code epilog */
    prolog_start_address = next_xlate_address;

    /* Set up some temporary values for base_register and scratch_reg */
    /* to use only while saving off the Aria entry state              */
    base_register = 12;
    scratch_reg   = 11;

    /********************************************************/
    /** Save off all of Aria's important (clobbered) state **/
    /********************************************************/
    /* Load GPRbase with the address of our ARIA state saving space */          
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);  next_xlate_address++; 

    /* Store the processor state save space address in our reserved work area */
    *next_xlate_address = STORE_REG( 3, PROC_STATE_OFFSET,  base_register);      next_xlate_address++;

    /* If we are calculating not-taken addresses, store the not_taken address receiver address */
    if (not_taken_address_required_flag == 1) {
      *next_xlate_address = STORE_REG( 4, NOT_TAKEN_RVAL_OFFSET, base_register);   next_xlate_address++;
    }
  
    /** If we are doing dynamic tracing AND have dynamic data in the user BB **/
    /**  Create a pointer to the trace buffer **/
    if ( (do_dynamic_trace_generation) && (dynamic_data_in_block || block_ending_branch) ) {
      unsigned static_hi = ((unsigned)microtrace_buffer_address & 0xffff0000) >> 16;
      unsigned static_lo = ((unsigned)microtrace_buffer_address & 0x0000ffff);
      *next_xlate_address = OPC_ASM_ADDIS( 4, 0, static_hi);                          next_xlate_address++;
      *next_xlate_address = OPC_ASM_ORI  ( 4, 4, static_lo);                          next_xlate_address++;
      *next_xlate_address = STORE_REG( 4, PROC_DYN_TRBUF_OFFSET, base_register);  next_xlate_address++;
    }

    /********************************************************************************************/
    /** Insert the code to save the Aria state (that may be altered by user program execution) **/
    /********************************************************************************************/
    {
      unsigned gprs = the_full_block.touch_vectors[0];
      unsigned fprs = the_full_block.touch_vectors[1];
      unsigned crs  = the_full_block.touch_vectors[2];

      #include "save_aria_state_code.c"
    }

    /* Recover the settings for the selected base, scratch and dynamic trace buffer ptrs */
    base_register = original_base_register;
    scratch_reg   = original_scratch_reg;
    trbf_base_reg      = original_trbf_base_reg;

    /* Copy the GPR3 (user processor state ptr) value to the base register */
    if (base_register != 3) {
      /* Copy GPR 3 value to the base register */
      *next_xlate_address = OPC_ASM_OR(base_register,  3,  3);  next_xlate_address++;
    }
  
    /****************************************************************************/
    /** Insert the bb prolog, which inserts instructions to restore user state **/
    /****************************************************************************/
    {
      unsigned gpr_uses   = top_visible_reg_use_vectors[0];
      unsigned fpr_uses   = top_visible_reg_use_vectors[1];
      unsigned cr_touches = the_full_block.touch_vectors[2];
      unsigned spr_uses   = the_full_block.use_vectors[3];
      unsigned spr_defs   = the_full_block.def_vectors[3];
      
      #include "bb_prolog_code.c"
    }

    /** Calculate the number of instructions in the translated BB prolog **/
    instr_in_prolog = (next_xlate_address - prolog_start_address);
  }
  else {
    /*********************************************/
    /** COPY the previously-defined prolog code **/
    /*********************************************/
    unsigned iid;
    for (iid = 0; iid < instr_in_prolog; iid++) {
      *next_xlate_address = *(prolog_start_address + iid);  next_xlate_address++;
    }

    /* Recover the settings for the selected base, scratch and dynamic trace buffer ptrs */
    base_register = original_base_register;
    scratch_reg   = original_scratch_reg;
    trbf_base_reg      = original_trbf_base_reg;
  }
  
  /** Output the prolog (if appropriate) **/
  if (do_no_output_during_runs == 0) {
    warn_xlate("* (prolog) ----------------------------------------------------------------\n");
    DUMP_DESCRIPTION(temp_ptr, next_xlate_address);    
    warn_xlate("* (translated body) -------------------------------------------------------\n");
    temp_ptr = next_xlate_address;
  }

  /******************************************************************/
  /** Now generate a translated copy of the original (user         **/
  /** instructions) basic block                                    **/
  /******************************************************************/
  
  /** If doing dynamic trace generation, set up the dynamic trace buffer pointer **/
  if ( (do_dynamic_trace_generation) && (dynamic_data_in_block) ) {
    /** Get the pointer (base address) of the aria save space **/
    *next_xlate_address = OPC_ASM_ADDIS(scratch_reg,  0, ARIA_STATE_SAVE_HI_16);   next_xlate_address++;
    /** Load the dynamic trace buffer pointer from it **/
    *next_xlate_address = LOAD_REG(trbf_base_reg, PROC_DYN_TRBUF_OFFSET, scratch_reg);  next_xlate_address++;
    
    /* print out the translation if user desires it */
    if (do_no_output_during_runs == 0) {
      DUMP_DESCRIPTION(temp_ptr, next_xlate_address);
      temp_ptr = next_xlate_address;
    }
  }
  
  /*******************************************************************************/
  /** Do the copying of the original basic block instructions.                  **/
  /**  These instructions are modified according to the translation_type flag:  **/
  /**    TAKEN_XLATE     : non-speculative execution path (copy verbatim)       **/
  /**    NOT_TAKEN_XLATE : speculative execution path (alter memory operations) **/
  /*******************************************************************************/

  /* First, set hte flag that indicates the condition register must be saved to 0 */
  /*  A 0 value indicates the proper user CR value is in the CR; in a not-taken   */
  /*  translation, the CR is used during the evaluation of the data address (to   */
  /*  determine whether a load will actually be executed or not), so we will      */
  /*  need to do some CR swapping as it is used by the user code and xl code...   */
  CR_must_be_saved = 0;

  /* Now copy over the actual user instructions (unmodified) */
  /*  and insert dynamic trace code if requested             */
  for (instr_num = 0; instr_num < total_instr_in_bb; instr_num++) {

    /* If the last instruction is a branch, hop out of this loop */
    if ( (instr_num == (total_instr_in_bb-1)) && (OPC_IS_A_BRANCH(instr_xlate_info[instr_num].ikind)) ) {
      goto exit_the_taken_path_copy_user_code_loop;
    }
    
    /* Insert the instruction-index-dependent code required to generate a translation */
    /*  Note that this includes:                                                   */
    /*    setting the data access length in the trace buffer (for memory accesses) */
    /*    inserting any dynamic trace generation code required by the instruction  */
    /*    inserting either the original iword or a not-taken path code sequence    */
    #ifdef DO_64_BIT_COMPILE
      #include "opc_idx_full_xlate_code_64.gen" 
    #else
      #include "opc_idx_full_xlate_code.gen" 
    #endif
  } 
exit_the_taken_path_copy_user_code_loop:
  ;

  /* Restore the last correct CR register value if necessary      */
  /*  If the speculation code messed with the condition register, */
  /*  and the condition register will be saved, get the correct   */
  /*  value to save in the CR (so the bb_epilog code is cleaner). */
  if (translation_type == NOT_TAKEN_XLATE) {
    if ( (the_full_block.touch_vectors[2] != 0x0) && (CR_must_be_saved == 0) ) {
      *next_xlate_address = LOAD_REG(scratch_reg, CR_SAVE_OFFSET, base_register);    next_xlate_address++;
      *next_xlate_address = OPC_ASM_MTCRF (0xff, scratch_reg);                          next_xlate_address++;
    }
    
    if (do_no_output_during_runs == 0) {
      DUMP_FAKE_DESCRIPTION(temp_ptr, next_xlate_address);
      temp_ptr = next_xlate_address;
    }
  }
  
  /* Output the epilog marker */
  if (do_no_output_during_runs == 0) {
    warn_xlate("* (epilog) --------------------------------------------------------------------------------\n", next_xlate_address);
    temp_ptr = next_xlate_address;
  }

  if (first_xlation) {
    /* Store (temporarily) the start of the translated code epilog */
    epilog_start_address = next_xlate_address;

    /*****************************************************************************/
    /** Insert the bb epilog, which inserts instructions to save new user state **/
    /*****************************************************************************/
    {
      unsigned   gpr_defs   = the_full_block.def_vectors[0];
      unsigned   fpr_defs   = the_full_block.def_vectors[1];
      unsigned   cr_touches = the_full_block.touch_vectors[2];
      unsigned   spr_uses   = the_full_block.use_vectors[3];
      unsigned   spr_defs   = the_full_block.def_vectors[3];
    
      #include "bb_epilog_code.c"
    }

    if ( (do_no_output_during_runs == 0) ) {
      unsigned iword;
      opc_idx  index;
      char buf[512] = "*illegal instruction (data?)*";
      DUMP_DESCRIPTION(temp_ptr, next_xlate_address);
      warn_xlate("* (exit) ----------------------------------------------------------------------------------\n");
      iword = instr_xlate_info[total_instr_in_bb-1].iword;
      index = instr_xlate_info[total_instr_in_bb-1].index;
      opc_dis(iword, buf);

      if (OPC_IS_A_BRANCH(opc_itype(index))) {
	warn_xlate("      translating %08x %08x = %s to:\n", last_orig_bb_instr_ptr, iword, buf);
      }
      else {
	warn_xlate("      block broken at instr: %08x %08x = %s\n", last_orig_bb_instr_ptr, iword, buf);
      }
      temp_ptr = next_xlate_address;
    }
  
    /*********************************************************/
    /** Insert code to analyze and save the branch behavior **/
    /*********************************************************/
    {
      unsigned * orig_br_addr = original_bb_start_address + total_instr_in_bb - 1;
      if (OPC_IS_A_BRANCH(instr_xlate_info[total_instr_in_bb - 1].ikind)) {
	next_xlate_address = analyze_branch(*last_orig_bb_instr_ptr, orig_br_addr,
					    &base_register, scratch_reg, instr_num);
      }
      else {
	if ( instr_xlate_info[total_instr_in_bb - 1].index == OPC_SC ) {
	  warn("ARIA: ERROR: block includes an SCA instruction\n");
	}
	next_xlate_address = bb_fallthrough(orig_br_addr, &base_register);    
      }
      temp_ptr = next_xlate_address;
    }

    /********************************/
    /** Now restore the Aria state **/
    /********************************/
    scratch_reg   = 11;
    
    if (scratch_reg == base_register) {
      scratch_reg--;
    }

    {
      unsigned gprs = the_full_block.touch_vectors[0];
      unsigned fprs = the_full_block.touch_vectors[1];
      unsigned crs  = the_full_block.touch_vectors[2];
    
      #include "restore_aria_state_code.c"
    }

    /***************************************************************/
    /* Return through the link register (end xlated BB subroutine) */
    /***************************************************************/
    *next_xlate_address = OPC_ASM_BCLR (0x14, 0);        next_xlate_address++;

    /** Calculate the number of instructions in the translated BB prolog **/
    instr_in_epilog = (next_xlate_address - epilog_start_address);

  }
  else {
    /*********************************************/
    /** COPY the previously-defined epilog code **/
    /*********************************************/
    unsigned iid;
    for (iid = 0; iid < instr_in_epilog; iid++) {
      *next_xlate_address = *(epilog_start_address + iid); next_xlate_address++;
    }
    
    if ( (do_no_output_during_runs == 0) ) {
      DUMP_DESCRIPTION(temp_ptr, next_xlate_address);
      warn_xlate("* (back to aria) --------------------------------------------------------------------------\n");
      temp_ptr = next_xlate_address;
    }
  }

  /* Determine the total instructions in the translated BB code */
  total_instr_in_xl_bb = (next_xlate_address - translated_bb_start_address);

  /* Output the translated basic block information exit code */
  if (do_no_output_during_runs == 0) {
    DUMP_DESCRIPTION(temp_ptr, next_xlate_address);
    warn_xlate("===============================================================================================\n");
    warn_xlate("    end of BB %d [%u user, %u xlated instr] ; oa=%08x xa = %08x\n", 
	       global_bb_index+1, total_instr_in_bb,
	       total_instr_in_xl_bb, original_bb_start_address, translated_bb_start_address);
  }


  /***************************************************************/
  /* Update the addresses for the program and translation space **/
  /***************************************************************/
  #include "xlation_alignment_code.c"
  
  if ( (do_no_output_during_runs == 0) ) { 
    int iii;
    char dis_buf[256];
    warn_xlate("\nThe trace buffer is initialized to: (addr = 0x%08x)\n", microtrace_buffer_address); 
    for (iii = 0; iii < total_instr_in_bb; iii++) {
      opc_dis(aria_Trace_Iword(microtrace_buffer_address, iii), dis_buf);
      warn_xlate("  0x%08x  0x%08x  %5u  0x%08x  %3u  =  %s\n", 
		 aria_Trace_Iaddr(microtrace_buffer_address, iii),
		 aria_Trace_Iword(microtrace_buffer_address, iii),
		 aria_Trace_Ikind(microtrace_buffer_address, iii),
		 aria_Trace_Daddr(microtrace_buffer_address, iii),
		 aria_Trace_Dlen(microtrace_buffer_address, iii),
		 dis_buf);
    }
  }

  IFDEBUG(warn("[do_xlate] Finished with the translation... setting up the Hash Table entry\n"); 
	  warn("[do_xlate]   the_hash_entry_ptr = 0x%016LLx\n", the_hash_entry_ptr);
	  fflush(mystderr); 
	  );

  /**********************************************************/
  /* add a hash table entry for the newly translated block **/
  /**********************************************************/
  if (do_speculative_path_xlation == 0) {
    HashTable_xlation_address(*the_hash_entry_ptr) = translated_bb_start_address;
    HashTable_xlated_bb_size(*the_hash_entry_ptr)  = total_instr_in_xl_bb;
    /* Set the prolog and epilog sizes */
    HashTable_prolog_size(*the_hash_entry_ptr) = instr_in_prolog;
    HashTable_epilog_size(*the_hash_entry_ptr) = instr_in_epilog;
  }
  else {
    HashTable_NT_xlation_address(*the_hash_entry_ptr) = translated_bb_start_address;
    HashTable_NT_xlated_bb_size(*the_hash_entry_ptr)  = total_instr_in_xl_bb;
    /* Set the prolog and epilog sizes */
    HashTable_prolog_size(*the_hash_entry_ptr) = instr_in_prolog;
    HashTable_epilog_size(*the_hash_entry_ptr) = instr_in_epilog;
  }

exit_from_do_one_translation:

  /****************************************************/
  /** Dump the translated code out of the data cache **/
  /****************************************************/
  /** (*synch_cache_range)(translated_bb_start_address, next_xlate_address); **/
  _sync_cache_range(translated_bb_start_address, 4*(next_xlate_address - translated_bb_start_address));

  IFDEBUG(warn("[do_xlate] Done with do_one_translation ... returning\n");
	  fflush(mystderr); 
	  );
  
  return next_xlate_address;
}


