#include <limits.h>
#include <posix_stat.h>
#include <errno.h>
#ifdef BUILD_FFGZ
#include "zlib.h"
#endif

#include "aix_ldr.h"

#include "globtypes.h"
#include "opc.h"

#include "ansi_stdio.h"
#include "ansi_stdlib.h"
#include "ansi_string.h"

extern int do_xor_iaddress;
extern int do_xor_daddress;
extern int do_xor_seg13;

void print_usage(int argc, 
		 char * argv[]
		 );

int  read_input_flags(int argc, 
		      char * argv[], 
		      int * trace_posn,
		      int * trace_count, 
		      int * model_posn,
		      int * model_count
		      );
void ffinit( FILE * );

#if( BUILD_FFGZ )
#define FF_MARKER "--FFGZTRACE"
#else
#define FF_MARKER "--FFTRACE"
#endif

int
main(
	int argc,
	char * argv[],
	char * envp
	)
{
  char *	argv0_name;
  func_ptr     	model_ptr;
  char *	model_name;

  int    	model_argc = 0;
  char ** 	model_argv = NULL;
  char  * 	model_envp = NULL;
  int		model_arg_posn = 0;

  char *	trace_name;
  FILE *	trace_fp;
  int		trace_arg_posn = 0;
  int		trace_argc = 0;
  char **	trace_argv = NULL;

  int		n;
  int		status;
  int		return_value; 
  struct stat	file_status;

  /** Check the number of arguments **/
  argv0_name = argv[0];
  if( argc < 2 ) {
    print_usage(argc, argv);
    return EXIT_FAILURE;
  }

  /** Read in the input (command line) switch values **/
  status = read_input_flags(argc, argv,
			&trace_arg_posn, &trace_argc,
			&model_arg_posn, &model_argc);
  if( status != 0) {
    fprintf(stderr, "Problem reading the command-line input... Exiting\n");
    print_usage(argc, argv);
    return EXIT_FAILURE;
  }

  /** Set pointers into argv[] for trace and model sections **/
  model_name = argv[model_arg_posn];
  model_argv = &argv[model_arg_posn];
  model_envp = envp;

  /** Set up the disassembler **/
  opc_init();

  /** Check that these files do not have global access permissions; **/
  /**  NOTE: global permissions allow the file to be loaded into    **/
  /**   segment 0xd, and do not unload the program nicely at exit.  **/
  if (stat(model_name, &file_status) == 0) {
    if( (file_status.st_mode & S_IROTH) ) {
      fprintf(stderr,
	"ARIA: ERROR: You should not run Aria with a globally-read--enabled timer model (%s)\n",
	model_name);
      return EXIT_FAILURE;
    }
  }
  else {
    fprintf(stderr, "ARIA: ERROR: Cannot check the status of xcoff file %s\n", model_name);
    return EXIT_FAILURE;
  }

  /***********************************************/
  /** Get the extent of the trace program	**/
  /***********************************************/
  trace_argv = &argv[trace_arg_posn];
  trace_name = argv[trace_arg_posn];
  
  #if( BUILD_FFGZ )
    trace_fp = gzopen( trace_name, "rb");
  #else
    trace_fp = fopen( trace_name, "rb");
  #endif

  if( trace_fp == 0 ) {
    fprintf( stderr, "%s: could not open trace %s\n",
	     argv0_name, trace_name);
    return EXIT_FAILURE;
  }
  ffinit( trace_fp );

  /**********************************/
  /** Now do the load of the model **/
  /**********************************/
  model_ptr = load( model_name, 1, 0);

  if( model_ptr == 0 ) {
    fprintf( stderr, "%s: could not load model %s, errno=%d\n",
	     argv0_name, model_name, errno);
    return EXIT_FAILURE;
  }

  /** Invoke the user model (driver) program **/
  return_value = (int) model_ptr(model_argc, model_argv, model_envp);

  return return_value;
}

/*******************************************************************
  PROGRAM NAME..... print_usage

  FUNCTION ........ Dumps out information about the invocation of 
     aria, and the options it supports.

  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

void
print_usage(int argc,
	    char * argv[]
	    )
{
  fprintf(stderr, "usage: %s <options> --MODEL <model> %s <fftrace>\n", argv[0], FF_MARKER);
  fprintf(stderr, "\n  Where the\n");
  fprintf(stderr, "      <model> is the consumer of the aria instruction trace buffer data\n");
  fprintf(stderr, "      <fftrace> is the trace to be processed, and \n");
  fprintf(stderr, "      <options> can be any of:\n");

  fprintf(stderr, "%30s : xor segment ids\n",    "-xor");
  fprintf(stderr, "%30s : xor segment ids on data\n",    "-ixor");
  fprintf(stderr, "%30s : xor segment ids on insn\n",    "-dxor");
  fprintf(stderr, "%30s : do not xor segment 13 for insn\n",    "-noixor13");
}


/*******************************************************************
  PROGRAM NAME..... read_input_flags

  FUNCTION ........ Parses the input command-line arguments, and 
     sets (global) variables to indicate the flag conditions that 
     are requested by the user.

  NOTES............
  ASSUMPTIONS... 
  GLOBAL-VARS... 
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... 0 = good, other = error in parsing
  CALLED-ROUTINES.. None
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... None

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... April, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/

int
read_input_flags(int    argc, 
		 char * argv[],
		 int  * trace_arg_posn,
		 int  * trace_argc,
		 int  * model_arg_posn,
		 int  * model_argc
		 ) 
{
  int ret_val = 0;
  int flag_type = 0;
  int trace_argc_count  = 0;
  int model_argc_count = 0;

  int i = 1;

  while (i < argc) {
    switch (flag_type) {
    case 1:
      if (strcmp(argv[i], FF_MARKER) == 0) {
	i++;
	if (i < argc) {
	  *trace_arg_posn = i;
	  trace_argc_count = 1;
	  flag_type = 2;
	}
	else {
	  fprintf(stderr, "MUST specify the trace name after the %s flag\n", FF_MARKER);
	  ret_val = 1;
	  goto exit_this_while_loop;
	}
      }
      else {
	model_argc_count++;
      }
      break;

    case 2:
      trace_argc_count++;
      break;

    case 0:
      if (strcmp(argv[i], "--MODEL") == 0) {
	i++;
	if (i < argc) {
	  *model_arg_posn = i;
	  flag_type = 1;
	  model_argc_count = 1;
	}
	else {
	  fprintf(stderr, "MUST specify the model name after the --MODEL flag\n");
	  ret_val = 1;
	  goto exit_this_while_loop;
	}
      }
      else if( strcmp(argv[i], "-xor") == 0 ) {
	do_xor_iaddress = 1;
	do_xor_daddress = 1;
      }
      else if( strcmp(argv[i], "-ixor") == 0 ) {
	do_xor_iaddress = 1;
      }
      else if( strcmp(argv[i], "-dxor") == 0 ) {
	do_xor_daddress = 1;
      }
      else if( strcmp(argv[i], "-noixor13") == 0 ) {
	do_xor_seg13 = 0;
      }
      else {
        fprintf(stderr, "Unrecognized switch in Aria read_input_flags: '%s'\n",
		argv[i]);
	ret_val = 1;
	goto exit_this_while_loop;
      }
      break;

    default:
      fprintf(stderr, "Illegal flag type in read_input_flags of Aria\n");
      fprintf(stderr, "   the flag is: '%s'\n", argv[i]);
      ret_val = 1;
      break;
    }
    i++;
  }

exit_this_while_loop:

  if (model_argc_count == 0) {
    fprintf(stderr, "NOTE: Did not find a --MODEL entry and model name...\n");
    ret_val = 1;
  }

  if (trace_argc_count == 0) {
    fprintf(stderr, "NOTE: Did not find a %s entry and trace program name...\n", FF_MARKER);
    ret_val = 1;
  }

  *model_argc = model_argc_count;
  *trace_argc  = trace_argc_count;

  return ret_val;
}
