/*                                                                     
** COMPONENT_NAME: MET / Aria
**                                                                      
** Licensed Materials - Property of IBM                                  
**                                                                     
** IBM Confidential (IBM Confidential-Restricted when combined          
**                   with aggregated modules for this product)          
**                                                                     
**                                                                     
** OCO Source Materials                                                
**                                                                     
**                                                                     
** GA-0100 (C) Copyright IBM Corp. 1997                                 
**                                                                     
** US Government Users Restricted RIghts - Use, duplication or         
** disclosure restricted by GSA ADP Schedule Contract with IBM Corp.   
**                                                                     
** The source code for this program is not published or otherwise      
** divested of its trade secrets, irrespective of what has been        
** deposited within the U.S. Copyright Office.                         
**                                                                     
*/

 
/*******************************************************************
  PROGRAM NAME..... get_null_sc_timing
  FUNCTION ........ Creates a translation wrapper for a user-program
     supervisor call.  This wrapper has a predefined form for both 
     the taken-path and not-taken path codes.  The taken-path code
     will check whether this is a call to exit from the program, and
     otherwise set up the user state, and invoke the underlying 
     supervisor function.  When control returns to this translated
     code, the Aria and user states are swapped and control is 
     handed back to Aria.  To be safe, ALL user process state is 
     restored prior to the supervisor call, and ALL state is 
     saved after the supervisor call returns.  

  NOTES............
  ASSUMPTIONS... The area at TEMP_SPOT off the user stack (in this 
     case -216 bytes down from the current user stack pointer) is 
     NOT in use by the user program; it is overwritten during the 
     swapping of the user and aria states.
  GLOBAL-VARS... None
  RESTRICTIONS.. None
  DEVIATIONS.... None
  PROGRAM TYPE..... C
  RETURN CODES..... the next available translation address
  CALLED-ROUTINES.. 
  CALLED-BY........ None
  INCLUDED-BY...... None
  MACROS-USED...... None
  INCLUDES......... <see below> plus 
     xlation_alignment_code.c, 
     save_aria_state_code.c, bb_prolog_code.c, bb_epilog_code.c, 
     restore_aria_state_code.c

  AUTHOR........... John-David Wellman, IBM Yorktown
  CREATION DATE.... June, 1997
  MODIFIED......... Who       Date      What
  
  ********************************************************************/
#include <stdio.h>
#include <sys/time.h>

#include "aria_xlate_macros.h"

#include "opc.h"

#include "proc_state.h"
#include "aria_state.h"

#include "io_msgs.h"

#include "dump_description.h"

#include "globtypes.h"
#include "global_variables.h"

#include "get_null_sc_timing.h"

#include "display_reg_usedef.h"

#include "interface_globals.h"

#define TEMP_SPOT  -216

#define NUMBER_OF_TIMER_OVERHEAD_TRIALS 50

int
get_null_sc_timing_overhead(proc_state_type * dummy_proc_state_ptr)
{
  unsigned * xlate_ptr;
  unsigned * xl_start_address;

  unsigned timebase_struct_address;
  unsigned * ta_ptr;

  unsigned base_register;
  unsigned scratch_reg;

  unsigned * microtrace_buffer_address;

  /** Set up a dummy microtrace buffer address so later code can access it **/
  {
    unsigned br_addr = 0x00000000;
    unsigned iword   = OPC_ASM_BCLR(0x14, 0);
    unsigned iword_index = opc_index(iword);
    unsigned iword_kind  = opc_itype(iword_index);
    unsigned iword_daddr    = 0xdeadbeef;
    unsigned iword_dlength  = 0;
    
    *next_xlate_address  = 0x00000000;        next_xlate_address++;
    
    microtrace_buffer_address = next_xlate_address;    
    
    *next_xlate_address  = br_addr;           next_xlate_address++;
    *next_xlate_address  = iword;             next_xlate_address++;
    *next_xlate_address  = iword_kind;        next_xlate_address++;
    *next_xlate_address  = iword_daddr;       next_xlate_address++;
    *next_xlate_address  = iword_dlength;     next_xlate_address++;
    *next_xlate_address  = 0x00000000;        next_xlate_address++;
  }

  /** Set up a space for the xlation address (it's used below) **/
  *next_xlate_address = 0x00000000; next_xlate_address ++; 
  next_xlate_address += 10;
  *next_xlate_address = 0x00000000; next_xlate_address ++; 

  /* Create a space for the timebasestruct information on calls to read_real_time **/
  /* define the time base structure space (with appropriate markers) */
  ta_ptr = next_xlate_address;
  *next_xlate_address = 0x00ffff00; next_xlate_address++;
  timebase_struct_address = (unsigned) next_xlate_address;
  next_xlate_address += (TIMEBASE_SZ>>2);
  *next_xlate_address = 0x00ffff00; next_xlate_address++;
  
  if (do_no_output_during_runs == 0) {
    int iii;
    warn_xlate("\nThe space for the timebase structure is:\n");
    warn_xlate("\t0x%08x  %08x\n", ta_ptr, *ta_ptr);  ta_ptr++;
    for (iii = 0; iii < (TIMEBASE_SZ>>2); iii++) {
      warn_xlate("\t0x%08x  %08x\n", ta_ptr, *ta_ptr);  ta_ptr++;
    }
    warn_xlate("\t0x%08x  %08x\n", ta_ptr, *ta_ptr);  ta_ptr++;      
    warn_xlate("\n");
  }

  /** Update the addresses for the program and translation space **/
  #include "xlation_alignment_code.c"

  xl_start_address = next_xlate_address;

  /********************************************************/
  /** Save off all of Aria's important (clobbered) state **/
  /********************************************************/

  base_register = 12;
  scratch_reg   = 11;
  
  /* Load GPR base_register with the address of our ARIA state saving space */          
  *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16); next_xlate_address++; 
  
  /* Save off the input parameters */
  *next_xlate_address = STORE_REG ( 3, PROC_STATE_OFFSET,  base_register);     next_xlate_address++;
  
  if (not_taken_address_required_flag == 1) { 
    *next_xlate_address = STORE_REG ( 4, NOT_TAKEN_RVAL_OFFSET, base_register);  next_xlate_address++;
  }
  
  /** Create a pointer to the trace buffer **/
  if ( (do_dynamic_trace_generation) ) {
    unsigned static_hi = ((unsigned)microtrace_buffer_address & 0xffff0000) >> 16;
    unsigned static_lo = ((unsigned)microtrace_buffer_address & 0x0000ffff);
    *next_xlate_address = OPC_ASM_ADDIS( 4, 0, static_hi);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_ORI  ( 4, 4, static_lo);  next_xlate_address++;
    *next_xlate_address = STORE_REG ( 4, PROC_DYN_TRBUF_OFFSET, base_register);  next_xlate_address++;
  }

  /*************************************/
  /** Do the saving of the Aria state **/
  /*************************************/
  {
    unsigned gprs = 0xffffffff;
    unsigned fprs = 0xffffffff;
    unsigned crs  = 0xffffffff;
    
    #include "save_aria_state_code.c"
  }
  
  if (do_no_output_during_runs == 0) {
    warn_xlate(" (aria prolog) ------------------------------------------------------------\n");
    DUMP_DESCRIPTION(xl_start_address, next_xlate_address);    
    xlate_ptr = next_xlate_address;
  }

  /** If the user wants the times for the execution, get them... **/
  if (global_get_system_call_execution_times == 1) {
    unsigned address;
    unsigned addr_hi;
    unsigned addr_lo;
    unsigned * Aaddr = (unsigned *) (ARIA_STATE_SAVE_SPACE + ARIA_READ_REAL_TIME);
    
    /* Set the CTR to branch through */
    address = *(Aaddr);
    addr_hi = (address & 0xffff0000) >> 16;
    addr_lo = (address & 0x0000ffff);
    *next_xlate_address = OPC_ASM_ADDIS(3, 0, addr_hi);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_ORI  (3, 3, addr_lo);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_MTSPR(9, 3);           next_xlate_address++;
    
    /* Set the TOC value */
    address = *(Aaddr+1);
    addr_hi = (address & 0xffff0000) >> 16;
    addr_lo = (address & 0x0000ffff);
    *next_xlate_address = OPC_ASM_ADDIS(2, 0, addr_hi);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_ORI  (2, 2, addr_lo);  next_xlate_address++;
    
    /* Set the first parameter */
    address = timebase_struct_address;
    addr_hi = (address & 0xffff0000) >> 16;
    addr_lo = (address & 0x0000ffff);
    *next_xlate_address = OPC_ASM_ADDIS(3, 0, addr_hi);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_ORI  (3, 3, addr_lo);  next_xlate_address++;
    
    /* Set the second parameter */
    *next_xlate_address = OPC_ASM_ADDI(4, 0, TIMEBASE_SZ);  next_xlate_address++;
    
    /* branch and link */
    *next_xlate_address = OPC_ASM_BCCTRL(20, 0);  next_xlate_address++;
    
    /* Save the returned data in the aria state */
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16); next_xlate_address++; 
    *next_xlate_address = LOAD_REG(5, 0, 4);  next_xlate_address++;
    *next_xlate_address = STORE_REG(5, ARIA_SAVE_START_TIME, base_register);  next_xlate_address++;
    *next_xlate_address = LOAD_REG(5, 4, 4);  next_xlate_address++;
    *next_xlate_address = STORE_REG(5, ARIA_SAVE_START_TIME+4, base_register);  next_xlate_address++;
    *next_xlate_address = LOAD_REG(5, 8, 4);  next_xlate_address++;
    *next_xlate_address = STORE_REG(5, ARIA_SAVE_START_TIME+8, base_register);  next_xlate_address++;
    
    if (do_no_output_during_runs == 0) {
      warn_xlate(" (time base prolog) -------------------------------------------------------\n");
      DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
      warn_xlate("read_real_time = 0x%08x TOC = %08x tbasestruct = 0x%08x\n",
		 *(Aaddr), *(Aaddr+1), timebase_struct_address);
      xlate_ptr = next_xlate_address;
    }
  }

  /* Load base_register with the processor state save space address from the reserved work area */
  *next_xlate_address = LOAD_REG ( base_register,  0, base_register);              next_xlate_address++;  
  
  /****************************************************************************/
  /** Insert the bb prolog, which inserts instructions to restore user state **/
  /****************************************************************************/
  {
    unsigned gpr_uses   = 0xffffffff;
    unsigned fpr_uses   = 0xffffffff;
    unsigned cr_touches = 0xffffffff;
    unsigned spr_uses   = 0xffffffff;
    unsigned spr_defs   = 0xffffffff;
    
    #include "bb_prolog_code.c"
  }
    
  if (do_no_output_during_runs == 0) {
    warn_xlate("* (prolog) ----------------------------------------------------------------\n");
    DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);    
    xlate_ptr = next_xlate_address;
    warn_xlate("* (svc prep translated body) ==============================================\n");
  }
  
  /*********************************************************/
  /** Now generate some code for the SVC call basic block **/
  /*********************************************************/

  /*******************************************************************/
  /** Check the value in GPR 2 (the system call routine identifier) **/
  /**  to see if the program is trying to exit; the exit routine    **/
  /**  flags differs by operating system version, so check for any. **/
  /*******************************************************************/
  /** Save GPR 12 on the stack (so we can use it for some case checking) **/
  *next_xlate_address = STORE_REG  ( 12,  TEMP_SPOT, STACK_PTR);     next_xlate_address++;
  *next_xlate_address = OPC_ASM_MFCR ( 12);                            next_xlate_address++;
  *next_xlate_address = OPC_ASM_CMPI (  7,0,  2, the_exit_code_value); next_xlate_address++;
  
  /********************************************************/
  /** If the GPR2 code is not EXIT, go to the BLA code   **/
  /**     (i.e. skip the exit code below)                **/
  /** The branch is inserted at the end of the exit code **/
  /**  (below) in space saved at branch_over_address     **/
  /********************************************************/
  {
    unsigned * branch_over_address = next_xlate_address;
    unsigned instrs_in_exit_code = 0;
    
    /** Skip where the conditional branch will be placed **/
    next_xlate_address++;
    
    /** Insert the bb epilog, which inserts instructions to save new user state **/
    {
      unsigned   gpr_defs   = 0xffffffff;
      unsigned   fpr_defs   = 0xffffffff;
      unsigned   cr_touches = 0xffffffff;
      unsigned   spr_uses   = 0xffffffff;
      unsigned   spr_defs   = 0xffffffff;
      
      #include "bb_epilog_code.c"
    }
    
    /** Load the scratch register with the link register value **/
    *next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, base_register);    next_xlate_address++;
    if ( (do_dynamic_trace_generation) && (do_iaddr_fixup) ) {
      unsigned hi_offset = (global_user_iaddr_offset & 0xffff0000) >> 16;
      unsigned lo_offset = (global_user_iaddr_offset & 0x0000ffff);
      *next_xlate_address = OPC_ASM_RLWINM(scratch_reg, scratch_reg, 4, 28, 31);       next_xlate_address++;
      *next_xlate_address = OPC_ASM_CMPI ( 7, 0, scratch_reg, ((global_user_program_iaddr_mask >> 28) & 0xf));                    next_xlate_address++;
      *next_xlate_address = OPC_ASM_BC   ( 0x4, 30, 5);                                next_xlate_address++;
      *next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, base_register);  next_xlate_address++;
      *next_xlate_address = OPC_ASM_ADDIS( scratch_reg, scratch_reg, -hi_offset);      next_xlate_address++;
      *next_xlate_address = OPC_ASM_ADDI ( scratch_reg, scratch_reg, -lo_offset);      next_xlate_address++;
      *next_xlate_address = OPC_ASM_B    ( 2);                                         next_xlate_address++;
      *next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, base_register);  next_xlate_address++;      
    }
    
    /********************************/
    /** Now restore the Aria state **/
    
    /* Load GPRbase with the address of our ARIA state saving space    */               
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);               next_xlate_address++; 
    
    /* Store the dynamic trace information (if user wants it) */
    if (do_dynamic_trace_generation) {
      *next_xlate_address = LOAD_REG  (base_register, PROC_DYN_TRBUF_OFFSET, base_register);  next_xlate_address++;
      *next_xlate_address = STORE_REG  (scratch_reg, 12, base_register);                       next_xlate_address++;
      *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);             next_xlate_address++; 
    }
    
    /******************************************/
    /** Do the restoration of the Aria state **/
    /******************************************/
    {
      unsigned gprs = 0xffffffff;
      unsigned fprs = 0xffffffff;
      unsigned crs  = 0xffffffff;
      
      #include "restore_aria_state_code.c"
    }
    
    /**********************************************************************************************/
    /* Set the next_address return value to be address 0x00000000 (indicates user program exits) **/ 
    *next_xlate_address = OPC_ASM_ADDIS( 3, 0, 0);           next_xlate_address++; 
    
    if (not_taken_address_required_flag == 1) { 
      /* And the not-taken next address */
      *next_xlate_address = LOAD_REG  (10, NOT_TAKEN_RVAL_OFFSET, 12);  next_xlate_address++; 
      *next_xlate_address = STORE_REG  ( 3, 0, 10);                      next_xlate_address++; 
    }
    
    /* Replace the branch in the original code with code to return through link register branch */
    *next_xlate_address = OPC_ASM_BCLR (0x14, 0);        next_xlate_address++;
    
    /** Determine the number of instructions to branch over this exit code **/
    instrs_in_exit_code = (next_xlate_address - branch_over_address);
    *branch_over_address = OPC_ASM_BC ( 0x05, 7*4+2, instrs_in_exit_code);
    
    if (do_no_output_during_runs == 0) {
      warn_xlate("* (exit branch code) ------------------------------------------------------\n");
      DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
      xlate_ptr = next_xlate_address;
    }
  }
  
  /***********************************************************/
  /** ELSE, not an EXIT call, so do the underlying SVC call **/
  /***********************************************************/
  
  *next_xlate_address = OPC_ASM_MTCRF(0xff, 12);                             next_xlate_address++; 
  *next_xlate_address = LOAD_REG  ( 12, TEMP_SPOT, STACK_PTR);            next_xlate_address++;
  *next_xlate_address = OPC_ASM_B  ( 1 );                                    next_xlate_address++;
  
  if (do_no_output_during_runs == 0) {
    warn_xlate("* (sc body) ---------------------------------------------------------------\n");
    DUMP_FAKE_DESCRIPTION(xlate_ptr, next_xlate_address);
    xlate_ptr = next_xlate_address;
  }
  
  *next_xlate_address = STORE_REG  ( 12,  TEMP_SPOT, STACK_PTR);        next_xlate_address++;  
  *next_xlate_address = OPC_ASM_ADDIS( 12,  0, ARIA_STATE_SAVE_HI_16);    next_xlate_address++;  
  *next_xlate_address = LOAD_REG  ( 12,  0, 12);                       next_xlate_address++;  
  *next_xlate_address = LOAD_REG  ( 12, LR_SAVE_OFFSET, 12);           next_xlate_address++;  
  *next_xlate_address = OPC_ASM_MTSPR(  8, 12);                           next_xlate_address++; 
  *next_xlate_address = LOAD_REG  ( 12,  TEMP_SPOT, STACK_PTR);        next_xlate_address++;  
  
  /*****************************************************************************/
  /** Insert the bb epilog, which inserts instructions to save new user state **/
  /*****************************************************************************/
  
  {
    unsigned   gpr_defs   = 0xffffffff;
    unsigned   fpr_defs   = 0xffffffff;
    unsigned   cr_touches = 0xffffffff;
    unsigned   spr_uses   = 0xffffffff;
    unsigned   spr_defs   = 0xffffffff;
    
#include "bb_epilog_code.c"
  }
  
  if (do_no_output_during_runs == 0) {
    warn_xlate("* (BB epilog) -------------------------------------------------------------\n");
    DUMP_FAKE_DESCRIPTION(xlate_ptr, next_xlate_address);
    xlate_ptr = next_xlate_address;
  }
  
  /** If the user wants the times for the execution, get them... **/
  if (global_get_system_call_execution_times == 1) {
    unsigned address;
    unsigned addr_hi;
    unsigned addr_lo;
    unsigned * Aaddr = (unsigned *) (ARIA_STATE_SAVE_SPACE + ARIA_READ_REAL_TIME);
      
    /* Set the CTR to branch through */
    address = *(Aaddr);
    addr_hi = (address & 0xffff0000) >> 16;
    addr_lo = (address & 0x0000ffff);
    *next_xlate_address = OPC_ASM_ADDIS(3, 0, addr_hi);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_ORI  (3, 3, addr_lo);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_MTSPR(9, 3);           next_xlate_address++;

      /* Set the TOC value */
    address = *(Aaddr+1);
    addr_hi = (address & 0xffff0000) >> 16;
    addr_lo = (address & 0x0000ffff);
    *next_xlate_address = OPC_ASM_ADDIS(2, 0, addr_hi);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_ORI  (2, 2, addr_lo);  next_xlate_address++;

    /* Set the first parameter */
    address = timebase_struct_address;
    addr_hi = (address & 0xffff0000) >> 16;
    addr_lo = (address & 0x0000ffff);
    *next_xlate_address = OPC_ASM_ADDIS(3, 0, addr_hi);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_ORI  (3, 3, addr_lo);  next_xlate_address++;
      
    /* Set the second parameter */
    *next_xlate_address = OPC_ASM_ADDI(4, 0, TIMEBASE_SZ);  next_xlate_address++;

    /* branch and link */
    *next_xlate_address = OPC_ASM_BCCTRL(20, 0);  next_xlate_address++;
      
    /* Save the returned data in the aria state */
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16); next_xlate_address++; 
    *next_xlate_address = LOAD_REG(5, 0, 4);  next_xlate_address++;
    *next_xlate_address = STORE_REG(5, ARIA_SAVE_FINISH_TIME, base_register);  next_xlate_address++;
    *next_xlate_address = LOAD_REG(5, 4, 4);  next_xlate_address++;
    *next_xlate_address = STORE_REG(5, ARIA_SAVE_FINISH_TIME+4, base_register);  next_xlate_address++;
    *next_xlate_address = LOAD_REG(5, 8, 4);  next_xlate_address++;
    *next_xlate_address = STORE_REG(5, ARIA_SAVE_FINISH_TIME+8, base_register);  next_xlate_address++;

    /* Get back saved proc state addressibility */
    *next_xlate_address = LOAD_REG(base_register, PROC_STATE_OFFSET, base_register); next_xlate_address++; 

    if (do_no_output_during_runs == 0) {
      warn_xlate(" (time base prolog) -------------------------------------------------------\n");
      DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
      warn_xlate("read_real_time = 0x%08x TOC = %08x tbasestruct = 0x%08x\n",
		 *(Aaddr), *(Aaddr+1), timebase_struct_address);
      xlate_ptr = next_xlate_address;
    }
  }

  /** Load the scratch register with the link register value **/
  *next_xlate_address = LOAD_REG (scratch_reg, LR_SAVE_OFFSET, base_register);     next_xlate_address++;
  if ( (do_dynamic_trace_generation) && (do_iaddr_fixup) ) {
    unsigned hi_offset = (global_user_iaddr_offset & 0xffff0000) >> 16;
    unsigned lo_offset = (global_user_iaddr_offset & 0x0000ffff);
    *next_xlate_address = OPC_ASM_RLWINM(scratch_reg, scratch_reg, 4, 28, 31);        next_xlate_address++;
    *next_xlate_address = OPC_ASM_CMPI ( 7, 0, scratch_reg, ((global_user_program_iaddr_mask >> 28) & 0xf));                     next_xlate_address++;
    *next_xlate_address = OPC_ASM_BC   ( 0x4, 30, 5);                                 next_xlate_address++;
    *next_xlate_address = LOAD_REG  (scratch_reg, LR_SAVE_OFFSET, base_register);  next_xlate_address++;
    *next_xlate_address = OPC_ASM_ADDIS( scratch_reg, scratch_reg, -hi_offset);       next_xlate_address++;
    *next_xlate_address = OPC_ASM_ADDI ( scratch_reg, scratch_reg, -lo_offset);       next_xlate_address++;  
    *next_xlate_address = OPC_ASM_B    ( 2);                                          next_xlate_address++;
    *next_xlate_address = LOAD_REG  (scratch_reg, LR_SAVE_OFFSET, base_register);  next_xlate_address++;
  }

  /* Load GPRbase with the address of our ARIA state saving space    */               
  *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);      next_xlate_address++; 

  /* Store the dynamic trace information (if appropriate) */
  if (do_dynamic_trace_generation) {
    *next_xlate_address = LOAD_REG (base_register, PROC_DYN_TRBUF_OFFSET, base_register);  next_xlate_address++;
    *next_xlate_address = STORE_REG (scratch_reg, 12, base_register);                       next_xlate_address++;
    *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);            next_xlate_address++; 
  }
    
  /** Store the (possibly adjusted) link register value as the taken and not-taken addresses **/
  *next_xlate_address = LOAD_REG  (base_register, PROC_STATE_OFFSET, base_register);  next_xlate_address++; 
  *next_xlate_address = STORE_REG  (scratch_reg, NEXT_TARGET_OFFSET, base_register);   next_xlate_address++;
  if (not_taken_address_required_flag == 1) {   
    *next_xlate_address = STORE_REG  (scratch_reg, NOT_TAKEN_OFFSET, base_register);     next_xlate_address++;
  }
    
  if (do_no_output_during_runs == 0) {
    warn_xlate("* (exit) ------------------------------------------------------------------\n");
    DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
    xlate_ptr = next_xlate_address;
  }

  /********************************/
  /** Now restore the Aria state **/
  /********************************/
    
    /* Load GPRbase_register with the address of our ARIA state saving space    */               
  *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);         next_xlate_address++; 

  /******************************************/
  /** Do the restoration of the Aria state **/
  /******************************************/
  {
    unsigned gprs = 0xffffffff;
    unsigned fprs = 0xffffffff;
    unsigned crs  = 0xffffffff;
    
    #include "restore_aria_state_code.c"
  }

  /**********************************************************/
  /* Determine the next basic block address and the         */
  /*   next not taken basic block address.                  */
  /**********************************************************/
  base_register = 12;
  *next_xlate_address = OPC_ASM_ADDIS(base_register,  0, ARIA_STATE_SAVE_HI_16);               next_xlate_address++;

  if (not_taken_address_required_flag == 1) { 
    *next_xlate_address = LOAD_REG  (10, NOT_TAKEN_RVAL_OFFSET, base_register);             next_xlate_address++; 
    *next_xlate_address = LOAD_REG  (base_register,  PROC_STATE_OFFSET, base_register);     next_xlate_address++;
    *next_xlate_address = LOAD_REG  (11, NOT_TAKEN_OFFSET, base_register);                  next_xlate_address++;
    *next_xlate_address = STORE_REG (11, 0, 10);                                            next_xlate_address++; 
  }
  else {
    *next_xlate_address = LOAD_REG  (base_register,  PROC_STATE_OFFSET, base_register);     next_xlate_address++;
  }

  *next_xlate_address = LOAD_REG  ( 3, NEXT_TARGET_OFFSET, base_register);                  next_xlate_address++;

    /********************************************************************************************/
    /* Replace the branch in the original code with code to return through link register branch */
    /********************************************************************************************/
  *next_xlate_address = OPC_ASM_BCLR (0x14, 0);        next_xlate_address++;

  /* Output the translated basic block exit code */
  if (do_no_output_during_runs == 0) {
    DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
    xlate_ptr = next_xlate_address;
  }

  if (do_no_output_during_runs == 0) {
    warn_xlate("* (Aria restore) ----------------------------------------------------------\n");
    DUMP_DESCRIPTION(xlate_ptr, next_xlate_address);
    warn_xlate("===============================================================================\n");
    warn_xlate("    end of BB %d [%u user, %u xlated instr] ; oa=%08x xa = %08x\n", 
	       global_bb_index, 0,
	       (next_xlate_address - xl_start_address),
	       0, xl_start_address);
  }

  /** Dump the translated code out of the data cache **/
  _sync_cache_range(xl_start_address, 4*(next_xlate_address - xl_start_address));

  /**********************************************************/
  /** Execute the translated basic block several times and **/
  /**  determine a timing overhead estimate.               **/
  /**********************************************************/
  pc_func_desc(prog_desc) = xl_start_address;
  
  /** The copy to run when debugging output desired **/
  if ( (output_after_dynamic_bb_execution <= global_bb_executions) 
       && (no_output_after_dynamic_bb_execution  >= global_bb_executions) 
       && (output_after_static_bb_translation <= global_bb_index) 
       && (no_output_after_static_bb_translation >= global_bb_index) ) {
    warn_trace("\nExecuting the timing-overhead translated code:\n");
    warn_exec("\nExecuting the timing-overhead translated code:\n");
    warn_exec("     Starting address %08x:\n",   pc_func_desc(prog_desc));
    warn_exec("     TOC address      %08x:\n",   toc_func_desc(prog_desc));
    warn_exec("     Other data       %08x:\n", other_func_desc(prog_desc));
    warn_processor_state(dummy_proc_state_ptr, show_proc_float_state);
  }

  {
    timebasestruct_t * Saddr  = (timebasestruct_t *) (ARIA_STATE_SAVE_SPACE + ARIA_SAVE_START_TIME);      
    timebasestruct_t * Faddr  = (timebasestruct_t *) (ARIA_STATE_SAVE_SPACE + ARIA_SAVE_FINISH_TIME);      
    timebasestruct_t * Naddr  = (timebasestruct_t *) (ARIA_STATE_SAVE_SPACE + ARIA_SAVE_NULL_TIME);      
    int t1[NUMBER_OF_TIMER_OVERHEAD_TRIALS];
    int t2[NUMBER_OF_TIMER_OVERHEAD_TRIALS];
    int i;
    int w1, w2;

    /* set a non-exit value for the GPR 2 value */
    proc_state_GPR(dummy_proc_state_ptr, 2) = the_exit_code_value - 1;
    for (i = 0; i < NUMBER_OF_TIMER_OVERHEAD_TRIALS; i++) {
      unsigned * taken_path;
      unsigned * not_taken_path;
      taken_path = (unsigned *)prog_ptr(dummy_proc_state_ptr, &not_taken_path);

      if ( (output_after_dynamic_bb_execution <= global_bb_executions) && 
	   (no_output_after_dynamic_bb_execution  >= global_bb_executions) && 
	   (output_after_static_bb_translation <= global_bb_index) &&
	   (no_output_after_static_bb_translation >= global_bb_index) ) {
	warn_exec("  ... Done with run %d\n", i);
	warn_exec("Saddr = %08x %08x %08x\n", Saddr->flag, Saddr->tb_high, Saddr->tb_low);
	warn_exec("Faddr = %08x %08x %08x\n", Faddr->flag, Faddr->tb_high, Faddr->tb_low);
      }
      time_base_to_time(Saddr, TIMEBASE_SZ);
      time_base_to_time(Faddr, TIMEBASE_SZ);

      w1 = Faddr->tb_high - Saddr->tb_high;
      w2 = Faddr->tb_low  - Saddr->tb_low;
      if (w2 < 0) { 
	w1 --;
	w2 += 1000000000;
      }
      
      if (w1 < 0) {
	w1 = 0;
      }

      t1[i] = w1;
      t2[i] = w2;
    }
    warn_trace("  ... Done with %d runs\n", i);

    /***************************************/
    /** Select the representative offsets **/
    /***************************************/
    
    /* find the minimum execution time required */
    for (i = 0; i < NUMBER_OF_TIMER_OVERHEAD_TRIALS; i++) {
      if (t1[i] < w1) { 
	w1 = t1[i];
      }
      if (t2[i] < w2) { 
	w2 = t2[i];
      }
    }

    /* then average those near that minimum */
    /**
      for (i = 0; i < NUMBER_OF_TIMER_OVERHEAD_TRIALS; i++) {
      if ((t1[i] <= 2*w1) && (t2[i] <= 2*w2)) {
      warn("   (hi, low) = ( %08x , %08x )\n", t1[i], t2[i]);
      hi_total  += (double)t1[i];
      low_total += (double)t2[i];
      num_included++;
      }
      else {
      warn(" * (hi, low) = ( %08x , %08x )\n", t1[i], t2[i]);
      }
      }
      
      Naddr->tb_high = (int) (hi_total/num_included);
      Naddr->tb_low  = (int) (low_total/num_included);
      **/

    /* For now, take the minimum value so that we never get a negative time req'd */    
    Naddr->tb_high = w1;
    Naddr->tb_low  = w2;

    warn_trace("Selecting null overhead values of: %08x %08x\n", Naddr->tb_high, Naddr->tb_low);
  }
  
  return 0;
}






