/******************************************************************************
 This file reads in an xcoff using ravi's tool and dumps out the info required
 to read and expand an xtrace stream
******************************************************************************/

/******************************************************************************
  INCLUDES
******************************************************************************/

/*** STANDARD INCLUDES ***/
#include <cassert>
#include <cstdlib>
#include <iostream>
using namespace std; // jdonald

#if( 0 )
/*** RAVI'S INCLUDES ***/
#include "xcoff.H"
#include "ppcinstr.H"
#include "itype.H"
#include "deplist.H"
#endif

/*** LOCAL INCLUDES ***/
extern "C" {
#include "text_xcoff.h"
#include "opc.h"
#include "dep_prep_process.h"
#include "dep_prep_info.h"
}

/******************************************************************************
  LOCAL
******************************************************************************/

static const char * out_suffix = ".dep";


/******************************************************************************
  FUNCTIONS
******************************************************************************/

void
usage(
	char * program_name
	)
{
  cerr << "usage: " << program_name <<
  " [-a lg2arch] [-h] [-c] [-g] [-gm2|-gm3] [-d(a|c|f) ifetch param] [-(a|l)(a|b|c|d|e|f|l|m|s|x|D|F) avail/latency] [-o output] [-s(allcrl||fxldu|fpldu|fxstu|fpstu|fxst|fpst|mult|spr|string|stu2|show)]* [-u(b|f|l|s|x|L|X)] [-p [-pdist <dist>] [-pshow] [-pmode <[01]>] [--] <xcoff_file>"
  << endl;

  cerr << "\t-aa: extra load avail for arithmetic; default =" <<  ARITHMETIC_LOAD_AVAIL <<endl;
  cerr << "\t-ab: branch avail; default =" <<  BRANCH_AVAIL <<endl;
  cerr << "\t-ac: compare avail; default =" <<  CMP_AVAIL <<endl;
  cerr << "\t-ad: idiv avail; default =" << IDIV_AVAIL << endl;
  cerr << "\t-ae: extra load avail for sub words; default =" <<  SUBWORD_LOAD_AVAIL <<endl;
  cerr << "\t-af: fpu avail; default =" <<  FPU_AVAIL <<endl;
  cerr << "\t-al: load avail; default =" <<  LOAD_AVAIL <<endl;
  cerr << "\t-am: imul avail; default =" << IMUL_AVAIL << endl;
  cerr << "\t-as: store avail; default =" <<  STORE_AVAIL <<endl;
  cerr << "\t-ax: fix avail; default =" << FIX_AVAIL << endl;
  cerr << "\t-aD: fdiv avail; default =" << FDIV_AVAIL << endl;
  cerr << "\t-aF: extra load avail for floats; default =" << FLOAT_LOAD_AVAIL << endl;
  cerr << "\t-aL: logical avail; default =" << CRL_AVAIL << endl;

  cerr << "\t-da: fetch block alignment in words; default =" << IFETCH_ALIGN << endl;
  cerr << "\t-dc: cache line size in words; default =" << ICACHE_ALIGN << endl;
  cerr << "\t-df: fetch block size in words; default =" << IFETCH_MAX << endl;

  cerr << "\t-la: extra load latency for arithmetic; default =" <<  ARITHMETIC_LOAD_LATENCY <<endl;
  cerr << "\t-lb: branch latency; default =" <<  BRANCH_LATENCY <<endl;
  cerr << "\t-lc: compare latency; default =" <<  CMP_LATENCY <<endl;
  cerr << "\t-ld: idiv latency; default =" << IDIV_LATENCY << endl;
  cerr << "\t-le: extra load latency for sub words; default =" <<  SUBWORD_LOAD_LATENCY <<endl;
  cerr << "\t-lf: fpu latency; default =" <<  FPU_LATENCY <<endl;
  cerr << "\t-ll: load latency; default =" <<  LOAD_LATENCY <<endl;
  cerr << "\t-lm: imul latency; default =" << IMUL_LATENCY << endl;
  cerr << "\t-ls: store latency; default =" <<  STORE_LATENCY <<endl;
  cerr << "\t-lx: fix latency; default =" << FIX_LATENCY << endl;
  cerr << "\t-lD: fdiv latency; default =" << FDIV_LATENCY << endl;
  cerr << "\t-lF: extra load latency for floats; default =" << FLOAT_LOAD_LATENCY << endl;
  cerr << "\t-lL: logical latency; default =" << CRL_LATENCY << endl;

  cerr << "\t-sall: do not split any operations" << endl;
  cerr << "\t-scrl: do not split crlogical operations" << endl;
  cerr << "\t-sfxldu: do not split fix-point load-update operations" << endl;
  cerr << "\t-sfpldu: do not split floating-point load-update operations" << endl;
  cerr << "\t-sfxstu: do not split fix-point store-update operations" << endl;
  cerr << "\t-sfpstu: do not split floating-point store-update operations" << endl;
  cerr << "\t-sfxst: do not split fix-point store operations" << endl;
  cerr << "\t-sfpst: do not split floating-point store operations" << endl;
  cerr << "\t-smult: do not split multi-source/target operations" << endl;
  cerr << "\t-sspr: do not split spr operations" << endl;
  cerr << "\t-sstring: do not split string operations" << endl;
  cerr << "\t-sstu2: when splitting store-update operations, split into 2 [instead of 3]" << endl;
  cerr << "\t-sshow: show all splits" << endl;

  cerr << "\t-ub: unit for branch ops" << endl;
  cerr << "\t-uf: unit for float ops" << endl;
  cerr << "\t-ul: unit for load ops" << endl;
  cerr << "\t-us: unit for store ops" << endl;
  cerr << "\t-ux: unit for fix ops" << endl;
  cerr << "\t-uL: unit for logical ops" << endl;
  cerr << "\t-uX: unit for complex ops" << endl;

  cerr << "\t-a: lg2 of max arch registers; default =" << LG2ARCH_MAX << endl;
  cerr << "\t-g: turn off all the gp specific tweaks" << endl;
  cerr << "\t-gm2: turn on all gp, as specified by model-2" << endl;
  cerr << "\t-gm3: turn on all gp, as specified by model-3" << endl;

  cerr << "\t-p: turn on predication" << endl;
  cerr << "\t-pdist: set predication distance" << endl;
  cerr << "\t-pmode: set predication mode" << endl;
  cerr << "\t-pshow: show predication" << endl;
}

int
get_flag(
	char *	prog,
	int	argc,
	char *	argv[],
	int	i,
	char	test
	)
{
  assert( argv[i][0] == '-' );
  if( argv[i][1] == test ) {
    if( argv[i][2] != 0 ) {
      usage( prog );
      exit( EXIT_FAILURE );
    }
    return 1;
  }
  return 0; 
}

int
get_flag(
	char *	prog,
	int	argc,
	char *	argv[],
	int&	i,
	const char *	test
	)
{
  assert( argv[i][0] == '-' );
  if( strcmp( test, &argv[i][1]) == 0 ) {
    return 1;
  }
  return 0; 
}

int
get_flag_arg(
	char *	prog,
	int	argc,
	char *	argv[],
	int&	i,
	const char *	test,
	unsigned&	result
	)
{
  unsigned len = strlen(test);

  assert( argv[i][0] == '-' );
  if( strncmp( test, &argv[i][1], len) == 0 ) {
    if( argv[i][len+1] == 0 ) {
      if( i == argc - 1 ) {
	usage( prog );
	exit( EXIT_FAILURE );
      }
      i++;
      result = atoi(argv[i]);
    }
    else {
      result = atoi(&argv[i][len+1]);
    }
    return 1;
  }
  return 0; 
}

int
get_flag_arg(
	char *	prog,
	int	argc,
	char *	argv[],
	int&	i,
	const char *	test,
	unsigned char&	result
	)
{
  unsigned len = strlen(test);

  assert( argv[i][0] == '-' );
  if( strncmp( test, &argv[i][1], len) == 0 ) {
    if( argv[i][len+1] == 0 ) {
      if( i == argc - 1 ) {
	usage( prog );
	exit( EXIT_FAILURE );
      }
      i++;
      result = atoi(argv[i]);
    }
    else {
      result = atoi(&argv[i][len+1]);
    }
    return 1;
  }
  return 0; 
}



int
get_flag(
	char *	prog,
	int	argc,
	char *	argv[],
	int&	i,
	char 	test,
	char *&	result
	)
{
  assert( argv[i][0] == '-' );
  if( argv[i][1] == test ) {
    if( argv[i][2] == 0 ) {
      if( i == argc - 1 ) {
	usage( prog );
	exit( EXIT_FAILURE );
      }
      i++;
      result = argv[i];
    }
    else {
      result = &argv[i][2];
    }
    return 1;
  }
  return 0; 
}

int
main(
	int argc,
	char * argv[]
	)
{
  int		i;
  int		j;
  int		out_length;
  char *	out_name = 0;
  char *	prog = argv[0];

/****************************************************************/
/* Process command line arguments				*/
/****************************************************************/
  char *	p_out_name = 0;
  char *	p_lg2arch_max = 0;
  unsigned	seen_ddash = 0;
  unsigned	no_input = 0;

  for( i = 1; i < argc; i++ ) {
    if( !seen_ddash && argv[i][0] == '-' ) {
      if( get_flag(prog, argc, argv, i, '-' ) ) {
	seen_ddash = 1;
      }
      else if( get_flag(prog, argc, argv, i, 'h' ) ) {
	usage(argv[0]);
	return EXIT_SUCCESS;
      }
      else if( get_flag(prog, argc, argv, i, 'c' ) ) {
	xdep_prep_ctl.fix_unit = xdep_prep_ctl.load_unit;
	xdep_prep_ctl.imul_unit = xdep_prep_ctl.load_unit;
	xdep_prep_ctl.idiv_unit = xdep_prep_ctl.load_unit;
      }
      else if( get_flag(prog, argc, argv, i, "cf" ) ) {
	xdep_prep_ctl.fix_unit = xdep_prep_ctl.load_unit;
	xdep_prep_ctl.imul_unit = xdep_prep_ctl.load_unit;
	xdep_prep_ctl.idiv_unit = xdep_prep_ctl.load_unit;
	xdep_prep_ctl.fpu_unit = xdep_prep_ctl.load_unit;
      }
      else if( get_flag(prog, argc, argv, i, "gm2") ) {
        xdep_prep_ctl.flag_use_gp = 2;
cerr << "use gp = " << int(xdep_prep_ctl.flag_use_gp) << endl;
      }
      else if( get_flag(prog, argc, argv, i, "gm3") ) {
        xdep_prep_ctl.flag_use_gp = 3;
cerr << "use gp = " << int(xdep_prep_ctl.flag_use_gp) << endl;
      }
      else if( get_flag(prog, argc, argv, i, 'g' ) ) {
        xdep_prep_ctl.flag_use_gp = 0;
cerr << "use gp = " << int(xdep_prep_ctl.flag_use_gp) << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ux", xdep_prep_ctl.fix_unit ) ) {
cerr << "fix unit = " << int(xdep_prep_ctl.fix_unit) << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "uf", xdep_prep_ctl.fpu_unit ) ) {
cerr << "fpu unit = " << int(xdep_prep_ctl.fpu_unit) << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ul", xdep_prep_ctl.load_unit ) ) {
cerr << "load unit = " << int(xdep_prep_ctl.load_unit) << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "us", xdep_prep_ctl.store_unit ) ) {
cerr << "store unit = " << int(xdep_prep_ctl.store_unit) << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ub", xdep_prep_ctl.branch_unit ) ) {
cerr << "branch unit = " << int(xdep_prep_ctl.branch_unit) << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "uL", xdep_prep_ctl.crl_unit ) ) {
cerr << "crl unit = " << int(xdep_prep_ctl.crl_unit) << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "uX", xdep_prep_ctl.imul_unit ) ) {
cerr << "imul/idiv unit = " << int(xdep_prep_ctl.imul_unit) << endl;
	xdep_prep_ctl.idiv_unit = xdep_prep_ctl.imul_unit;
      }
      else if( get_flag_arg(prog, argc, argv, i, "dc", xdep_prep_ctl.cache_align) ) {
cerr << "cache align = " << xdep_prep_ctl.cache_align << endl;
	xdep_prep_ctl.cache_mask = xdep_prep_ctl.cache_align - 1;
      }
      else if( get_flag_arg(prog, argc, argv, i, "da", xdep_prep_ctl.fetch_align) ) {
cerr << "fetch align = " << xdep_prep_ctl.fetch_align << endl;
	xdep_prep_ctl.fetch_align_mask = xdep_prep_ctl.fetch_align - 1;
      }
      else if( get_flag_arg(prog, argc, argv, i, "df", xdep_prep_ctl.fetch_max) ) {
cerr << "fetch max = " << xdep_prep_ctl.fetch_max << endl;
      }

      else if( get_flag_arg(prog, argc, argv, i, "lm", xdep_prep_ctl.imul_latency) ) {
cerr << "imul latency = " << (int)xdep_prep_ctl.imul_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ld", xdep_prep_ctl.idiv_latency) ) {
cerr << "idiv latency = " << (int)xdep_prep_ctl.idiv_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "lD", xdep_prep_ctl.fdiv_latency) ) {
cerr << "fdiv latency = " << (int)xdep_prep_ctl.fdiv_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "lc", xdep_prep_ctl.cmp_latency) ) {
cerr << "cmp latency = " << (int)xdep_prep_ctl.cmp_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "le", xdep_prep_ctl.subword_load_latency) ) {
cerr << "subword_load latency = " << (int)xdep_prep_ctl.subword_load_latency << endl;
	xdep_prep_ctl.arithmetic_subword_load_latency =
		max( xdep_prep_ctl.arithmetic_load_latency, xdep_prep_ctl.subword_load_latency);
      }
      else if( get_flag_arg(prog, argc, argv, i, "la", xdep_prep_ctl.arithmetic_load_latency) ) {
cerr << "arithmetic_load latency = " << (int)xdep_prep_ctl.arithmetic_load_latency << endl;
	xdep_prep_ctl.arithmetic_subword_load_latency =
		max( xdep_prep_ctl.arithmetic_load_latency, xdep_prep_ctl.subword_load_latency);
      }
      else if( get_flag_arg(prog, argc, argv, i, "lF", xdep_prep_ctl.float_load_latency) ) {
cerr << "float_load latency = " << (int)xdep_prep_ctl.float_load_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "lf", xdep_prep_ctl.fpu_latency) ) {
cerr << "fpu latency = " << (int)xdep_prep_ctl.fpu_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "lx", xdep_prep_ctl.fix_latency) ) {
cerr << "fix latency = " << (int)xdep_prep_ctl.fix_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ll", xdep_prep_ctl.load_latency) ) {
cerr << "load latency = " << (int)xdep_prep_ctl.load_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "lL", xdep_prep_ctl.crl_latency) ) {
cerr << "crl latency = " << (int)xdep_prep_ctl.crl_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ls", xdep_prep_ctl.store_latency) ) {
cerr << "store latency = " << (int)xdep_prep_ctl.store_latency << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "lb", xdep_prep_ctl.branch_latency) ) {
cerr << "branch latency = " << (int)xdep_prep_ctl.branch_latency << endl;
      }

      else if( get_flag_arg(prog, argc, argv, i, "am", xdep_prep_ctl.imul_avail) ) {
cerr << "imul avail = " << (int)xdep_prep_ctl.imul_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ad", xdep_prep_ctl.idiv_avail) ) {
cerr << "idiv avail = " << (int)xdep_prep_ctl.idiv_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "aD", xdep_prep_ctl.fdiv_avail) ) {
cerr << "fdiv avail = " << (int)xdep_prep_ctl.fdiv_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ac", xdep_prep_ctl.cmp_avail) ) {
cerr << "cmp avail = " << (int)xdep_prep_ctl.cmp_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ae", xdep_prep_ctl.subword_load_avail) ) {
cerr << "subword_load avail = " << (int)xdep_prep_ctl.subword_load_avail << endl;
	xdep_prep_ctl.arithmetic_subword_load_avail =
		max( xdep_prep_ctl.arithmetic_load_avail, xdep_prep_ctl.subword_load_avail);
      }
      else if( get_flag_arg(prog, argc, argv, i, "aa", xdep_prep_ctl.arithmetic_load_avail) ) {
cerr << "arithmetic_load avail = " << (int)xdep_prep_ctl.arithmetic_load_avail << endl;
	xdep_prep_ctl.arithmetic_subword_load_avail =
		max( xdep_prep_ctl.arithmetic_load_avail, xdep_prep_ctl.subword_load_avail);
      }
      else if( get_flag_arg(prog, argc, argv, i, "aF", xdep_prep_ctl.float_load_avail) ) {
cerr << "float_load avail = " << (int)xdep_prep_ctl.float_load_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "af", xdep_prep_ctl.fpu_avail) ) {
cerr << "fpu avail = " << (int)xdep_prep_ctl.fpu_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ax", xdep_prep_ctl.fix_avail) ) {
cerr << "fix avail = " << (int)xdep_prep_ctl.fix_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "al", xdep_prep_ctl.load_avail) ) {
cerr << "load avail = " << (int)xdep_prep_ctl.load_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "aL", xdep_prep_ctl.crl_avail) ) {
cerr << "crl avail = " << (int)xdep_prep_ctl.crl_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "as", xdep_prep_ctl.store_avail) ) {
cerr << "store avail = " << (int)xdep_prep_ctl.store_avail << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "ab", xdep_prep_ctl.branch_avail) ) {
cerr << "branch avail = " << (int)xdep_prep_ctl.branch_avail << endl;
      }

      else if( get_flag(prog, argc, argv, i, 'o', p_out_name) ) {
	out_length = strlen(p_out_name);
	out_name = (char *) malloc(out_length+1);
	if( out_name == 0 ) {
	  cerr << "out of memory; needed " << out_length+1 << "*1" << endl;
	  return EXIT_FAILURE;
	}
	strcpy(out_name, p_out_name);
      }
      else if( get_flag(prog, argc, argv, i, 'a', p_lg2arch_max) ) {
	xdep_prep_ctl.flag_lg2arch_max = atoi(p_lg2arch_max);
cerr << "flag_lg2arch_max = " << (int)xdep_prep_ctl.flag_lg2arch_max << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sall") ) {
	xdep_prep_ctl.flag_all_split  = 0;
cerr << "flag_all_split = " << (int)xdep_prep_ctl.flag_all_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sfxldu") ) {
	xdep_prep_ctl.flag_fxldu_split  = 0;
cerr << "flag_fxldu_split = " << (int)xdep_prep_ctl.flag_fxldu_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sfpldu") ) {
	xdep_prep_ctl.flag_fpldu_split  = 0;
cerr << "flag_fpldu_split = " << (int)xdep_prep_ctl.flag_fpldu_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sfxstu") ) {
	xdep_prep_ctl.flag_fxstu_split  = 0;
cerr << "flag_fxstu_split = " << (int)xdep_prep_ctl.flag_fxstu_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sfpstu") ) {
	xdep_prep_ctl.flag_fpstu_split  = 0;
cerr << "flag_fpstu_split = " << (int)xdep_prep_ctl.flag_fpstu_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sspr") ) {
	xdep_prep_ctl.flag_spr_split  = 0;
cerr << "flag_spr_split = " << (int)xdep_prep_ctl.flag_spr_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sstu2") ) {
	xdep_prep_ctl.flag_stu3_split  = 0;
cerr << "flag_stu3_split = " << (int)xdep_prep_ctl.flag_stu3_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sfxst") ) {
	xdep_prep_ctl.flag_fxst_split  = 0;
cerr << "flag_fxst_split = " << (int)xdep_prep_ctl.flag_fxst_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sfpst") ) {
	xdep_prep_ctl.flag_fpst_split  = 0;
cerr << "flag_fpst_split = " << (int)xdep_prep_ctl.flag_fpst_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "smult") ) {
	xdep_prep_ctl.flag_mult_split  = 0;
cerr << "flag_mult_split = " << (int)xdep_prep_ctl.flag_mult_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "scrl") ) {
	xdep_prep_ctl.flag_crl_split  = 0;
cerr << "flag_crl_split = " << (int)xdep_prep_ctl.flag_crl_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sstring") ) {
	xdep_prep_ctl.flag_string_split  = 0;
cerr << "flag_string_split = " << (int)xdep_prep_ctl.flag_string_split << endl;
      }
      else if( get_flag(prog, argc, argv, i, "sshow") ) {
	xdep_prep_ctl.flag_show_split  = 1;
cerr << "flag_show_split = " << (int)xdep_prep_ctl.flag_show_split << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "pdist", xdep_prep_ctl.pred_dist) ) {
cerr << "pred_dist = " << (int)xdep_prep_ctl.pred_dist << endl;
      }
      else if( get_flag_arg(prog, argc, argv, i, "pmode", xdep_prep_ctl.pred_mode) ) {
cerr << "pred_mode = " << (int)xdep_prep_ctl.pred_mode << endl;
      }
      else if( get_flag(prog, argc, argv, i, "pshow") ) {
	xdep_prep_ctl.flag_show_pred  = 1;
cerr << "flag_show_pred = " << (int)xdep_prep_ctl.flag_show_pred << endl;
      }
      else if( get_flag(prog, argc, argv, i, 'p') ) {
	xdep_prep_ctl.pred_do  = 1;
cerr << "pred_do = " << (int)xdep_prep_ctl.pred_do << endl;
      }
      else {
	usage( prog );
	return EXIT_FAILURE;
      }
      if( i == argc  ) {
	no_input = 1;
      }
    }
    else if( i != argc -1  ) {
      usage( prog );
      return EXIT_FAILURE;
    }
  } /* for( command line arguments ) */

  if( no_input || argc < 2 ) {
    usage( prog );
    return EXIT_FAILURE;
  }

/****************************************************************/
/* Open output files 						*/
/****************************************************************/

  if( out_name == 0 ) {
    out_length = strlen(argv[argc-1]) + strlen(out_suffix);
    out_name = (char *) malloc(out_length+1);
    if( out_name == 0 ) {
      cerr << "out of memory; needed " << out_length+1 << "*1" << endl;
      return EXIT_FAILURE;
    }
    sprintf(out_name, "%s%s", argv[argc-1], out_suffix);
  }

  FILE * out_fp = fopen( out_name, "wb");
  if( out_fp == 0 ) {
    cerr << "could not open file " << out_name << " for writing" << endl;
    return EXIT_FAILURE;
  }

/****************************************************************/
/* Open input (xcoff) file					*/
/****************************************************************/
  char * in_name = argv[argc-1];

#if( USING_FF )
  FILE * in_fp = fopen(in_name, "rb" );
  if( in_fp == 0 ) {
    cerr << "could not open input file " << in_name << endl;
    return EXIT_FAILURE;
  }
#elif( USING_ROCH )
  FILE * in_fp = fopen(in_name, "rb" );
  if( in_fp == 0 ) {
    cerr << "could not open input file " << in_name << endl;
    return EXIT_FAILURE;
  }
#elif( 1 )
  struct xcoffhdr *	xcoffhdr;
  struct scnhdr *	secns;
  unsigned		nsecns;
  struct text_section *	texts;
  unsigned 		ntexts;
  FILE * in_fp = fopen(in_name, "rb" );

  get_text_xcoff( in_fp, &xcoffhdr, &nsecns, &secns, &ntexts, &texts);
  unsigned first_addr = text_start_aouthdr(aouthdr_xcoffhdr(xcoffhdr));
cerr << "first address is " << hex << first_addr << dec << endl;

  unsigned max_addr = 0;
  unsigned min_addr = UINT_MAX;
  for( i = 0; i < ntexts; i++ ) {
    struct text_section *	text = &texts[i];
    unsigned lo_addr = address_text_section(text);
    unsigned hi_addr = lo_addr + size_text_section(text);

    if( (hi_addr%4) ) {
      hi_addr += 4-(hi_addr%4);
    }

    if( hi_addr > max_addr ) {
      max_addr = hi_addr;
    }
    if( lo_addr < min_addr ) {
      min_addr = lo_addr;
    }
  }

  cerr << "address range is " << hex << min_addr << ", " << max_addr << dec << endl;
#else
  Xcoff xcoff(in_name);

  int num_entries = xcoff.NumEntries();

  if( num_entries <= 0 ) {
    cerr << "no entries in xcoff file "  << in_name << endl;
    return EXIT_FAILURE;
  }

  int * first_addr = xcoff.EntryAddress();
  cerr << "entry address is " << first_addr << endl;

  int * max_addr = 0;
  int * min_addr = (int *) UINT_MAX;
  for( i = 0; i < num_entries; i++ ) {
    Entry& entry = xcoff.XcoffEntry(i);
    int * lo_addr = entry.StartAddress();
    int * hi_addr = entry.EndAddress();
    if( hi_addr > max_addr ) {
      max_addr = hi_addr;
    }
    if( lo_addr < min_addr ) {
      min_addr = lo_addr;
    }
  }

  cerr << "address range is " << min_addr << ", " << max_addr << endl;
#endif

/****************************************************************/
/* Setup the instruction info array				*/
/****************************************************************/
  opc_init();

#if( USING_FF )
  unsigned	first_addr;
  int		status;
  status = fread( &first_addr, sizeof(unsigned), 1, in_fp);
  if( status != 1 ) {
    cerr << "error reading first addr from file " << in_name << endl;
    return EXIT_FAILURE;
  }

  xdep_prep_info_table _info_table;
  xdep_prep_info_table * info_table = &_info_table;
  read_xdep_prep_info_table(in_fp, info_table);

  unsigned		size = size_xdep_prep_info_table(info_table);
  xdep_prep_info *	array = array_xdep_prep_info_table(info_table);
  unsigned *		addrs = addrs_xdep_prep_info_table(info_table);
#elif( USING_ROCH )
  int status;

  unsigned size;
  status = fread( &size, sizeof(unsigned), 1, in_fp);
  if( status != 1 ) {
    cerr << "error reading size from file " << in_name << endl;
    return EXIT_FAILURE;
  }

  unsigned *	insns = (unsigned *) calloc( sizeof(unsigned), size);
  if( insns == 0 ) {
    cerr << "out of memory; needed " << size << "*" << sizeof(unsigned) << endl;
    return EXIT_FAILURE;
  }

  status = fread( insns, sizeof(unsigned), size, in_fp);
  if( status != size ) {
    cerr << "error reading " << size << " words from file " << in_name << endl;
    return EXIT_FAILURE;
  }

  xdep_prep_info * array = (xdep_prep_info *) calloc(sizeof(xdep_prep_info), size);
  if( array == 0 ) {
    cerr << "out of memory; needed " << size << "*" << sizeof(xdep_prep_info) << endl;
    return EXIT_FAILURE;
  }

#elif( 1 )
  assert( (max_addr % 4) == 0);
  assert( (min_addr % 4) == 0);
  unsigned size	= ((max_addr - min_addr)>>2);
  xdep_prep_info * array = (xdep_prep_info *) calloc(sizeof(xdep_prep_info), size);
  if( array == 0 ) {
    cerr << "out of memory; needed " << size << "*" << sizeof(xdep_prep_info) << endl;
    return EXIT_FAILURE;
  }

#else
  unsigned size = max_addr - min_addr;
  xdep_prep_info * array = (xdep_prep_info *) calloc(sizeof(xdep_prep_info), size);
  if( array == 0 ) {
    cerr << "out of memory; needed " << size << "*" << sizeof(xdep_prep_info) << endl;
    return EXIT_FAILURE;
  }
#endif

  unsigned osize = 0;
  unsigned xsize = 0;

/********************************************************/
/* Here begins the real loop				*/
/********************************************************/

#if( USING_FF )
  for( i = 0; (unsigned) i < size; i++ ) {
    xdep_prep_info * 	entry = &array[i];
    unsigned		iaddr = addrs[i];
    unsigned		iword = xdep_prep_info_iword(entry);
    unsigned		nexp;

    process_iword( (unsigned) iaddr, iword, entry);
    if( xdep_prep_info_is_over(entry) ) {
      osize += xdep_prep_info_ins(entry) + xdep_prep_info_outs(entry);
    }
    nexp = xdep_prep_info_nexpands(entry);
    xsize += nexp;
    for( j = 0; (unsigned) j < nexp; j++ ) {
      xdep_prep_info *	xentry = &xdep_prep_info_expands(entry)[j];
      if( xdep_prep_info_is_over(xentry) ) {
	osize += xdep_prep_info_ins(xentry) + xdep_prep_info_outs(xentry);
      }
    }
  }
#elif( USING_ROCH )
  for( i = 0; (unsigned) i < size; i++ ) {
    xdep_prep_info * 	entry = &array[i];
    unsigned		iword = insns[i];
    unsigned		nexp;

    process_iword( 0, iword, entry);
    if( xdep_prep_info_is_over(entry) ) {
      osize += xdep_prep_info_ins(entry) + xdep_prep_info_outs(entry);
    }
    nexp = xdep_prep_info_nexpands(entry);
    xsize += nexp;
    for( j = 0; j < nexp; j++ ) {
      xdep_prep_info *	xentry = &xdep_prep_info_expands(entry)[j];
      if( xdep_prep_info_is_over(xentry) ) {
	osize += xdep_prep_info_ins(xentry) + xdep_prep_info_outs(xentry);
      }
    }
  }
#elif( 1 )
  for( i = 0; i < ntexts; i++ ) {
    struct text_section * text = &texts[i];
    unsigned	lo_addr = address_text_section(text);
    unsigned	hi_addr = lo_addr + size_text_section(text);
    unsigned *	iwords = insns_text_section(text);
    unsigned	addr;

    for( addr = lo_addr, j = 0; addr < hi_addr; addr += 4, j++ ) {
      unsigned		index = (addr - min_addr)>>2;
      xdep_prep_info *	entry = &array[index];
      unsigned		iaddr = addr;
      unsigned		iword = iwords[j];
      unsigned		nexp;
      unsigned		k;

      assert( index < size );
      process_iword( iaddr, iword, entry);
      if( xdep_prep_info_is_over(entry) ) {
	osize += xdep_prep_info_ins(entry) + xdep_prep_info_outs(entry);
      }
      nexp = xdep_prep_info_nexpands(entry);
      xsize += nexp;
      for( k = 0; k < nexp; k++ ) {
	xdep_prep_info *	xentry = &xdep_prep_info_expands(entry)[k];
	if( xdep_prep_info_is_over(xentry) ) {
	  osize += xdep_prep_info_ins(xentry) + xdep_prep_info_outs(xentry);
	}
      }
    }
  }

#else
  for( i = 0; i < num_entries; i++ ) {
    Entry& entry = xcoff.XcoffEntry(i);
    int * lo_addr = entry.StartAddress();
    int * hi_addr = entry.EndAddress();
    int * addr;
    for( addr = lo_addr; addr < hi_addr; addr++ ) {
      unsigned		index = addr - min_addr;
      xdep_prep_info *	entry = &array[index];
      unsigned		iaddr = (unsigned) addr;
      unsigned		iword = xcoff.InstructionAtAddress(addr);
      unsigned		nexp;

      process_iword( iaddr, iword, entry);
      if( xdep_prep_info_is_over(entry) ) {
	osize += xdep_prep_info_ins(entry) + xdep_prep_info_outs(entry);
      }
      nexp = xdep_prep_info_nexpands(entry);
      xsize += nexp;
      for( j = 0; j < nexp; j++ ) {
	xdep_prep_info *	xentry = &xdep_prep_info_expands(entry)[j];
	if( xdep_prep_info_is_over(xentry) ) {
	  osize += xdep_prep_info_ins(xentry) + xdep_prep_info_outs(xentry);
	}
      }
    }
  }
#endif

/********************************************************/
/* Perform the fixup for xcoffs...			*/
/********************************************************/
#if( USING_FF )
#elif( USING_ROCH )
#elif( 1 )
#else
  for( i = 0; i < size; i++ ) {
    xdep_prep_info *	entry = &array[i];
    unsigned		iaddr =(unsigned)(min_addr + i );
    if( xdep_prep_info_fetch(entry) != 0 ) {
      continue;
    }
    /* not processed yet */
    process_illegal(iaddr, entry);
    /* cerr << "illegal instruction " << hex << iaddr << " -> " << i << dec << endl; */
  }
#endif

/********************************************************/
/* For xcoffs, if predication is true, expand branches	*/
/********************************************************/

#if( USING_FF )
  if( xdep_prep_ctl.pred_do ) {
    cerr << "ignoring predication" << endl;
  }
#elif( USING_ROCH )
  if( xdep_prep_ctl.pred_do ) {
    cerr << "ignoring predication" << endl;
  }
#elif( 1 )
  if( xdep_prep_ctl.pred_do ) {
    unsigned		dist = 4 + 4*xdep_prep_ctl.pred_dist;
    cerr << "going to predicate " << dist << endl;
    for( i = 0; i < ntexts; i++ ) {
      struct text_section * text = &texts[i];
      unsigned		lo_addr = address_text_section(text);
      unsigned		hi_addr = lo_addr + size_text_section(text);
      unsigned *	iwords = insns_text_section(text);
      unsigned		addr;
      unsigned		top_addr = hi_addr - dist;

      for( 	addr = lo_addr, j = 0;
		addr < hi_addr;
		addr += 4, j++ ) {
	unsigned		index = (addr - min_addr)>>2;
	xdep_prep_info *	entry = &array[index];
	unsigned		iaddr = addr;
	unsigned		iword = iwords[j];
	unsigned		nexp;
	unsigned		k;

	assert( index < size );
	/* only bcs */
	if( (iword >> 26) != 16 ) {
	  continue;
	}
	/* not bca, bcl, bcla */
	if( iword & 0x3 ) {
	  continue;
	}
	/* short bcs */
	if( (iword & 0xffff) > dist ) {
	  continue;
	}
	/* not ctr */
	if( (iword & 0x00800000) == 0 )  {
	  continue;
	}
	/* uses CR */
	if( (iword & 0x02000000) != 0 )  {
	  continue;
	}
	/* only marked */
	if( (iword & 0x00400000) == 0 ) {
	  continue;
	}


	/* at this point, bc NOT bca, bcl, bcla, and */
	/* branches by less than specified distance */

	assert( xdep_prep_info_nexpands(entry) == 0 );
	process_branch( iaddr, iword, entry, xdep_prep_ctl.pred_dist, &iwords[j+1], &array[index+1]);

	nexp = xdep_prep_info_nexpands(entry);
	xsize += nexp;
	for( k = 0; k < nexp; k++ ) {
	  xdep_prep_info *	xentry = &xdep_prep_info_expands(entry)[k];
	  if( xdep_prep_info_is_over(xentry) ) {
	    osize += xdep_prep_info_ins(xentry) + xdep_prep_info_outs(xentry);
	  }
	}
      }
    }
  }
#else
#endif


/********************************************************/
/* Create the expand array				*/
/********************************************************/
  xdep_prep_info * expand;
  unsigned nexp = 0;

  if( xsize == 0 ) {
    expand = 0;
  }
  else {
    expand = (xdep_prep_info *)malloc(xsize * sizeof( xdep_prep_info ));
	
    if( expand == 0 ) {
      cerr << "out of memory; needed " << xsize << "*" << sizeof(xdep_prep_info) << endl;
      return EXIT_FAILURE;
    }

    unsigned xat = 0;

    for( i = 0; (unsigned) i < size; i++ ) {
      xdep_prep_info * entry = &array[i];
      unsigned is_exp = xdep_prep_info_is_expand(entry);
      if( is_exp ) {
	nexp++;
	xdep_prep_info * exp = xdep_prep_info_expands(entry);
	unsigned nexps = xdep_prep_info_nexpands(entry);

	xdep_prep_info_expands(entry) = &expand[xat];

	unsigned j;
	for( j = 0; j < nexps; j++, xat++ ) {
	  assert( xat < xsize );
	  expand[xat] = exp[j];
	}
      }
    }
  }


/********************************************************/
/* Create the overflow array				*/
/********************************************************/
  xarch * over; 
  unsigned oat = 0;
  unsigned nover = 0;
  unsigned n4reg = 0;

  if( osize == 0 ) {
    over = 0;
  }
  else {
    over = (xarch *)malloc( osize * sizeof( xarch ));
    if( over == 0 ) {
      cerr << "out of memory; needed " << osize << "*1" << endl;
      return EXIT_FAILURE;
    }

    for( i = 0; (unsigned) i < size; i++ ) {
      xdep_prep_info * entry = &array[i];
      unsigned is_over = xdep_prep_info_is_over(entry);
      if( is_over ) {
	nover++;

	xarch * overflow = xdep_prep_info_overflow(entry);
	unsigned ins = xdep_prep_info_ins(entry);
	unsigned outs = xdep_prep_info_outs(entry);
	unsigned total = ins + outs;

	n4reg += (total <= 4)?1:0;
	xdep_prep_info_overflow(entry) = &over[oat];

	unsigned j;
	for( j = 0; j < total; j++, oat++ ) {
	  assert( oat < osize );
	  over[oat] = overflow[j];
	}
      }
    }

    for( i = 0; (unsigned) i < xsize; i++ ) {
      xdep_prep_info * entry = &expand[i];
      unsigned is_over = xdep_prep_info_is_over(entry);
      assert( entry != NULL );
      if( is_over ) {
	nover++;

	xarch * overflow = xdep_prep_info_overflow(entry);
	unsigned ins = xdep_prep_info_ins(entry);
	unsigned outs = xdep_prep_info_outs(entry);
	unsigned total = ins + outs;

	n4reg += (total <= 4)?1:0;
	xdep_prep_info_overflow(entry) = &over[oat];

	unsigned j;
	for( j = 0; j < total; j++, oat++ ) {
	  assert( oat < osize );
	  over[oat] = overflow[j];
	}
      }
    }
  }

/********************************************************/
/* Finally, the writing out of the prep stuff		*/
/********************************************************/

#if( USING_FF )
  write_ffdep_prep(out_fp, &xdep_prep_ctl, xdep_prep_ctl.fetch_max, (unsigned) first_addr,
		info_table, xsize, expand, osize, over );
#elif( USING_ROCH )
  write_rochdep_prep(out_fp, size, array, osize, over, xsize, expand);
#else
  write_dep_prep(out_fp, &xdep_prep_ctl, (unsigned) first_addr,
		(unsigned) min_addr, size, array, osize, over, xsize, expand);
#endif

cerr << "number overflows = " << nover << " that use 4 regs or less " << n4reg << endl;
cerr << "number expands = " << nexp << " expanding into " << xsize << endl;
cerr << "total = " << 5*sizeof(unsigned) << "+" << size+xsize << "(size+xsize)*" <<
	sizeof( xdep_prep_info) << "+" << osize << "(over)*" << sizeof( xarch ) <<
	" = " << 5*sizeof(unsigned)+(size+xsize)*sizeof(xdep_prep_info) + osize*sizeof(xarch ) << endl;

  return EXIT_SUCCESS;
}
