/******************************************************************************
 This file reads in an xcoff using ravi's tool and dumps out the info required
 to read and expand an xtrace stream
******************************************************************************/

/******************************************************************************
  INCLUDES
******************************************************************************/

/*** STANDARD INCLUDES ***/
#include <assert.h>
#include <stdlib.h>
#include <iostream.h>

/*** RAVI'S INCLUDES ***/
#include "xcoff.H"
#include "iprops.H"
#include "ppcinstr.H"
#include "itype.H"

/*** LOCAL INCLUDES ***/
extern "C" {
#include "reader_prep_info.h"
}

/******************************************************************************
  LOCAL
******************************************************************************/

static const char * out_suffix = ".prep";

/******************************************************************************
  FUNCTIONS
******************************************************************************/

void
usage(
	char * program_name,
	unsigned id
	)
{
  cerr << "usage [" << id << "]: " << program_name << " [-h] [-o output] [--] <xcoff_file>" << endl;
}

int
main(
	int argc,
	char * argv[]
	)
{
  int		i;
  unsigned	seen_ddash = 0;
  unsigned	no_input = 0;
  int		out_length = 0;
  char *	out_name = 0;

  for( i = 1; i < argc; i++ ) {
    if( !seen_ddash && argv[i][0] == '-' ) {
      if( argv[i][1] == '-' ) {
	if( argv[i][2] != 0 ) {
	  usage( argv[0], 1 );
	  return EXIT_FAILURE;
	}
	seen_ddash = 1;
      }
      else if( argv[i][1] == 'h' ) {
	if( argv[i][2] != 0 ) {
	  usage( argv[0], 2);
	  return EXIT_FAILURE;
	}
	usage( argv[0], 0);
	return EXIT_SUCCESS;
      }
      else if( argv[i][1] == 'o' ) {
cerr << "processing -o" << endl;
	if( argv[i][2] == 0 ) {
	  if( i == argc - 1 ) {
	    usage(argv[0], 3);
	    return EXIT_FAILURE;
	  }
	  i++;
cerr << "processing " << argv[i] << endl;
	  out_length = strlen( argv[i] );
	  out_name = (char *) malloc( out_length + 1 );
	  if( out_name == 0 ) {
	    cerr << "out of memory; needed " << out_length+1 <<  "*1" << endl;
	    return EXIT_FAILURE;
	  }
	  sprintf( out_name, "%s", argv[i] );
	}
	else {
cerr << "processing " << argv[i] << endl;
	  out_length = strlen(argv[i]);
	  out_name = (char *) malloc( out_length - 1);
	  if( out_name == 0 ) {
	    cerr << "out of memory; needed " << out_length+1 <<  "*1" << endl;
	    return EXIT_FAILURE;
	  }
	  sprintf( out_name, "%s", &argv[i][2]);
	}
      }
      else {
	usage( argv[0], 6 );
	return EXIT_FAILURE;
      }
      if( i == argc ) {
	no_input = 1;
      }
    }
    else if( i != argc - 1 ){
      usage( argv[0], 4 );
      return EXIT_FAILURE;
    }
  }

  if( no_input || argc < 2 ) {
    usage(argv[0], 5);
    return EXIT_FAILURE;
  }

if( out_name ) {
  cerr << "name is " <<  out_name << endl;
}

  if( out_name == 0 ) {
    out_length = strlen(argv[argc-1]) + strlen(out_suffix);
    out_name = (char *) malloc(out_length+1);
    if( out_name == 0 ) {
      cerr << "out of memory; needed " << out_length+1 << "*1" << endl;
      return EXIT_FAILURE;
    }
    sprintf(out_name, "%s%s", argv[argc-1], out_suffix);
  }

  FILE * out_fp = fopen( out_name, "wb");
  if( out_fp == 0 ) {
    cerr << "could not open file " << out_name << " for writing" << endl;
    return EXIT_FAILURE;
  }

  char * xcoff_name = argv[argc-1];
  Xcoff xcoff(xcoff_name);

  int num_entries = xcoff.NumEntries();

  if( num_entries <= 0 ) {
    cerr << "no entries in xcoff file "  << xcoff_name << endl;
    return EXIT_FAILURE;
  }

  int *	first_addr = xcoff.EntryAddress();
  cerr << "entry address is "  << first_addr  << endl;

  int * max_addr = 0;
  int *	min_addr = (int*) UINT_MAX;
  for( i = 0; i < num_entries; i++ ) {
    Entry& entry = xcoff.XcoffEntry(i);
    int * lo_addr = entry.StartAddress();
    int * hi_addr = entry.EndAddress();
    if( hi_addr > max_addr ) {
      max_addr = hi_addr;
    }
    if( lo_addr < min_addr ) {
      min_addr = lo_addr;
    }
  }

  cerr << "address range is " << min_addr << ", " << max_addr << endl;

  unsigned size = max_addr - min_addr;
  xreader_prep_info * info = (xreader_prep_info *) calloc(sizeof(xreader_prep_info), size);
  if( info == 0 ) {
    cerr << "out of memory; needed " << size << "*" << sizeof(xreader_prep_info)<<endl;
    return EXIT_FAILURE;
  }

  assert( first_addr >=  min_addr );
  assert( first_addr <  max_addr );

  for( i = 0; i < num_entries; i++ ) {
    Entry& entry = xcoff.XcoffEntry(i);
    int * lo_addr = entry.StartAddress();
    int * hi_addr = entry.EndAddress();

    int * addr;
    for( addr = lo_addr; addr < hi_addr; addr++ ) {
      unsigned iword = xcoff.InstructionAtAddress(addr);
      unsigned index = addr - min_addr;
      unsigned kind;

      assert( index < size );

      if( IsAbsoluteBranch(iword) ) {
	kind = xreader_prep_info_is_brabs;
      }
      else if( IsUnconditionalBranch(iword) ) {
	kind = xreader_prep_info_is_branch;
      }
      else if( IsRegisterBranch(iword) ) {
	kind = xreader_prep_info_is_brreg;
      }
      else if( IsConditionalBranch(iword) ) {
	kind = xreader_prep_info_is_brcond;
      }
      else {
	PPCInstruction ppc(iword);
	assert( !IsBranch(ppc));
	if( IsLoadStore(ppc) ) {
	  kind = xreader_prep_info_is_memop;
	}
	else {
	  kind = xreader_prep_info_is_other;
	}
      }
      xreader_prep_info_kind(info[index]) = kind;
      xreader_prep_info_iword(info[index]) = iword;
    }
  }

cerr << "size is " << size  << endl;

  write_reader_prep(out_fp, (unsigned) first_addr,
		(unsigned) min_addr, size, info);

  return EXIT_SUCCESS;
}
