// array.c

// implementation of a general array structure, appliable to cache, TLB, etc. zhigangh@1/12/2003
// latency, backup, and port control are not handled here. 
// see stage_ifetch.macros for icache, and stage_mem_exec for dcache.

// jdonald: totally revised

/* creating a new array structure with appropriate parameters */
inline struct xarray *shared_array_create(char *name, unsigned n_set, unsigned n_way, unsigned blk_size)
{
    struct xarray *array;
    unsigned i_set, i_blk;

    array = (struct xarray *)malloc(sizeof(struct xarray));

    if (array == NULL)
    {
		fprintf(the_stderr, "Not enough memory to create array %s at [%s:%d]\n", name, __FILE__, __LINE__);
		exit( EXIT_FAILURE );
    }
    array->name = name;
    array->n_set = n_set;
    array->n_way = n_way;
    array->blk_size = blk_size;

    array->tag_shift = lg2(n_set) + lg2(blk_size);
    array->set_shift = lg2(blk_size);
    array->set_mask = (~(~0<<lg2(n_set)));
    array->offset_mask = (~(~0<<lg2(blk_size)));

    array->sets = (struct xarray_set *)malloc(sizeof(struct xarray_set)*n_set);

    if (array->sets == NULL)
    {
		fprintf(the_stderr, "Not enough memory to create sets of array %s at [%s:%d]\n", name, __FILE__, __LINE__);
		exit( EXIT_FAILURE );
    }

    array->blks = (struct xarray_blk *)malloc(sizeof(struct xarray_blk)*n_set*n_way);

    if (array->blks == NULL)
    {
		fprintf(the_stderr, "Not enough memory to create blks of array %s at [%s:%d]\n", name, __FILE__, __LINE__);
		exit( EXIT_FAILURE );
    }

    for (i_set = 0; i_set < n_set; i_set++)
    {
        struct xarray_blk *p_blk = &(array->blks[i_set * n_way]);
        array->sets[i_set].blks = p_blk;

        for (i_blk = 0; i_blk < n_way; i_blk++, p_blk++)
        {
            p_blk->valid = FALSE;
            p_blk->dirty = FALSE;
            p_blk->shared = FALSE; // jdonald
            p_blk->prev_dirty = FALSE;
            p_blk->tag = 0;
            p_blk->thread_id = -1;
            p_blk->processor_id = -1;

            if (i_blk == n_way - 1)
            {
            	p_blk->next = NULL;
            }else
            {
            	p_blk->next = &(array->sets[i_set].blks[i_blk + 1]);
            }
        }/* for i_blk */
     }/* for i_set */
     return array;

}/* shared_array_create */

#define array_create(name, n_set, n_way, blk_size) \
	shared_array_create(name, n_set, n_way, blk_size)
// jdonald got rid of the copypasted function

/* access an array with a key, and an operation type (lookup/update/probe) */
inline int shared_array_access(int processor_id, int thread_id, struct xarray *array, unsigned key, unsigned op, struct xarray_blk **pp_blk)
{
	unsigned tag = key >> array->tag_shift;
	unsigned set = (key >> array->set_shift) & array->set_mask;
	// unsigned offset = key & array->offset_mask; // unused?
	
	// unsigned i_blk; // unused?
	struct xarray_blk *p_blk;
	struct xarray_blk *prev_blk, *prev_prev_blk;
	boolean array_hit = FALSE; // jdonald
	
	for (p_blk = array->sets[set].blks, prev_blk = NULL, prev_prev_blk = NULL; p_blk; prev_prev_blk = prev_blk, prev_blk = p_blk, p_blk = p_blk->next)
	{
		if (p_blk->tag == tag && (shared_memory ||
								  (p_blk->thread_id == thread_id && p_blk->processor_id == processor_id))) // jdonald
		{
			array_hit = TRUE;
			break;
		}
	}/* for p_blk */
	if (!array_hit) {
		if ((op & ARRAY_MISS_UPDATE) == 0){
		    return ARRAY_MISS;
	        }
		
		/* reuse the victim blk to store the new guy */
		prev_blk->valid = TRUE;
		prev_blk->prev_dirty = prev_blk->dirty;
		prev_blk->dirty = (op & ARRAY_WRITE)?1:0;
		prev_blk->tag = tag;
		prev_blk->thread_id = thread_id;
		prev_blk->processor_id = processor_id; /* LYM CMP */
		prev_blk->shared = (op & ARRAY_SHARE)?1:0;
		assert(!(op & ARRAY_INVALIDATE)); // we wouldn't invalidate something not in the cache
		
		if (array->n_way >= 2)
		{
			prev_prev_blk->next = NULL;
			prev_blk->next = array->sets[set].blks;
			array->sets[set].blks = prev_blk;
		}
	
		/* pass back the address of the blk for possible further handling */
		if (pp_blk != NULL) *pp_blk = prev_blk;
		return ARRAY_MISS;
	}
	else { // array_hit
		p_blk->dirty |= (op & ARRAY_WRITE)?1:0;
		if (pp_blk != NULL) /* pass back the address of the blk for possible further handling */
			*pp_blk = p_blk; // moving this up here, since i need it even on ARRAY_LOOKUP -jdonald
		if (op & ARRAY_INVALIDATE)  // broadcast received, we have been called by some core that doesn't own the array
		{
			assert(!(op & ARRAY_WRITE)); // hopefully noone will want to write and invalidate simultaneously
			// could be shared && !dirty, in the usual case
			// or !shared && dirty, in the case of a write-allocate miss to a would-be shared line
			p_blk->valid = FALSE;
			p_blk->prev_dirty = p_blk->dirty;
			p_blk->dirty = FALSE;
			p_blk->tag = 0;
			p_blk->thread_id = -1;
			p_blk->processor_id = -1;
			return ARRAY_HIT; // successful invalidate
		}

		if ((op & ARRAY_HIT_UPDATE) == 0){
			assert(!(op & ARRAY_SHARE)); // we would always want to proceed if this is for share updating
		    return ARRAY_HIT;
		}
		
		p_blk->shared |= (op & ARRAY_SHARE)?1:0; // remember if this is a share setting we may be called by some core that doesn't own the array
		/* only update LRU list when necessary in a set-associative array */
		if (p_blk != array->sets[set].blks)
		{
			prev_blk->next = p_blk->next;
			p_blk->next = array->sets[set].blks;
			array->sets[set].blks = p_blk;
		}
		
		return ARRAY_HIT;
	}
}/* shared_array_access */

#define array_access(thread_id, array, key, op, pp_blk) \
	shared_array_access(processor_id, thread_id, array, key, op, pp_blk)
// jdonald doing this so that we no longer need processor_id to use thread-local storage	

inline void free_array(struct xarray* array) // jdonald playing it clean
{
	free(array->sets);
	free(array->blks);
	free(array);
}

#if TEST_CACHE
int main(void)
{
	struct xarray *d1, *d2;
	int i, j, tag;
	
	d2 = array_create("d2cache", /* n_set */65536, /* n_way */4, /* blk_size */64);
	
	d1 = array_create("d1cache", /* n_set */1024, /* n_way */4, /* blk_size */32);

	/* write the whole L2 cache */	
	for (i = 0; i < 65536; i++)
	{
		for (j = 0; j < 4; j++)
		{
			array_access(d2, /* key */MK_ADDR(d2, j, i, 0), ARRAY_UPDATE, /* pp_blk */NULL);
		}
	}
	
	array_access(d1, /* key */MK_ADDR(d2, 1, 1, 0), ARRAY_LOOKUP, /* pp_blk */NULL);
	array_access(d1, /* key */MK_ADDR(d2, 65520, 3, 1), ARRAY_UPDATE, /* pp_blk */NULL);
	array_access(d1, /* key */MK_ADDR(d2, 65520, 3, 1), ARRAY_LOOKUP, /* pp_blk */NULL);
	
	free_array(d1);
	free_array(d2);
}/* main */
#endif /* TEST_CACHE */

