// thread_synchro.c
//
// by jdonald

// Our cores must call this by the proper ordering restricted by l2sync.macros.
// Ordering of the SMT threads within those cores is already deterministic
// so we need not worry.
//
// check_thread_synchro() is called by stage_ifetch.macros
// print_synchro_data() is called at the end of main() in turandot.c

#ifndef MAX_LOCKS
#define MAX_LOCKS 16384 // we'll keep a hard limit for now to avoid memory explosions

// 200 is too small for some odd-numbered barnes runs
// and 2048 is too small for radiosity!!
// It turns out the radiosity (8t) runs range as high as 3503
#endif

syncrec_t* locks = NULL; // when these have higher numbers, they are getting unlocked

int num_locks = 0, locks_allocated = 0;
syncrec_t requests[MAX_CORES][MAX_THREADS] = {{{0}}};
int sync_records_count, sync_bypass_count = 0, illegal_sync_count = 0;

void print_synchro_data() {
	if (sync_records_count > 0) {
		int i;
		fprintf(the_stderr, "synchronization escape records: %d\n", sync_records_count);
		fprintf(the_stderr, "synchronization point addresses: %d\n", num_locks);
		fprintf(the_stderr, "   %8s   %5s   %s\n", "addr", "count", "final signal");
		for (i = 0; i < num_locks; i += 1)
			fprintf(the_stderr, "   %08x   %5d   %s\n", locks[i].address, locks[i].count,
				(locks[i].code == SYNC_SIGNAL ? "SYNC_SIGNAL" : "SYNC_BROADCAST_SIGNAL"));
		fprintf(the_stderr, "detected reorderings: %d\n", sync_bypass_count);
		fprintf(the_stderr, "detected illegal bypasses: %d\n", illegal_sync_count);
	}
}

boolean check_thread_synchro(const syncrec_t* psyncrec, int core, int thread, int cycle) {
	// This function is non-reentrant and must be checked in the same manner
	// as L2 cache synchronization
	boolean retval = FALSE;
	int i;
	boolean is_repeat = FALSE;
	static int num_total_threads = -1; // an upper bound to judge illegal syncs
	static boolean stalled[MAX_CORES][MAX_THREADS] = {{FALSE}}; // to watch for deadlocks
	
	if (locks == NULL) {
		locks_allocated = 64;
		locks = malloc(locks_allocated * sizeof(syncrec_t));
		assert(locks != NULL);

		assert(num_total_threads == -1); // first-time initialization
		num_total_threads = 0;
		for (i = 0; i < processor_num; i += 1)
			num_total_threads += cmp_conf[i];
	}
	
	if (requests[core][thread].code != psyncrec->code ||
		requests[core][thread].address != psyncrec->address ||
		requests[core][thread].count != psyncrec->count) {
			sync_records_count += 1;
			requests[core][thread] = *psyncrec;
	}
	else
		is_repeat = TRUE;
	if (!is_repeat)
		fprintf(stderr, "%d,%d fetch: %x, %08x, %d\n", core, thread, psyncrec->code, psyncrec->address, psyncrec->count);
	switch(psyncrec->code) {
		case SYNC_SIGNAL:
		case SYNC_BROADCAST_SIGNAL:
			for (i = 0; i < num_locks; i += 1) {
				if (locks[i].address == psyncrec->address) {
					break;
				}
			}
			if (i >= num_locks) {
				if (num_locks >= locks_allocated) {
					locks_allocated *= 2; // grow the array size by powers of 2
					if (locks_allocated > MAX_LOCKS) {
						fprintf(the_stderr, "attempting to allocate %d locks, but MAX_LOCKS is limited at %d\n", locks_allocated, MAX_LOCKS);
						fatal("too many thread synchronization addresses\n");
					}
					locks = realloc((void*) locks, locks_allocated*sizeof(syncrec_t));
					assert(locks != NULL);
				}

				num_locks += 1;
				memset((void*) &locks[i], 0, sizeof(syncrec_t)); // a new lock must be zero'd out
			}
			if (psyncrec->count >= locks[i].count) // should be >, will fix this soon i hope
				locks[i] = *psyncrec;
			else if (!is_repeat) {
				// fprintf(stderr, "warning: lock misorder at address %08x (%d < %d)\n", psyncrec->address, psyncrec->count, locks[i].count);
				assert(psyncrec->count != locks[i].count);
				sync_bypass_count += 1;
				if (locks[i].count - psyncrec->count > num_total_threads) {
					fprintf(parent_stderr, "warning: sync address %08x bypassed illegally (signal 0x%x:%d < 0x%x:%d)\n",
						psyncrec->address, psyncrec->code, psyncrec->count, locks[i].code, locks[i].count);
					illegal_sync_count += 1;
				}
			}
				
			retval = TRUE;
			break;
		case SYNC_WAIT: {
			boolean found_lock = FALSE;
			for (i = 0; i < num_locks; i += 1) {
				if (locks[i].address == psyncrec->address) {
					found_lock = TRUE;
					if (psyncrec->count <= locks[i].count) {
						if (psyncrec->count < locks[i].count && !is_repeat && psyncrec->count != 1) { // strange bug in cholesky for count == 1
							sync_bypass_count += 1;
							if (locks[i].count - psyncrec->count > num_total_threads) {
								fprintf(parent_stderr, "warning: sync address %08x bypassed illegally (0x%x:%d < 0x%x:%d)\n",
									psyncrec->address, psyncrec->code, psyncrec->count, locks[i].code, locks[i].count);
								illegal_sync_count += 1;
							}
						}
						retval = TRUE;
					}
					// else retval = FALSE;
					break;
				}
			}
			if (!found_lock && psyncrec->count == 0)
				retval = TRUE; // for the first guy to lock (SYNC_WAIT) on the mutex
			break;
		}
		case SYNC_TRY_WAIT: { // apparently blastp from the NCBI BLAST package uses pthread_mutex_trylock()
			boolean found_lock = FALSE;
			for (i = 0; i < num_locks; i += 1) {
				if (locks[i].address == psyncrec->address) {
					found_lock = TRUE;
					if (psyncrec->count <= locks[i].count) {
						if (psyncrec->count < locks[i].count && !is_repeat && psyncrec->count != 1) { // strange bug in cholesky for count == 1
							sync_bypass_count += 1;
							if (locks[i].count - psyncrec->count > num_total_threads) {
								fprintf(parent_stderr, "warning: sync address %08x bypassed illegally (0x%x:%d < 0x%x:%d)\n",
									psyncrec->address, psyncrec->code, psyncrec->count, locks[i].code, locks[i].count);
								illegal_sync_count += 1;
							}
						}
					}
					break;
				}
			}
			retval = TRUE; // a successful pthread_mutex_trylock() takes ownership of the mutex
			break;
		}
		default:
			fatal("unrecognized thread synchronization record\n");
	}

	// this deadlock-checking routine isn't super-efficient, but oh well
	// neither is the lock-searching code above
	if (retval) {
		// reset all stall-records if any progress is seen
		for (i = 0; i < processor_num; i += 1) {
			int j;
			for (j = 0; j < cmp_conf[i]; j += 1)
				stalled[i][j] = FALSE;
		}
	}
	else {
		boolean found_progress = FALSE;
		stalled[core][thread] = TRUE;

		for (i = 0; i < processor_num; i += 1) {
			if (!finished_sync[i]) {
				int j;
				for (j = 0; j < cmp_conf[i]; j += 1) {
					if (!stalled[i][j]) {
						found_progress = TRUE;
						break;
					}
				}
				if (found_progress) break;
			}
		}
		if (!found_progress)
			fatal("deadlocked synchronization. Did you properly specify -chipconf all trace files?\n");
	}

	return retval;
}
