/*
** COMPONENT_NAME: MET / Parallel Turandot CMP
**
** GA-0100 (C) Copyright IBM Corp. 1997 - 2003
**         under development at Princeton University 2004 - 2006
**
** The source code for this program is not published or otherwise
** divested of its trade secrets, irrespective of what has been
** deposited within the U.S. Copyright Office.
**
*/

// turandot.c

#define FALSE		0
#define TRUE		1

/*****************************************************************************
  INCLUDES
*****************************************************************************/

#include "gp_parms.h" // generated by compile.smt.pl upon every new build

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
/* #include "config_power4.h" */

#if( !defined(STATS) )
  #define STATS         1
#endif

#if( !defined(STAT0) )
  #define STAT0         1
#endif

#if !defined(POWER_MODE) // PowerTimer
	#define POWER_MODE 1
#endif

// POWER_PERIODIC: periodically report power results, default disabled
#if !defined(POWER_PERIODIC) // unknown whether this works properly at the moment -jdonald
	#define POWER_PERIODIC 0
#endif

/* power delivery system model, default disabled */
#if !defined(PDS_SIM) // Russ Joseph's work on dI/dt
	#define PDS_SIM	0
#endif

#if !defined(FIXED_FREQUENCY) // cycle counts vs scalable time values
	#define FIXED_FREQUENCY 1
#endif

#if !defined(VARIATION)
	#define VARIATION 0
#endif

#if !defined(TEMPERATURE) // HotSpot
	#define	TEMPERATURE 1
#endif

#if !defined(RELIABILITY) // many variables declared, but much of Jayanth's code is missing
	#define RELIABILITY	0
#endif

#if !defined(POSTPROCESS) // this involves postprocess.c and other things, by dbrooks
	#define POSTPROCESS 0
#endif

#if !defined(DUMPSTATUS) // this seems to be just like David's binary input/output, except it has stats for the entire CMP
	#if ( POWER_MODE )
		#define DUMPSTATUS 1
	#else
		#define DUMPSTATUS 0
	#endif
#endif

#if !defined(RETRO_PARMS) // various old parameters for 0.18u experiments
	#define RETRO_PARMS 0
#endif

#if !defined(DYNAMIC_PROFILING) // a mechanism for dynamic SMT fetch/rename for jdonald TACS '05
	#define DYNAMIC_PROFILING 0
#endif

#if !defined(COARSE_STATS) // event counter profiling for input to dynamic thermal simulator jdonald ISCA '06 submission
	#define COARSE_STATS 0
#endif

/*
#if( !defined(STATS) )
  #define STATS		1
#endif

#if( !defined(STAT0) )
  #define STAT0		0
#endif // LYM
*/

#if !defined(MAX_CORES)
	#define MAX_CORES 16 // jdonald
#endif

#if !defined(SMT_MODE)
	#define SMT_MODE TRUE
#endif

#if !defined(MAX_THREADS)
	#define MAX_THREADS 4 // jdonald, still need to add in register configurations to prevent deadlock with 3 or more threads
#endif

/* power models are called every POWER_CYCLES cycles to calculate power dissiplated in this quanta */
// POWER_CYCLES is needed even when we're not in POWER_MODE, because it's the synchronization frequency
// necessary for cycle_reset and other features to happen in a timely manner
#if !defined(POWER_CYCLES)
	#define POWER_CYCLES 100000 // not 0x010000, and my default is to 100000 not 10000, now back to 10000 for the parallel paper
#endif

#if( FIXED_FREQUENCY )
	#define CYCLE_OR_TIMESTAMP cycle
#else // !FIXED_FREQUENCY
	// #define POWER_INTERVAL 0.000006 // 6 us
	#define POWER_INTERVAL (POWER_CYCLES * 1e-9 / REAL_FREQ)
	#define CYCLE_OR_TIMESTAMP timestamp
#endif

#if !defined(POWERMODELDIR)
	/* #define POWERMODELDIR "/gsa/watgsa/projects/t/turandot/turandot_cmp/Sources/turandot/src/power-models" */
	#define POWERMODELDIR "power-models"
#endif

#ifdef __GNUC__
#define ESC "\e"
#else
// This here is to conform better with ISO C standards /jdonald
#define ESC "\033"
#endif

#if 1 // color
#define RED		ESC "[1;31;40m"
#define GREEN	ESC "[1;32;40m"
#define BLUE	ESC "[1;34;40m"
#define CYAN	ESC "[1;36;40m"
#define NORMAL	ESC "[0m"
#else // no color
#define RED		""
#define GREEN	""
#define BLUE	""
#define CYAN	""
#define NORMAL	""
#endif

// types e.g. boolean, xcycle moved up here so I can use them early. -jdonald

/*****************************************************************************
  TYPES
*****************************************************************************/

typedef unsigned char		boolean;
typedef unsigned long long 	xcycle;
typedef unsigned 			xthread_id;
typedef unsigned char		xp2bit;
typedef unsigned short		xread_idx;
typedef struct xiq_info		xiq_info;
typedef unsigned char		xiq_idx;
typedef unsigned char		xiq_cnt;
typedef unsigned short		xiq_regidx;

#include "controls.h" // created a separate file from control_defs.macros
#include <pthread.h> // jdonald

void* run_processor(void*);
inline boolean tempsync_ready();
void benchmark_name(char*, const char*);

void set_frequency_scale(double, int); // available even without POWER_MODE on and FIXED_FREQUENCY off

#if ( POWER_MODE )
void calculate_vscale_penalty(double, double, double*, double*, double*);
double power_rating(double, double);
#endif

#if ( VARIATION )
void gaussian_approx(double*, int, double, double);
void reverse_array(double*, int);
	#if ( POWER_MODE ) // && VARIATION
		#if ( TEMPERATURE )
		void lognormal_approx(double*, int, double, double);
		#endif // TEMPERATURE && POWER_MODE && VARIATION
	void print_variations(int, int);
	int mean_distance_compare(const void*, const void*);
	void median_sort(int, int);
	void profile_sort(int, int);
	double estimate_power_rating(int, int);
	double profile_power_rating(int, int);
	void swap_properties(int, int);
	#endif // POWER_MODE && VARIATION
#endif // VARIATION

// managed heap-allocated strings support by jdonald
char* stralloc(int);
void strfree(char*);
void free_strings();

#define fatal(s) {\
	fprintf(stderr, RED "fatal: %s: line %d: %s: %s" NORMAL, __FILE__, __LINE__, __FUNCTION__, s); \
	exit(EXIT_FAILURE);	\
	}

/*****************************************************************************
  STRUCTURES
*****************************************************************************/

typedef struct {  // comm_data_type
    int access;   // These were created by Yingmin, although with our new
	int end;      // synchronization method no longer needed.
	int realend;  //

// jdonald changing the next three to not require TEMPERATURE
	double    elapsed_time;
	double    total_elapsed_time;
	double    rel_period;

#if( TEMPERATURE && POWER_MODE )
	double *hotspotpower;
	double fxu_temp; // jdonald 2-16-2005
	double fpu_temp;
#endif

	int insn;

#if( DYNAMIC_PROFILING )
	long fxu_accesses[MAX_THREADS];
	long fpu_accesses[MAX_THREADS];
	long fetches[MAX_THREADS];
	long cyclez;
	double fxu_intensity[MAX_THREADS];
	double fpu_intensity[MAX_THREADS];
#if( COARSE_STATS )
	int total_cyclez; // 32 bit integer should be enough for these values
	int total_fetches;
	int total_fxu_accesses;
	int total_fpu_accesses;
	int total_retires;
	int retires;
#endif // COARSE_STATS

#endif /* DYNAMIC_PROFILING */
} comm_data_type; // since comm_data is shared, the above needs to be fixed to handle multiple cores

int processor_num; // total core count, made global, taken from main()

pthread_cond_t  tempsync_cond	 		= PTHREAD_COND_INITIALIZER;
pthread_mutex_t tempsync_cond_mutex 	= PTHREAD_MUTEX_INITIALIZER;

boolean tempsync_ready_flags[MAX_CORES]; // protected by tempsync_cond_mutex
boolean tempsync_done = FALSE; // protected by tempsync_cond_mutex

pthread_cond_t  l2sync_cond 			= PTHREAD_COND_INITIALIZER;
pthread_mutex_t l2sync_cond_mutex 		= PTHREAD_MUTEX_INITIALIZER;

#if( FIXED_FREQUENCY )
long long l2_sync[MAX_CORES]; // no longer volatile, similar array, protected by l2sync_cond_mutex, now using a small array rather than an unnecessary malloc
#else // !FIXED_FREQUENCY
double l2_sync[MAX_CORES];
#endif

long long global_insns_retired[MAX_CORES][MAX_THREADS] = { { 0 } }; // jdonald
xcycle global_cycle[MAX_CORES] = { 0 }; // l2_sync may be a time quantity so we can't count on it for stat aggregation

boolean finished_sync[MAX_CORES]; // also protected by l2sync_cond_mutex jdonald

#ifdef NO_TLS
pthread_mutex_t aria_mutex = PTHREAD_MUTEX_INITIALIZER;
// aria_mutex is to protect against non-reentrant functions such as
// process_iword(), process_branch(), opc_dis(), trace_ffreader_fill_buf(),
// get_entry_trans(), opc_init,  most of which are not thread safe
// because they call aria's functions which use some global storage.
// We cover most of these calls by locking the bodies of reader_fill_buf()
// and get_entry_trans()
#define ARIA_MUTEX_LOCK() pthread_mutex_lock(&aria_mutex)
#define ARIA_MUTEX_UNLOCK() pthread_mutex_unlock(&aria_mutex)
#else
#define ARIA_MUTEX_LOCK()
#define ARIA_MUTEX_UNLOCK()
#endif // thread-local storage doesn't need aria_mutex
 
#if defined(_AIX42) // for AIX's odd pthread implementation
pthread_cond_t join_cond = PTHREAD_COND_INITIALIZER;
pthread_mutex_t join_mutex = PTHREAD_MUTEX_INITIALIZER;
int unjoined_count;
#endif

unsigned regfile_scale; // to force register counts of SMT proportions

boolean goal_attained[MAX_CORES] = {0};
boolean end_simulation = FALSE;

comm_data_type* comm_data; // no longer volatile
int l2_id;
unsigned* l2_probe; // no longer volatile
unsigned* l2_miss;  // no longer volatile
unsigned* l2_snoop; // jdonald
unsigned new_l2usage, old_l2usage;

#define THE_MAX 1000000000000ULL // LYM STRANGE

#if( POWER_MODE )
#include "power_defs.macros" // contains REAL_VDD, REAL_FREQ, flp_cfg, cmp_flp_cfg, etc.

double total_dynamic_power[MAX_CORES] = { 0.0 }; // for reporting at the end
double l2_dynamic_power = 0.0;
#endif

#ifndef REAL_FREQ // in case we're not in POWER_MODE
	#define REAL_FREQ 2.6
#endif

double base_proc_freq = REAL_FREQ * 1e9; // REAL_FREQ is defined in power_defs.macros
double cur_time=0.0, prev_time=0.0; // these have been moved out of init.temp.macros, so we can use with power and freq scaling -jdonald
double elapsed_time = 0.0, total_elapsed_time = 0.0; // also moved here although from turandot.c

#if ( VARIATION )
double timing_sigma_mu = 0.0; // sigma/mu ratio (stdev/average)
double timing_variation[MAX_CORES] = { 1.0 };

boolean variation_median_sort = FALSE;
boolean variation_profile_sort = FALSE;
double power_profiles[MAX_CORES*2];

	#if( TEMPERATURE ) // && VARIATION
	double leakage_sigma_mu = 0.0; // leakage sigma/mu ratio
	double leakage_variation[MAX_CORES] = { 1.0 };
	#endif // TEMPERATURE
#endif // VARIATION

#if( !(FIXED_FREQUENCY) )
double next_tempsync = 0+POWER_INTERVAL; // had to make sure to define this after power_defs.macros
// double next_tempsync = 0+POWER_CYCLES; // had to make sure to define this after power_defs.macros
#endif // !FIXED_FREQUENCY

#if( TEMPERATURE && POWER_MODE )
double total_leakage_power[MAX_CORES] = { 0.0 };
double l2_leakage_power = 0.0;
#endif // TEMPERATURE

#if ( !(FIXED_FREQUENCY) )
// These next three arrays can be filled up beyond processor_num, since
// they may hold entries for idle cores that appear in the floorplan but
// are not being simulated actively /jdonald
double frequency_scale[MAX_CORES] = { 1.0 };
// I decided to make frequency_scale and dynamic_power_vscale as arrays anyway
// since they'll very likely be modified
// However, remember that modifications to any of these must
// only be done at tempsync points /jdonald
#endif // !FIXED_FREQUENCY

#if ( POWER_MODE )
double dynamic_power_vscale[MAX_CORES] = { 1.0 };
#if( TEMPERATURE ) // we only calculate leakage when we have temperature data
double leakage_power_vscale[MAX_CORES] = { 1.0 };

double *unit_area; // a basic TEMPERATURE variable, unlike its siblings

#if( RELIABILITY )
#define T_base 353.0

/*UnitRel *rel_unit;*/
double total_struct_area;
double total_fits;
int rel_counter;
long double access_counter;
double globalcountdown;

/* All the significant reliability variables */

double *old_temp,*currently_positive,*final_diff,*old_set,*delta_T,*delta_counter,*TC_new_cycles,*non_startup,*fits,*TC_fits,*unit_act,*unit_res,*power_density,*mean,*stdev,*countdown,*frequency,*firsttime,*nocycles,*tc_counter,*TC_cycles_allowed;

float act_ratio;
float temp_diff;
float rel_exp;
float fits_ratio;
float temp_ratio;

float TBDratio;
float base_exp;
float new_exp;
float fits_ratio;
#endif // RELIABILITY

#include "init.temp.macros" // probably should be named something like temp_defs.macros
#include "temperature.c"
#endif // TEMPERATURE
#endif // POWER_MODE

#if( DYNAMIC_PROFILING )
  #ifndef PROFILE_MEMORY
  #define PROFILE_MEMORY 100
#endif
	long old_fxu_accesses[MAX_THREADS][PROFILE_MEMORY] = { { 0 } };
	long old_fpu_accesses[MAX_THREADS][PROFILE_MEMORY] = { { 0 } };
	long old_fetches[MAX_THREADS][PROFILE_MEMORY] = { { 0 } };
	long old_cyclez[PROFILE_MEMORY] = { 0 };

	long last_fxu_accesses[MAX_THREADS] = { 0 };
	long last_fpu_accesses[MAX_THREADS] = { 0 };
	long last_fetches[MAX_THREADS] = { 0 };
	long last_cyclez = 0;

	unsigned old_loop = 0;
#endif // DYNAMIC_PROFILING

#if defined(CONTEXT_SWITCH)
unsigned performed_context_switch = 0;
#endif /* CONTEXT_SWITCH */

// unknown what these OS-dependent empty preprocessor constructs do
#if defined(linux) || defined(_AIX) || defined(_MSC_VER) || defined(__CYGWIN32__)
// Linux, AIX, or Cygwin
#endif
#if !defined(linux) && !defined(sparc) && !defined(hpux) && !defined(__hpux) && !defined(__CYGWIN32__) && !defined(ultrix)
// anything other than Linux, SPARC, HP-UX, Cygwin, pux, Cygwin, or Ultrix
#endif
#if defined(__svr4__) || defined(__osf__) || defined(sparc) && defined(__unix__)
// Solaris or Alpha or SPARC or SysV unix
#endif
#if defined(__APPLE__) && defined(__MACH__)
// Mac OS X /jdonald
#endif

FILE*			the_stderr;  		// global here, local in run_processor()
FILE*			parent_stderr;		// so children can send global (usu. temperature) data to the right output
FILE*			tempout_fp;
boolean			stdin_used = FALSE; // for input

char*			trace_fname[MAX_CORES][MAX_THREADS] = { { 0 } }; // jdonald making this global

/* for stopping at any cycle */
unsigned		cycle_stop = 0;

// variables made global since needed by utils_cmdline.macros and used later
unsigned		trace_from;
/* support for trace_to, zhigangh@1/10/2003 */
unsigned		trace_to;

/* detailed monitoring, zhigangh@1/10/2003 */
unsigned		detail_from, detail_to, detail_step;

// reader/insn_info.h contains typedefs and macros but no instantiated variables
#include "insn_info.h"

#if( !defined(USING_IM ) )
#define USING_IM 0
#endif

#define lg2(x)	( (((x)>>1) == 0) ? 0 : \
			  (((x)>>2) == 0) ? 1 : \
			  (((x)>>3) == 0) ? 2 : \
			  (((x)>>4) == 0) ? 3 : \
			  (((x)>>5) == 0) ? 4 : \
			  (((x)>>6) == 0) ? 5 : \
			  (((x)>>7) == 0) ? 6 : \
			  (((x)>>8) == 0) ? 7 : \
			  (((x)>>9) == 0) ? 8 : \
			  (((x)>>10) == 0) ? 9 : \
			  (((x)>>11) == 0) ? 10 : \
			  (((x)>>12) == 0) ? 11 : \
			  (((x)>>13) == 0) ? 12 : \
			  (((x)>>14) == 0) ? 13 : \
			  (((x)>>15) == 0) ? 14 : \
			  (((x)>>16) == 0) ? 15 : \
			  (((x)>>17) == 0) ? 16 : \
			  (((x)>>18) == 0) ? 17 : \
			  (((x)>>19) == 0) ? 18 : \
			  (((x)>>20) == 0) ? 19 : \
			  (((x)>>21) == 0) ? 20 : \
			  (((x)>>22) == 0) ? 21 : \
			  (((x)>>23) == 0) ? 22 : \
			  (((x)>>24) == 0) ? 23 : \
			  (((x)>>25) == 0) ? 24 : \
			  (((x)>>26) == 0) ? 25 : \
			  (((x)>>27) == 0) ? 26 : \
			  (((x)>>28) == 0) ? 27 : \
			  (((x)>>29) == 0) ? 28 : \
			  (((x)>>30) == 0) ? 29 : \
			  (((x)>>31) == 0) ? 30 : \
			  31)

/** maximum number of architected registers per class **/
#if( !defined(ARCH_MAX) )
#define		ARCH_MAX		32
#endif
#if( !defined(LG2ARCH_MAX) )
#define		LG2ARCH_MAX		lg2(ARCH_MAX)
#else
  #if( lg2(ARCH_MAX) != LG2ARCH_MAX )
    #error "lg2(ARCH_MAX) != LG2ARCH_MAX"
  #endif
#endif

// problem: what if these files have global variables and non-reentrant functions? -jdonald

#if( USING_FF ) // jdonald: only this one is being tested
  #include "trace_ffreader.h"  // ffreader/trace_ffreader.h contains no variables
  #include "trace_decompress.h"
  #include "dep_prep_info.h"

#elif( USING_ROCH )
  #include "trace_rochreader.h"
  #include "dep_prep_info.h"

#elif( USING_ARIA )
  #include "reader_prep_info.h"
  #include "dep_prep_info.h"
  #include "aria_reader.h"

#elif( USING_IM )
  #include "trace_imreader.h"
  #include "dep_prep_info.h"

#else
  #include "trace_reader.h"
#endif

#include "trauma.h"
#include "opc.h"

#if( !defined(FIX_UNIT) )
  #define FIX_UNIT	0
#endif
#if( !defined(FPU_UNIT) )
  #define FPU_UNIT	1
#endif
#if( !defined(MEM_UNIT) )
  #define MEM_UNIT	2
#endif
#if( !defined(BR_UNIT) )
  #define BR_UNIT	3
#endif
#if( !defined(LOG_UNIT) )
  #define LOG_UNIT	4
#endif
#if( !defined(CMPLX_UNIT) )
  #define CMPLX_UNIT	5
#endif


#if( COMBINE_FIX_MEM )
  #undef FIX_UNIT
  #define FIX_UNIT	MEM_UNIT
#endif

#if( !defined( MONITOR_CYCLES ) )

  #if( TRACE )
    #define MONITOR_CYCLES 0x010000
  #elif( USING_ARIA )
    #define MONITOR_CYCLES 0x0040000
  #else
    #define MONITOR_CYCLES 0x0040000
  #endif

#endif

/*****************************************************************************
  DEBUG
*****************************************************************************/
#if( !defined(DEBUG) )
  #define DEBUG	0
#endif

// the below code is disabled, since NDEBUG should stay undefined (not defined as zero to confuse assert.h)
#if 0
#if( !defined(NDEBUG) && !DEBUG  )
  #define NDEBUG	1
#endif
#endif

#if( !defined(VERIFY) )
  #define VERIFY	0
#endif

#if( !defined(GETCPI) )
  #define GETCPI	1
#endif

#if( !defined(TIMELINE) )
  #define TIMELINE	1
#endif

#if( !defined(GETCPI_REMAINING) )
  #define GETCPI_REMAINING	0
#endif

#if( DEBUG )
  #define DBG(x)	((void)(x))
#else
  #define DBG(x)	((void)(0))
#endif

#if( TRACE )
  #define TRCOUT(x)	((cycle>trace_from && cycle < trace_to)?(void)(x):(void)0)
//  #define TRCOUT(x) ((void)(x))
// going back to original problems in dep_prep_process.c
  #define TRC(x)	((void)(x))
#else
  #define TRCOUT(x)	((void)0)
  #define TRC(x)	((void)0)
#endif

#if( STATS )
  #define STT(x)	((void)(x))
#else
  #define STT(x)	((void)(0))
#endif

#if( STAT0 )
  #define ST0(x)	((void)(x))
#else
  #define ST0(x)	((void)(0))
#endif

#if( GET_SPEC_STATS )
  #define STATSPEC(x)	((void)(x))
#else
  #define STATSPEC(x)	((void)(0))
#endif

#if( MAPOUT )
  #define MPO(x)	((void)(x))
#else
  #define MPO(x)	((void)(0))
#endif

#if( GETCPI || GETCPI_REMAINING )
  #define CPI(x)	((void)(x))
#else
  #define CPI(x)	((void)(0))
#endif

#if( TIMELINE )
  #define SETTL(thread_,i,n,c)	(xiq_tl_status(iq[thread_][i])[ (n) % tl_size ] = (c))
  #define TL(x)		((void)(x))
#else
  #define SETTL(thread_,i,n,c)		((void)0)
  #define TL(x)		((void)0)
#endif

#define fprint_unsigned(fp,size,n)		\
  {						\
    int		_i;				\
    char	_buf[16];			\
    unsigned	_num = n;			\
    unsigned	_mod;				\
						\
    _buf[15] = 0;				\
    _buf[14] = '0';				\
    for( _i = 1; _i < size; _i++ ) {		\
      _buf[14-_i] = ' ';			\
    }						\
						\
    for( _i = 0; _num != 0; _i++ ) {		\
      _mod = _num % 10;				\
      _buf[14-_i] = _mod + '0';			\
      _num = (_num-_mod)/10;			\
    }						\
    if( _i >= size ) {				\
      fputs(&_buf[15-_i], fp);			\
    }						\
    else {					\
      fputs(&_buf[15-size], fp);		\
    }						\
  }

#if( 0 || (!NDEBUG && VERIFY) )
  #define AFTER_EACH_STAGE()			\
  {						\
    int	i_;					\
    int	j_;					\
    int	k_;					\
    int	thread_;					\
    int	outs_;					\
    int	ins_;					\
    int idx_;					\
    xarch *		out_regs_;		\
    unsigned short	phys_;			\
    unsigned short	arch_;			\
						\
    assert( ibuf_left <= IBUF_MAX );		\
    verifyQ( fix_n, fix_hd, fix_tl);		\
    verifyQ( fpu_n, fpu_hd, fpu_tl);		\
    verifyQ( br_n, br_hd, br_tl);		\
    verifyQ( mem_n, mem_hd, mem_tl);		\
    verifyQ( dmiss_n, dmiss_hd, dmiss_tl);	\
    verifySortedQ( dmiss_n, dmiss_hd, dmiss_tl);	\
    assert( dmiss_n != 0 || dmiss_loaded != DCACHE_SECTORS || dmiss_early == THE_MAX );	\
    assert( dmiss_n == 0 || dmiss_loaded != DCACHE_SECTORS || xiq_cycle( iq[ dmiss_hd ] ) <= dmiss_early ); \
    assert( freeq_tl[GPR_CLASS] - GPR_CLASS*PHYS_MAX >= 0 ); \
    assert( freeq_tl[GPR_CLASS] - GPR_CLASS*PHYS_MAX <= gpr_phys - GPR_ARCH ); \
    assert( freeq_tl[FPR_CLASS] - FPR_CLASS*PHYS_MAX >= 0 ); \
    assert( freeq_tl[FPR_CLASS] - FPR_CLASS*PHYS_MAX <= fpr_phys - FPR_ARCH ); \
    assert( freeq_tl[CCR_CLASS] - CCR_CLASS*PHYS_MAX >= 0 ); \
    assert( freeq_tl[CCR_CLASS] - CCR_CLASS*PHYS_MAX <= ccr_phys - CCR_ARCH ); \
    assert( freeq_tl[SPR_CLASS] - SPR_CLASS*PHYS_MAX >= 0 ); \
    assert( freeq_tl[SPR_CLASS] - SPR_CLASS*PHYS_MAX <= spr_phys - SPR_ARCH ); \
    assert( freeq_tl[FPR_CLASS] - FPR_CLASS*PHYS_MAX == ((freeq_avail_vec >> (8*FPR_CLASS)) &0xff)); \
    assert( freeq_tl[CCR_CLASS] - CCR_CLASS*PHYS_MAX == ((freeq_avail_vec >> (8*CCR_CLASS)) &0xff)); \
    assert( freeq_tl[SPR_CLASS] - SPR_CLASS*PHYS_MAX == ((freeq_avail_vec >> (8*SPR_CLASS)) &0xff)); \
    if( freeq_tl[GPR_CLASS] - GPR_CLASS*PHYS_MAX != ((freeq_avail_vec >> (8*GPR_CLASS)) &0xff) ) { \
      fprintf(the_stderr, "freeq_tl = %d, avail_vec = %d\n", freeq_tl[GPR_CLASS], freeq_avail_vec); \
    } \
    assert( freeq_tl[GPR_CLASS] - GPR_CLASS*PHYS_MAX == ((freeq_avail_vec >> (8*GPR_CLASS)) &0xff)); \
    assert( branch_pred >= 0 );			\
    assert( branch_pred_delta >= 0 );		\
    assert( branch_pred > branch_pred_delta || branch_pred == 0 || ifetch_mispredict_done ); \
    for(thread_ = 0; thread_ < n_threads; thread_++){							   \
	    for( i_ = 0; i_ < GPR_ARCH; i_++ ) {	\
	      assert( (rgrename[thread_][(GPR_CLASS<<LG2ARCH_MAX)|i_]>>LG2PHYS_MAX) ==  GPR_CLASS );	\
	      assert( (inorder[thread_][(GPR_CLASS<<LG2ARCH_MAX)|i_] >>LG2PHYS_MAX) ==  GPR_CLASS );	\
	    }						\
	    for( i_ = 0; i_ < FPR_ARCH; i_++ ) {	\
	      assert( (rgrename[thread_][(FPR_CLASS<<LG2ARCH_MAX)|i_]>>LG2PHYS_MAX) ==  FPR_CLASS );	\
	      assert( (inorder[thread_][(FPR_CLASS<<LG2ARCH_MAX)|i_] >>LG2PHYS_MAX) ==  FPR_CLASS );	\
	    }						\
	    for( i_ = 0; i_ < SPR_ARCH; i_++ ) {	\
	      assert( (rgrename[thread_][(SPR_CLASS<<LG2ARCH_MAX)|i_]>>LG2PHYS_MAX) ==  SPR_CLASS );	\
	      assert( (inorder[thread_][(SPR_CLASS<<LG2ARCH_MAX)|i_] >>LG2PHYS_MAX) ==  SPR_CLASS );	\
	    }						\
	    for( i_ = 0; i_ < CCR_ARCH; i_++ ) {	\
	      assert( (rgrename[thread_][(CCR_CLASS<<LG2ARCH_MAX)|i_]>>LG2PHYS_MAX) ==  CCR_CLASS );	\
	      assert( (inorder[thread_][(CCR_CLASS<<LG2ARCH_MAX)|i_] >>LG2PHYS_MAX) ==  CCR_CLASS );	\
	    }						\
	}						\
    for( i_ = iq_retire; i_ != iq_ren; i_ = (i_+1)%IQ_SIZE ) {	\
      thread_ = xiq_thread_id( iq[i_] );		\
      outs_ = xiq_outs( iq[i_] );		\
      out_regs_ = xiq_out_regs( iq[i_] );	\
      ins_ = xiq_ins( iq[i_] );			\
      assert( ins_ + outs_ < IQ_LARGE_REG );	\
      idx_ = xiq_reg_idx( iq[i_] );		\
								\
      for( j_ = 0; j_ < outs_; j_++ ) {				\
	phys_ = iq_reg[idx_ + ins_ + j_];			\
	arch_ = out_regs_[j_];					\
	assert( (phys_>>LG2PHYS_MAX) == (arch_>>LG2ARCH_MAX) );	\
								\
	for( k_ = 0; k_ < ins_; k_++ ) {			\
	  assert( iq_reg[idx_ + k_ ] != phys_ );		\
	}							\
								\
	if( GPR_CLASS == (phys_>>LG2PHYS_MAX) ) {		\
	  for( k_ = GPR_CLASS*PHYS_MAX; k_ < freeq_tl[GPR_CLASS]; k_++ ) { \
	    assert( freeq[k_] != phys_ && 1);	\
	  }					\
	  for( k_ = 0; k_ < GPR_ARCH; k_++ ) { \
	    if( phys_ == inorder[thread_][(GPR_CLASS<<LG2ARCH_MAX)|k_] ) {	\
	      fprintf(the_stderr, "i,j,k = %d,%d,%d phys/arch %d,%d\n",	\
		i_,j_,k_,phys_,arch_);					\
	    }								\
	    assert( (phys_ != inorder[thread_][(GPR_CLASS<<LG2ARCH_MAX)|k_]) && 5);		\
	  }						\
	}					\
      }						\
      for( j_ = 0; j_ < outs_; j_++ ) {		\
	phys_ = iq_reg[idx_ + j_];		\
	if( GPR_CLASS == (phys_>>LG2PHYS_MAX) ) {		\
	  for( k_ = GPR_CLASS*PHYS_MAX; k_ < freeq_tl[GPR_CLASS]; k_++ ) { \
	    assert( freeq[k_] != phys_ && 4);	\
	  }					\
	}					\
      }						\
    }						\
									\
    for(thread_ = 0; thread_ < n_threads; thread_++){							   \
	    for( i_ = 0; i_ < GPR_ARCH; i_ ++ ) {				\
	      arch_ = rgrename[thread_][(GPR_CLASS<<LG2ARCH_MAX)|i_];			\
	      for( j_ = GPR_CLASS*PHYS_MAX; j_ < freeq_tl[GPR_CLASS]; j_++ ) {	\
		phys_ = freeq[j_];						\
		if( arch_ == phys_ ) { 	\
		fprintf(the_stderr, "i,j %d %d arch phys %d %d\n", i_, j_, arch_, phys_); \
		}	\
		assert( arch_ != phys_ && 2);					\
	      }									\
	    }									\
										\
	    for( i_ = 0; i_ < GPR_ARCH; i_ ++ ) {	\
	      arch_ = inorder[thread_][(GPR_CLASS<<LG2ARCH_MAX)|i_];	\
	      for( j_ = GPR_CLASS*PHYS_MAX; j_ < freeq_tl[GPR_CLASS]; j_++ ) { \
		phys_ = freeq[j_];			\
		assert( arch_ != phys_ && 3);		\
	      }						\
	    }						\
	}						\
    for( i_ = iq_hd; i_ != iq_tl; i_ = (i_+1)%IQ_SIZE ) {	\
      if( xiq_trauma(iq[i_]) == TR_ILLEGAL ) {		\
	fprintf(the_stderr, "%s %d: found illegal trauma, %d\n", __FILE__, __LINE__, i_); \
      }						\
      assert( xiq_trauma(iq[i_]) != TR_ILLEGAL );\
    }						\
    {									\
      unsigned	i_;							\
      unsigned	j_;							\
      unsigned	wait_;							\
      unsigned	at_;							\
      unsigned	at1_;							\
      unsigned	id_;							\
								    	\
      assert( len_memq(storeq_lo, storeq_hi, STOREQ_SIZE) + storeq_left == STOREQ_LENGTH ); \
      assert( len_memq(reorderq_lo, reorderq_hi, REORDERQ_SIZE) + reorderq_left == REORDERQ_LENGTH ); \
      for( i_ = iq_hd; i_ != iq_tl; i_ = (i_+1)%IQ_SIZE ) {		\
	assert( !xiq_is_store(iq[i_]) || !xiq_is_load(iq[i_]) );	\
      } 								\
      for( i_ = storeq_retired; i_ != storeq_hi; i_ = (i_+1)%STOREQ_SIZE ) {\
	id_ = storeq[i_][2];						\
	assert( in_iq( id_ ) );						\
	for( j_ = 0, wait_= storeq[i_][4]; wait_; j_++, wait_ >>= 1 ) {	\
	  if( wait_ & 1 ) {						\
	    assert( in_memq( j_, reorderq_lo, reorderq_hi ) );		\
	    at_ = reorderq[j_][2];					\
	    assert( in_iq( at_ ) );					\
	    assert( !xiq_is_store( iq[at_] ) );				\
	  }								\
	}								\
      }									\
      for( i_ = reorderq_lo; i_ != reorderq_hi; i_ = (i_+1)%REORDERQ_SIZE ) { \
	at1_ = reorderq[i_][2];						\
	assert( in_iq( at1_ ) );					\
	assert( !xiq_is_store( iq[at1_] ) );				\
      }									\
    }									\
    {									\
      unsigned	i_;							\
      unsigned	at_;							\
      for( i_ = reorderq_lo; i_ != reorderq_hi; i_ = (i_+1)%REORDERQ_SIZE ) { \
	at_ = reorderq[i_][2];						\
	if( !(xiq_is_busy(iq[at_]) || xiq_reorder( iq[at_] ) != 0) ) {	\
	  fprintf(the_stderr, "reorderq idx = %d at=%d\n", i_, at_);	\
	}								\
	assert( xiq_is_memop(iq[at_]) );				\
	assert( !xiq_is_store( iq[at_] ) );				\
	assert( xiq_is_load( iq[at_] ) );				\
	assert( !xiq_is_notaddr( iq[at_] ) );				\
	assert( xiq_is_busy(iq[at_]) || xiq_reorder( iq[at_] ) != 0 );	\
      }									\
    }									\
  }									\
  {								\
    unsigned	i_;						\
    unsigned	bn_= 0;						\
    unsigned	bp_= 0;						\
    int		_is_branch;					\
    int		_is_predop;					\
    int		_is_bc;						\
    int		_is_brcrg;					\
    int		_is_bcctr;					\
    int		_is_bclr;					\
    xdep_prep_info *	entry_;					\
								\
    for( i_ = iq_retire; i_ != iq_tl; i_ = (i_ +1)%IQ_SIZE ) {	\
      entry_ = xiq_entry( iq[i_] );				\
      _is_branch = xdep_prep_info_is_branch( entry_ );		\
      _is_predop = USE_PREDICATION && xdep_prep_info_is_pred(entry_);	\
      _is_bc = xdep_prep_info_is_brcond( entry_ );		\
      _is_brcrg = xdep_prep_info_is_brcrg( entry_ );		\
      _is_bcctr = xdep_prep_info_is_brctr( entry_ );		\
      _is_bclr = xdep_prep_info_is_brlink( entry_ );		\
								\
      if( (xiq_is_busy( iq[i_] ) || xiq_cycle(iq[i_]) == THE_MAX) &&	\
	  !PERFECT_BRANCH_PREDICTION && _is_branch && !_is_predop &&	\
	   ( ( _is_bc && !PERFECT_COND_BRANCHES ) ||		\
	    ( _is_bclr && !PERFECT_LINK_BRANCHES ) ||		\
	    ( _is_bcctr && !PERFECT_CTR_BRANCHES ) ) ) {	\
	bn_++;							\
	bp_ += xiq_is_mispredicted(iq[i_]);			\
      }								\
								\
    }								\
    if( bn_ != branch_pred || bp_ != branch_pred_delta ) {	\
      fprintf(the_stderr, "n,p=%d %d pred,delta=%d %d\n",	\
	bn_, bp_, branch_pred, branch_pred_delta);		\
    }								\
    assert( bn_ == branch_pred );				\
    assert( bp_ == branch_pred_delta );				\
  }								\
  {								\
    xiq_idx	i_;						\
    int		n_ = 0;						\
    if( iq_dec != iq_tl ) {					\
      i_ = (iq_tl + IQ_SIZE -1 )%IQ_SIZE;			\
      assert( xiq_fchblk(iq[i_]));				\
    }								\
    if( iq_dec == iq_tl ) {					\
      if( FCHBLK_MAX != fchblk_left ) {				\
	fprintf(the_stderr, "MAX=%d, left=%d\n", FCHBLK_MAX, fchblk_left);	\
      }								\
      assert( FCHBLK_MAX == fchblk_left );			\
    }								\
    for( i_ = iq_dec; i_ != iq_tl; i_ = (i_+1)%IQ_SIZE ) {	\
      n_ += xiq_fchblk(iq[i_]);					\
    }								\
    assert( FCHBLK_MAX - n_ == fchblk_left );			\
  }
#else
  #define AFTER_EACH_STAGE()	((void)0)
#endif

/*****************************************************************************
  DEFINES
*****************************************************************************/


#if( !defined(min) )
#define min(a,b)	(((a)<(b))?(a):(b))
#endif

#if( !defined(max) )
#define max(a,b)	(((a)>(b))?(a):(b))
#endif

/*** Registers ***/
/* check if all operand regs are ready by cycle_, don't record trauma */
#define checkInRegs( thread_, idx_, ins_, cycle_, not_found_label_) \
  {					\
    unsigned	j_;			\
    unsigned	tmp_reg_;		\
    xcycle	tmp_cycle_; /* 	LYM */		\
					\
    for( j_ = 0; j_ < ins_; j_++ ) {	\
      assert( (unsigned)(idx_ + j_) < IQ_TOTAL_REG );	\
      tmp_reg_ = iq_reg[thread_][ idx_ + j_ ];	\
      assert( tmp_reg_ < PHYS_TOTAL );	\
      tmp_cycle_ = phys[thread_][ tmp_reg_ ];	\
      if( tmp_cycle_ > cycle_ ) {	\
	DBG( fprintf(the_stderr,		\
	      "reg %d wait till %u\n",	\
	      tmp_reg_, tmp_cycle_) );	\
	goto not_found_label_;		\
      }					\
    }					\
  }

/* check if all operand regs are ready by cycle_, record trauma */
#define checkInRegsCPI(thread_, iq_, idx_, ins_, cycle_, not_found_label_) \
  {					\
    unsigned	j_;			\
    unsigned	tmp_reg_;		\
    xcycle	tmp_cycle_;	/* LYM */	\
					\
    for( j_ = 0; j_ < ins_; j_++ ) {	\
      assert( (unsigned)(idx_ + j_) < IQ_TOTAL_REG );	\
      tmp_reg_ = iq_reg[thread_][ idx_ + j_ ];	\
      assert( tmp_reg_ < PHYS_TOTAL );	\
      tmp_cycle_ = phys[thread_][ tmp_reg_ ];	\
      if( tmp_cycle_ > cycle_ ) {	\
	DBG( fprintf(the_stderr,		\
	      "reg %d wait till %u\n",	\
	      tmp_reg_, tmp_cycle_) );	\
	assert( tmp_reg_ < PHYS_TOTAL );	\
	bounds( iq_, IQ_SIZE );	\
	CPI( xiq_trauma( iq[thread_][iq_] ) = src_phys[thread_][ tmp_reg_ ] ); \
	goto not_found_label_;		\
      }					\
    }					\
  }

/* set time of ready for output registers */
#define setOutRegs(thread_, idx_, ins_, outs_, cycle_ ) \
  {					\
    unsigned	j_;			\
    unsigned	tmp_reg_;		\
    unsigned	tmp_idx_;		\
					\
    tmp_idx_ = idx_ + ins_;		\
    for( j_ = 0; j_ < outs_; j_++ ) {	\
      assert( tmp_idx_ + j_ < IQ_TOTAL_REG );	\
      tmp_reg_ = iq_reg[thread_][tmp_idx_ + j_];	\
      assert( tmp_reg_ < PHYS_TOTAL );	\
      phys[thread_][ tmp_reg_ ] = cycle_;	\
      DBG( fprintf(the_stderr,		\
	      "reg %d set to %u\n",	\
	      tmp_reg_, cycle_) );	\
    }					\
  }

/*** instruction Qs ***/
/* remove elem_ from queue(head_, tail_, size_) knowing prev_ and next_ */
#define rem_nextQ(thread_, elem_, prev_, next_, size_, head_, tail_, done_label_)\
  {					\
					\
    size_--;				\
    if( prev_ == IQ_NULL ) {		\
      head_ = next_;			\
    }					\
    else {				\
      xiq_next( iq[thread_][bounds(prev_,IQ_SIZE)] ) = next_;	\
    }					\
    if( elem_ == tail_ ) {		\
      tail_ = prev_;			\
      goto done_label_;			\
    }					\
  }

/* remove elem_ from queue(head_, tail_, size_) knowing only prev_ */
#define remQ(thread_, elem_, prev_, size_, head_, tail_, done_label_)	\
  {					\
    unsigned tmp_next_;			\
					\
    size_--;				\
    bounds( elem_, IQ_SIZE );		\
    tmp_next_ = xiq_next( iq[thread_][elem_] );	\
    if( prev_ == IQ_NULL ) {		\
      head_ = tmp_next_;		\
    }					\
    else {				\
      bounds( prev_, IQ_SIZE );	\
      xiq_next( iq[thread_][prev_] ) = tmp_next_;\
    }					\
    if( elem_ == tail_ ) {		\
      tail_ = prev_;			\
      goto done_label_;			\
    }					\
  }

/* add elem_ to an empty queue (head_, tail_, size_) */
#define add0Q(thread_, elem_, size_, head_, tail_ ) \
  {					\
    assert( size_ == 0 );		\
    head_ = elem_;			\
    tail_ = elem_;			\
  }

/* add elem_ to an non-empty queue (head_, tail_, size_) */
#define addNQ(thread_, elem_, size_, head_, tail_ ) \
  {					\
    assert( size_ != 0 );		\
    bounds( tail_, IQ_SIZE );		\
    xiq_next( iq[thread_][tail_] ) = elem_;	\
    tail_ = elem_;			\
  }

#define addQ(thread_, elem_, size_, head_, tail_ ) \
  {					\
    if( size_ == 0 ) {			\
      add0Q(thread_, elem_, size_, head_, tail_ ); \
    }					\
    else {				\
      addNQ(thread_, elem_, size_, head_, tail_ ); \
    }					\
    size_++;				\
  }

#define verifyQ(thread_, size_, head_, tail_ )	\
  {					\
    unsigned	i_;			\
    unsigned	at_;			\
					\
    assert( size_ >= 0 );		\
    if( size_ != 0 ) {			\
      at_ = head_;			\
      i_ = 1;				\
      while( at_ != tail_ && i_ < size_) {	\
	bounds( at_, IQ_SIZE );	\
	at_ = xiq_next( iq[thread_][at_] );	\
	i_++;				\
      }					\
      assert( i_ == size_ );		\
      assert( at_ == tail_ );		\
    }					\
  }

#define verifySortedQ( thread_, size_, head_, tail_ )	\
  {						\
    unsigned	at_;				\
    unsigned	next_;				\
    if( size_ > 0 ) {				\
      at_ = head_;				\
      while( at_ != tail_ ) {			\
	bounds( at_, IQ_SIZE );		\
	next_ = xiq_next( iq[thread_][ at_ ] );		\
	bounds( next_, IQ_SIZE );		\
	assert( xiq_cycle( iq[thread_][at_ ] ) <= xiq_cycle( iq[thread_][ next_ ] )); \
	at_ = next_;				\
      }						\
    }						\
  }


/*** arrays, statistics, misc ***/
#define clearArray( size_, array_ )		\
  {						\
    unsigned	j_;				\
    for( j_ = 0; j_ < size_; j_++ ) {		\
      array_[j_] = 0;				\
    }						\
  }

/* dump histograms */
#define dumpStats( total_, size_, array_, tag_ )	\
  {						\
    unsigned	tmp_total_;			\
    unsigned	num_;				\
    unsigned	div_;				\
    unsigned	hi_;				\
    unsigned	frac_;				\
    unsigned	n_;				\
    unsigned	j_;				\
    unsigned	scale_;				\
    unsigned	weight_;			\
						\
    scale_ = total_/10000;			\
    for( n_ = (size_); n_ > 0; n_ -- ) {	\
      if( array_[n_] ) {			\
	break;					\
      }						\
    }						\
    tmp_total_ = 0;				\
    weight_ = 0;				\
    fprintf(the_stderr, "@@ %s\n", tag_);		\
    for( j_ = 0; j_ <= n_; j_++ ) {		\
      num_ = array_[j_];			\
      if( total_ < (1<<18) ) {			\
	div_ = (num_*10000)/total_;		\
      }						\
      else {					\
	div_ = num_/scale_;			\
      }						\
      frac_ = div_%100;				\
      hi_ = (div_ - frac_);			\
      hi_ = hi_/100;				\
      fprintf(the_stderr, "@@ \t%2u:%-8u\t%2d.%02d\n", j_, num_, hi_, frac_);\
      tmp_total_ += num_;			\
      weight_ += num_*j_;			\
    }						\
    fprintf(the_stderr, "@@ = %u %u\n", tmp_total_, weight_);	\
    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" ); \
  }

#if( TRACE )

/* ************************* WARNING *******************************************/
/* Warning:  If you change anything in this enumerated type,
   you must coordinate the changes with the structures in trace.macros 	*/
/* ************************* WARNING *******************************************/


  enum xstatus {
    XSTATUS_UNKNOWN,	/* This is also the invalide status to start with */
    XSTATUS_IBUF,
    XSTATUS_DECODE,
    XSTATUS_DECODE_1,
    XSTATUS_DECODE_2,
    XSTATUS_DECODE_3,
    XSTATUS_DECODE_4,
    XSTATUS_RENAME,
    XSTATUS_RESV_BR,
    XSTATUS_RESV_LOG,
    XSTATUS_RESV_CMPLX,
    XSTATUS_RESV_FIX,
    XSTATUS_RESV_FIX1,
    XSTATUS_RESV_MEM,
    XSTATUS_RESV_MEM1,
    XSTATUS_RESV_FPU,
    XSTATUS_RESV_FPU1,
    XSTATUS_EXEC_BR,
    XSTATUS_EXEC_LOG,
    XSTATUS_EXEC_CMPLX,
    XSTATUS_EXEC_FIX,
    XSTATUS_EXEC_FIX1,
    XSTATUS_EXEC_MEM,
    XSTATUS_EXEC_MEM1,
    XSTATUS_EXEC_FPU,
    XSTATUS_EXEC_FPU1,
    XSTATUS_STQ_DATA,
    XSTATUS_BANK_CONFLICT,
    XSTATUS_DCACHE_MISS,
    XSTATUS_L2CACHE_MISS,
    XSTATUS_MISS_COMPLETE,
    XSTATUS_WRITEBACK,
    XSTATUS_RETIRE,
    XSTATUS_REALLY_RETIRE,
    XSTATUS_STORE_QUEUE,
    XSTATUS_LOAD_QUEUE,
    XSTATUS_NUM
  };

  typedef enum xstatus xstatus;

  static char * xstatus_name[] = {
    "unknown   ",
    "ibuf      ",
    "decode    ",
    "decode_1  ",
    "decode_2  ",
    "decode_3  ",
    "decode_4  ",
    "rename    ",
    "resv_br   ",
    "resv_log  ",
    "resv_cmplx",
    "resv_fix  ",
    "resv_fx1  ",
    "resv_mem  ",
    "resv_mm1  ",
    "resv_fpu  ",
    "resv_fpu1 ",
    "exec_br   ",
    "exec_log  ",
    "exec_cmplx",
    "exec_fix  ",
    "exec_fx1  ",
    "exec_mem  ",
    "exec_mm1  ",
    "exec_fpu  ",
    "exec_fpu1 ",
    "stq_data  ",
    "conflict  ",
    "dmiss     ",
    "l2miss    ",
    "did_miss  ",
    "exec_fin  ",
    "writback  ",
    "retire    ",
    "store_q   ",
    "load_q    "
  };


  static char * xregclass_name[] = {
    "r", "f", "c", "s"
  };
#endif

#include "iq.h" // defines structs and all sorts of obscure macros, but no variables

typedef unsigned char	xregclass;

#if( WIDE_ARCH_REG )
typedef unsigned short	xarch_reg;
#else
typedef unsigned char	xarch_reg;
#endif

typedef unsigned short	xphys_reg;
typedef unsigned short	xphys_cnt;
typedef unsigned char	xunit;
typedef unsigned char	xresv_cnt;

/* for easier debugging of dep_prep_process.c, zhigangh@1/11/2003 */
#include "utils_trans.macros"
// for some reason the program segfaults when this is #included locally in run_processor()
// will look into it

// AIX-specific "assert.h" no longer used
#include <assert.h>
inline int bounds(unsigned i, unsigned bound) {
	assert( i < bound );
	return i;
}
// #define bounds(i,bound)	__builtin_choose_expr( sizeof(i) <= 1 && bound >= 256, i, (assert( (unsigned)(i)<(bound) ), i) )
// I ended up turning this macro into an inline function so gcc would stop warning about bound >= 256 comparisons on char types

boolean		shared_memory = FALSE; // for parallel programs, -shared option, default off
boolean		thread_synchronization = TRUE; // -sync option, default on (since it gets ignored if there are no sync points anyway)
#if( USE_INSN_CUTOFF )
boolean		use_insn_cutoff = TRUE;
#else
boolean		use_insn_cutoff = FALSE;
#endif
boolean		group_cutoff = TRUE;
#if( !(FIXED_FREQUENCY) )
double dvfs_scale_knob = 1.0; // for now we choose one scaling frequency statically chosen across all cores
#endif

/* new array modeling, zhigangh@1/14/2003 */
#include "array.h" // contains plenty of macros and protos
#include "array.c" // contains function definitions

#if( !defined(STOREQ_SIZE) )
#define         STOREQ_SIZE             (STOREQ_LENGTH+1)
#endif

#if( USING_LOG_UNIT )
  #if( !defined( LOG_NUM ) )
  #define               LOG_NUM                 1
  #endif
#endif

/* CP FROM MEMQ.MACROS */

/* type def for status dump LYM */
/* LYM CMP */

#if( POWER_MODE && DUMPSTATUS )
typedef struct { // dumpstatus_type
      xcycle cycle;

	  // why are some of these unsigned but in power_defs.macros the non-struct variables were all long long?
      unsigned insns[MAX_THREADS];
      unsigned memop[MAX_THREADS];
      unsigned insns_retired[MAX_THREADS];
      unsigned insns_total_retired[MAX_THREADS];
      unsigned insns_total_execd[MAX_THREADS];
      unsigned insns_total_dispd[MAX_THREADS];

      long long stats_iprobe[MAX_THREADS];  /* LYM */
      long long stats_imiss[MAX_THREADS];  /* LYM */
      long long stats_iprefetch;  /* LYM */ // put this back in since it's a member
      unsigned stats_iprefetch_hits;
      unsigned stats_i15miss[MAX_THREADS];
      unsigned stats_il2miss[MAX_THREADS];
      long long stats_itmiss[MAX_THREADS];  /* LYM */
      long long stats_i2tmiss[MAX_THREADS]; /* LYM */

      long long stats_dprobe[MAX_THREADS];    /* LYM */
      unsigned stats_dlines[MAX_THREADS];
      unsigned stats_d1lines[MAX_THREADS];
      unsigned stats_d2lines[MAX_THREADS];
      unsigned stats_dmiss_taken[MAX_THREADS];
      long long stats_dmiss[MAX_THREADS];   /* LYM */
      unsigned stats_dl2miss_taken[MAX_THREADS];
      unsigned stats_dl2miss[MAX_THREADS];
      unsigned stats_dtrail[MAX_THREADS];
      unsigned stats_dcast[MAX_THREADS];

   /*  long long stats_dprobe[MAX_THREADS]; */
     /* unsigned stats_dlines[MAX_THREADS]; */
      unsigned stats_store_d1lines[MAX_THREADS];
   /*   unsigned stats_d2lines[MAX_THREADS]; */
      unsigned stats_store_dmiss_taken[MAX_THREADS];
      unsigned stats_store_dmiss[MAX_THREADS];
    /*  unsigned stats_dl2miss_taken[MAX_THREADS]; */
    /*  unsigned stats_dl2miss[MAX_THREADS];*/
   /*   unsigned stats_dtrail[MAX_THREADS]; */
    /*  unsigned stats_dcast[MAX_THREADS]; */

      unsigned stats_dtmiss_taken[MAX_THREADS];
      long long stats_dtmiss[MAX_THREADS]; /* LYM */
      unsigned stats_d2tmiss_taken[MAX_THREADS];
      long long stats_d2tmiss[MAX_THREADS]; /* LYM */

      unsigned stats_dinter_conflicts;
      unsigned stats_dinter_equals;

    #if( STAT0 )
      unsigned stat0_ifetch_tlb[MAX_THREADS];
      unsigned stat0_ifetch_cache[MAX_THREADS];
      unsigned stat0_ifetch_nfa[MAX_THREADS];
      unsigned stat0_ifetch_mispredict[MAX_THREADS];
      unsigned stat0_ifetch_ibuf[MAX_THREADS];
      unsigned stat0_ifetch_fchblk[MAX_THREADS];
      unsigned stat0_ifetch_sync[MAX_THREADS];
      unsigned stat0_ifetch_pred[MAX_THREADS];
      unsigned stat0_ifetch_partial[MAX_THREADS];
      unsigned stat0_ifetch_inflight[MAX_THREADS];
      unsigned stat0_ifetch_end[MAX_THREADS];

      long long stat0_decode_empty[MAX_THREADS]; /* LYM */
      unsigned stat0_decode_stall[MAX_THREADS];
      unsigned stat0_decode_multi[MAX_THREADS];
      unsigned stat0_decode_string[MAX_THREADS];

      unsigned stat0_dispatch_groups[MAX_THREADS];            /* LYM */
    #endif

   /*   unsigned stats_iprefetch_hits; */

      unsigned stats_nexpands[MAX_THREADS];
      unsigned stats_expands_to[MAX_THREADS];

      unsigned stats_strings_total[MAX_THREADS];
      unsigned stats_strings_expands[MAX_THREADS];
      unsigned stats_strings_stalls[MAX_THREADS];

      long long bp_bclr_total[MAX_THREADS];  /* LYM */
      unsigned nfa_total[MAX_THREADS];
      unsigned nfa_miss[MAX_THREADS];
      unsigned nfa_miss_buried[MAX_THREADS];

    #if( 1 || !PERFECT_BRANCH_PREDICTION )
      long long branch_total[MAX_THREADS];    /* LYM */
      unsigned branch_stalls[MAX_THREADS];
      unsigned branch_pred_stalls[MAX_THREADS];
      long long bp_bc_miss[MAX_THREADS];   /* LYM */
      unsigned bp_bc_total[MAX_THREADS];
      unsigned bp_bclr_miss[MAX_THREADS];
   /*   long long bp_bclr_total[MAX_THREADS];*/
      unsigned bp_bcctr_miss[MAX_THREADS];
      long long bp_bcctr_total[MAX_THREADS]; /* LYM */
      unsigned branch_miss_total[MAX_THREADS];
   /*   unsigned branch_stalls[MAX_THREADS]; */
      unsigned bp_bc_miss_miss[MAX_THREADS];
      unsigned bp_bc_miss_total[MAX_THREADS];
      unsigned bp_bclr_miss_miss[MAX_THREADS];
      unsigned bp_bclr_miss_total[MAX_THREADS];
      unsigned bp_bcctr_miss_miss[MAX_THREADS];
      unsigned bp_bcctr_miss_total[MAX_THREADS];

      unsigned bp_exec_total[MAX_THREADS];
      unsigned bp_exec_pred[MAX_THREADS];
      unsigned bp_exec_miss[MAX_THREADS];
      unsigned bp_exec_bc_miss[MAX_THREADS];
      unsigned bp_exec_bc_total[MAX_THREADS];
      unsigned bp_exec_bclr_miss[MAX_THREADS];
      unsigned bp_exec_bclr_total[MAX_THREADS];   /* LYM */
      unsigned bp_exec_bcctr_miss[MAX_THREADS];
      unsigned bp_exec_bcctr_total[MAX_THREADS]; /* LYM */

      unsigned stats_reissue_flushes[MAX_THREADS];
      unsigned stats_reissue_decode[MAX_THREADS];
      unsigned stats_recirc[MAX_THREADS];
      long long stats_groups_dispatched[MAX_THREADS]; /* LYM */
    #endif

    #if( GET_SPEC_STATS )
      unsigned stats_spec_ifetch[MAX_THREADS];
      unsigned stats_spec_nexpands[MAX_THREADS];
      unsigned stats_spec_nexpands_to[MAX_THREADS];
      unsigned stats_spec_dispatch[MAX_THREADS];
      unsigned stats_spec_br_execd[MAX_THREADS];
      unsigned stats_spec_fix_execd[MAX_THREADS];
      unsigned stats_spec_fpu_execd[MAX_THREADS];
      unsigned stats_spec_mem_execd[MAX_THREADS];

      unsigned stats_spec_dmiss[MAX_THREADS];
      long long stats_dmiss[MAX_THREADS];
      unsigned stats_spec_dlines[MAX_THREADS];
      unsigned stats_spec_dflush[MAX_THREADS];
      unsigned stats_spec_dsatisfy[MAX_THREADS];
      unsigned stats_spec_extra_hits[MAX_THREADS];
      unsigned stats_spec_extra_misses[MAX_THREADS];
      unsigned stats_spec_src_hits[MAX_THREADS];
      unsigned stats_spec_reexec_hits[MAX_THREADS];
      unsigned stats_spec_used_once[MAX_THREADS];
      unsigned stats_spec_imiss[MAX_THREADS];
      long long stats_imiss[MAX_THREADS];  /* LYM */
      unsigned stats_spec_extra_ihits[MAX_THREADS];
      unsigned stats_spec_extra_imisses[MAX_THREADS];
      unsigned stats_spec_src_ihits[MAX_THREADS];
      unsigned stats_spec_iused_once[MAX_THREADS];
    #endif

      unsigned stats_ibuf_stall[MAX_THREADS];
      unsigned stats_inflight_stall[MAX_THREADS];
      unsigned stats_dmiss_stall;
      unsigned stats_dcastout_stall;
      unsigned stats_storeq_stall[MAX_THREADS];
      unsigned stats_reorderq_stall[MAX_THREADS];
      unsigned stats_resv_full_stall[MAX_THREADS];
      unsigned stats_rename_stall[MAX_THREADS];

      unsigned stats_storeq_total[MAX_THREADS];
      long long stats_storeq_forward[MAX_THREADS];  /* LYM */
      unsigned stats_storeq_partials[MAX_THREADS];
      unsigned stats_storeq_data_not_ready[MAX_THREADS];

      unsigned stats_reorderq_total[MAX_THREADS];
      unsigned stats_reorderq_conflict[MAX_THREADS];

      unsigned stats_mtsr[MAX_THREADS];
      unsigned stats_mtsr_ipurge[MAX_THREADS];
      unsigned stats_mtsr_dpurge[MAX_THREADS];

      unsigned stats_nops_dispd[MAX_THREADS];
      unsigned stats_nops_retd[MAX_THREADS];


      long long stats_num_stores[MAX_THREADS];
      long long stats_num_loads[MAX_THREADS];      /* LYM */

    #if( GET_SPEC_STATS )
      unsigned stats_spec_dflush_histo[DMISS_MAX + MEM_NUM +1];
      unsigned stats_spec_over_histo[DMISS_MAX + MEM_NUM +1];
      unsigned stats_spec_dflush_over_histo[DMISS_MAX + MEM_NUM +1];
    #endif

    #if( STATS )
      unsigned stats_ibuf[MAX_THREADS][IBUF_MAX +1];
      unsigned stats_fchblk[MAX_THREADS][FCHBLK_MAX +1];

      unsigned stats_retireq[MAX_THREADS][RETIREQ_MAX +1];
      unsigned stats_fix_resv[MAX_THREADS][FIX_RESV +1];

      #if( CLUSTER )
         unsigned stats_fix1_resv[MAX_THREADS][FIX_RESV +1];
      #endif

      unsigned stats_fpu_resv[MAX_THREADS][FPU_RESV +1];

      #if( CLUSTER_FPU )
        unsigned stats_fpu1_resv[MAX_THREADS][FPU_RESV +1];
      #endif

      unsigned stats_mem_resv[MAX_THREADS][MEM_RESV +1];
      #if( CLUSTER )
        unsigned stats_mem1_resv[MAX_THREADS][MEM_RESV +1];
      #endif
      unsigned stats_br_resv[MAX_THREADS][BR_RESV +1];

      #if( USING_LOG_UNIT )
      unsigned stats_log_resv[MAX_THREADS][LOG_RESV +1];
      #endif

      #if( USING_CMPLX_UNIT )
        unsigned stats_cmplx_resv[MAX_THREADS][CMPLX_RESV +1];
      #endif

      unsigned stats_resv[MAX_THREADS][(FIX_RESV + FPU_RESV + MEM_RESV + BR_RESV + LOG_RESV + CMPLX_RESV )*CLUSTERS_MAX + 1];

      unsigned stats_gpr[MAX_THREADS][GPR_PHYS_MAX +1];
      unsigned stats_fpr[MAX_THREADS][FPR_PHYS_MAX +1];
      unsigned stats_spr[MAX_THREADS][SPR_PHYS_MAX +1];
      unsigned stats_ccr[MAX_THREADS][CCR_PHYS_MAX +1];

      unsigned stats_storeq[MAX_THREADS][STOREQ_SIZE +1];
      unsigned stats_reorderq[MAX_THREADS][REORDERQ_LENGTH +1];
      unsigned stats_dmissq[MAX_THREADS][DMISS_MAX+MEM_NUM +1];
      unsigned stats_lines[MAX_THREADS][DMISS_MAX+MEM_NUM +1];
      unsigned stats_l2dmissq[MAX_THREADS][DMISS_MAX+MEM_NUM +1];
      unsigned stats_l2_lines[MAX_THREADS][DMISS_MAX+MEM_NUM +1];
      unsigned stats_dcastoutq[DCASTOUT_MAX + MEM_NUM+1];

      unsigned stats_st_commit[MAX_THREADS][DCACHE_PORTS + 1];

      unsigned stats_retired[MAX_THREADS][RETIRE_MAX + 1];
      unsigned stats_st_retired[MAX_THREADS][DCACHE_PORTS + 1];
      unsigned stats_fix_execd[MAX_THREADS][FIX_NUM + 1];

      #if( CLUSTER )
        unsigned stats_fix1_execd[MAX_THREADS][FIX_NUM + 1];
      #endif

      unsigned stats_fpu_execd[MAX_THREADS][FPU_NUM + 1];

      #if( CLUSTER_FPU )
        unsigned stats_fpu1_execd[MAX_THREADS][FPU_NUM + 1];
      #endif

      unsigned stats_mem_execd[MAX_THREADS][MEM_NUM + 1];

      #if( CLUSTER )
        unsigned stats_mem1_execd[MAX_THREADS][MEM_NUM + 1];
      #endif

      unsigned stats_br_execd[MAX_THREADS][BR_NUM + 1];

      #if( USING_LOG_UNIT )
      unsigned stats_log_execd[MAX_THREADS][LOG_NUM + 1];
      #endif

      #if( USING_CMPLX_UNIT )
      unsigned stats_cmplx_execd[MAX_THREADS][CMPLX_NUM + 1];
      #endif
      unsigned stats_execd[MAX_THREADS][(FIX_NUM+FPU_NUM+MEM_NUM+BR_NUM+LOG_NUM+CMPLX_NUM)*CLUSTERS_MAX+1];

      unsigned stats_dispatched[MAX_THREADS][DISPATCH_MAX + 1];
      unsigned stats_fix_dispatched[MAX_THREADS][DISPATCH_MAX + 1];

      #if( CLUSTER )
        unsigned stats_fix1_dispatched[MAX_THREADS][DISPATCH_MAX + 1];
      #endif
      unsigned stats_fpu_dispatched[MAX_THREADS][DISPATCH_MAX + 1];

      #if( CLUSTER_FPU )
        unsigned stats_fpu1_dispatched[MAX_THREADS][DISPATCH_MAX + 1];
      #endif
      unsigned stats_mem_dispatched[MAX_THREADS][DISPATCH_MAX + 1];

      #if( CLUSTER )
        unsigned stats_mem1_dispatched[MAX_THREADS][DISPATCH_MAX + 1];
      #endif
      unsigned stats_br_dispatched[MAX_THREADS][DISPATCH_MAX + 1];

      unsigned stats_renamed[MAX_THREADS][IBUF_MAX + 1];
      unsigned stats_fetched[MAX_THREADS][IFETCH_MAX + 1];
      unsigned stats_inflight[MAX_THREADS][INFLIGHT_MAX + 1];
    #endif
      unsigned stats_groups[MAX_THREADS][GROUPS_MAX + 1];

      /* LYM, pstats */
      long long pstats_br_resv[MAX_THREADS]; /* LYM */
#if ( !( COMBINE_FIX_MEM ) ) // jdonald 
      unsigned pstats_fix_resv[MAX_THREADS];
#endif
      long long pstats_fpu_resv[MAX_THREADS]; /* LYM */
      long long pstats_fpu1_resv[MAX_THREADS]; /* LYM */
      long long pstats_mem_resv[MAX_THREADS]; /* LYM */
      long long pstats_mem1_resv[MAX_THREADS]; /* LYM */
      long long pstats_mem_execd[MAX_THREADS]; /* LYM */
      long long pstats_mem1_execd[MAX_THREADS]; /* LYM */
      long long pstats_fpu_execd[MAX_THREADS]; /* LYM */
      long long pstats_fpu1_execd[MAX_THREADS]; /* LYM */
      long long pstats_gpr[MAX_THREADS]; /* LYM */
      long long pstats_fpr[MAX_THREADS]; /* LYM */
      long long pstats_spr[MAX_THREADS]; /* LYM */
      long long pstats_ccr[MAX_THREADS]; /* LYM */
      long long pstats_groups[MAX_THREADS]; /* LYM */
      long long pstats_dmissq[MAX_THREADS]; /* LYM */
      long long pstats_reorderq[MAX_THREADS]; /* LYM */
      long long pstats_storeq[MAX_THREADS]; /* LYM */
      double fxu_intensity[MAX_THREADS];
	  double fpu_intensity[MAX_THREADS]; // jdonald
} dumpstatus_type;

  dumpstatus_type *dumpstatus; // no longer volatile
#endif // DUMPSTATUS

  unsigned		sample_skip = 0;
  unsigned		sample_number = ~0;

  xcycle		skip = 0;
  xcycle		cutoff = 5000;
  xcycle		cycle_reset = 0;   // protected by tempsync_cond_mutex
  xcycle		saved_cycle_reset = 0; // for use with statistics output, after cycle_reset has been cleared
#if( !( FIXED_FREQUENCY ) )
  double		saved_timestamp_reset = 0.0;
#endif
  int			reset_counter = 0; // sort of a semaphore, also protected by tempsync_cond_mutex

#if ( POWER_MODE )  // all moved to global, these are needed by power_defs.macros or init.power.macros or something
  double no_cg[7];
  double both_cg[7];
  double hold_cg[7];
  double logic_cg[7];

  double FO4,FO4_SCALE,FO4_FREQ_SCALE, FO4_LATCH_SCALE,FO4_LOGIC_SCALE,FO4_LEAKAGE_SCALE;
  double logic_power,cpam_full_power;

  double latch_scale, latch_over, latch_ratio, latch_fraction,glitch_factor;
  double freq_scale, freq_over;
  double FO4_overhead = 3.0;
  double leakage_factor, latch_area;

  double leakage_over, glitch_over;
  
  FILE* ifu_file;
  FILE* idu_file;
  FILE* isu_file;
  FILE* fxu_file;
  FILE* fpu_file;
  FILE* lsu_file;
  FILE* vmx_file;
#endif // if ( POWER_MODE )

#if(PDS_SIM)
{
  /* Should consider moving this stuff to init.power.macros... */

  pds_mult = 1.0 / (1000.0 * REAL_VDD);
  pds_offset = 12202.5; /* FIX ME -- calculate the real min power */

  pds_volt_hist_min = 1.17;
  pds_volt_hist_max = 1.43;
  pds_volt_hist_size = 200;
  pds_volt_hist_scale = ((double)pds_volt_hist_size) /
    (pds_volt_hist_max-pds_volt_hist_min);
}
#endif  /* (PDS_SIM) */

// l2cache moved here, since it must be global
// a small section copied from block_dcache.macros
#if( !defined( L2CACHE_LG2ENTRIES ) )
#define		L2CACHE_LG2ENTRIES	14
#endif

#if( !defined( L2CACHE_LG2SIZE ) )
#define		L2CACHE_LG2SIZE		7
#endif
#if( !defined( L2CACHE_ASSOC ) )
#define		L2CACHE_ASSOC		4
#endif

#define		L2CACHE_ENTRIES		(1<<L2CACHE_LG2ENTRIES)
#define		L2CACHE_SET		(L2CACHE_ENTRIES/L2CACHE_ASSOC)

struct		xarray *l2cache; // shared among cores
#if( !INFINITE_DCACHE )
struct		xarray *dcache_local[MAX_CORES]; // array of pointers for cache snooping and broadcast
#endif

int			waitpidcount=0; // made global, taken from main()
unsigned	cmp_conf[MAX_CORES] = {0}; // same
char*		the_stderr_fname[MAX_CORES * MAX_THREADS] = {0};

#if( POSTPROCESS )
char*		output_fname[MAX_CORES] = { 0 }; // output file for binary stats
char*		input_fname[MAX_CORES] = { 0 };
#endif // POSTPROCESS

int			thecutoff = 1; // this seems to be strongly related to POSTPROCESS

#if ( TEMPERATURE && POWER_MODE )
flp_t*		flp_temp;
// elapsed_time and total_elapsed_time now on by default
#endif // TEMPERATURE

#if ( DUMPSTATUS && POWER_MODE )
FILE*		dumpstatus_fp = 0;
#endif

#include "thread_synchro.c" // dependent thread synchronization -jdonald

#if defined(_AIX) || (defined(__APPLE__) && defined(__MACH__)) // AIX or OS X
// this is structured mainly from sample code on Apple's developer site /jdonald
#define REQUIRED_STACK_SIZE 1024*1024

int pthread_create_stacksafe(pthread_t * thread, pthread_attr_t * pattr, void * (*start_routine)(void *), void * arg)
{
	int err = 0;
	pthread_attr_t stacksize_attr;
	size_t stacksize = 0;
	if (pattr != NULL)
		stacksize_attr = *pattr;
	else {
		err = pthread_attr_init(&stacksize_attr);
		if (err)
			fatal("pthread_attr_init() returned an error\n");
	}

	err = pthread_attr_getstacksize(&stacksize_attr, &stacksize);
	if (err)
		fatal("pthread_attr_getstacksize() returned an error\n");
	if (stacksize < REQUIRED_STACK_SIZE) {
		fprintf(the_stderr, "yo d00d my stack is too small\n");
		err = pthread_attr_setstacksize(&stacksize_attr, REQUIRED_STACK_SIZE);
		if (err)
			fatal("pthread_attr_setstacksize() returned an error\n");
	}

	err = pthread_create(thread, &stacksize_attr, start_routine, arg);
	return err;
}

#else // stable pthread stack
#define pthread_create_stacksafe(thread, attr, start_routine, arg) \
	pthread_create(thread, attr, start_routine, arg)
#endif // stable pthread stack

/*****************************************************************************
  MAIN()
*****************************************************************************/

int main(int argc, char* argv[])
{
    // int 		status; // unused?
  	int			i;
#if( POWER_MODE && DUMPSTATUS )
	/* added by dbrooks for binary output of stats */
	char*       dumpstatus_fname = NULL;
#endif // DUMPSTATUS
	int errorcode;
  	pthread_t	pthread_id[MAX_CORES] = { 0 };
	// This is needed even before utils_cmdline.macros
#if ( TEMPERATURE && POWER_MODE )
	char*       temperature_fname = NULL;
	// init_file is declared in init.temp.macros
	char*		old_flp;
	old_flp = flp_cfg;
#endif // TEMPERATURE

	the_stderr = stderr; /* LYM */ // global one set here, local in run_processor()
	parent_stderr = stderr; // needed by child threads to output global (temperature) data properly

  	#include "utils_cmdline.macros" // goes inside main(), NOT run_processor()

  // jdonald changing all the nodev and mmap stuff to simple malloc() calls
  // none of these should be volatile anymore. nodev was removed anyway.

  comm_data = (comm_data_type *) malloc(sizeof(comm_data_type) * processor_num);
  memset((void*)comm_data, 0, sizeof(comm_data_type) * processor_num);

  l2_probe = (unsigned *)malloc(sizeof(unsigned)*MAX_THREADS*processor_num);
  memset((void*)l2_probe, 0, sizeof(unsigned)*MAX_THREADS*processor_num);

  l2_miss = (unsigned *)malloc(sizeof(unsigned)*MAX_THREADS*processor_num);
  memset((void*)l2_miss, 0, sizeof(unsigned)*MAX_THREADS*processor_num);
  
  l2_snoop = (unsigned *)malloc(sizeof(unsigned)*processor_num);
  memset((void*)l2_snoop, 0, sizeof(unsigned)*processor_num);

#if ( POWER_MODE && DUMPSTATUS )
  dumpstatus = (dumpstatus_type *)malloc(sizeof(dumpstatus_type)*processor_num);
  memset((void*)dumpstatus, 0, sizeof(dumpstatus_type)*processor_num);
#endif

/***** l2cache *****/
#if( !defined( L2CACHE_LG2ENTRIES ) )
#define		L2CACHE_LG2ENTRIES	14
#endif

#if( !defined( L2CACHE_LG2SIZE ) )
#define		L2CACHE_LG2SIZE		7
#endif
#if( !defined( L2CACHE_ASSOC ) )
#define		L2CACHE_ASSOC		4
#endif

#define		L2CACHE_ENTRIES		(1<<L2CACHE_LG2ENTRIES)
#define		L2CACHE_SET		(L2CACHE_ENTRIES/L2CACHE_ASSOC)


#if( !INFINITE_L2CACHE )
  // TODO: make this shared array normal malloc instead of an mmap() call
  l2cache = shared_array_create("l2cache", L2CACHE_ENTRIES / L2CACHE_ASSOC, L2CACHE_ASSOC, 1 << L2CACHE_LG2SIZE); /* LYM */
#endif

  #if( TIMELINE )
    tl_output = the_stderr;
  #endif
  fprintf(the_stderr, CYAN "Parallel Turandot CMP, built %s %s" NORMAL "\n", __DATE__, __TIME__ );
  fprintf(the_stderr, "\n");

#if( TRACE )
   	fprintf(the_stderr, "Trace parameters: cutoff=%lld, trace_from=%d, trace_to=%d, detail_from=%d, detail_to=%d, detail_step=%d\n",
   		cutoff, trace_from,trace_to,detail_from, detail_to, detail_step);
#else
   	fprintf(the_stderr, "Trace parameters: cutoff=%lld\n", cutoff);
#endif // TRACE
#if ( TEMPERATURE && POWER_MODE )
	if (old_flp == flp_cfg) {
		fprintf(the_stderr, "No HotSpot floorplan specified; using default layout from %s, %s, %s\n",
			flp_cfg, cmp_flp_cfg, cmp_attach_flp_cfg);
	}
#endif // TEMPERATURE

#if ( POWER_MODE ) // trying it here in main() instead of run_processor()

// earlier bugs largely caused by mistaknely putting this inside run_processor()
  FO4 = latch_over = freq_over = leakage_over = glitch_over = 0.0;
  latch_scale = freq_scale = 1.0;
  latch_fraction = 1.0;
  latch_ratio = 0.0;
  glitch_factor = 0.0;
  leakage_factor = 0.0;
  latch_area = 0.3;

  FO4 = 19.0;  /* OK */
  latch_over = 0.0;
  freq_over = 0.0;
  glitch_over = 0.0;
  leakage_over = 0.0;
  latch_scale = 1.1;
  latch_ratio = 0.7;
  latch_fraction = 1.0;
  glitch_factor = 0.3;
  leakage_factor = 0.1;

  /* Use CPAM defined latch ratio */
  latch_ratio = 0.0;

  /* linear scaling for frequency */
  FO4_FREQ_SCALE = pow(19.0/FO4,freq_scale);

  /* linear scaling for latch bloat */
  /* pow(a,b,m,c) ==> c = ( a ^ b ) mod m */
  /*  pow(19.0/FO4,latch_scale,1,FO4_LATCH_SCALE)*/
  FO4_LATCH_SCALE = latch_fraction * pow((19.0-FO4_overhead)/(FO4-FO4_overhead),latch_scale);

  FO4_LOGIC_SCALE = (1.0/(1.0 + glitch_factor)) * ( 1 + (glitch_factor * ((FO4-FO4_overhead)/(19.0-FO4_overhead))));

  FO4_LEAKAGE_SCALE = pow((19.0-FO4_overhead)/(FO4-FO4_overhead),latch_scale);

  // adding in latch_over and CPAM_SCALE to display at start just like init.power.macros
  fprintf(the_stderr, "latch_ratio = %f, ", latch_ratio);
  fprintf(the_stderr, "CPAM_SCALE = %f, ", CPAM_SCALE);
  // making this less cluttered -jdonald
  fprintf(the_stderr,"latch_over = %f, ",latch_over);
  fprintf(the_stderr,"freq_over = %f, ",freq_over);
  fprintf(the_stderr,"glitch_over = %f, ",glitch_over);
  fprintf(the_stderr,"leakage_over = %f, ",leakage_over);

  if(latch_over != 0.0)
    FO4_LATCH_SCALE = latch_over;

  if(glitch_over != 0.0)
    FO4_LOGIC_SCALE = glitch_over;

  if(leakage_over != 0.0)
    FO4_LEAKAGE_SCALE = leakage_over;

  if(freq_over != 0.0)
    FO4_FREQ_SCALE = freq_over;

  /*        FO4_FREQ_SCALE = 1.0;   */
  /*        FO4_LATCH_SCALE = 1.0;  */

  fprintf(the_stderr,"FO4_FREQ_SCALE = %f, ",FO4_FREQ_SCALE);
  fprintf(the_stderr,"FO4_LATCH_SCALE = %f, ",FO4_LATCH_SCALE);
  fprintf(the_stderr,"FO4_LOGIC_SCALE = %f, ",FO4_LOGIC_SCALE);
  fprintf(the_stderr,"FO4_LEAKAGE_SCALE = %f, ",FO4_LEAKAGE_SCALE);
  fprintf(the_stderr,"FO4_SCALE = %f\n",FO4_FREQ_SCALE*FO4_LATCH_SCALE*FO4_LOGIC_SCALE);

  #include "init.power.macros" // for inside main()
  // having separated out into init.power.macros and init.power.core.macros
#endif

#if ( TEMPERATURE && POWER_MODE )
  flp_temp = read_flp(flp_cfg); // core file input also later sent to read_cmpflp()

  for (waitpidcount = 0; waitpidcount < processor_num; waitpidcount++) {
    // uses the whole n_units to match the size of the hotspottemp array, apparently
    comm_data[waitpidcount].hotspotpower = (double *) malloc (sizeof(double)*flp_temp->n_units);
	memset((void*)comm_data[waitpidcount].hotspotpower, 0, sizeof(double)*flp_temp->n_units);
	// initialize this data since it expects to be clean just after any power-calc cycle
  }

  #include "sim.init.temp.macros" // for inside main()
  if (temperature_fname != NULL) {
	  tempout_fp = fopen(temperature_fname, "w");
	  if (tempout_fp == NULL) {
		  fprintf(the_stderr, "unable to open file '%s' for temperature output\n", temperature_fname);
		  exit(EXIT_FAILURE);
	  }
  }
  else {
	fprintf(the_stderr, "temperature output defaulting to stderr\n");
  	tempout_fp = the_stderr;
  }
#endif

#if ( VARIATION && POWER_MODE ) // even timing variation is meaningless without power measurements
	{
#if ( TEMPERATURE )
		int num_flp_cores;
		
		for (i = 0; i < processor_num; i += 1) {
			timing_variation[i] = 1.0;  // default values
			leakage_variation[i] = 1.0;
		}
		num_flp_cores = (flp->n_units-flp->attach_units) / flp->core_units;
		gaussian_approx(&timing_variation[temperature_core_skip], num_flp_cores, 1.0, timing_sigma_mu);
		reverse_array(&timing_variation[temperature_core_skip], num_flp_cores);
		lognormal_approx(&leakage_variation[temperature_core_skip], num_flp_cores, 1.0, leakage_sigma_mu);
#else
		gaussian_approx(timing_variation, processor_num, 1.0, timing_sigma_mu);
		reverse_array(&timing_variation, processor_num);
#endif // TEMPERATURE
	}
#endif // VARIATION

  // before this gets called, we assume the timing_variation and
  // leakage_variation variables were set in utils_cmdline.macros
  // well, it's only the timing_variation variable that matters
#if ( !( FIXED_FREQUENCY ) ) // && TEMPERATURE
  // This part must be after utils_cmdline.macros, so that variations are already set
  for (i = 0; i < processor_num; i += 1) // even coreignore'd cores need to have their frequency settings
	set_frequency_scale(dvfs_scale_knob, i);
#if ( TEMPERATURE ) // we need a floorplan if temperature_core_skip matters...
  for (i = processor_num; i < (flp->n_units-flp->attach_units)/flp->core_units + temperature_core_skip; i += 1) {
  	// make sure these are covered for variation calculations
  	// or leakage adjustments, if we decide to put idle leakage on unused cores
  	set_frequency_scale(dvfs_scale_knob, i);
  }
#endif	// TEMPERATURE
#else // FIXED_FREQUENCY
  for (i = 0; i < processor_num; i += 1)
  	set_frequency_scale(1.0, i); // set_frequency_scale() will exist regardless of mode
  								 // without POWER_MODE or !FIXED_FREQUENCY it doesn't do anything though
#endif // FIXED_FREQUENCY

#if ( VARIATION && POWER_MODE )
	if (variation_median_sort && variation_profile_sort)
		fatal("You cannot specify both -coremedian and -coresort\n");
	if (variation_median_sort) {
		#if ( TEMPERATURE )
		median_sort(temperature_core_skip, (flp->n_units-flp->attach_units)/flp->core_units);
		#else
		median_sort(0, processor_num);
		#endif // !TEMPERATURE
	}
	if (variation_profile_sort) {
		#if ( TEMPERATURE )
		profile_sort(temperature_core_skip, processor_num);
		profile_sort(temperature_core_skip, processor_num);
		profile_sort(temperature_core_skip, processor_num);
		profile_sort(temperature_core_skip, processor_num);
		profile_sort(temperature_core_skip, processor_num);
		profile_sort(temperature_core_skip, processor_num);
		#else
		profile_sort(0, processor_num);
		#endif // !TEMPERATURE (no temperature_core_skip)
	}
#if ( TEMPERATURE )
	print_variations(temperature_core_skip, (flp->n_units-flp->attach_units)/flp->core_units);
#else // no temperature_core_skip
	print_variations(0, processor_num);
#endif // temperature_core_skip

#endif // VARIATION 
  	fprintf(the_stderr, "begin simulation\n");
#ifdef NO_TLS
	opc_init(); // otherwise we call this in run_processor() for each thread /jdonald
#endif
  	// srand(time(NULL)); // scrambled startup order for testing pthread lock system robustness
#if defined(_AIX42)
	unjoined_count = processor_num;
#endif
  	for (i = 1; i < processor_num; i += 1) {
	  	// int r;
	  	// while (pthread_id[r = rand() % processor_num] == 0);
	  	// errorcode = pthread_create_stacksafe( &pthread_id[r], NULL, &run_processor, (void*) r );
		errorcode = pthread_create_stacksafe( &pthread_id[i], NULL, &run_processor, (void*) i );
		if (errorcode != 0)
			fatal("pthread_create() returned error\n");
	}
	run_processor(0); // parent thread handles core #0

// this part needs to be moved into the mutex call over in run_processor()
/* LYM CMP TEMP */

// finish by calling pthread_join() to confirm termination of all threads
#if !defined(_AIX42) // use regular pthread_join()
	for (i = 1; i < processor_num; i += 1) {
		errorcode = pthread_join( pthread_id[i], NULL );
		if (errorcode != 0)
			fatal("pthread_join() returned error\n");
		pthread_id[i] = 0;
	}
#else // _AIX42 detached pthreads
	pthread_mutex_lock( &join_mutex );
	while (unjoined_count > 0)
		pthread_cond_wait( &join_cond, &join_mutex );
	pthread_mutex_unlock( &join_mutex );
#endif // _AIX42

	/* Now that all threads are complete I can print the final result.     */
	/* Without the join I could be printing a value before all the threads */
	/* have been completed.                                                */

/* // removing a lot of redundant temperature crap by me, jdonald
#if ( TEMPERATURE && POWER_MODE )
	    {
	    int n; // jdonald
		n=flp->n_units;

        // get steady state temperatures
        // steady_state_temp(overalls_power, steady_temp, flp->n_units);

        // dump temperatures if needed
        {
		   	fprintf(stderr, "dumping to %s, %s\n", dumped_file, transient_file);
	        dump_temp (flp, steady_temp, processor_num, dumped_file);
	        dump_temp (flp, hotspottemp, processor_num, transient_file);
        }

// print the results here, and other stuff from their original main()

		print_flp_fig(flp); // jdonald 2-22-2004 hey the README lied. it's not in sim-template

        // print results
	fprintf(stderr, "Unit\tSteady \tTransient\n");
        		steady_state_temp(overalls_power, steady_temp, flp->n_units);
	for (i=0; i < 14; i++) {
		char foo[64];
		strcpy(foo, flp->units[i].name);
		if (strstr(foo, "_p1") != NULL)
			foo[strlen(foo)-3] = '\0';
		fprintf(stderr, "%s\t%.2f\t%.2f\n", foo, steady_temp[get_blk_index(flp_temp, foo)]-273.15, hotspottemp[get_blk_index(flp_temp, foo)]-273.15);
		if (i == 3)
		fprintf(stderr,"Processor: %d STEADY: IFU_cache: %.2f\tIFU_B1: %.2f\tIFU_B2: %.2f\tIDU: %.2f\tLSU_cache: %.2f\tLSU_B1: %.2f\tLSU_B2: %.2f\tFXU_reg: %.2f\tFXU_B1: %.2f\tFXU_B2: %.2f\tISU: %.2f\tFPU_reg: %.2f\tFPU_B1: %.2f\tBXU: %.2f\t\n",0,steady_temp[get_blk_index(flp_temp,"IFU_cache.a")],steady_temp[get_blk_index(flp_temp,"IFU_B1.a")],steady_temp[get_blk_index(flp_temp,"IFU_B2.a")],steady_temp[get_blk_index(flp_temp,"IDU.a")],steady_temp[get_blk_index(flp_temp,"LSU_cache.a")],steady_temp[get_blk_index(flp_temp,"LSU_B1.a")],steady_temp[get_blk_index(flp_temp,"LSU_B2.a")],steady_temp[get_blk_index(flp_temp,"FXU_reg.a")],steady_temp[get_blk_index(flp_temp,"FXU_B1.a")],steady_temp[get_blk_index(flp_temp,"FXU_B2.a")],steady_temp[get_blk_index(flp_temp,"ISU.a")],steady_temp[get_blk_index(flp_temp,"FPU_reg.a")],steady_temp[get_blk_index(flp_temp,"FPU_B1.a")],steady_temp[get_blk_index(flp_temp,"BXU.a")]);
	}

	for (i=0; i < n; i++)
		fprintf(stderr, "interface_%s\t%.2f\t%.2f\n", flp->units[i].name, steady_temp[IFACE*n+i]-273.15, hotspottemp[IFACE*n+i]-273.15);
	for (i=0; i < n; i++)
		fprintf(stderr, "spreader_%s\t%.2f\t%.2f\n", flp->units[i].name, steady_temp[HSP*n+i]-273.15, hotspottemp[HSP*n+i]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "spreader_west", steady_temp[NL*n+SP_W]-273.15, hotspottemp[NL*n+SP_W]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "spreader_east", steady_temp[NL*n+SP_E]-273.15, hotspottemp[NL*n+SP_E]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "spreader_north", steady_temp[NL*n+SP_N]-273.15, hotspottemp[NL*n+SP_N]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "spreader_south", steady_temp[NL*n+SP_S]-273.15, hotspottemp[NL*n+SP_S]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "spreader_bottom", steady_temp[NL*n+SP_B]-273.15, hotspottemp[NL*n+SP_B]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "sink_west", steady_temp[NL*n+SINK_W]-273.15, hotspottemp[NL*n+SINK_W]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "sink_east", steady_temp[NL*n+SINK_E]-273.15, hotspottemp[NL*n+SINK_E]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "sink_north", steady_temp[NL*n+SINK_N]-273.15, hotspottemp[NL*n+SINK_N]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "sink_south", steady_temp[NL*n+SINK_S]-273.15, hotspottemp[NL*n+SINK_S]-273.15);
	fprintf(stderr, "%s\t%.2f\t%.2f\n", "sink_bottom", steady_temp[NL*n+SINK_B]-273.15, hotspottemp[NL*n+SINK_B]-273.15);
	{
		int i,j;
		for (i=0;i<processor_num;i++){

                fprintf(stderr,"Processor: %d TEMPERATURE: IFU_cache: %.2f\tIFU_B1: %.2f\tIFU_B2: %.2f\tIDU: %.2f\tLSU_cache: %.2f\tLSU_B1: %.2f\tLSU_B2: %.2f\tFXU_reg: %.2f\tFXU_B1: %.2f\tFXU_B2: %.2f\tISU: %.2f\tFPU_reg: %.2f\tFPU_B1: %.2f\tBXU: %.2f\t\n",i,hotspottemp[get_blk_index(flp_temp,"IFU_cache.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"IFU_B1.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"IFU_B2.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"IDU.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"LSU_cache.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"LSU_B1.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"LSU_B2.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"FXU_reg.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"FXU_B1.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"FXU_B2.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"ISU.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"FPU_reg.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"FPU_B1.a")+i*flp_temp->n_units],hotspottemp[get_blk_index(flp_temp,"BXU.a")+i*flp_temp->n_units]);
        		steady_state_temp(overalls_power, steady_temp, flp->n_units);
                fprintf(stderr,"Processor: %d STEADY: IFU_cache: %.2f\tIFU_B1: %.2f\tIFU_B2: %.2f\tIDU: %.2f\tLSU_cache: %.2f\tLSU_B1: %.2f\tLSU_B2: %.2f\tFXU_reg: %.2f\tFXU_B1: %.2f\tFXU_B2: %.2f\tISU: %.2f\tFPU_reg: %.2f\tFPU_B1: %.2f\tBXU: %.2f\t\n",i,steady_temp[get_blk_index(flp_temp,"IFU_cache.a")],steady_temp[get_blk_index(flp_temp,"IFU_B1.a")],steady_temp[get_blk_index(flp_temp,"IFU_B2.a")],steady_temp[get_blk_index(flp_temp,"IDU.a")],steady_temp[get_blk_index(flp_temp,"LSU_cache.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"LSU_B1.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"LSU_B2.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"FXU_reg.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"FXU_B1.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"FXU_B2.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"ISU.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"FPU_reg.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"FPU_B1.a")+i*flp_temp->n_units],steady_temp[get_blk_index(flp_temp,"BXU.a")+i*flp_temp->n_units]);
		}
	if (flp->n_units > flp_temp->n_units*processor_num) {
        fprintf(stderr,"L2 TEMPERATURE: %.2f\n",hotspottemp[flp_temp->n_units*processor_num]);
		steady_state_temp(overalls_power, steady_temp, flp->n_units);
        fprintf(stderr,"L2 STEADY: %.2f\n",steady_temp[flp_temp->n_units*processor_num]);
	}
	}
  }
#endif // TEMPERATURE
*/

#if ( DUMPSTATUS )
	if (dumpstatus_fname != NULL) {
		if (thecutoff != 0)
			dumpstatus_fp = fopen(dumpstatus_fname,"wb");
		else
			dumpstatus_fp = fopen(dumpstatus_fname,"rb"); // and do what? good thing this never happens. -jdonald
		if (dumpstatus_fp != NULL) {
			int hotspoti;
	   		for (hotspoti = 0; hotspoti < processor_num; hotspoti += 1) {
		   		fwrite(&hotspoti,sizeof(int), 1, dumpstatus_fp);
	   	    	fwrite(dumpstatus+hotspoti, sizeof(dumpstatus_type), 1, dumpstatus_fp);
		   	}
	  	   	fwrite(l2_miss,sizeof(unsigned),processor_num*MAX_THREADS,dumpstatus_fp);
	  	   	fwrite(l2_probe,sizeof(unsigned),processor_num*MAX_THREADS,dumpstatus_fp);
	  	   	// l2_snoop not yet added in, until someone needs it -jdonald
	  		fclose(dumpstatus_fp);
	  		dumpstatus_fp = NULL;
	  		fprintf(the_stderr, "%d bytes written to %s\n", (int) (processor_num *
	  			(sizeof(int) + sizeof(dumpstatus_type) + 2 * sizeof(unsigned) * MAX_THREADS)), dumpstatus_fname);
  		}
  		else
  			fprintf(the_stderr, "unable to open %s for writing\n", dumpstatus_fname);
	}
#endif // DUMPSTATUS

#if ( TEMPERATURE && POWER_MODE )
	if (steady_file != 0) {
	   	fprintf(the_stderr, "dumping steady-state temperatures to %s\n", steady_file);
		dump_temp(flp, steady_temp, steady_file);
    }
    if (transient_file != 0) {
		fprintf(the_stderr, "dumping final temperatures to %s\n", transient_file);
		dump_temp(flp, hotspottemp, transient_file);
	}
#endif

/*****************************************************************************
  SUMMARY DATA (all cores)
*****************************************************************************/
  {
	int j, smt_columns; // for my code /jdonald
#if ( POWER_MODE )
	double dynamic_total = 0.0; // for output stats
#if ( TEMPERATURE ) // && POWER_MODE
	double leakage_total = 0.0; // for output stats
#endif
#endif // POWER_MODE
#if ( FIXED_FREQUENCY )
	xcycle max_cycle = 0;
#else // !FIXED_FREQUENCY
	double max_timestamp = 0.0;
#endif
	long long total_retired = 0;

	print_synchro_data(); // defined in thread_synchro.c
	
    smt_columns = 0;
    for (i = 0; i < processor_num; i += 1) {
	    if (cmp_conf[i] > smt_columns)
	    	smt_columns = cmp_conf[i];
    }
    // we'll try a bit of color here too, for the time being
    fputs(GREEN, the_stderr);

    fprintf(the_stderr, "@@ %4s %10s ", "core", "cycle");
#if !( FIXED_FREQUENCY )
	fprintf(the_stderr, "%10s ", "timestamp");
#endif
    for (j = 0; j < smt_columns; j += 1) {
	    char column[16];
	    sprintf(column, "bench[%d]", j);
	    fprintf(the_stderr, "%10s ", column);
    }
    for (j = 0; j < smt_columns; j += 1) {
    	char column[16];
    	sprintf(column, "retir[%d]", j);
    	fprintf(the_stderr, "%10s ", column);
	}
#if ( POWER_MODE )
	fprintf(the_stderr, "%10s ", "dynamic W");
#if ( TEMPERATURE)
	fprintf(the_stderr, "%10s %10s ", "leakage W", "steady\260C"); // degree symbol
#endif // TEMPERATURE
#endif // POWER_MODE
	fputs("\n", the_stderr);
    for (i = 0; i < processor_num; i += 1) {
	    fputs("@@ ", the_stderr);
	    fprint_unsigned(the_stderr, 4, i);
    	fputc(' ', the_stderr);
    	fprint_unsigned(the_stderr, 10, global_cycle[i] - saved_cycle_reset);
#if ( FIXED_FREQUENCY )
    	if (global_cycle[i] > max_cycle)
    		max_cycle = global_cycle[i];
#else
		fputc(' ', the_stderr);
		fprintf(the_stderr, "%10.2e", l2_sync[i] - saved_timestamp_reset); // timestamp
		if (l2_sync[i] > max_timestamp)
			max_timestamp = l2_sync[i];
		// fprintf(the_stderr, "%10.2e", l2_sync[i] / REAL_FREQ / 1e9);
#endif
    	fputc(' ', the_stderr);
    	for (j = 0; j < smt_columns; j += 1) {
	    	if (j < cmp_conf[i]) {
		    	char benchname[16];
		    	benchmark_name(benchname, trace_fname[i][j]);
	    		fprintf(the_stderr, "%10s ", benchname);
    		}
	    	else
	    		fprintf(the_stderr, "%10s ", ""); // 11 spaces
    	}
    	for (j = 0; j < smt_columns; j += 1) {
	    	if (j < cmp_conf[i]) {
	    		fprint_unsigned(the_stderr, 10, global_insns_retired[i][j]);
	    		total_retired += global_insns_retired[i][j];
	    		fputc(' ', the_stderr);
    		}
	    	else
	    		fprintf(the_stderr, "%10s ", ""); // 11 spaces
    	}
#define ROUND_UP(x) ( (x) + POWER_CYCLES - 1 - ((x)-1) % POWER_CYCLES )

#if( POWER_MODE )
#if ( FIXED_FREQUENCY )
		fprintf(the_stderr, "%10.2f ", total_dynamic_power[i] / 1000.0 * POWER_CYCLES / ROUND_UP(global_cycle[i] - saved_cycle_reset)); // was in Powertimer's milliwatts
	#if ( TEMPERATURE ) // temperature_core_skip
		if (i >= temperature_core_skip) {
			dynamic_total += total_dynamic_power[i] / 1000.0 * POWER_CYCLES / ROUND_UP(global_cycle[i] - saved_cycle_reset);
		}
	#else
		dynamic_total += total_dynamic_power[i] / 1000.0 * POWER_CYCLES / ROUND_UP(global_cycle[i] - saved_cycle_reset);
	#endif // temperature_core_skip
#else // vscale
		fprintf(the_stderr, "%10.2f ", total_dynamic_power[i] / 1000.0 * POWER_INTERVAL / (l2_sync[i] - saved_timestamp_reset));
	#if ( TEMPERATURE ) // temperature_core_skip
		if (i >= temperature_core_skip) {
			dynamic_total += total_dynamic_power[i] / 1000.0 * POWER_INTERVAL / (l2_sync[i] - saved_timestamp_reset);
		}
	#else
		dynamic_total += total_dynamic_power[i] / 1000.0 * POWER_INTERVAL / (l2_sync[i] - saved_timestamp_reset);
	#endif // temperature_core skip
#endif // !FIXED_FREQUENCY
#if ( TEMPERATURE ) // && POWER_MODE
#if ( FIXED_FREQUENCY ) // && TEMPERATURE && POWER_MODE
		if (i >= temperature_core_skip) {
			fprintf(the_stderr, "%10.2f ", total_leakage_power[i] * POWER_CYCLES / ROUND_UP(global_cycle[i] - saved_cycle_reset)); // was in HotSpot's watts
			leakage_total += total_leakage_power[i] * POWER_CYCLES / ROUND_UP(global_cycle[i] - saved_cycle_reset);
		}
		else
			fprintf(the_stderr, "%10s ", "");
#else // vscale
		if (i >= temperature_core_skip) {
			fprintf(the_stderr, "%10.2f ", total_leakage_power[i] * POWER_INTERVAL / (l2_sync[i] - saved_timestamp_reset));
			leakage_total += total_leakage_power[i] * POWER_INTERVAL / (l2_sync[i] - saved_timestamp_reset);
		}
		else
			fprintf(the_stderr, "%10s ", "");
#endif // !FIXED_FREQUENCY
    	// total_dynamic_power[i] + total_leakage_power[i] ~= calculate_core_power(i, overalls_power, flp))
		if (i >= temperature_core_skip)
			fprintf(the_stderr, "%10.2f ", calculate_core_temperature(i, steady_temp, flp) - 273.15);
		else
			fprintf(the_stderr, "%10s ", "");
#endif // TEMPERATURE
#endif // POWER_MODE
	    fputs("\n", the_stderr);
	}
// L2 cache info
#if ( POWER_MODE )
	fprintf(the_stderr, "@@ %4s ", "L2");
	fprintf(the_stderr, "%10s ", ""); // one blank for the cycle column
#if !( FIXED_FREQUENCY )
	fprintf(the_stderr, "%10s ", ""); // another blank for timestamp column
#endif
	for (i = 0; i < smt_columns; i += 1)
		fprintf(the_stderr, "%10s %10s ", "", ""); // blanks for benchmark names and retire counts
#if ( FIXED_FREQUENCY )
	fprintf(the_stderr, "%10.2f ", l2_dynamic_power * POWER_CYCLES / ROUND_UP(max_cycle - saved_cycle_reset));
	dynamic_total += l2_dynamic_power * POWER_CYCLES / (ROUND_UP(max_cycle) - saved_cycle_reset);
#else // !FIXED_FREQUENCY
	fprintf(the_stderr, "%10.2f ", l2_dynamic_power * POWER_INTERVAL / (total_elapsed_time - saved_timestamp_reset));
	dynamic_total += l2_dynamic_power * POWER_INTERVAL / (total_elapsed_time - saved_timestamp_reset);
#endif
	// fprintf(the_stderr, "%10.2f ", l2_dynamic_power * POWER_CYCLES / (total_elapsed_time * base_proc_freq));
	// dynamic_total += l2_dynamic_power * POWER_CYCLES / (total_elapsed_time * base_proc_freq);
#if ( TEMPERATURE )
#if ( FIXED_FREQUENCY )
	fprintf(the_stderr, "%10.2f ", l2_leakage_power * POWER_CYCLES / ROUND_UP(max_cycle - saved_cycle_reset));
	leakage_total += l2_leakage_power * POWER_CYCLES / (ROUND_UP(max_cycle) - saved_cycle_reset);
#else // !FIXED_FREQUENCY
	fprintf(the_stderr, "%10.2f ", l2_leakage_power * POWER_INTERVAL / (total_elapsed_time - saved_timestamp_reset));
	leakage_total += l2_leakage_power * POWER_INTERVAL / (total_elapsed_time - saved_timestamp_reset);
#endif
	// fprintf(the_stderr, "%10.2f ", l2_leakage_power * POWER_CYCLES / (total_elapsed_time * base_proc_freq));
	// leakage_total += l2_leakage_power * POWER_CYCLES / (total_elapsed_time * base_proc_freq);
	fprintf(the_stderr, "%10.2f ", calculate_l2cache_temperature(steady_temp, flp) - 273.15);
#endif // TEMPERATURE
	fprintf(the_stderr, "\n");
#endif // POWER_MODE L2

	fprintf(the_stderr, "\n");
	fprintf(the_stderr, "@@ totals");
#if ( FIXED_FREQUENCY )
	fprintf(the_stderr, "%9lld ", max_cycle);
#else // !FIXED_FREQUENCY
	fprintf(the_stderr, "%9s %10.2e ", "", max_timestamp);
#endif
    for (j = 0; j < smt_columns; j += 1) { // for each benchmark name
   		fprintf(the_stderr, "%10s ", ""); // 11 spaces
    }
	fprintf(the_stderr, "%10lld ", total_retired);
    for (j = 1; j < smt_columns; j += 1) { // for instruction count columns
   		fprintf(the_stderr, "%10s ", ""); // 11 spaces
    }
#if ( POWER_MODE )
#if ( TEMPERATURE )
	fprintf(the_stderr, "%10.2f  + %7.2f  = %7.2f", dynamic_total, leakage_total, dynamic_total + leakage_total);
#else // no leakage
	fprintf(the_stderr, "%10.2f", dynamic_total);
#endif // TEMPERATURE
#endif // POWER_MODE
	fprintf(the_stderr, "\n");

	fputs(NORMAL, the_stderr);
  }

#define FREE_AND_NULLIFY(addr) \
	free(addr);	\
	addr = NULL;
#define FREE_VECTOR_AND_NULLIFY(addr) \
	free_vector(addr);	\
	addr = NULL;
	
    // free the HotSpot structures created by sim.init.temp.macros
#if ( TEMPERATURE && POWER_MODE )
	if (temperature_fname != NULL) {
		fclose(tempout_fp);
		tempout_fp = NULL;
	}
	cleanup_hotspot(flp->n_units); // free: b, c, inva, invb // jdonald
	free_flp(flp);
	free_flp(flp_temp);
	flp = NULL;
	flp_temp = NULL;
    FREE_VECTOR_AND_NULLIFY(hotspottemp)
    FREE_VECTOR_AND_NULLIFY(hotspotpower)
    FREE_VECTOR_AND_NULLIFY(steady_temp)
    FREE_VECTOR_AND_NULLIFY(overall_power)
    FREE_VECTOR_AND_NULLIFY(templeakage)
	FREE_VECTOR_AND_NULLIFY(overalls_power)
	// TODO: remove any of the variables that are unnecessary creations of sim.init.temp.macros
	// errata: we need to know which ones for-sure are reliability-specific
	// The compiler hasn't been warning us since they are malloc()'d and free()'d
#if( RELIABILITY )
	FREE_VECTOR_AND_NULLIFY(old_temp)
	FREE_VECTOR_AND_NULLIFY(currently_positive)
	FREE_VECTOR_AND_NULLIFY(final_diff)
	FREE_VECTOR_AND_NULLIFY(old_set)
	FREE_VECTOR_AND_NULLIFY(delta_T)
	FREE_VECTOR_AND_NULLIFY(delta_counter)
	FREE_VECTOR_AND_NULLIFY(TC_new_cycles)
	FREE_VECTOR_AND_NULLIFY(non_startup)
	FREE_VECTOR_AND_NULLIFY(fits)
	FREE_VECTOR_AND_NULLIFY(mean)
	FREE_VECTOR_AND_NULLIFY(stdev)
	FREE_VECTOR_AND_NULLIFY(countdown)
	FREE_VECTOR_AND_NULLIFY(frequency)
	FREE_VECTOR_AND_NULLIFY(firsttime)
	FREE_VECTOR_AND_NULLIFY(nocycles)
	FREE_VECTOR_AND_NULLIFY(tc_counter)
	FREE_VECTOR_AND_NULLIFY(TC_cycles_allowed)
	FREE_VECTOR_AND_NULLIFY(power_density)
	FREE_VECTOR_AND_NULLIFY(unit_act)
	FREE_VECTOR_AND_NULLIFY(unit_res)
#endif // RELIABILITY

	FREE_VECTOR_AND_NULLIFY(unit_area) // unit_area is used regardless of RELIABILITY
	
	// free all other memory allocations done in turandot.c
	for (i = 0; i < processor_num; i += 1)
		free(comm_data[i].hotspotpower);
#endif // TEMPERATURE

#if(DUMPSTATUS)
	FREE_AND_NULLIFY(dumpstatus);
#endif

	FREE_AND_NULLIFY(comm_data)
	FREE_AND_NULLIFY(l2_probe)
	FREE_AND_NULLIFY(l2_miss)
	FREE_AND_NULLIFY(l2_snoop)
	free_array(l2cache);
	l2cache = NULL;
	free_strings();

	pthread_cond_destroy( &tempsync_cond );
	pthread_cond_destroy( &l2sync_cond );
	pthread_mutex_destroy( &tempsync_cond_mutex );
	pthread_mutex_destroy( &l2sync_cond_mutex );
#ifdef NO_TLS
	pthread_mutex_destroy( &aria_mutex );
#endif // NO_TLS
#ifdef _AIX42 // for detached pthreads under AIX 4.2
	pthread_cond_destroy( &join_cond );
	pthread_mutex_destroy(&join_mutex );
#endif // AIX42
		
	return EXIT_SUCCESS;
} // main()


/*****************************************************************************
  RUN_PROCESSOR()
*****************************************************************************/

void* run_processor(void* arg) {
	
/***** files *****/
  FILE			*the_stderr;  // global variable normally, local for run_processor()
  // int		the_stderr_des;
  FILE *		trace_fp[MAX_THREADS];

  #include "control_defs.macros" // originally a piece of controls.h
  //#include "utils_trans.macros" // now included here and dep_prep_process.c is included earlier

  /* a general variable for receiving blks from array_access */
  struct		xarray_blk *p_blk; // p_blk ??

#if 0
  double		fpu_intensity[MAX_THREADS]; // jdonald, later need to move this to per-core 2-18-2005
  double		fxu_intensity[MAX_THREADS];
#endif

  int			thread_finish[MAX_THREADS]={0};
  xcycle		thread_cycles[MAX_THREADS] = {0}; /* stage_ifetch.macros LYM */
  // int 	    trace_num[MAX_THREADS] = {0}; // unused?
  // int   	    trace_id[MAX_THREADS] = {0}; // unused??
  // char      	trace_name[1024]; // unused?

  unsigned 		insns_retired[MAX_THREADS] = {0};

  #if( POWER_MODE )
    unsigned 		old_insns_retired[MAX_THREADS] = {0}; /* LYM */ // this seems to only be used in power.macros
    #include "power_defs.core.macros" // jdonald
  #endif

#include "utils_reader.macros"

/***** preprocessed information ******/
  unsigned		prep_first[MAX_THREADS];
  // unsigned		prep_ovsize; // unused?
  // xarch *		prep_over; // unused?
  // unsigned		prep_xsize; // unused?
  #if( (USING_FF || USING_IM))
  #elif( USING_ROCH )
    unsigned		prep_size;
    xdep_prep_info *	prep;

  #else
    unsigned		prep_size;
    xdep_prep_info *	prep;
    unsigned		prep_min;
  #endif

/***** instruction queue *****/
#if( !defined( IQ_SIZE ) )
  #define		IQ_SIZE			256
#endif
#if( IQ_SIZE > 256 )
/*  #error	"IQ_SIZE should be less than or equal to 256" */ /* LYM */
#endif
#if( ((IQ_SIZE - 1) & IQ_SIZE) != 0 )
  #error	"IQ_SIZE should be a power of 2"
#endif

#define in_iq(idx_)				\
 ( ( iq_hd <= idx_ && idx_ < iq_tl ) || 	\
   ( idx_ < iq_tl && iq_tl <  iq_hd ) || 	\
   ( iq_tl < iq_hd && iq_hd <= idx_ ) )

#define between_iq( idx_, hi_, lo_ )		\
 ( ( hi_ <= idx_ && idx_ < lo_ ) ||		\
   ( idx_ < lo_ && lo_ <  hi_ ) || 		\
   ( lo_ < hi_ && hi_ <= idx_ ) )

  #define		IQ_NULL			(IQ_SIZE+1)
  xiq_info		iq[MAX_THREADS][IQ_SIZE] = {{{0}}};
  xiq_idx		iq_hd[MAX_THREADS] = {0};
  xiq_idx		iq_tl[MAX_THREADS] = {0};
  unsigned		iq_n[MAX_THREADS] = {0};

  /* global iq in SMT mode */
  // xiq_info		*global_iq; // unused?
  // xiq_idx		*global_iq_hd; // unused?
  // xiq_idx		*global_iq_tl; // unused?
  // unsigned		global_iq_n = 0; // unused?

/* normally each insn can reference 3-4 physical registers (1-2 for operand, 1-2 for output). millicode can have as much as 70 registers */
  #define		IQ_SMALL_REG		4
  #define		IQ_LARGE_REG		70
  #define		IQ_TOTAL_REG		((IQ_SMALL_REG+IQ_LARGE_REG)*IQ_SIZE)
  xphys_reg		iq_reg[MAX_THREADS][IQ_TOTAL_REG];	/* stores input and output phy reg ids for each instruction */

  #include "utils_linked_list.macros"
  #include "block_prediction.macros"

  #include "block_nfa.macros"

/* each insn is 4-byte, so ifetch_stride = ifetch_max * 4 */
  #define IFETCH_STRIDE		(IFETCH_MAX*4)
  #define IFETCH_STRIDE_MASK	(IFETCH_STRIDE-1)

  #include "block_icache.macros" // changed the 7 occurrences back to normal

  #include "block_prefetch.macros" // changed back to normal

  #include "block_dcache.macros" // changed the 13 occurrences back

  #include "block_memq.macros" // changed the 19 occurences back

  #include "block_bus.macros"

/***** ifetch *****/
  #if( !defined(IFETCH_MAX) )
  #define IFETCH_MAX				4
  #endif

  #if( !defined(IFETCH_MAX_COND_BRANCH) )
  #define		IFETCH_MAX_COND_BRANCH	1
  #endif

  #define		IFETCH_STALL_FINISHED	((xcycle)(THE_MAX))
  #define		IFETCH_STALL_MISPREDICT	((xcycle)(THE_MAX-1))
  #define		IFETCH_STALL_CACHE	((xcycle)(THE_MAX-2))
  #define		IFETCH_STALL_PRED	((xcycle)(THE_MAX-3))
  #define		IFETCH_STALL_SYNC	((xcycle)(THE_MAX-4))
  #define		IFETCH_STALL_SPECIAL	IFETCH_STALL_SYNC

/* ifetch status variables for each thread */
  xcycle		ifetch_stalled_till[MAX_THREADS];
  boolean		ifetch_mispredict_done[MAX_THREADS];
  boolean		ifetch_mispredict_done_early[MAX_THREADS];
  unsigned		ifetch_mispredicted_branch[MAX_THREADS];
  unsigned		ifetch_mispredicted_branch_backup[MAX_THREADS];
  unsigned		ifetch_in_mispredicted[MAX_THREADS];
  unsigned		ifetch_trauma[MAX_THREADS];
#if( USE_PREDICATION )
#define PREDICATION_IQBUF_SIZE	512
  unsigned		ifetch_pred_surpress = 0;
#if( USE_PREDICATION_MODE1 )
  xdep_prep_info *	pred_ifbuf;
  unsigned		pred_ifbuf_at;
  unsigned		pred_ifbuf_sz;
#endif
#endif

  unsigned		ifetch_branch_history[MAX_THREADS] = {0};
  unsigned		ifetch_mispredicted_history_backup[MAX_THREADS] = {0};
  // unsigned		ifetch_pred_count[MAX_THREADS] = {0}; // unused?

/***** ibuf *****/
  #if( !defined( IBUF_MAX ))
  #define		IBUF_MAX		24
  #endif
  xiq_cnt		ibuf_left[MAX_THREADS];
  // xiq_cnt		global_ibuf_left; // unused?

  #if( !defined( FCHBLK_MAX) )
  #define		FCHBLK_MAX		0
  #endif
  unsigned		fchblk_left[MAX_THREADS] = {0};
  // unsigned		global_fchblk_left; // unused?

/***** ifetch'd till *****/
  xiq_idx		iq_ifetch[MAX_THREADS];
  // xiq_info		*global_iq_ifetch; // unused?

/***** decode *****/
  #if( !defined(DECODE_MAX))
  #define		DECODE_MAX		4
  #endif
  xcycle		decode_stalled_till[MAX_THREADS];
  xiq_idx		iq_dec[MAX_THREADS];
  xiq_idx		iq_dec_1[MAX_THREADS];
  xiq_idx		iq_dec_2[MAX_THREADS];
  xiq_idx		iq_dec_3[MAX_THREADS];
  xiq_idx		iq_dec_4[MAX_THREADS];

/***** rename *****/
  #if( !defined(RENAME_MAX) )
  #define		RENAME_MAX		4
  #endif

/* maximum number of physical registers per class */
  #if( !defined(PHYS_MAX))
    #define		PHYS_MAX		128
  #endif
  #define		LG2PHYS_MAX		lg2(PHYS_MAX)

  #define		CLASS_MAX		4
  #define		ARCH_TOTAL		(CLASS_MAX*ARCH_MAX)
  #define		PHYS_TOTAL		(CLASS_MAX*PHYS_MAX)

  xphys_reg		rgrename_backup[MAX_THREADS][ARCH_TOTAL];
  xphys_reg		rgrename[MAX_THREADS][ARCH_TOTAL];
  xphys_reg		inorder[MAX_THREADS][ARCH_TOTAL];

  xiq_idx		iq_ren[MAX_THREADS];

/***** free register queue *****/
  #if(!defined(GPR_PHYS_MAX) )
  #define		GPR_PHYS_MAX		80
  #endif
  #if( !defined(GPR_ARCH) )
  #define		GPR_ARCH		32
  #endif
  #define		GPR_CLASS		0
  #if( GPR_PHYS_MAX > PHYS_MAX )
  #error		"GPR_PHYS_MAX should be <= PHYS_MAX"
  #endif
  #if( GPR_ARCH > ARCH_MAX )
  #error		"GPR_PHYS_MAX should be <= ARCH_MAX"
  #endif

  #if( !defined(FPR_PHYS_MAX) )
  #define		FPR_PHYS_MAX		80
  #endif
  #if( !defined(FPR_ARCH) )
  #define		FPR_ARCH		32
  #endif
  #define		FPR_CLASS		1
  #if( FPR_PHYS_MAX > PHYS_MAX )
  #error		"FPR_PHYS_MAX should be <= PHYS_MAX"
  #endif
  #if( FPR_ARCH > ARCH_MAX )
  #error		"FPR_PHYS_MAX should be <= ARCH_MAX"
  #endif

  #if( !defined(CCR_PHYS_MAX))
  #define		CCR_PHYS_MAX		32
  #endif
  #if( !defined(CCR_ARCH))
  #define		CCR_ARCH		8
  #endif
  #define		CCR_CLASS		2
  #if( CCR_PHYS_MAX > PHYS_MAX )
  #error		"CCR_PHYS_MAX should be <= PHYS_MAX"
  #endif
  #if( CCR_ARCH > ARCH_MAX )
  #error		"CCR_PHYS_MAX should be <= ARCH_MAX"
  #endif

  #if( !defined(SPR_PHYS_MAX))
  #define		SPR_PHYS_MAX		64
  #endif
  #if( !defined(SPR_ARCH))
  #define		SPR_ARCH		32
  #endif
  #define		SPR_CLASS		3
  #if( SPR_PHYS_MAX > PHYS_MAX )
  #error		"SPR_PHYS_MAX should be <= PHYS_MAX"
  #endif
  #if( SPR_ARCH > ARCH_MAX )
  #error		"SPR_PHYS_MAX should be <= ARCH_MAX"
  #endif

// jdonald making register counts flexible

	unsigned gpr_phys, fpr_phys, spr_phys, ccr_phys;

/* reclaim phys from insn iq[at] when flush */
#define free_iq_freeq(thread_, at)					\
  {								\
    xdep_prep_info *	entry_ = xiq_entry( iq[thread_][at] );		\
    unsigned		res_ = xdep_prep_info_res(entry_);	\
    unsigned		ins_ = xiq_ins( iq[thread_][at] );		\
    unsigned		reg_idx_ = xiq_reg_idx( iq[thread_][at] );	\
    unsigned		idx_ = ins_ + reg_idx_;			\
    unsigned		outs_ = xiq_outs( iq[thread_][at] );		\
    xarch *		out_regs_ = xiq_out_regs( iq[thread_][at] );	\
    unsigned		arch_;					\
    unsigned		reg_;					\
    unsigned		j_;					\
    unsigned		tl_;					\
    unsigned		class_;					\
								\
    bounds(at, IQ_SIZE );					\
    assert( ((~res_&0x7f7f7f7f) +0x01010101) ==			\
		xdep_prep_info_res_compl(entry_) );		\
								\
    freeq_avail_vec += res_;					\
								\
    for( j_ = 0; j_ < outs_; j_++ ) {				\
      assert( idx_ + j_ < IQ_TOTAL_REG );			\
      reg_ = iq_reg[thread_][idx_+j_];					\
      arch_ = out_regs_[j_];					\
      phys[thread_][reg_] = /* UINT_MAX;	*/ /* LYM 64bits adaptation */ THE_MAX;				\
      class_ = reg_ >> LG2PHYS_MAX;				\
      assert( (arch_ >> LG2ARCH_MAX) == class_ );		\
      tl_ = freeq_tl[thread_][class_];					\
      freeq[thread_][tl_] = reg_;					\
      freeq_tl[thread_][class_] = tl_ + 1;				\
    }								\
  }

/* reclaim phys from insn iq[at] when retire, and update inorder map */
#define free_inorder_iq_freeq(thread_, at)				\
  {								\
    xdep_prep_info *	entry_ = xiq_entry( iq[thread_][at] );		\
    unsigned		res_ = xdep_prep_info_res(entry_);	\
    unsigned		ins_ = xiq_ins( iq[thread_][at] );		\
    unsigned		reg_idx_ = xiq_reg_idx( iq[thread_][at] );	\
    unsigned		idx_ = ins_ + reg_idx_;			\
    unsigned		outs_ = xiq_outs( iq[thread_][at] );		\
    xarch *		out_regs_ = xiq_out_regs( iq[thread_][at] );	\
    unsigned		arch_;					\
    unsigned		free_;					\
    unsigned		reg_;					\
    unsigned		j_;					\
    unsigned		tl_;					\
    unsigned		class_;					\
								\
    bounds( at, IQ_SIZE );					\
    assert( ((~res_&0x7f7f7f7f) +0x01010101) ==			\
		xdep_prep_info_res_compl(entry_) );		\
								\
    freeq_avail_vec += res_;					\
								\
    for( j_ = 0; j_ < outs_; j_++ ) {				\
      assert( idx_ + j_ < IQ_TOTAL_REG );			\
      reg_ = iq_reg[thread_][idx_+j_];					\
      arch_ = out_regs_[j_];					\
      assert( arch_ < ARCH_TOTAL );				\
      free_ = inorder[thread_][arch_];					\
      assert( free_ < PHYS_TOTAL );				\
								\
      inorder[thread_][ arch_] = reg_;					\
      phys[thread_][free_] = /* UINT_MAX; */ THE_MAX; /* LYM 64 bits adaptation */					\
      class_ = reg_ >> LG2PHYS_MAX;				\
      assert( class_ < CLASS_MAX);				\
      assert( (arch_ >> LG2ARCH_MAX) == class_ );		\
      assert( (free_ >> LG2PHYS_MAX) == class_ );		\
      tl_ = freeq_tl[thread_][class_];					\
      freeq[thread_][tl_] = free_;					\
      freeq_tl[thread_][class_] = tl_ + 1;				\
    }								\
  }

/***** dispatch *****/
  #if( !defined(DISPATCH_MAX))
  #define		DISPATCH_MAX		4
  #endif
  xiq_idx		iq_dis[MAX_THREADS];
  xcycle		dispatch_stalled_till[MAX_THREADS] = {0}; /* LYM */

  #if( GETCPI )
    unsigned		cpiq_n[MAX_THREADS];
    unsigned		cpiq_unit[MAX_THREADS][DISPATCH_MAX];
    unsigned		cpiq_idx[MAX_THREADS][DISPATCH_MAX];
  #endif /* #if( GETCPI ) */

/***** miss queues *****/
  #if( !defined(DMISS_MAX))
  #define		DMISS_MAX		7
  #endif

  #define		DMISS_DEFAULT_ADDR	(~0)

  xiq_idx		dmiss_hd[MAX_THREADS];
  xiq_idx		dmiss_tl[MAX_THREADS];
  xresv_cnt		dmiss_n[MAX_THREADS];

  xiq_info		dmiss_hd_node, dmiss_tl_node, *global_dmiss_hd = NULL, *global_dmiss_tl = NULL; // initialize 2x
  xresv_cnt		global_dmiss_n = 0; // initialize

  unsigned		stats_lines_n[MAX_THREADS];
  unsigned		stats_l2_dmiss_n[MAX_THREADS];
  unsigned		stats_l2_lines_n[MAX_THREADS];

/* these variables are temp variables shared among threads */
  unsigned		dmiss_address = 0; // initialize
  boolean		dmiss_is_store = 0; // initialize
  unsigned		dmiss_loaded = 0; // initialize
  unsigned		dmiss_thread = 0;	/* currently thread being loaded */

  // unsigned		dmiss_spec_hdspec; // unused?
  // unsigned		dmiss_spec_hdaddr; // unused?

/***** castout queues *****/
  #if( !defined(DCASTOUT_MAX) )
  #define		DCASTOUT_MAX		DMISS_MAX
  #endif

  #define		DCASTOUT_OVERHEAD	5
  xresv_cnt		dcastout_n = 0;

/* bus */
  xcycle		dmiss_early[MAX_THREADS] = {0};
  xcycle		dcastout_early = 0;
  xcycle		ifetch_miss_till[MAX_THREADS] = {0};

/***** reservation stations *****/

  #if( USING_CMPLX_UNIT )
  #define		UNITS_MAX		6
  #elif( USING_LOG_UNIT )
  #define		UNITS_MAX		5
  #else
  #define		UNITS_MAX		4
  #endif

  xresv_cnt		resv_left[MAX_THREADS][UNITS_MAX*CLUSTERS_MAX];

  #if( !defined(FIX_RESV) )
  #define		FIX_RESV		20
  #endif
  #if( !defined(FIX_NUM))
  #define		FIX_NUM			3
  #endif
  xresv_cnt		fix_n[MAX_THREADS];
  xiq_idx		fix_hd[MAX_THREADS];
  xiq_idx		fix_tl[MAX_THREADS];

  xiq_info		fix_hd_node, fix_tl_node, *global_fix_hd, *global_fix_tl;
  xresv_cnt		global_fix_n;

  #if( USE_NONPIPE_FIX_OPS )
  unsigned		fix_nonpipe_in_progress = 0; // initialize
  xcycle		fix_nonpipe_till[FIX_NUM]; /* LYM */
  #endif

  #if( CLUSTER && CLUSTERS_MAX > 1 )
  xresv_cnt		fix1_n[MAX_THREADS];
  xiq_idx		fix1_hd[MAX_THREADS];
  xiq_idx		fix1_tl[MAX_THREADS];

  xiq_info		fix1_hd_node, fix1_tl_node, *global_fix1_hd, *global_fix1_tl;
  xresv_cnt		global_fix1_n;
  #endif

  #if( USE_NONPIPE_FIX_OPS )
  unsigned		fix1_nonpipe_in_progress = 0; // initialize
  xcycle		fix1_nonpipe_till[FIX_NUM]; /* LYM */
  #endif

  #if( !defined(FPU_RESV) )
  #define		FPU_RESV		20
  #endif
  #if( !defined(FPU_NUM) )
  #define		FPU_NUM			2
  #endif
  xresv_cnt		fpu_n[MAX_THREADS];
  xiq_idx		fpu_hd[MAX_THREADS];
  xiq_idx		fpu_tl[MAX_THREADS];

  xiq_info		fpu_hd_node, fpu_tl_node, *global_fpu_hd = NULL, *global_fpu_tl = NULL; // initialize 2x
  xresv_cnt		global_fpu_n = 0; // initialize

  #if( USE_NONPIPE_FPU_OPS )
  unsigned		fpu_nonpipe_in_progress = 0; // initialize
  xcycle		fpu_nonpipe_till[FPU_NUM] = { 0 }; /* LYM */ // initialize
  #endif

  #if( CLUSTER && CLUSTERS_MAX > 1 )
  xresv_cnt		fpu1_n[MAX_THREADS];
  xiq_idx		fpu1_hd[MAX_THREADS];
  xiq_idx		fpu1_tl[MAX_THREADS];

  xiq_info		fpu1_hd_node, fpu1_tl_node, *global_fpu1_hd = NULL, *global_fpu1_tl = NULL; // initialize 2x
  xresv_cnt		global_fpu1_n = 0; // initialize

  #if( USE_NONPIPE_FPU_OPS )
  unsigned		fpu1_nonpipe_in_progress = 0; // initialize
  xcycle		fpu1_nonpipe_till[FPU_NUM] = { 0 }; /* LYM */ // initialize
  #endif
  #endif

  #if( !defined(MEM_RESV))
  #define		MEM_RESV		20
  #endif
  #if( !defined(MEM_NUM))
  #define		MEM_NUM			2
  #endif
  xresv_cnt		mem_n[MAX_THREADS];
  xiq_idx		mem_hd[MAX_THREADS];
  xiq_idx		mem_tl[MAX_THREADS];
  xcycle		mem_exec_busy_till = 0; /* LYM */ // initialize

  xiq_info		mem_hd_node, mem_tl_node, *global_mem_hd = NULL, *global_mem_tl = NULL; // initialize 2x
  xresv_cnt		global_mem_n = 0; // initialize

  #if( CLUSTER && CLUSTERS_MAX > 1 )
  xresv_cnt		mem1_n[MAX_THREADS];
  xiq_idx		mem1_hd[MAX_THREADS];
  xiq_idx		mem1_tl[MAX_THREADS];
  xcycle		mem1_exec_busy_till; /* LYM */

  xiq_info		mem1_hd_node, mem1_tl_node, *global_mem1_hd = NULL, *global_mem1_tl = NULL; // initialize 2x
  xresv_cnt		global_mem1_n = 0; // initialize
  #endif

  #if( !defined( BR_RESV ) )
  #define		BR_RESV			12
  #endif
  #if( !defined( BR_NUM ) )
  #define		BR_NUM			2
  #endif
  xresv_cnt		br_n[MAX_THREADS];
  xiq_idx		br_hd[MAX_THREADS];
  xiq_idx		br_tl[MAX_THREADS];

  xiq_info		br_hd_node, br_tl_node, *global_br_hd = NULL, *global_br_tl = NULL; // initialize
  xresv_cnt		global_br_n = 0; // initialize

  #if( CLUSTER && CLUSTERS_MAX > 1 )
  xresv_cnt		br1_n[MAX_THREADS];
  xiq_idx		br1_hd[MAX_THREADS];
  xiq_idx		br1_tl[MAX_THREADS];

  xiq_info		br1_hd_node, br1_tl_node, *global_br1_hd, *global_br1_tl;
  xresv_cnt		global_br1_n;
  #endif

#if( USING_LOG_UNIT )
  #if( !defined( LOG_RESV ) )
  #define		LOG_RESV		12
  #endif
  #if( !defined( LOG_NUM ) )
  #define		LOG_NUM			1
  #endif
  xresv_cnt		log_n[MAX_THREADS];
  xiq_idx		log_hd[MAX_THREADS];
  xiq_idx		log_tl[MAX_THREADS];

  xiq_info		log_hd_node, log_tl_node, *global_log_hd = NULL, *global_log_tl = NULL; // intialize
  xresv_cnt		global_log_n = 0; // initialize

  #if( CLUSTER && CLUSTERS_MAX > 1 )
  xresv_cnt		log1_n[MAX_THREADS];
  xiq_idx		log1_hd[MAX_THREADS];
  xiq_idx		log1_tl[MAX_THREADS];

  xiq_info		log1_hd_node, log1_tl_node, *global_log1_hd, *global_log1_tl;
  xresv_cnt		global_log1_n;
  #endif
#else
  #if( !defined( LOG_RESV ) )
  #define		LOG_RESV		0
  #endif
  #if( !defined( LOG_NUM) )
  #define		LOG_NUM			0
  #endif
  #define		log_n			0
#endif

#if( USING_CMPLX_UNIT )
  #if( !defined( CMPLX_RESV ) )
  #define		CMPLX_RESV		16
  #endif
  #if( !defined( CMPLX_NUM ) )
  #define		CMPLX_NUM		1
  #endif
  xresv_cnt		cmplx_n[MAX_THREADS];
  xiq_idx		cmplx_hd[MAX_THREADS];
  xiq_idx		cmplx_tl[MAX_THREADS];

  xiq_info		cmplx_hd_node, cmplx_tl_node, *global_cmplx_hd, *global_cmplx_tl;
  xresv_cnt		global_cmplx_n;
  #if( CLUSTER && CLUSTERS_MAX > 1 )
  xresv_cnt		cmplx1_n[MAX_THREADS];
  xiq_idx		cmplx1_hd[MAX_THREADS];
  xiq_idx		cmplx1_tl[MAX_THREADS];

  xiq_info		cmplx1_hd_node, cmplx1_tl_node, *global_cmplx1_hd, *global_cmplx1_tl;
  xresv_cnt		global_cmplx1_n;
  #endif
#else
  #define		CMPLX_RESV		0
  #define		CMPLX_NUM		0
/* #define		cmplx_n			0 */
  xresv_cnt		cmplx_n[MAX_THREADS] = {0};
#endif

/***** registers *****/
/* physical registers are shared among threads */
  xcycle		phys[MAX_THREADS][PHYS_TOTAL];
  xphys_reg		freeq[MAX_THREADS][PHYS_TOTAL];
  xphys_cnt		freeq_tl[MAX_THREADS][CLASS_MAX];
  unsigned		freeq_avail_vec = 0; // initialize

  #if( GETCPI )
  unsigned char		src_phys[MAX_THREADS][PHYS_TOTAL];
  #endif

/***** retirement queue *****/
  #if( !defined( RETIRE_MAX ) )
  #define		RETIRE_MAX		6
  #endif
  #if( !defined( RETIREQ_MAX ) )
  #define		RETIREQ_MAX		128
  #endif
  #if( RETIREQ_MAX + IBUF_MAX > 256 )
  #error		"RETIREQ_MAX + IBUF_MAX should be <= 256"
  #endif

  xiq_idx		iq_retire[MAX_THREADS] = {0};
  xiq_cnt		retireq_left[MAX_THREADS] = {0};
  unsigned		groups_left[MAX_THREADS] = {0};
  unsigned		groups_gid[MAX_THREADS] = {0};
  unsigned		decode_reissue_groups[MAX_THREADS] = {0};
  unsigned		retire_branch_history[MAX_THREADS] = {0};
  // unsigned		retire_pred_count[MAX_THREADS] = {0}; // unused?


  xiq_idx		iq_flush[MAX_THREADS] = {0};
  int			do_flush[MAX_THREADS] = {0};

/**** POWERTIMER STATS ADDED BY DBROOKS */
/*  unsigned		stats_num_stores[MAX_THREADS] = {0}; */
/* unsigned		stats_num_loads[MAX_THREADS] = {0}; */
  long long            stats_num_stores[MAX_THREADS] = {0}; /* LYM long long */
  long long            stats_num_loads[MAX_THREADS] = {0}; /* LYM long long */
  boolean		is_load;

/***** others *****/
  unsigned 		insns[MAX_THREADS] = {0};
  unsigned 		insns_total_execd[MAX_THREADS] = {0};
  unsigned		insns_total_dispd[MAX_THREADS] = {0};
  unsigned		insns_total_retired[MAX_THREADS] = {0};
  unsigned		memop[MAX_THREADS] = {0};

  unsigned		bp_exec_total[MAX_THREADS] = {0};
  unsigned		bp_exec_pred[MAX_THREADS] = {0};
  unsigned		bp_exec_miss[MAX_THREADS] = {0};
  unsigned		bp_exec_bc_miss[MAX_THREADS] = {0};
  unsigned		bp_exec_bc_total[MAX_THREADS] = {0};
  unsigned		bp_exec_bclr_miss[MAX_THREADS] = {0};
  long long             bp_exec_bclr_total[MAX_THREADS] = {0}; /* LYM long long */
  unsigned		bp_exec_bcctr_miss[MAX_THREADS] = {0};
  long long            bp_exec_bcctr_total[MAX_THREADS] = {0}; /* LYM long long */


  unsigned		stats_dinter_conflicts = 0;
  unsigned		stats_dinter_equals = 0;

  unsigned		stats_strings_total[MAX_THREADS] = {0};
  unsigned		stats_strings_expands[MAX_THREADS] = {0};
  unsigned		stats_strings_stalls[MAX_THREADS] = {0};

  unsigned		stats_total_execd[MAX_THREADS] = {0};
  unsigned		stats_nexpands[MAX_THREADS] = {0};
  unsigned		stats_expands_to[MAX_THREADS]  = {0};
  unsigned		stats_nops_dispd[MAX_THREADS] = {0};
  unsigned		stats_nops_retd[MAX_THREADS] = {0};

  unsigned		stats_mtsr[MAX_THREADS] = {0};
  unsigned		stats_mtsr_ipurge[MAX_THREADS] = {0};
  unsigned		stats_mtsr_dpurge[MAX_THREADS] = {0};


  /* unsigned	stats_dtmiss[MAX_THREADS];
  unsigned		stats_d2tmiss[MAX_THREADS];
  unsigned		stats_dmiss[MAX_THREADS]; */
  long long		stats_dtmiss[MAX_THREADS] = {0};
  long long		stats_d2tmiss[MAX_THREADS] = {0};
  long long		stats_dmiss[MAX_THREADS] = {0}; /* LYM long long */
  unsigned		stats_store_dmiss[MAX_THREADS] = {0};
  unsigned		stats_store_dmiss_taken[MAX_THREADS] = {0};
  unsigned		stats_dlines[MAX_THREADS] = {0};
  unsigned		stats_d1lines[MAX_THREADS] = {0};
  unsigned		stats_store_d1lines[MAX_THREADS] = {0};
  unsigned		stats_d2lines[MAX_THREADS] = {0};

  unsigned		stats_dtrail[MAX_THREADS] = {0};
  unsigned		stats_dmiss0[MAX_THREADS] = {0};
  /* unsigned	stats_dprobe[MAX_THREADS]; */
  long long		stats_dprobe[MAX_THREADS] = {0}; /* LYM long long */
  unsigned		stats_dcast[MAX_THREADS] = {0};
  unsigned		stats_dl2miss[MAX_THREADS] = {0};

  unsigned		stats_dtmiss_taken[MAX_THREADS] = {0};
  unsigned		stats_d2tmiss_taken[MAX_THREADS] = {0};
  unsigned		stats_dmiss_taken[MAX_THREADS] = {0};
  unsigned		stats_dl2miss_taken[MAX_THREADS] = {0};

  /* unsigned	stats_itmiss[MAX_THREADS] = {0};
  unsigned		stats_i2tmiss[MAX_THREADS] = {0};
  unsigned		stats_imiss[MAX_THREADS] = {0}; */
  long long		stats_itmiss[MAX_THREADS] = {0};
  long long		stats_i2tmiss[MAX_THREADS] = {0};
  long long		stats_imiss[MAX_THREADS] = {0}; /* LYM long long */
  unsigned		stats_imiss0[MAX_THREADS] = {0};
  /* unsigned	stats_iprobe[MAX_THREADS] = {0}; */
  long long		stats_iprobe[MAX_THREADS] = {0}; /* LYM long long */
  unsigned		stats_i15miss[MAX_THREADS] = {0};
  unsigned		stats_il2miss[MAX_THREADS] = {0};

#if( USE_PREDICATION )
  unsigned		stats_pred_branch = 0;
  unsigned		stats_pred_exps = 0;
  unsigned		stats_pred_ops = 0;
  unsigned		stats_pred_taken = 0;
#endif

/* not updated yet 'cause trace mode has no spec insns */
#if( GET_SPEC_STATS )
  unsigned		stats_spec_ifetch = 0;

  unsigned		stats_spec_expands_to = 0;
  unsigned		stats_spec_nexpands = 0;

  unsigned		stats_spec_dispatch = 0;

  unsigned		stats_spec_fix_execd = 0;
  unsigned		stats_spec_fpu_execd = 0;
  unsigned		stats_spec_mem_execd = 0;
  unsigned		stats_spec_br_execd = 0;

  unsigned		stats_spec_imiss = 0;
  unsigned		stats_spec_dmiss = 0;
  unsigned		stats_spec_dlines = 0;
  unsigned		stats_spec_dflush = 0;
  unsigned		stats_spec_dsatisfy = 0;

  unsigned		stats_spec_extra_hits = 0;
  unsigned		stats_spec_extra_misses = 0;
  unsigned		stats_spec_src_hits = 0;
  unsigned		stats_spec_reexec_hits = 0;
  unsigned		stats_spec_used_once = 0;

  unsigned		stats_spec_extra_ihits = 0;
  unsigned		stats_spec_extra_imisses = 0;
  unsigned		stats_spec_src_ihits = 0;
  unsigned		stats_spec_iused_once = 0;

  unsigned		stats_spec_dflush_histo[ DMISS_MAX + MEM_NUM + 1] = {0};
  unsigned		stats_spec_over_histo[ DMISS_MAX + MEM_NUM + 1];
  unsigned		stats_spec_dflush_over_histo[ DMISS_MAX + MEM_NUM + 1] = {0};
#endif /* #if( GET_SPEC_STATS ) */

#if( EVALUATE_MISPREDICTED_BRANCHES )
  unsigned		stats_eval_mispredict;
#endif

#if( GETCPI )
  unsigned		cpi_counters[MAX_THREADS][RETIRE_MAX][TR_NUM] = {{{0}}};
  unsigned		cpi_max = 0; // initialize
#endif

#if( GETCPI_REMAINING )
  unsigned		cpi_rem_counters[RETIRE_MAX][TR_NUM] = {{0}};
  unsigned		cpi_rem_max;
#endif

#if( STATS )
  unsigned 		stats_retired[MAX_THREADS][RETIRE_MAX+1] = {{0}};
  unsigned 		stats_st_retired[MAX_THREADS][DCACHE_PORTS+1] = {{0}};
  unsigned 		stats_st_commit[MAX_THREADS][DCACHE_PORTS+1] = {{0}};

  unsigned		stats_fix_execd[MAX_THREADS][ FIX_NUM + 1] = {{0}};
  unsigned		stats_fpu_execd[MAX_THREADS][ FPU_NUM + 1] = {{0}};
  unsigned		stats_mem_execd[MAX_THREADS][ MEM_NUM + 1] = {{0}};
  unsigned		stats_br_execd[MAX_THREADS][ BR_NUM + 1] = {{0}};
  #if( USING_LOG_UNIT )
    unsigned		stats_log_execd[MAX_THREADS][ LOG_NUM + 1] = {{0}};
  #endif
  #if( USING_CMPLX_UNIT )
    unsigned		stats_cmplx_execd[MAX_THREADS][ CMPLX_NUM + 1] = {{0}};
  #endif

  #if( CLUSTER )
    unsigned		stats_fix1_execd[MAX_THREADS][ FIX_NUM + 1] = {{0}};
    unsigned		stats_fpu1_execd[MAX_THREADS][ FPU_NUM + 1] = {{0}};
    unsigned		stats_mem1_execd[MAX_THREADS][ MEM_NUM + 1] = {{0}};
    unsigned		stats_br1_execd[MAX_THREADS][ BR_NUM + 1] = {{0}};
  #endif
  #if( USING_LOG_UNIT )
    unsigned		stats_log1_execd[MAX_THREADS][ LOG_NUM + 1] = {{0}};
  #endif

  #if( USING_CMPLX_UNIT )
    unsigned		stats_cmplx1_execd[MAX_THREADS][ CMPLX_NUM + 1] = {{0}};
  #endif

  unsigned		stats_execd[MAX_THREADS][(FIX_NUM+FPU_NUM+MEM_NUM+BR_NUM+LOG_NUM+CMPLX_NUM)*CLUSTERS_MAX + 1] = {{0}};

  unsigned		stats_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};

  unsigned		stats_fix_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};
  unsigned		stats_fpu_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};
  unsigned		stats_mem_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};
  unsigned		stats_br_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};
  #if( USING_LOG_UNIT )
    unsigned		stats_log_dispatched[MAX_THREADS][ DISPATCH_MAX + 1] = {{0}};
  #endif
  #if( USING_CMPLX_UNIT )
    unsigned		stats_cmplx_dispatched[MAX_THREADS][ DISPATCH_MAX + 1] = {{0}};
  #endif

  #if( CLUSTER )
    unsigned		stats_fix1_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};
    unsigned		stats_fpu1_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};
    unsigned		stats_mem1_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};
    unsigned		stats_br1_dispatched[MAX_THREADS][ DISPATCH_MAX + 1 ] = {{0}};
    #if( USING_LOG_UNIT )
      // unsigned		stats_log1_dispatched[MAX_THREADS][ DISPATCH_MAX + 1]; // unused?
    #endif
    #if( USING_CMPLX_UNIT )
      unsigned		stats_cmplx1_dispatched[MAX_THREADS][ DISPATCH_MAX + 1] = {0};
    #endif
  #endif

  unsigned		stats_renamed[MAX_THREADS][ IBUF_MAX + 1 ] = {{0}};
  unsigned		stats_fetched[MAX_THREADS][ IFETCH_MAX + 1 ] = {{0}};

  unsigned		stats_ibuf[MAX_THREADS][ IBUF_MAX + 1] = {{0}};
  unsigned		stats_fchblk[MAX_THREADS][ FCHBLK_MAX + 1] = {{0}};
  unsigned 		stats_retireq[MAX_THREADS][RETIREQ_MAX+1] = {{0}};
  unsigned		stats_fix_resv[MAX_THREADS][ FIX_RESV + 1] = {{0}};
  unsigned		stats_fpu_resv[MAX_THREADS][ FPU_RESV + 1] = {{0}};
  unsigned		stats_mem_resv[MAX_THREADS][ MEM_RESV + 1] = {{0}};
  unsigned		stats_br_resv[MAX_THREADS][ BR_RESV + 1] = {{0}};

/*  unsigned		pstats_br_resv[MAX_THREADS] = {0}; */
#if ( !( COMBINE_FIX_MEM ) ) // jdonald
  unsigned		pstats_fix_resv[MAX_THREADS] = {0}; // this is unused outside of putting another 0 in dumpstatus -jdonald
#endif
/*  unsigned		pstats_fpu_resv[MAX_THREADS] = {0}; */
/* unsigned		pstats_fpu1_resv[MAX_THREADS] = {0}; */
/* unsigned		pstats_mem_resv[MAX_THREADS] = {0}; */
/* unsigned		pstats_mem1_resv[MAX_THREADS] = {0}; */

/*  unsigned		pstats_mem_execd[MAX_THREADS] = {0}; */
/*  unsigned		pstats_mem1_execd[MAX_THREADS] = {0}; */
/*  unsigned		pstats_fpu_execd[MAX_THREADS] = {0}; */
/*  unsigned		pstats_fpu1_execd[MAX_THREADS] = {0}; */

/*  unsigned		pstats_gpr[MAX_THREADS] = {0}; */
/*  unsigned		pstats_fpr[MAX_THREADS] = {0}; */
/*  unsigned		pstats_spr[MAX_THREADS] = {0}; */
/*  unsigned		pstats_ccr[MAX_THREADS] = {0}; */
/* unsigned		pstats_groups[MAX_THREADS] = {0}; */

/*  unsigned		pstats_dmissq[MAX_THREADS] = {0}; */
/*  unsigned		pstats_reorderq[MAX_THREADS] = {0}; */
/*  unsigned		pstats_storeq[MAX_THREADS] = {0}; */

    long long           pstats_br_resv[MAX_THREADS] = {0};
    long long           pstats_fpu_resv[MAX_THREADS] = {0};
    long long           pstats_fpu1_resv[MAX_THREADS] = {0};
    long long           pstats_mem_resv[MAX_THREADS] = {0};
    long long           pstats_mem1_resv[MAX_THREADS] = {0};

    long long            pstats_mem_execd[MAX_THREADS] = {0};
    long long            pstats_mem1_execd[MAX_THREADS] = {0};
    long long            pstats_fpu_execd[MAX_THREADS] = {0};
    long long            pstats_fpu1_execd[MAX_THREADS] = {0};

    long long            pstats_gpr[MAX_THREADS] = {0};
    long long            pstats_fpr[MAX_THREADS] = {0};
    long long            pstats_spr[MAX_THREADS] = {0};
    long long           pstats_ccr[MAX_THREADS] = {0};
    long long            pstats_groups[MAX_THREADS] = {0};

    long long            pstats_dmissq[MAX_THREADS] = {0};
    long long           pstats_reorderq[MAX_THREADS] = {0};
    long long            pstats_storeq[MAX_THREADS] = {0}; /* LYM long long */


  #if( USING_LOG_UNIT )
    unsigned		stats_log_resv[MAX_THREADS][ LOG_RESV + 1] = {{0}};
  #endif
  #if( USING_CMPLX_UNIT )
    unsigned		stats_cmplx_resv[MAX_THREADS][ CMPLX_RESV + 1] = {{0}};
  #endif

  #if( CLUSTER )
    unsigned		stats_fix1_resv[MAX_THREADS][ FIX_RESV + 1] = {{0}};
    unsigned		stats_fpu1_resv[MAX_THREADS][ FPU_RESV + 1] = {{0}};
    unsigned		stats_mem1_resv[MAX_THREADS][ MEM_RESV + 1] = {{0}};
    unsigned		stats_br1_resv[MAX_THREADS][ BR_RESV + 1] = {{0}};
    #if( USING_LOG_UNIT )
      unsigned		stats_log1_resv[MAX_THREADS][ LOG_RESV + 1] = {{0}};
    #endif
    #if( USING_CMPLX_UNIT )
      unsigned		stats_cmplx1_resv[MAX_THREADS][ CMPLX_RESV + 1] = {{0}};
    #endif
  #endif

  unsigned		stats_resv[MAX_THREADS][ (FIX_RESV + FPU_RESV + MEM_RESV + BR_RESV + LOG_RESV + CMPLX_RESV )*CLUSTERS_MAX + 1] = {{0}};

  unsigned		stats_gpr[MAX_THREADS][ GPR_PHYS_MAX + 1] = {{0}};
  unsigned		stats_fpr[MAX_THREADS][ FPR_PHYS_MAX + 1 ] = {{0}};
  unsigned		stats_spr[MAX_THREADS][ SPR_PHYS_MAX + 1 ] = {{0}};
  unsigned		stats_ccr[MAX_THREADS][ CCR_PHYS_MAX + 1 ] = {{0}};
  unsigned		stats_dmissq[MAX_THREADS][ DMISS_MAX + MEM_NUM + 1] = {{0}};
  unsigned		stats_l2dmissq[MAX_THREADS][ DMISS_MAX + MEM_NUM + 1] = {{0}};
  unsigned		stats_lines[MAX_THREADS][ DMISS_MAX + MEM_NUM + 1] = {{0}};
  unsigned		stats_l2_lines[MAX_THREADS][ DMISS_MAX + MEM_NUM + 1] = {{0}};
  unsigned		stats_dcastoutq[ DCASTOUT_MAX + MEM_NUM + 1] = {0};
  unsigned		stats_inflight[MAX_THREADS][INFLIGHT_MAX + 1] = {{0}};
  unsigned		stats_groups[MAX_THREADS][GROUPS_MAX + 1] = {{0}};

  unsigned		stats_dmiss_stall = 0;
  unsigned		stats_dcastout_stall = 0;
  unsigned		stats_resv_full_stall[MAX_THREADS] = {0};
  unsigned		stats_rename_stall[MAX_THREADS] = {0};
  unsigned		stats_ibuf_stall[MAX_THREADS] = {0};
  unsigned		stats_inflight_stall[MAX_THREADS] = {0};
  unsigned		stats_reissue_decode[MAX_THREADS] = {0};
  unsigned		stats_reissue_flushes[MAX_THREADS] = {0};
  unsigned		stats_recirc[MAX_THREADS] = {0};
/* unsigned		stats_groups_dispatched[MAX_THREADS] = {0}; */
  long long            stats_groups_dispatched[MAX_THREADS] = {0}; /* LYM long long */
#endif /* STATS */

#if( STAT0 )
  unsigned		stat0_ifetch_reason[MAX_THREADS] = {0};
  // unsigned		stat0_ifetch_other[MAX_THREADS] = {0}; // unused?
  unsigned		stat0_ifetch_tlb[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_cache[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_nfa[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_ibuf[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_fchblk[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_sync[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_end[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_partial[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_mispredict[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_inflight[MAX_THREADS] = {0};
  unsigned		stat0_ifetch_pred[MAX_THREADS] = {0};

/*  unsigned		stat0_decode_empty[MAX_THREADS] = {0}; */
  long long            stat0_decode_empty[MAX_THREADS] = {0}; /* LYM long long */
  unsigned		stat0_decode_stall[MAX_THREADS] = {0};
  unsigned		stat0_decode_multi[MAX_THREADS] = {0};
  unsigned		stat0_decode_string[MAX_THREADS] = {0};

  unsigned		stat0_dispatch_groups[MAX_THREADS] = {0};
#endif /* STAT0 */

/***** temporary variables for decomposing current instruction *****/
  unsigned		iaddr;
  unsigned		iword;
  xdep_prep_info *	entry;		/* pre-decoded information for the current instruction, Zhigang Hu@IBMUS, 1/8/2003 */
  unsigned char		ins;
  unsigned char		outs;
  unsigned		daddr;
  unsigned char		dlength;
  boolean		is_over;
  boolean		is_store;
  // boolean		is_multi; // unused?
  xarch *		regs;
  xarch *		out_regs;
  // unsigned		exp_off; // unused?
  unsigned		nexpands;
  xiq_regidx		reg_idx;
  xunit			unit;
  xcycle		latency;
  xcycle		avail;

#if( TRACE )
/***** trace variables ******/
  // char			dtlb_trace_buf[16]; // unused?
  // char			itlb_trace_buf[16]; // unused?
  char *		trace_ifetch_status;
  char *		trace_mem_status = NULL; // initialize
#endif /* #if( TRACE ) */

#if( TIMELINE )
  unsigned		tl_start = 0;
  unsigned		tl_size = 100;	/* HZG - was 80 */
  FILE *		tl_output;
#endif /* #if( TIMELINE ) */

/***** counters *****/
  unsigned		i;
  unsigned		j;
  // unsigned		k; // unused?
  unsigned		thread_;	/* just for thread */

  xcycle cycle; // no longer a __thread variable
#if( !(FIXED_FREQUENCY) )
  double timestamp;
#endif

#if(POSTPROCESS)
  FILE *		output_fp = 0; // unused?
  FILE *  	    input_fp = 0;
#endif // POSTPROCESS

  int processor_id; // no longer a __thread variable

  pthread_mutex_lock( &tempsync_cond_mutex ); // avoid non-reentrant problems with I/O on init.macros or init.power.macros, or opc_init()
  processor_id = (int) arg; // jdonald

/* LYM CMP */
  /* if (the_stderr_setup == 0) */
  the_stderr = stderr; // can change this later to output different cores to different files

  #include "init.macros"  // used to be inside main(), but now it's inside run_processor()

#if( POWER_MODE )
  #include "init.power.core.macros"
#endif // POWER_MODE

/* LYM CMP TEMP END */

#ifndef NO_TLS
  trans_overs = malloc(sizeof(xarch) * TRANS_OVERS_NUM);
  memset(trans_overs, 0, sizeof(xarch) * TRANS_OVERS_NUM);
  trans_deps = malloc(sizeof(xdep_prep_info) * TRANS_DEPS_NUM);
  memset(trans_deps, 0, sizeof(xdep_prep_info) * TRANS_DEPS_NUM);
  trans_ptrs = malloc(sizeof(trans_ptrs_t) * TRANS_PTRS_NUM);
  memset(trans_ptrs, 0, sizeof(trans_ptrs_t) * TRANS_PTRS_NUM);
  trans_array = malloc(sizeof(xdep_prep_info**) * TRANS_ARRAY_NUM);
  memset(trans_array, 0, sizeof(xdep_prep_info**) * TRANS_ARRAY_NUM);

  opc_init(); // this region is still protected by aria_mutex
#endif // otherwise, opc_init() should be called back in main()

  /* skip some instructions in the trace if required */
  if( skip )
  {
    unsigned	tmp_left = skip;
    unsigned	tmp_n;

    while( 1 )
    {
      reader_fill_buf(thread_ifetch, tmp_n);
      if( tmp_n > tmp_left )
      {
	reader_pop_n(thread_ifetch, tmp_left);
	break;
      }
      reader_flush(thread_ifetch);
      tmp_left -= tmp_n;
    }/* while */
    fprintf(the_stderr, "skipped %llu...\n", skip);
  }/* if( skip ) */

  DBG( fprintf(the_stderr, "tl = %d,%d,%d,%d\n",
	freeq_tl[GPR_CLASS], freeq_tl[FPR_CLASS], freeq_tl[CCR_CLASS], freeq_tl[SPR_CLASS]));
  TRC( fprintf(the_stderr, "beginning simulation\n") );

#if( TRACE )
    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
    fprintf(the_stderr, "@@ CYCLES DATA\n");
    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n");
    fprintf(the_stderr, "@@ %10s %10s %10s %10s %10s %10s %10s [%4s %4s]\n",
	"cycle",
	"fetch",
	"total",
	"retir",
	"mispr",
	"ldstc",
	"dlines",
	"hd",
	"tl");
#endif
  dcache_ports = DCACHE_PORTS;

  	AFTER_EACH_STAGE();

	l2_sync[processor_id] = CYCLE_OR_TIMESTAMP;

	fprintf(the_stderr, "begin simulation\n");
  /* main pipeline simulation loop */

  pthread_mutex_unlock( &tempsync_cond_mutex ); // now that things should be safely independent

  while( 1 ) // main loop
  {
  if (!finished_sync[processor_id]) {

  AFTER_EACH_STAGE();

/*****************************************************************************
  STORE COMMIT STAGE
*****************************************************************************/
#if( COMMIT_STORES_LATE )
// begin_commit_stage:;
{
  unsigned	tmp_n = 0;

// pthread_mutex_lock(&mutex1);
#include "stage_commit.macros"
// end_commit_stage:;
// pthread_mutex_unlock(&mutex1);

    assert( size_storeq(thread_commit) < STOREQ_SIZE + 1 );
    assert( tmp_n < DCACHE_PORTS + 1);
    STT( stats_st_commit[thread_commit][tmp_n]++ );
}
#endif
  AFTER_EACH_STAGE();


/*****************************************************************************
  RETIRE STAGE
*****************************************************************************/
// begin_retire_stage:;
{
    #if( STATS )
    unsigned	tmp_retired = 0;
    unsigned	tmp_st_retired = 0;
    #endif

#include "stage_retire.macros"
end_retire_stage:;

    STT( assert( tmp_retired <RETIRE_MAX+1 ));
    STT( stats_retired[thread_retire][tmp_retired]++ );
    STT( assert( tmp_st_retired < DCACHE_PORTS + 1 ));
    STT( stats_st_retired[thread_retire][tmp_st_retired]++ );
    assert( retireq_left[thread_retire] < RETIREQ_MAX + 1);
    STT( stats_retireq[thread_retire][RETIREQ_MAX-retireq_left[thread_retire]]++ );
    assert( size_storeq(thread_retire) < STOREQ_SIZE + 1 );
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "in retire stage\n") );

  STT( stats_total_execd[thread_retire] = 0 );

#if( USING_BYPASS_DISPATCH )
/*****************************************************************************
  DISPATCH STAGE
*****************************************************************************/
// begin_dispatch_stage:;
{
    unsigned	tmp_diff = 0;
    unsigned	tmp_fix_diff = 0;
    unsigned	tmp_fix1_diff = 0;
    unsigned	tmp_fpu_diff = 0;
    unsigned	tmp_fpu1_diff = 0;
    unsigned	tmp_mem_diff = 0;
    unsigned	tmp_mem1_diff = 0;
    unsigned	tmp_log_diff = 0;
    unsigned	tmp_cmplx_diff = 0;
    unsigned	tmp_br_diff = 0;
    unsigned	tmp_execd = 0;

// pthread_mutex_lock(&mutex1);
#include "stage_dispatch.macros"
end_dispatch_stage:;

    assert( tmp_diff < DISPATCH_MAX + 1 );
    assert( size_storeq(thread_dispatch) < STOREQ_SIZE + 1 );
    STT( stats_fix_dispatched[ tmp_fix_diff ]++);
    STT( stats_fix1_dispatched[ tmp_fix1_diff ]++);
    STT( stats_fpu_dispatched[ tmp_fpu_diff ]++);
    STT( stats_fpu1_dispatched[ tmp_fpu1_diff ]++);
    STT( stats_mem_dispatched[ tmp_mem_diff ]++);
    STT( stats_mem1_dispatched[ tmp_mem1_diff ]++);
    #if( USING_LOG_UNIT )
    STT( stats_log_dispatched[ tmp_log_diff ]++);
    #endif
    #if( USING_CMPLX_UNIT )
    STT( stats_cmplx_dispatched[ tmp_cmplx_diff ]++);
    #endif
    STT( stats_br_dispatched[ tmp_br_diff ]++);
    insns_total_dispd += tmp_execd;
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with dispatch stage\n") );
#endif

/*****************************************************************************
  FIX EXECUTE
*****************************************************************************/
#if( !COMBINE_FIX_MEM )
begin_fix_execute_stage:;
{
    #if( STATS )
    unsigned	tmp_execd[MAX_THREADS] = {0};
    #endif

#include "stage_fix_exec.macros"
end_fix_execute_stage:; // should be here, and not in the loop below
	
  for (i = 0; i < n_threads; i++) {
    STT( stats_total_execd[i] += tmp_execd[i] );
    assert( tmp_execd[i] < FIX_NUM + 1 );
    STT( stats_fix_execd[i][ tmp_execd[i]]++ );
    assert( fix_n[i] < FIX_RESV + 1 );
    STT( stats_fix_resv[i][ fix_n[i] ]++ );
    STT( pstats_fix_resv[i] += fix_n[i] );
  }/* for i */
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with fix stage\n") );
#endif

/*****************************************************************************
  FIX1 EXECUTE
*****************************************************************************/
#if( !COMBINE_FIX_MEM && CLUSTER )
begin_fix1_execute_stage:;
{
    #if( STATS )
    unsigned	tmp_execd[MAX_THREADS] = {0};
    #endif

#include "stage_fix1_exec.macros"

end_fix1_execute_stage:;
    for (i = 0; i < n_threads; i++)
    {
    STT( stats_total_execd[i] += tmp_execd[i] );
    assert( tmp_execd[i] < FIX_NUM + 1 );
    STT( stats_fix1_execd[i][ tmp_execd[i]]++ );
    assert( fix1_n[i] < FIX_RESV + 1 );
    STT( stats_fix1_resv[i][ fix1_n[i] ]++ );
    }
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with fix stage\n") );
#endif

#if( USING_CMPLX_UNIT )
/*****************************************************************************
  CMPLX EXECUTE
*****************************************************************************/
begin_cmplx_execute_stage:;
{
    #if( STATS )
    unsigned	tmp_execd[MAX_THREADS] = {0};
    #endif

#include "stage_cmplx_exec.macros"

end_cmplx_execute_stage:;
    for (i = 0; i < n_threads; i++)
    {
    STT( stats_total_execd[i] += tmp_execd[i] );
    assert( tmp_execd[i] < CMPLX_NUM + 1 );
    STT( stats_cmplx_execd[i][ tmp_execd[i]]++ );
    assert( cmplx_n[i] < CMPLX_RESV + 1 );
    STT( stats_cmplx_resv[i][ cmplx_n[i] ]++ );
    }
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with cmplx stage\n") );
#endif


/*****************************************************************************
  FPU EXECUTE
*****************************************************************************/
// begin_fpu_execute_stage:;
{
    #if( STATS )
    unsigned	tmp_execd[MAX_THREADS] = {0};
    #endif

#include "stage_fpu_exec.macros"

end_fpu_execute_stage:;
    for (i = 0; i < n_threads; i++)
    {
    STT( stats_total_execd[i] += tmp_execd[i] );
    STT(assert( tmp_execd[i] < FPU_NUM + 1 ));
    STT( stats_fpu_execd[i][ tmp_execd[i]]++ );
    STT( pstats_fpu_execd[i] += tmp_execd[i] );
    assert( fpu_n[i] < FPU_RESV + 1 );
    STT( stats_fpu_resv[i][ fpu_n[i] ]++ );
    STT( pstats_fpu_resv[i] += fpu_n[i] );
    }
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with fpu stage\n") );

/*****************************************************************************
  FPU1 EXECUTE
*****************************************************************************/
#if( CLUSTER_FPU )
// begin_fpu1_execute_stage:;
{
    #if( STATS )
    unsigned	tmp_execd[MAX_THREADS] = {0};
    #endif

#include "stage_fpu1_exec.macros"

end_fpu1_execute_stage:;
    for (i = 0; i < n_threads; i++)
    {
    STT( stats_total_execd[i] += tmp_execd[i] );
    STT( assert( tmp_execd[i] < FPU_NUM + 1 ));
    STT( stats_fpu1_execd[i][ tmp_execd[i]]++ );
    STT( pstats_fpu1_execd[i] += tmp_execd[i]);
    assert( fpu1_n[i] < FPU_RESV + 1 );
    STT( stats_fpu1_resv[i][ fpu1_n[i] ]++ );
    STT( pstats_fpu1_resv[i] += fpu1_n[i] );
    }
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with fpu1 stage\n") );
#endif


/*****************************************************************************
  DCACHE MISS EXECUTE
*****************************************************************************/
  dcache_ports = DCACHE_PORTS;

#if( !INFINITE_DCACHE )
// begin_dmiss_stage:;
{
// pthread_mutex_unlock(&mutex1);
#include "stage_dmiss_exec.macros"
end_dmiss_stage:;

    for (i = 0; i < n_threads; i++)
    {
    assert( dmiss_n[i] < DMISS_MAX + MEM_NUM );
    STT( stats_dmissq[i][ dmiss_n[i] ]++ );
    STT( pstats_dmissq[i] += dmiss_n[i] );
    STT( stats_l2dmissq[i][ stats_l2_dmiss_n[i] ]++ );
    #if( DMISS_COUNT_LINES )
      STT( stats_lines[i][stats_lines_n[i]]++ );
      STT( stats_l2_lines[i][stats_l2_lines_n[i]]++ );
    #endif
    }
}/* end dmiss stage */
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with dmiss stage\n") );
#endif /* #if( !INFINITE_DCACHE ) */

/*****************************************************************************
  DCASTOUT EXECUTE
*****************************************************************************/
#if( !INFINITE_DCACHE )
// begin_dcastout_stage:;
{
    if( is_ok_dcastout_bus() && dcastout_n != 0) {
    	initiated_dcastout_bus();
	    dcastout_n--;
    }
// end_dcastout_stage:;

    STT( stats_dcastoutq[ dcastout_n ]++ );
}
#endif

  #if( DCACHE_IS_INTERLEAVED )
    dinter_n = 0;
  #endif

/*****************************************************************************
  MEM EXECUTE
*****************************************************************************/
// begin_mem_execute_stage:;
{
    unsigned tmp_execd[MAX_THREADS] = {0};
#include "stage_mem_exec.macros"
end_mem_execute_stage:;

    for (i = 0; i < n_threads; i++)
    {
    STT( stats_total_execd[i] += tmp_execd[i] );
    assert( tmp_execd[i] < MEM_NUM + 1 );
    STT( stats_mem_execd[i][ tmp_execd[i]]++ );
    STT( pstats_mem_execd[i] += tmp_execd[i] );
    assert( mem_n[i] < MEM_RESV + 1 );
    STT( stats_mem_resv[i][ mem_n[i]]++ );
    STT( pstats_mem_resv[i] += mem_n[i] );
    }/* for i */
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with mem exec stage\n") );

/*****************************************************************************
  MEM1 EXECUTE
*****************************************************************************/
#if( CLUSTER )
// begin_mem1_execute_stage:;
{
    unsigned tmp_execd[MAX_THREADS] = {0};
#include "stage_mem1_exec.macros"

end_mem1_execute_stage:;
    for (i = 0; i < n_threads; i++)
    {
    STT( stats_total_execd[i] += tmp_execd[i] );
    assert( tmp_execd[i] < MEM_NUM + 1 );
    STT( stats_mem1_execd[i][ tmp_execd[i]]++ );
    STT( pstats_mem1_execd[i] += tmp_execd[i] );
    assert( mem1_n[i] < MEM_RESV + 1 );
    STT( stats_mem1_resv[i][ mem1_n[i]]++ );
    STT( pstats_mem1_resv[i] += mem1_n[i] );
    }/* for i */
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with mem exec stage\n") );
#endif

assert( size_storeq(thread_exec) < STOREQ_SIZE + 1 );
STT( stats_storeq[thread_exec][ size_storeq(thread_exec) ]++ );
STT( pstats_storeq[thread_exec] += size_storeq(thread_exec) );
assert( size_reorderq(thread_exec) < REORDERQ_SIZE + 1);
STT( stats_reorderq[thread_exec][ size_reorderq(thread_exec) ]++ );
STT( pstats_reorderq[thread_exec] += size_reorderq(thread_exec) );
  
#if( USING_LOG_UNIT )
/*****************************************************************************
  LOG EXECUTE
*****************************************************************************/
// begin_log_execute_stage:;
{
    #if( STATS )
    unsigned	tmp_execd[MAX_THREADS] = {0};
    #endif

#include "stage_log_exec.macros"
end_log_execute_stage:;

	for (i = 0; i < n_threads; i++) {
		STT( stats_total_execd[i] += tmp_execd[i] );
		STT( assert( tmp_execd[i] < LOG_NUM + 1 ));
		STT( stats_log_execd[i][ tmp_execd[i]]++ );
		assert( log_n[i] < LOG_RESV + 1 );
		STT( stats_log_resv[i][ log_n[i]]++ );
	}/* for i */
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with log exec stage\n") );
#endif /* USING_LOG_UNIT */

/*****************************************************************************
  BR EXECUTE
*****************************************************************************/
// begin_br_execute_stage:;
{
    #if( STATS )
    unsigned	tmp_execd[MAX_THREADS] = {0};
    #endif

#include "stage_br_exec.macros"
end_br_execute_stage:;

	for (i = 0; i < n_threads; i++) {
		STT( stats_total_execd[i] += tmp_execd[i] );
		STT( assert( tmp_execd[i] < BR_NUM + 1 ));
		STT( stats_br_execd[i][ tmp_execd[i]]++ );
		assert( br_n[i] < BR_RESV + 1 );
		STT( stats_br_resv[i][ br_n[i]]++ );
		STT( pstats_br_resv[i] += br_n[i] );
	}/* for i */
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with br exec stage\n") );

  #if( CLUSTER || CLUSTER_FPU )
    assert( stats_total_execd[thread_exec] < (FIX_NUM+FPU_NUM+MEM_NUM+BR_NUM+LOG_NUM+CMPLX_NUM)*CLUSTERS_MAX+1);
    STT( stats_execd[thread_exec][stats_total_execd[thread_exec]]++ );
    assert( stats_total_execd[thread_exec] < (FIX_RESV+FPU_RESV+MEM_RESV+BR_RESV+LOG_RESV+CMPLX_RESV)*CLUSTERS_MAX+1);
    STT( stats_resv[thread_exec][fix_n[thread_exec] + fix1_n[thread_exec] + fpu_n[thread_exec] + fpu1_n[thread_exec] + br_n[thread_exec] + log_n[thread_exec] + cmplx_n[thread_exec] + mem_n[thread_exec] + mem1_n[thread_exec] ]++ );
  #else
    assert( stats_total_execd < (FIX_NUM+FPU_NUM+MEM_NUM+BR_NUM+LOG_NUM+CMPLX_NUM)+1);
    STT( stats_execd[thread_exec][stats_total_execd[thread_exec]]++ );
    assert( stats_total_execd < (FIX_RESV+FPU_RESV+MEM_RESV+BR_RESV+LOG_RESV+CMPLX_RESV)+1);
    STT( stats_resv[thread_exec][fix_n[thread_exec] + fpu_n[thread_exec] + br_n[thread_exec]  + log_n[thread_exec] + cmplx_n[thread_exec] + mem_n[thread_exec] ]++ );
  #endif

  insns_total_execd[thread_exec] += stats_total_execd[thread_exec];

#if( NO_DECODE_STAGE )
/*****************************************************************************
  DECODE STAGE: if there is none, i.e. is combined with the rename stage
*****************************************************************************/
#include "stage_decode.macros"
#endif

/*****************************************************************************
  RENAME STAGE
*****************************************************************************/
// begin_rename_stage:;
{
    #if( STATS )
    unsigned	tmp_renamed = 0;
    #endif

#include "stage_rename.macros"
end_rename_stage:;

    STT( assert( tmp_renamed < IBUF_MAX + 1 ));
    STT( stats_renamed[thread_exec][ tmp_renamed ]++ );
    {
      int	n_;

// LYM
      int       thread_idx;

      for (thread_idx = 0; thread_idx < n_threads; thread_idx++){
          n_ = (gpr_phys) - (freeq_tl[thread_idx][GPR_CLASS] - GPR_CLASS*PHYS_MAX);
          assert( n_ < gpr_phys + 1 );
          STT( stats_gpr[thread_exec][n_]++ );
          STT( pstats_gpr[thread_exec] += n_);
#if(DYNAMIC_PROFILING)
          comm_data->fxu_accesses[thread_exec] += n_;
#endif // jdonald 4-11-2005
          n_ = (fpr_phys) - (freeq_tl[thread_idx][FPR_CLASS] - FPR_CLASS*PHYS_MAX);
          assert( n_ < fpr_phys + 1 );
          STT( stats_fpr[thread_exec][n_]++ );
          STT( pstats_fpr[thread_exec] += n_);
#if(DYNAMIC_PROFILING)
          comm_data->fpu_accesses[thread_exec] += n_;
#endif /* jdonald 4-11-2005 */

          n_ = (spr_phys) - (freeq_tl[thread_idx][SPR_CLASS] - SPR_CLASS*PHYS_MAX);
          assert( n_ < spr_phys + 1 );
          STT( stats_spr[thread_exec][n_]++ );
          STT( pstats_spr[thread_exec] += n_);

          n_ = (ccr_phys) - (freeq_tl[thread_idx][CCR_CLASS] - CCR_CLASS*PHYS_MAX);
          assert( n_ < ccr_phys + 1 );
          STT( stats_ccr[thread_exec][n_]++ );
          STT( pstats_ccr[thread_exec] += n_);

      }

/*      n_ = (GPR_PHYS_MAX -GPR_ARCH) - (freeq_tl[thread_rename][GPR_CLASS] - GPR_CLASS*PHYS_MAX);
      assert( n_ < GPR_PHYS_MAX + 1 );
      STT( stats_gpr[thread_exec][n_]++ );
      STT( pstats_gpr[thread_exec] += n_);
      n_ = (FPR_PHYS_MAX -FPR_ARCH) - (freeq_tl[thread_rename][FPR_CLASS] - FPR_CLASS*PHYS_MAX);
      assert( n_ < FPR_PHYS_MAX + 1 );
      STT( stats_fpr[thread_exec][n_]++ );
      STT( pstats_fpr[thread_exec] += n_);
      n_ = (SPR_PHYS_MAX -SPR_ARCH) - (freeq_tl[thread_rename][SPR_CLASS] - SPR_CLASS*PHYS_MAX);
      assert( n_ < SPR_PHYS_MAX + 1 );
      STT( stats_spr[thread_exec][n_]++ );
      STT( pstats_spr[thread_exec] += n_);
      n_ = (CCR_PHYS_MAX -CCR_ARCH) - (freeq_tl[thread_rename][CCR_CLASS] - CCR_CLASS*PHYS_MAX);
      assert( n_ < CCR_PHYS_MAX + 1 );
      STT( stats_ccr[thread_exec][n_]++ );
      STT( pstats_ccr[thread_exec] += n_); */
    }
    assert( size_storeq(thread_rename) < STOREQ_SIZE + 1 );
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with rename stage\n") );

#if( !USING_BYPASS_DISPATCH )
/*****************************************************************************
  DISPATCH STAGE
*****************************************************************************/
// begin_dispatch_stage:;
{
    unsigned	tmp_diff = 0;
    unsigned	tmp_fix_diff = 0;
    unsigned	tmp_fix1_diff = 0;
    unsigned	tmp_fpu_diff = 0;
    unsigned	tmp_fpu1_diff = 0;
    unsigned	tmp_mem_diff = 0;
    unsigned	tmp_mem1_diff = 0;
    unsigned	tmp_log_diff = 0;
    // unsigned	tmp_cmplx_diff = 0; // unused?
    unsigned	tmp_br_diff = 0;
    unsigned	tmp_execd = 0;

#include "stage_dispatch.macros"

end_dispatch_stage:;
    assert( tmp_execd < DISPATCH_MAX + 1 );
    STT( stats_fix_dispatched[thread_dispatch][ tmp_fix_diff ]++);
    #if( CLUSTER && CLUSTERS_MAX > 1 )
      STT( stats_fix1_dispatched[thread_dispatch][ tmp_fix1_diff ]++);
    #endif
    STT( stats_fpu_dispatched[thread_dispatch][ tmp_fpu_diff ]++);
    #if( CLUSTER && CLUSTERS_MAX > 1 )
      STT( stats_fpu1_dispatched[thread_dispatch][ tmp_fpu1_diff ]++);
    #endif
    STT( stats_mem_dispatched[thread_dispatch][ tmp_mem_diff ]++);
    #if( CLUSTER && CLUSTERS_MAX > 1 )
      STT( stats_mem1_dispatched[thread_dispatch][ tmp_mem1_diff ]++);
    #endif
    #if( USING_LOG_UNIT )
      STT( stats_log_dispatched[thread_dispatch][ tmp_log_diff ]++);
    #endif
    #if( USING_CMPLX_UNIT )
      STT( stats_cmplx_dispatched[thread_dispatch][ tmp_cmplx_diff ]++);
    #endif
    STT( stats_br_dispatched[thread_dispatch][ tmp_br_diff ]++);
    STT( stats_dispatched[thread_dispatch][ tmp_execd ]++ );
    assert( size_storeq(thread_dispatch) < STOREQ_SIZE + 1 );
    insns_total_dispd[thread_dispatch] += tmp_execd;
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with dispatch stage\n") );
#endif

#if( !NO_DECODE_STAGE )
/*****************************************************************************
  DECODE STAGES; if they are separate, of course
*****************************************************************************/
#include "stage_decode.macros"
#endif
    assert( FCHBLK_MAX ==0 || fchblk_left[thread_decode] > 0 || iq_dec[thread_decode] != iq_tl[thread_decode] );
  
/*****************************************************************************
  IFETCH STAGE
*****************************************************************************/
// begin_ifetch_stage:;
{
#if( STATS )
    unsigned tmp_fetched = 0;
#endif

#include "stage_ifetch.macros"
end_ifetch_stage:;

    assert( FCHBLK_MAX ==0 || fchblk_left[thread_ifetch] > 0 || iq_dec[thread_ifetch] != iq_tl[thread_ifetch] );
    STT( stats_fetched[thread_ifetch][bounds( tmp_fetched, IFETCH_MAX + 1) ]++ );
    STT( stats_ibuf[thread_ifetch][bounds(IBUF_MAX-ibuf_left[thread_ifetch],IBUF_MAX + 1)]++ );
    STT( stats_fchblk[thread_ifetch][bounds(FCHBLK_MAX-fchblk_left[thread_ifetch],FCHBLK_MAX + 1)]++ );
}
  AFTER_EACH_STAGE();
  DBG( fprintf(the_stderr, "done with ifetch stage\n") );

  STT( stats_inflight[thread_ifetch][bounds(iq_n[thread_ifetch], INFLIGHT_MAX + 1)]++ );
  #if( RETIRE_GROUPS && DISPATCH_GROUPS )
  STT( stats_groups[thread_ifetch][bounds(GROUPS_MAX - groups_left[thread_ifetch], GROUPS_MAX+1)]++ );
  STT( pstats_groups[thread_ifetch] += bounds(GROUPS_MAX - groups_left[thread_ifetch], GROUPS_MAX+1));
  #endif

/*****************************************************************************
  FLUSH STAGE
  1. jump here from retire stage, due to reorder conflict, see stage_retire.macros
*****************************************************************************/
/*
  Alper's bug fix, best for two threads
  thread_flush = do_flush[1]*thread_ifetch+ ((do_flush[0]+1)%2)*thread_ifetch+((do_flush[0]+1)%2)*do_flush[1];
*/

restart_after_reset:;

  AFTER_EACH_STAGE();

  for (thread_flush = 0; thread_flush < n_threads; thread_flush++)
  {
// begin_flush_stage:
{
    unsigned	tmp_at;
    unsigned	tmp_reissue;

    /* any flush work to do? do_flush is set when a memory reorder conflict is detected */
    if( do_flush[thread_flush] )
    {
      STT( stats_reorderq_conflict[thread_flush]++ );
      do_flush[thread_flush] = 0;

      tmp_reissue = 2;

      #if( RETIRE_GROUPS )
	if( iq_flush[thread_flush] != iq_retire[thread_flush] )
	{
	  for( tmp_at = (iq_flush[thread_flush]+IQ_SIZE-1)%IQ_SIZE;
		tmp_at != iq_retire[thread_flush] && !xiq_is_group(iq[thread_flush][tmp_at]);
		tmp_at = (tmp_at +IQ_SIZE-1)%IQ_SIZE )
	  {
	    tmp_reissue++;
	  }

	  if( tmp_at == iq_retire[thread_flush] )
	  {
	    iq_flush[thread_flush] = iq_retire[thread_flush];
	  }
	  else
	  {
	    iq_flush[thread_flush] = (tmp_at + 1) %IQ_SIZE;
	  }
	}
      #endif

      if( iq_flush[thread_flush] != iq_retire[thread_flush] )
      {
	for( tmp_at = iq_flush[thread_flush];
	      tmp_at != iq_retire[thread_flush] && !xiq_is_first(iq[thread_flush][tmp_at]);
	      tmp_at = (tmp_at +IQ_SIZE-1)%IQ_SIZE )
	{
	  tmp_reissue++;
	}
	iq_flush[thread_flush] = tmp_at;
      }

#include "flush_arbitrary.macros"
#if( RETIRE_GROUPS )
	 decode_reissue_groups[thread_flush] = tmp_reissue;
#endif

      ifetch_stalled_till[thread_flush] = cycle + 7;
    }/* if( do_flush[thread_flush] )  */
// end_flush_stage:;
}
}/* for thread_flush */
  AFTER_EACH_STAGE();

/*****************************************************************************
  TRACING CODE
*****************************************************************************/

#if( TRACE )
// begin_trace:
  #include "trace.macros"
  end_trace:;
#endif
// pthread_mutex_unlock(&mutex1);

  cycle++;
#if( !(FIXED_FREQUENCY) )
  timestamp += 1e-9 / frequency_scale[processor_id] / REAL_FREQ;
  // timestamp += 1.0 / frequency_scale[processor_id];
#endif // !FIXED_FREQUENCY

#if(COARSE_STATS) || (DYNAMIC_PROFILING)
// jdonald 5-18-2005
  comm_data->cyclez++;
#endif // COARSE_STATS

	pthread_mutex_lock( &l2sync_cond_mutex );
	l2_sync[processor_id] = CYCLE_OR_TIMESTAMP;
	pthread_cond_broadcast( &l2sync_cond ); // in case anybody was waiting on this
	pthread_mutex_unlock( &l2sync_cond_mutex ); // will later make this an infrequent method

  /* set a stop point for debugging */
  if (cycle == cycle_stop)
  {
  	cycle_stop = 0;
  }


  if( (cycle % MONITOR_CYCLES) == 0 || (detail_step && cycle > detail_from && cycle < detail_to && cycle % detail_step == 0)) {
  /*
    fprintf(the_stderr, "@@ %10u %10u %10u %10u %10u %10u %10u [%4u %4u]\n",
	cycle, insns,
	insns - stats_nexpands + stats_expands_to, insns_retired,
	bp_bc_miss+bp_bclr_miss+bp_bcctr_miss,
	stats_reorderq_conflict, stats_d1lines, iq_hd, iq_tl );
    fflush( the_stderr );
  */

    fputs("@@ ", the_stderr);
    fflush( the_stderr);
    fprint_unsigned(the_stderr, 10, cycle - saved_cycle_reset);
    fputc(' ', the_stderr);
    fprint_unsigned(the_stderr, 10, insns[thread_trace]);
    fputc(' ', the_stderr);
    fprint_unsigned(the_stderr, 10, insns[thread_trace]-stats_nexpands[thread_trace]+stats_expands_to[thread_trace]);
    fputc(' ', the_stderr);
    fprint_unsigned(the_stderr, 10, insns_retired[thread_trace]);
    fputc(' ', the_stderr);
    fprint_unsigned(the_stderr, 10, bp_bc_miss[thread_flush]+bp_bclr_miss[thread_flush]+bp_bcctr_miss[thread_flush]);
    fputc(' ', the_stderr);
#if(STATS)
    fprint_unsigned(the_stderr, 10, stats_reorderq_conflict[thread_trace]);
    fputc(' ', the_stderr);
#endif
    fprint_unsigned(the_stderr, 10, stats_d1lines[thread_trace]);
    fputc(' ', the_stderr);
    fprint_unsigned(the_stderr, 4, iq_hd[thread_trace]);
    fputc(' ', the_stderr);
    fprint_unsigned(the_stderr, 4, iq_tl[thread_trace]);
    fputc('\n', the_stderr);
    fflush( the_stderr);

  } /* if( (cycle % MONITOR_CYCLES) == 0 ) */

	  if (use_insn_cutoff) {
      	for (i = 0; i < n_threads; i += 1)  { // group_cutoff check for threads within cores
	      	if (group_cutoff && insns_retired[i] < cutoff) // if group_cutoff, give up if anyone misses the threshold
      			break;
      		if (!group_cutoff && insns_retired[i] >= cutoff) // if !group_cutoff, signal if anyone breaks the threshold
      			break;
  		}
  	  }
  	  if (!goal_attained[processor_id] && cycle_reset == 0 &&
  	  	((use_insn_cutoff && group_cutoff && i == n_threads) ||
  	  	 (use_insn_cutoff && !group_cutoff && i != n_threads) ||
		 (!use_insn_cutoff && cycle >= cutoff )))
	  {
	    fprintf(the_stderr, "goal attained at %llu\n", cycle );
	    goal_attained[processor_id] = TRUE;
  	  }
  } // if (!finished_sync[processor_id])
	if (
#if ( FIXED_FREQUENCY )
			( cycle != 0 && (cycle % POWER_CYCLES) == 0)
#else // !FIXED_FREQUENCY
			( timestamp >= next_tempsync )
#endif
		  || finished_sync[processor_id] ) {
		pthread_mutex_lock( &tempsync_cond_mutex ); // power.macros needs to be locked since it increments a shared total

		if (cycle_reset != 0 && cycle >= cycle_reset) {
			reset_counter = processor_num;
			saved_cycle_reset = cycle; // since it'll get rounded up to the nearest POWER_CYCLES anyway
#if ( !( FIXED_FREQUENCY) )
			saved_timestamp_reset = timestamp;
#endif
			cycle_reset = 0;
			fprintf(parent_stderr, RED "stats reset signaled by core %d at cycle %lld" NORMAL "\n", processor_id, cycle);
		}

#if ( POWER_MODE )
    	#include "stats.power.macros"
		#include "power.macros"
#endif /* #if( POWER_MODE ) */
		global_cycle[processor_id] = cycle;
		for (j = 0; j < n_threads; j += 1) {
			global_insns_retired[processor_id][j] = insns_retired[j];
		}

		tempsync_ready_flags[processor_id] = TRUE;
		pthread_cond_broadcast( &tempsync_cond );
		
		while (!tempsync_ready()) {
			pthread_cond_wait( &tempsync_cond, &tempsync_cond_mutex );
			if (!tempsync_ready_flags[processor_id]) // someone cleared us out, so let's go back to normal
				break;
		}
		if (tempsync_ready()) { // nobody cleared us or anyone else, so let's assume responsibility for the processing
			int i;
			#include "temp_inner.sync.macros" // also does L2 power calculations

			for (i = 0; i < processor_num; i += 1)
				tempsync_ready_flags[i] = FALSE; // let everyone know that we're done so they don't repeat the work
#if ( !(FIXED_FREQUENCY) )
			next_tempsync += POWER_INTERVAL; // should be fine if someobody doesn't shrink POWER_CYCLES down to 1...
			// next_tempsync += POWER_CYCLES; // should be fine if someobody doesn't shrink POWER_CYCLES down to 1...
#endif
			for (i = 0; i < processor_num; i += 1) { // group cutoff check across all cores
				if (group_cutoff && !goal_attained[i])
					break;
				if (!group_cutoff && goal_attained[i])
					break;
			}
			if ((group_cutoff && i == processor_num) || // all cores attained goal
				(!group_cutoff && i != processor_num)) // at least one core attained goal
					end_simulation = TRUE;
			if (reset_counter > 0) { // in the rare event, end_simulation has priority over cycle_reset
				#include "reset_stats.macros"
			}
			pthread_cond_broadcast( &tempsync_cond );
		}
		if (reset_counter > 0) {
			#include "reset_stats.core.macros"
			reset_counter -= 1;
		}
		pthread_mutex_unlock( &tempsync_cond_mutex );
		if (end_simulation)
			break;
	}
  } /* while loop */

// apparently there is no way to get to the following few lines
#if ( TEMPERATURE && POWER_MODE )
	if (thecutoff == 0) // why?
		#include "power.status.macros"
#endif

// end_simulation:;

	// fprintf(stderr, "*** end_simulation reached for core #%d at cycle %lld, insns_retired[thread_trace] = %d\n", processor_id, cycle, insns_retired[thread_trace]);
	pthread_mutex_lock( &tempsync_cond_mutex ); // just in case, power.dump.macros could be dangerous

/*****************************************************************************
  FINAL CODE
*****************************************************************************/
  {
    unsigned tmp_total;
    unsigned tmp_cumm;
    unsigned tmp_insns;

#if(POSTPROCESS)
	if (thecutoff == 0)	{
		printf("!!!%s\n",output_fname[processor_id]);
		input_fp = fopen(output_fname[processor_id],"r");
	}
	if (thecutoff != 0) {
		output_fp = fopen(output_fname[processor_id],"w");
	}

	if (thecutoff == 0){
		#include "power.read.macros"
	}
	if (thecutoff != 0){
		#include "power.dump.macros"
	}
#endif // POSTPROCESS

    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
    fprintf(the_stderr, "@@ DEPPREP PARAMETERS DATA\n" );
    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );

#include "dep.macros"

    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
    fprintf(the_stderr, "@@ PARAMETERS DATA\n" );
    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );

#include "params.macros"

    for (thread_trace = 0; thread_trace < n_threads; thread_trace++)
    {
    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
    fprintf(the_stderr, "@@ SUMMARY DATA\n" );
    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
    fprintf(the_stderr, "@@ totals: cycle=%lld insns=%u memops=%u retired=%u totalret=%u totalexecd=%u totaldispd=%u\n",
		thread_cycles[thread_trace], insns[thread_trace], memop[thread_trace], insns_retired[thread_trace],
		insns_total_retired[thread_trace], insns_total_execd[thread_trace], insns_total_dispd[thread_trace] );

    fprintf(the_stderr, "@@ icache: probes=%lld l1miss=%lld prehit=%u l15miss=%u l2miss=%u tlb1=%lld tlb2=%lld\n",
		stats_iprobe[thread_trace], stats_imiss[thread_trace], stats_iprefetch_hits, stats_i15miss[thread_trace],
		stats_il2miss[thread_trace], stats_itmiss[thread_trace], stats_i2tmiss[thread_trace]);

    fprintf(the_stderr, "@@ dcache: probes=%lld l1lines_started=%u l1lines_fetched=%u l2lines=%u l1miss_all=%u lmiss_taken=%lld l2miss_all=%u l2miss_taken=%u trail=%u cast=%u\n",
		stats_dprobe[thread_trace], stats_dlines[thread_trace], stats_d1lines[thread_trace],
		stats_d2lines[thread_trace],
		stats_dmiss_taken[thread_trace], stats_dmiss[thread_trace],
		stats_dl2miss_taken[thread_trace], stats_dl2miss[thread_trace],
		stats_dtrail[thread_trace], stats_dcast[thread_trace]); /* ??? LYM I think dmiss_taken and dmiss should be exchanged */

    fprintf(the_stderr, "@@ dcache_store: probes=%lld l1lines_started=%u l1lines_fetched=%u l2lines=%u l1miss_all=%u lmiss_taken=%u l2miss_all=%u l2miss_taken=%u trail=%u cast=%u\n",
		stats_dprobe[thread_trace], stats_dlines[thread_trace], stats_store_d1lines[thread_trace],
		stats_d2lines[thread_trace],
		stats_store_dmiss_taken[thread_trace], stats_store_dmiss[thread_trace],
		stats_dl2miss_taken[thread_trace], stats_dl2miss[thread_trace],
		stats_dtrail[thread_trace], stats_dcast[thread_trace]);

    fprintf(the_stderr, "@@ dtlb: tlb1_all=%u tlb1_taken=%lld tlb2_all=%u tlb2_taken=%lld\n",
		stats_dtmiss_taken[thread_trace], stats_dtmiss[thread_trace],
		stats_d2tmiss_taken[thread_trace], stats_d2tmiss[thread_trace]);

    fprintf(the_stderr, "@@ interleave: conflicts=%u samebank=%u\n",
		stats_dinter_conflicts, stats_dinter_equals);
    #if( STAT0 )
      fprintf(the_stderr, "@@ ifetch0: tlb=%u cache=%u nfa=%u mispredict=%u ibuf=%u fchblk=%u sync=%u max_pred=%u partial=%u inflight=%u end=%u\n",
	stat0_ifetch_tlb[thread_trace],
	stat0_ifetch_cache[thread_trace],
	stat0_ifetch_nfa[thread_trace],
	stat0_ifetch_mispredict[thread_trace],
	stat0_ifetch_ibuf[thread_trace],
	stat0_ifetch_fchblk[thread_trace],
	stat0_ifetch_sync[thread_trace],
	stat0_ifetch_pred[thread_trace],
	stat0_ifetch_partial[thread_trace],
	stat0_ifetch_inflight[thread_trace],
	stat0_ifetch_end[thread_trace]);

      fprintf(the_stderr, "@@ decode0: empty=%lld stall=%u multi=%u string=%u\n",
	stat0_decode_empty[thread_trace],
	stat0_decode_stall[thread_trace],
	stat0_decode_multi[thread_trace],
	stat0_decode_string[thread_trace]);

      fprintf(the_stderr, "@@ dispatch0: groups=%u\n",
	stat0_dispatch_groups[thread_trace]);
    #endif

    fprintf(the_stderr, "@@ prefetch: hits=%u total=%lld\n", stats_iprefetch_hits, stats_iprefetch);
    fprintf(the_stderr, "@@ expands: insns=%u total=%u\n", stats_nexpands[thread_trace], stats_expands_to[thread_trace]);
    fprintf(the_stderr, "@@ strings: insns=%u total=%u stalls=%u\n", stats_strings_total[thread_trace], stats_strings_expands[thread_trace], stats_strings_stalls[thread_trace]);
    fprintf(the_stderr, "@@ nfa: link=%lld nfa=%u mis=%u mispred=%u\n", bp_bclr_total[thread_trace], nfa_total[thread_trace], nfa_miss[thread_trace], nfa_miss_buried[thread_trace]);

    #if( 1 || !PERFECT_BRANCH_PREDICTION )
    fprintf(the_stderr, "@@ branch: total=%lld stalls=%u full=%u cond_mis=%lld cond_tot=%u link_mis=%u link_tot=%lld ctr_mis=%u ctr_tot=%lld\n",
	  branch_total[thread_trace], branch_stalls[thread_trace], branch_pred_stalls[thread_trace],
	  bp_bc_miss[thread_trace], bp_bc_total[thread_trace],
	  bp_bclr_miss[thread_trace], bp_bclr_total[thread_trace], bp_bcctr_miss[thread_trace], bp_bcctr_total[thread_trace]);
    fprintf(the_stderr, "@@ actual_branch: total=%lld stalls=%u full=%u cond_mis=%lld cond_tot=%u link_mis=%u link_tot=%lld ctr_mis=%u ctr_tot=%lld\n",
	  branch_total[thread_trace]-branch_miss_total[thread_trace], branch_stalls[thread_trace], branch_pred_stalls[thread_trace],
	  bp_bc_miss[thread_trace]-bp_bc_miss_miss[thread_trace], bp_bc_total[thread_trace]- bp_bc_miss_total[thread_trace],
	  bp_bclr_miss[thread_trace]-bp_bclr_miss_miss[thread_trace], bp_bclr_total[thread_trace] - bp_bclr_miss_total[thread_trace],
	  bp_bcctr_miss[thread_trace]-bp_bcctr_miss_miss[thread_trace], bp_bcctr_total[thread_trace] - bp_bcctr_miss_total[thread_trace]);
	fprintf(the_stderr, "@@ exec_branch: total=%u stalls=%u full=%u cond_mis=%u cond_tot=%u link_mis=%u link_tot=%lld ctr_mis=%u ctr_tot=%lld\n",
	  bp_exec_total[thread_trace], bp_exec_pred[thread_trace], bp_exec_miss[thread_trace],
	  bp_exec_bc_miss[thread_trace], bp_exec_bc_total[thread_trace],
	  bp_exec_bclr_miss[thread_trace], bp_exec_bclr_total[thread_trace],
	  bp_exec_bcctr_miss[thread_trace], bp_exec_bcctr_total[thread_trace]);
    #endif

    #if( USE_PREDICATION )
    fprintf(the_stderr, "@@ predication: skips=%u taken=%u exps=%u ops=%u\n",
	stats_pred_branch[thread_trace], stats_pred_taken[thread_trace], stats_pred_exps[thread_trace], stats_pred_ops[thread_trace]);
    #endif

    #if( PREDICTION_IGNORES_BDCTR )
    fprintf(the_stderr, "@@ bdctr: mispredict=%u total=%u\n",
		stats_bdctr_miss[thread_trace], stats_bdctr[thread_trace]);
    #endif

    #if( EVALUATE_MISPREDICTED_BRANCHES )
    fprintf(the_stderr, "@@ prediction: early=%u mispredict=%lld total=%lld\n",
		stats_eval_mispredict, bp_bc_miss[thread_trace]+bp_bclr_miss[thread_trace]+bp_bcctr_miss[thread_trace],
		bp_bc_total[thread_trace]+bp_bclr_total[thread_trace]+bp_bcctr_total[thread_trace] );
    #endif

#if(STATS)
    fprintf(the_stderr, "@@ gp: flushes=%u reissue=%u recirc=%u groups=%lld\n",
	stats_reissue_flushes[thread_trace], stats_reissue_decode[thread_trace], stats_recirc[thread_trace], stats_groups_dispatched[thread_trace]);

    #if( GET_SPEC_STATS )
      fprintf(the_stderr,
	"@@ spec: fetch_insn=%u fetch=%u expand_insns=%u expand=%u dispatch=%u exec_br=%u exec_fix=%u exec_fpu=%u exec_mem=%u\n",
	stats_spec_ifetch[thread_trace], stats_spec_ifetch[thread_trace] - stats_spec_nexpands[thread_trace] + stats_spec_expands_to[thread_trace],
	stats_spec_nexpands[thread_trace], stats_spec_expands_to[thread_trace],
	stats_spec_dispatch[thread_trace],
	stats_spec_br_execd[thread_trace], stats_spec_fix_execd[thread_trace], stats_spec_fpu_execd[thread_trace], stats_spec_mem_execd[thread_trace]);

      fprintf(the_stderr,
	"@@ specdcache: spec_miss=%u tot_miss%lld lines=%u flush=%u satisfy=%u hits=%u miss=%u src=%u rehit=%u once=%u\n",
	stats_spec_dmiss[thread_trace], stats_dmiss[thread_trace],
	stats_spec_dlines[thread_trace], stats_spec_dflush[thread_trace], stats_spec_dsatisfy[thread_trace],
	stats_spec_extra_hits[thread_trace], stats_spec_extra_misses[thread_trace], stats_spec_src_hits[thread_trace],
	stats_spec_reexec_hits[thread_trace], stats_spec_used_once[thread_trace]);

      fprintf(the_stderr,
	"@@ specicache: spec_miss=%u tot_miss=%lld hits=%u miss=%u src=%u once=%u\n",
	stats_spec_imiss[thread_trace], stats_imiss[thread_trace],
	stats_spec_extra_ihits[thread_trace], stats_spec_extra_imisses[thread_trace], stats_spec_src_ihits[thread_trace], stats_spec_iused_once[thread_trace]
	);
    #endif

      fprintf(the_stderr, "@@ stalls: ibuf=%u inflight=%u dmissq=%u cast=%u storeq=%u reoderq=%u resv=%u rename=%u\n",
		  stats_ibuf_stall[thread_trace], stats_inflight_stall[thread_trace],
		  stats_dmiss_stall, stats_dcastout_stall,
		  stats_storeq_stall[thread_trace], stats_reorderq_stall[thread_trace],
		  stats_resv_full_stall[thread_trace],
		  stats_rename_stall[thread_trace]);
      fprintf(the_stderr, "@@ storeq: total=%u forward=%lld partial=%u data_wait=%u\n",
	stats_storeq_total[thread_trace], stats_storeq_forward[thread_trace],
	stats_storeq_partials[thread_trace], stats_storeq_data_not_ready[thread_trace]);
      fprintf(the_stderr, "@@ reorderq: total=%u conflicts=%u\n",
	stats_reorderq_total[thread_trace], stats_reorderq_conflict[thread_trace]);
      fprintf(the_stderr, "@@ mtsr: total=%u ipurge=%u dpurge=%u\n",
	stats_mtsr[thread_trace], stats_mtsr_ipurge[thread_trace], stats_mtsr_dpurge[thread_trace]);
      fprintf(the_stderr, "@@ nops: dispd=%u retd=%u\n",
	stats_nops_dispd[thread_trace], stats_nops_retd[thread_trace] );
      /**** POWERTIMER STATS ADDED BY DBROOKS */
      fprintf(the_stderr, "@@ extra powertimer stats: num_stores=%lld num_loads=%lld\n",
      	stats_num_stores[thread_trace],stats_num_loads[thread_trace]);
      /* track usage of decode information storage, zhigangh@1/12/2003 */
      fprintf(the_stderr, "@@ trans_deps_top=%u trans_overs_top=%u trans_ptrs_top=%u\n",
      	trans_deps_top, trans_overs_top, trans_ptrs_top);
#include "stats.power.macros"

#endif /* #if(STATS) */


    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
    fprintf(the_stderr, "@@ HISTOGRAMS DATA\n" );
    fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );

    #if( GET_SPEC_STATS )
      dumpStats( cycle, DMISS_MAX+MEM_NUM, stats_spec_dflush_histo, "spec_flush");
      dumpStats( cycle, DMISS_MAX+MEM_NUM, stats_spec_over_histo, "spec_over");
      dumpStats( cycle, DMISS_MAX+MEM_NUM, stats_spec_dflush_over_histo, "spec_flushover");
    #endif

    #if( STATS )
      dumpStats( cycle, IBUF_MAX, stats_ibuf[thread_trace], "ibuf");
      dumpStats( cycle, FCHBLK_MAX, stats_fchblk[thread_trace], "fchblk");
      dumpStats( cycle, RETIREQ_MAX, stats_retireq[thread_trace], "retireq");

      dumpStats( cycle, FIX_RESV, stats_fix_resv[thread_trace], "fix_rsv");
      #if( CLUSTER )
	dumpStats( cycle, FIX_RESV, stats_fix1_resv[thread_trace], "fix1_rsv");
      #endif
      dumpStats( cycle, FPU_RESV, stats_fpu_resv[thread_trace], "fpu_rsv");
      #if( CLUSTER_FPU )
	dumpStats( cycle, FPU_RESV, stats_fpu1_resv[thread_trace], "fpu1_rsv");
      #endif
      dumpStats( cycle, MEM_RESV, stats_mem_resv[thread_trace], "mem_rsv");
      #if( CLUSTER )
	dumpStats( cycle, MEM_RESV, stats_mem1_resv[thread_trace], "mem1_rsv");
      #endif
      dumpStats( cycle, BR_RESV, stats_br_resv[thread_trace], "br_rsv");

      #if( USING_LOG_UNIT )
      dumpStats( cycle, LOG_RESV, stats_log_resv[thread_trace], "log_rsv");
      #endif

      #if( USING_CMPLX_UNIT )
      dumpStats( cycle, CMPLX_RESV, stats_cmplx_resv[thread_trace], "cmplx_rsv");
      #endif

      dumpStats( cycle, (FIX_RESV+FPU_RESV+MEM_RESV+BR_RESV+LOG_RESV+CMPLX_RESV)*CLUSTERS_MAX, stats_resv[thread_trace], "tot_rsv");

      dumpStats( cycle, gpr_phys, stats_gpr[thread_trace], "gpr");
      dumpStats( cycle, fpr_phys, stats_fpr[thread_trace], "fpr");
      dumpStats( cycle, spr_phys, stats_spr[thread_trace], "spr");
      dumpStats( cycle, ccr_phys, stats_ccr[thread_trace], "ccr");

      dumpStats( cycle, STOREQ_LENGTH, stats_storeq[thread_trace], "storeq");
      dumpStats( cycle, REORDERQ_LENGTH, stats_reorderq[thread_trace], "reorderq");
      dumpStats( cycle, DMISS_MAX+MEM_NUM, stats_dmissq[thread_trace], "dmissq");
      dumpStats( cycle, DMISS_MAX+MEM_NUM, stats_lines[thread_trace], "lines");
      dumpStats( cycle, DMISS_MAX+MEM_NUM, stats_l2dmissq[thread_trace], "l2dmissq");
      dumpStats( cycle, DMISS_MAX+MEM_NUM, stats_l2_lines[thread_trace], "l2lines");
      dumpStats( cycle, DCASTOUT_MAX+MEM_NUM-1, stats_dcastoutq, "dcast");

      dumpStats( cycle, DCACHE_PORTS, stats_st_commit[thread_trace], "st_commit");

      dumpStats( cycle, RETIRE_MAX, stats_retired[thread_trace], "retired");
      dumpStats( cycle, DCACHE_PORTS, stats_st_retired[thread_trace], "st_retd");
      dumpStats( cycle, FIX_NUM, stats_fix_execd[thread_trace], "fix_exd");
      #if( CLUSTER )
	dumpStats( cycle, FIX_NUM, stats_fix1_execd[thread_trace], "fix1_exd");
      #endif
      dumpStats( cycle, FPU_NUM, stats_fpu_execd[thread_trace], "fpu_exd");
      #if( CLUSTER_FPU )
	dumpStats( cycle, FPU_NUM, stats_fpu1_execd[thread_trace], "fpu1_exd");
      #endif
      dumpStats( cycle, MEM_NUM, stats_mem_execd[thread_trace], "mem_exd");
      #if( CLUSTER )
	dumpStats( cycle, MEM_NUM, stats_mem1_execd[thread_trace], "mem1_exd");
      #endif
      dumpStats( cycle, BR_NUM, stats_br_execd[thread_trace], "br_exd");

      #if( USING_LOG_UNIT )
      dumpStats( cycle, LOG_NUM, stats_log_execd[thread_trace], "log_exd");
      #endif

      #if( USING_CMPLX_UNIT )
      dumpStats( cycle, CMPLX_NUM, stats_cmplx_execd[thread_trace], "cmplx_exd");
      #endif
      dumpStats( cycle, (FIX_NUM+FPU_NUM+MEM_NUM+BR_NUM+LOG_NUM+CMPLX_NUM)*CLUSTERS_MAX, stats_execd[thread_trace], "tot_exd");

      dumpStats( cycle, DISPATCH_MAX, stats_dispatched[thread_trace], "dispchd");
      dumpStats( cycle, DISPATCH_MAX, stats_fix_dispatched[thread_trace], "fix_dsp");
      #if( CLUSTER )
	dumpStats( cycle, DISPATCH_MAX, stats_fix1_dispatched[thread_trace], "fix1_dsp");
      #endif
      dumpStats( cycle, DISPATCH_MAX, stats_fpu_dispatched[thread_trace], "fpu_dsp");
      #if( CLUSTER_FPU )
	dumpStats( cycle, DISPATCH_MAX, stats_fpu1_dispatched[thread_trace], "fpu1_dsp");
      #endif
      dumpStats( cycle, DISPATCH_MAX, stats_mem_dispatched[thread_trace], "mem_dsp");
      #if( CLUSTER )
	dumpStats( cycle, DISPATCH_MAX, stats_mem1_dispatched[thread_trace], "mem1_dsp");
      #endif
      dumpStats( cycle, DISPATCH_MAX, stats_br_dispatched[thread_trace], "br_dsp");
      dumpStats( cycle, IBUF_MAX, stats_renamed[thread_trace], "renamed");
      dumpStats( cycle, IFETCH_MAX, stats_fetched[thread_trace], "fetched");

      dumpStats( cycle, INFLIGHT_MAX, stats_inflight[thread_trace], "inflight");
    #endif

#if(STATS)
    dumpStats( cycle, GROUPS_MAX, stats_groups[thread_trace], "groups");

    fprintf(the_stderr, "@@ TRAUMAS DATA\n" );
    #if( GETCPI )
      fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
      for( j = 0; j < TR_NUM; j++ ) {
	tmp_cumm = 0;
	fprintf(the_stderr, "@@ %8s:", trauma_names[j]);
	for( i = 0; i < RETIRE_MAX; i++ ) {
	  fprintf(the_stderr, "\t%6u", cpi_counters[thread_trace][i][j]);
	  tmp_cumm +=  cpi_counters[thread_trace][i][j];
	}
	fprintf(the_stderr, "\t%6u\t%6u\n", 0, tmp_cumm);
      }

      fprintf(the_stderr, "@@ %8s:", "cpi");
      tmp_total = 0;
      tmp_insns = 0;
      for( i = 0; i < RETIRE_MAX; i++ ) {
	tmp_cumm = 0;
	for( j = 0; j < TR_NUM; j++ ) {
	  tmp_cumm += cpi_counters[thread_trace][i][j];
	}
	fprintf(the_stderr, "\t%6u", tmp_cumm);
	tmp_total += tmp_cumm;
	tmp_insns += tmp_cumm*i;
      }
      fprintf(the_stderr, "\t%6u\t%6u\n", cpi_max, tmp_total);

      tmp_total += cpi_max;
      tmp_insns += cpi_max * RETIRE_MAX;
      fprintf(the_stderr, "@@ total   : %8u\n", tmp_total);
      fprintf(the_stderr, "@@ insns   : %8u\n", tmp_insns);
    #endif /* #if( GETCPI ) */

    #if( GETCPI_REMAINING )
      fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
      for( j = 0; j < TR_NUM; j++ ) {
	tmp_cumm = 0;
	fprintf(the_stderr, "@@ %8s:", trauma_names[j]);
	for( i = 0; i < RETIRE_MAX; i++ ) {
	  fprintf(the_stderr, "\t%6u", cpi_rem_counters[i][j]);
	  tmp_cumm +=  cpi_rem_counters[i][j];
	}
	fprintf(the_stderr, "\t%6u\t%6u\n", 0, tmp_cumm);
      }

      fprintf(the_stderr, "@@ %8s:", "cpi");
      tmp_total = 0;
      tmp_insns = 0;
      for( i = 0; i < RETIRE_MAX; i++ ) {
	tmp_cumm = 0;
	for( j = 0; j < TR_NUM; j++ ) {
	  tmp_cumm += cpi_rem_counters[i][j];
	}
	fprintf(the_stderr, "\t%6u", tmp_cumm);
	tmp_total += tmp_cumm;
	tmp_insns += tmp_cumm*i;
      }
      fprintf(the_stderr, "\t%6u\t%6u\n", cpi_rem_max, tmp_total);

      tmp_total += cpi_rem_max;
      tmp_insns += cpi_rem_max * RETIRE_MAX;
      fprintf(the_stderr, "@@ total   : %8u\n", tmp_total);
      fprintf(the_stderr, "@@ insns   : %8u\n", tmp_insns);
      fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
    #endif /* #if( GETCPI_REMAINING ) */
    }/* for thread_trace */
#endif /* #if(STATS) */

#if( POWER_MODE )
      // #include "utils_dump_stats.macros"
      // it seems this file no longer exists -jdonald

#if 0
      fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
      fprintf(the_stderr, "@@ UTILIZATION DATA \n");
      fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
      for(k=1;k<CLOCK_GATE_STYLES;k++) {
	PRINT_UTILIZATION(last_total_util, cycle);
      }
#endif
 {
/*   double cycle_inv = (double)POWER_CYCLES / cycle; */
   double cycle_inv = 1.0 / cycle * POWER_CYCLES; /* LYM */

      fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );
      fprintf(the_stderr, "@@ POWER DATA \n");
      fprintf(the_stderr, "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n" );

      PRINT_UARCH_POWER(total_power[real], cycle_inv);
      fprintf(the_stderr, "@@ min_power   : %5.2f\n",
	      min_quanta_power / POWER_CYCLES);
      fprintf(the_stderr, "@@ max_power   : %5.2f\n",
	      max_quanta_power / POWER_CYCLES);

/* LYM, printf detailed turandot power usage */
  for(j=bht;j<=ifu_misc;j++) {
    fprintf(the_stderr,"%s %5.2f\n",ifu_uarch[j].uarch_name,ifu_uarch[j].total_power[real]*cycle_inv);
  }

  for(j=idu_ucode_rom;j<=idu_misc;j++) {
    fprintf(the_stderr,"%s %5.2f\n",idu_uarch[j].uarch_name,idu_uarch[j].total_power[real]*cycle_inv);
  }

  for(j=crq;j<=isu_misc;j++) {
    fprintf(the_stderr,"%s %5.2f\n",isu_uarch[j].uarch_name,isu_uarch[j].total_power[real]*cycle_inv);
  }

  for(j=fxu_alu;j<=fxu_misc;j++) {
    fprintf(the_stderr,"%s %5.2f\n",fxu_uarch[j].uarch_name,fxu_uarch[j].total_power[real]*cycle_inv);
  }

  for(j=fpu0;j<=fpu_misc;j++) {
    fprintf(the_stderr,"%s %5.2f\n",fpu_uarch[j].uarch_name,fpu_uarch[j].total_power[real]*cycle_inv);
  }

  for(j=lsu_adder;j<=lsu_misc;j++) {
    fprintf(the_stderr,"%s %5.2f\n",lsu_uarch[j].uarch_name,lsu_uarch[j].total_power[real]*cycle_inv);
  }
fprintf(the_stderr,"@@ End of Power Data\n");
/* LYM, printf detailed turandot power usage, END */

#if(PDS_SIM)
{
   int pds_index;
   char pds_name[PDS_FILENAME_MAX + 16];

   for(pds_index = 0; pds_index < pds_count; pds_index++) {
     sprintf(pds_name, "voltage (%s)", pds_impulse_filename[pds_index]);
     dumpStats( cycle, pds_volt_hist_size, pds_volt_hist_arr[pds_index], pds_name);
   }
}
#endif /* PDS_SIM */
 }
#endif  /* POWER_MODE */
  }/* final code */
  comm_data[processor_id].end = 1;
  comm_data[processor_id].access = 1;
  
  free_array(bp_counter);
#if( !INFINITE_ITLB )
  free_array(itlb);
#endif /* !INFINITE_ITLB */
#if( !INFINITE_I2TLB && !UNIFIED_2TLB )
  free_array(i2tlb);
#endif /* !INFINITE_I2TLB && !UNIFIED_2TLB */
#if( !INFINITE_ICACHE )
  free_array(icache);
#endif /* !INFINITE_ICACHE */
#if( USE_I15CACHE  && !INFINITE_I15CACHE )
  free_array(i15cache);
#endif /* USE_I15CACHE  && !INFINITE_I15CACHE */
#if( !NFA_NEXT_SEQ )
  free_array(nfa);
#endif
#if( !INFINITE_DTLB )
  free_array(dtlb);
#endif /* !INFINITE_DTLB */
#if( !INFINITE_D2TLB )
  free_array(d2tlb);
#endif
#if( !INFINITE_DCACHE )
  free_array(dcache);
  dcache_local[processor_id] = NULL; // global variable needs to be cleaned up
#endif
  
	if (the_stderr_fname[processor_id] != NULL) { // early code is responsible for zeroing this out otherwise
		fclose(the_stderr); // was opened in init.macros
		the_stderr = NULL;
	}
	for (j = 0; j < n_threads; j += 1) {
		if (trace_fname[processor_id][j] != NULL) {
			int retval;
			retval = trace_close(trace_fp[j]);
			if (retval)
				fprintf(the_stderr, "warning: trace_close() returned %d for %s\n", retval,
					trace_fname[processor_id][j]);
		}
	}
  pthread_mutex_unlock( &tempsync_cond_mutex ); // just in case, power.dump.macros could be dangerous

#ifdef _AIX42 // _AIX detached pthreads
	pthread_mutex_lock( &join_mutex );
	unjoined_count -= 1;
	pthread_cond_signal( &join_cond );
	pthread_mutex_unlock( &join_mutex );
#endif // _AIX42

  return EXIT_SUCCESS; // not exit()
} // run_processor()

inline boolean tempsync_ready() { // jdonald
	// must be called when tempsync_cond_mutex is locked
	int i;
	for (i = 0; i < processor_num; i += 1) {
		if (!tempsync_ready_flags[i])
			return FALSE;
	}
	return TRUE;
}

#ifndef PI
const double PI = 3.14159265358979323846264338327;
#endif

#if ( POWER_MODE )
double power_rating(double dyn_ratio, double leak_ratio) {
	return 0.75 * dyn_ratio + 0.25 * leak_ratio;
}
#endif

void set_frequency_scale(double f, int processor_id) {
#if ( POWER_MODE )
	// must be called only in a typical tempsync safe region
	const double min_vdd = 0.6, max_vdd = 1.1; // let's block the 1.25X speedup thing for now
	const double min_vbs = -1.0, max_vbs = 0.450;
	const double num_intervals = 200; // we'll use something reasonably large, 40,000 combinations

	double min_freq;
	double delta_vdd, delta_vbs;
	double vdd, vbs;
	
	double best_dynamic = THE_MAX, best_leakage = THE_MAX, best_score = THE_MAX; // smaller is better
	double best_vdd, best_vbs;
	boolean found_qualify = FALSE;
	
	double baseline_dynamic, baseline_leakage;

	calculate_vscale_penalty(1.100000, -0.572250, &min_freq, &baseline_dynamic, &baseline_leakage);
	baseline_dynamic /= min_freq; // power reduction should not be cubic, just quadratic
	// baseline is set from experimentation, matches with the ld adjustment factor
#endif // POWER_MODE

#if ( FIXED_FREQUENCY )
	if (f != 1.0)
		fatal("set_frequency_scale() should only be used to adjust supply voltage and body biases, use a scaling factor of 1.0\n");
#else
	frequency_scale[processor_id] = f;
#endif // FIXED_FREQUENCY
#if( POWER_MODE )
	min_freq = f * REAL_FREQ * 1e9;

	delta_vdd = (max_vdd - min_vdd) / num_intervals;
	delta_vbs = (max_vbs - min_vbs) / num_intervals;
	
	for (vdd = min_vdd; vdd <= max_vdd; vdd += delta_vdd) {
		for (vbs = min_vbs; vbs <= max_vbs; vbs += delta_vbs) {
			double freq, dynamic, leakage, score;
			calculate_vscale_penalty(vdd, vbs, &freq, &dynamic, &leakage);
			dynamic /= freq; // scale for actual running frequency
#if( VARIATION )
			freq *= timing_variation[processor_id];
#endif
			// fprintf(stderr, "freq = %e, min_freq = %e\n", freq, min_freq);
			// fprintf(stderr, "REAL_FREQ = %f\n", REAL_FREQ);
			if (freq >= min_freq) {
				assert(dynamic > 0.0); // positive frequencies don't give negative power
				assert(leakage > 0.0);
				// the dynamic and leakage power values here on for a single transistor
				// this does not take into account the types of transistors, level of clock
				// gating, power, gating, etc., so we need to scale these values differently
				// assume the absolute ratio of baseline_dynamic to baseline_leakage is 2:1
				// for core logic
				//
				// the ratio of transistor dynamic/leakage need not resemble the aggregate ratio

#if ( VARIATION )
	#if ( FIXED_FREQUENCY )
				score = power_rating(dynamic / baseline_dynamic, leakage / baseline_leakage * leakage_variation[processor_id]);
	#else
				score = power_rating(dynamic / baseline_dynamic * frequency_scale[processor_id], leakage / baseline_leakage * leakage_variation[processor_id]);
	#endif
				// i guess this is confusing. each line right after calculate_vscale_penalty() already takes
				// timing_variation into account, but leakage_variation must be accounted for afterwards by
				// multiplying it with leakage_power_vscale whenever /jdonald
#else // no VARIATION
	#if ( FIXED_FREQUENCY )
				score = power_rating(dynamic / baseline_dynamic, leakage / baseline_leakage);
	#else
				score = power_rating(dynamic / baseline_dynamic * frequency_scale[processor_id], leakage / baseline_leakage);
	#endif
#endif
				if (score < best_score) { // smaller is better
					// fprintf(the_stderr, "score improved to %e with freq %e on core %d\n", score, freq, processor_id);
					best_score = score;
					best_dynamic = dynamic;
					best_leakage = leakage;
					best_vdd = vdd;
					best_vbs = vbs;
					found_qualify = TRUE;
				}
			}
		}
	}
	if (!found_qualify)
		fatal("could not find qualifying voltage scales for dvfs frequency\n");

	fprintf(the_stderr, "core %d: baseline_dynamic = %e, baseline_leakage = %e\n", processor_id, baseline_dynamic, baseline_leakage);
	fprintf(the_stderr, "        best_dynamic = %e, best_leakage = %e\n", best_dynamic, best_leakage);
	fprintf(the_stderr, "        best_vdd = %f, best_vbs = %f\n", best_vdd, best_vbs);  

	dynamic_power_vscale[processor_id] = best_dynamic / baseline_dynamic;
#if ( TEMPERATURE )
	leakage_power_vscale[processor_id] = best_leakage / baseline_leakage;
#endif

/*
	dynamic_power_vscale[processor_id] = 1.0;
#if ( TEMPERATURE )
	leakage_power_vscale[processor_id] = 1.0;
#endif
*/
#endif // POWER_MODE
}

#if ( POWER_MODE )
/*
    70 nm parameters:
	const double ceff = 2e-15;		// from 0.07um
	const double k1 = 0.063;		// from 0.07um
	const double k2 = 0.153;		// from 0.07um, fudge
	const double k3 = 5.38e-7 / 4;	// from 0.07um
	const double k4 = 1.83;			// from 0.07um
	const double k5 = 4.19;			// from 0.07um
	const double k6 = 5.26e-12 * 6;	// from 0.07um fudged here?
	// const double k7 = -0.144;	// from 0.07um // unused?
	const double ld = 10;			// from 0.07um, units ?
	const double vth1 = 0.244;		// from 0.07um
	const double ij = 4.80e-10;		// from 0.07um
*/

// These two are also used by lognormal_approx()
const double k2 = 0.166;	// 35nm estimate, logarithmic increase
const double k5 = 4.73;		// 35nm estimate, logarithmic increase

void calculate_vscale_penalty(double vdd, double vbs, double* freq, double* dynamic, double* leakage)
{
	const double ceff = 1e-15 * 11.4060490625;		// 35nm estimate, linear decrease /control param
	const double k1 = 0.073;			// 35nm estimate, logarithmic increase
	extern const double k2; // = 0.166;	// 35nm estimate, logarithmic increase
	const double k3 = 1.076e-6;		// 35nm estimate, linear increase
	const double k4 = 2.03;			// 35nm estimate, logarithmic increase
	extern const double k5; // = 4.73;	// 35nm estimate, logarithmic increase
	const double k6 = 2.63e-12;		// 35nm estimate, linear decrease
	const double ld = 25.0 * 5; // 27.78 * 5;		// 35nm estimate, linear decrease /control param
	const double vth1 = 0.230;		// 35nm estimate, small decrease
	const double ij = 9.6e-10;		// 35nm estimate, linear increase
	
	*freq = 1/(ld*k6)*((1 + k1)*vdd + k2*vbs - vth1); // alpha = 1
	*dynamic = (ceff * vdd*vdd * (*freq));
	*leakage = (vdd * k3 * exp(k4*vdd) * exp(k5*vbs)) + ((vbs >= 0.0 ? vbs : -vbs)*ij);
}

#endif // POWER_MODE

#if( VARIATION )

#if ( POWER_MODE ) // && VARIATION
double mean_rating;

void print_variations(int start, int num) {
	int i;
	fprintf(the_stderr, "%10s %10s ", "core", "freqvar");
#if ( POWER_MODE )
#if ( TEMPERATURE )
	fprintf(the_stderr, "%10s %10s %10s\n", "leakvar", "dynscale", "leakscale");
#else // no leakage
	fprintf(the_stderr, "%10s\n", "dynscale");
#endif
#else // !POWER_MODE
	fprintf(the_stderr, "\n");
#endif
	for (i = 0; i < num; i += 1) {
		fprintf(the_stderr, "%10d ", start+i);
#if ( FIXED_FREQUENCY )
		fprintf(the_stderr, "%10d ", 1);
#else
		fprintf(the_stderr, "%10.4f ", timing_variation[start+i]);
#endif
#if ( POWER_MODE )
#if ( TEMPERATURE )
		fprintf(the_stderr, "%10.4f %10.4f %10.4f\n",
			leakage_variation[start+i],
			dynamic_power_vscale[start+i],
			leakage_power_vscale[start+i]);
#else
		fprintf(the_stderr, "%10.4f\n",
			dynamic_power_vscale[start+i],
#endif
#else // !POWER_MODE
		fprintf(the_stderr, "\n");
#endif
	}
	fprintf(the_stderr, "\n");
}
			
int mean_distance_compare(const void* pa, const void* pb) {
	// This function is non-reentrant, so hopefully we call
	// it only once in the whole program!
	int a, b;
	double rating_a, rating_b, dist_a, dist_b;
	a = *((int*) pa);
	b = *((int*) pb);
#if( TEMPERATURE )
#if ( VARIATION)
	#if ( FIXED_FREQUENCY )
	rating_a = power_rating(dynamic_power_vscale[a], leakage_power_vscale[a] * leakage_variation[a]);
	rating_b = power_rating(dynamic_power_vscale[b], leakage_power_vscale[b] * leakage_variation[b]);
	#else
	rating_a = power_rating(dynamic_power_vscale[a] * frequency_scale[a], leakage_power_vscale[a] * leakage_variation[a]);
	rating_b = power_rating(dynamic_power_vscale[b] * frequency_scale[b], leakage_power_vscale[b] * leakage_variation[b]);
	#endif
#else
	rating_a = power_rating(dynamic_power_vscale[a] * frequency_scale[a], leakage_power_vscale[a]);
	rating_b = power_rating(dynamic_power_vscale[b] * frequency_scale[b], leakage_power_vscale[b]);
#endif // VARIATION
#else
	rating_a = dynamic_power_vscale[a];
	rating_b = dynamic_power_vscale[b];
#endif // TEMPERATURE
	// fprintf(the_stderr, "hey i am comparing %f and %f\n", rating_a, rating_b);
	dist_a = rating_a - mean_rating;
	if (dist_a < 0.0)
		dist_a = -dist_a;
	dist_b = rating_b - mean_rating;
	if (dist_b < 0.0)
		dist_b = -dist_b;
	// fprintf(the_stderr, "dists: %f < %f\n", dist_a, dist_b);
	if (dist_a < dist_b || (dist_a == dist_b && a < b))
		return -1;
	return 1;
}

void median_sort(int start, int num) {
#if ! ( FIXED_FREQUENCY )
	double new_frequency_scale[MAX_CORES];
	double new_timing_variation[MAX_CORES];
#endif
	double new_dynamic_power_vscale[MAX_CORES];
#if( TEMPERATURE )
	double new_leakage_power_vscale[MAX_CORES];
	double new_leakage_variation[MAX_CORES];
#endif
	int indices[MAX_CORES];
	int i;
	
	for (i = 0; i < num; i += 1) {
		indices[i] = i+start;
	}
	mean_rating = estimate_power_rating(start, num) / num;
	// mean_rating = power_rating(dynamic_power_vscale[4], leakage_power_vscale[4]);
	qsort(indices, num, sizeof(int), &mean_distance_compare);

	for (i = 0; i < num; i += 1) {
#if ( !( FIXED_FREQUENCY ) )
		new_frequency_scale[i] = frequency_scale[indices[i]];
		new_timing_variation[i] = timing_variation[indices[i]];
#endif
		new_dynamic_power_vscale[i] = dynamic_power_vscale[indices[i]];
#if( TEMPERATURE )
		new_leakage_power_vscale[i] = leakage_power_vscale[indices[i]];
		new_leakage_variation[i] = leakage_variation[indices[i]];
#endif
	}
#if ( !( FIXED_FREQUENCY ) )
	memcpy(&frequency_scale[start], new_frequency_scale, sizeof(double)*num);
	memcpy(&timing_variation[start], new_timing_variation, sizeof(double)*num);
#endif
	memcpy(&dynamic_power_vscale[start], new_dynamic_power_vscale, sizeof(double)*num);
#if( TEMPERATURE )
	memcpy(&leakage_power_vscale[start], new_leakage_power_vscale, sizeof(double)*num);
	memcpy(&leakage_variation[start], new_leakage_variation, sizeof(double)*num);
#endif
}

void profile_sort(int start, int num) {
	int i, j;
	fprintf(the_stderr, RED "starting profile_power_rating = %f\n" NORMAL, profile_power_rating(start, num));
	for (i = 0; i < num; i += 1) {
		for (j = 0; j < num; j += 1) {
			double total_a, total_b;
			if (i == j) continue;
#if ( TEMPERATURE )
#if ( VARIATION )
			total_a = dynamic_power_vscale[start+i] * power_profiles[i*2] +
					  leakage_power_vscale[start+i] * power_profiles[i*2+1] * leakage_variation[start+i] +
					  dynamic_power_vscale[start+j] * power_profiles[j*2] +
					  leakage_power_vscale[start+j] * power_profiles[j*2+1] * leakage_variation[start+j];
			total_b = dynamic_power_vscale[start+j] * power_profiles[i*2] +
					  leakage_power_vscale[start+j] * power_profiles[i*2+1] * leakage_variation[start+j] +
					  dynamic_power_vscale[start+i] * power_profiles[j*2] +
					  leakage_power_vscale[start+i] * power_profiles[j*2+1] * leakage_variation[start+i];
#else
			total_a = dynamic_power_vscale[start+i] * power_profiles[i*2] +
					  leakage_power_vscale[start+i] * power_profiles[i*2+1] +
					  dynamic_power_vscale[start+j] * power_profiles[j*2] +
					  leakage_power_vscale[start+j] * power_profiles[j*2+1];
			total_b = dynamic_power_vscale[start+j] * power_profiles[i*2] +
					  leakage_power_vscale[start+j] * power_profiles[i*2+1] +
					  dynamic_power_vscale[start+i] * power_profiles[j*2] +
					  leakage_power_vscale[start+i] * power_profiles[j*2+1];
#endif
#else // !TEMPERATURE (no leakage)
			total_a = dynamic_power_vscale[start+i] * power_profiles[i] +
					  dynamic_power_vscale[start+j] * power_profiles[j];
			total_b = dynamic_power_vscale[start+j] * power_profiles[i] +
					  dynamic_power_vscale[start+i] * power_profiles[j];
#endif
			if (total_b < total_a) // would swapping reduce power?
				swap_properties(start+i, start+j);
		}
	}	
	fprintf(the_stderr, RED "ending profile_power_rating = %f\n" NORMAL, profile_power_rating(start, num));
}

double estimate_power_rating(int start, int num) {
	int i;
	double total = 0.0;
	for (i = 0; i < num; i += 1) {
#if ( TEMPERATURE )
	#if ( FIXED_FREQUENCY )
		total += power_rating(dynamic_power_vscale[start+i], leakage_power_vscale[start+i] * leakage_variation[start+i]);
	#else
		total += power_rating(dynamic_power_vscale[start+i] * frequency_scale[start+i], leakage_power_vscale[start+i] * leakage_variation[start+i]);
	#endif
#else
		total += dynamic_power_vscale[start+i];
#endif
	}
	return total;
}

double profile_power_rating(int start, int num) {
	// this gives the total power rating based on the input dynamic/leakage ratios
	int i;
	double total = 0.0;
	// the 'start' offset does not offset power_profiles
	for (i = 0; i < num; i += 1) {
#if ( TEMPERATURE )
		total += dynamic_power_vscale[start+i] * power_profiles[i*2];
#if ( VARIATION )
		total += leakage_power_vscale[start+i] * leakage_variation[start+i] * power_profiles[i*2+1];
#else
		total += leakage_power_vscale[start+i] * power_profiles[i*2+1];
#endif // VARIATION
#else // !TEMPERATURE (no leakage)
		total += dynamic_power_vscale[start+i] * power_profiles[i];
#endif
	}
	return total;
}

void swap_properties(int a, int b) {
	double temp;
#if ( !( FIXED_FREQUENCY ) )
	temp = frequency_scale[a];
	frequency_scale[a] = frequency_scale[b];
	frequency_scale[b] = temp;
	temp = timing_variation[a];
	timing_variation[a] = timing_variation[b];
	timing_variation[b] = temp;
#endif
	temp = dynamic_power_vscale[a];
	dynamic_power_vscale[a] = dynamic_power_vscale[b];
	dynamic_power_vscale[b] = temp;
#if ( TEMPERATURE )
	temp = leakage_power_vscale[a];
	leakage_power_vscale[a] = leakage_power_vscale[b];
	leakage_power_vscale[b] = temp;
	temp = leakage_variation[a];
	leakage_variation[a] = leakage_variation[b];
	leakage_variation[b] = temp;
#endif // TEMPERATURE (leakage)
}

void reverse_array(double* arr, int num) {
	// used to reverse any array of doubles
	// in particular, this reverses timing_variation to go from fastest to slowest
	int i, back;
	double temp;
	back = num-1;
	for (i = 0; i < num/2; i += 1) {
		temp = arr[i];
		arr[i] = arr[back];
		arr[back] = temp;
		back -= 1;
	}
}
#endif // POWER_MODE (&& VARIATION)

void gaussian_approx(double* nums, int n, double mean, double sigma_mu_ratio) {
	double stdev, delta, interval, sum, x;
	int i, mid;
	
	// stdev is the actual standard deviation
	stdev = sigma_mu_ratio * mean;
	if (stdev == 0.0) {
		for (i = 0; i < n; i += 1)
			nums[i] = mean;
		return;
	}

	delta = stdev / 100.0; // arbitrary precision
	interval = 1.0 / (n+2); // n+1 would _just_ make it at the precision level limit
	sum = 0.0;
	x = mean;
	if (n % 2 == 1) {
		mid = n/2;
		nums[mid] = mean;
	}
	else
		mid = n/2 - 1; // actually 0.5 less than mid
		
	i = mid + 1;
	while (i < n) {
		sum += delta * 1/stdev / sqrt(2*PI) * exp(-(x-mean)*(x-mean)/(2*stdev*stdev));
		if ((n % 2 == 0) ?
			(sum > (i - mid + 0.5) * interval) :
			(sum > (i - mid + 1) * interval)) {
			nums[i] = x;
			nums[n-i-1] = 2*mean - x;
			i += 1;
		}
		x += delta;
		// fprintf(the_stderr, "x = %f, sum = %f\n", x, sum);
	}
	/*
	for (i = 0; i < n; i += 1) {
		fprintf(the_stderr, "%10.3f\n", nums[i]);
	}
	fprintf(the_stderr, "sum = %10.3f\n", sum);
	fprintf(the_stderr, "\n");
	*/
}

#if ( POWER_MODE && TEMPERATURE ) // && VARIATION
void lognormal_approx(double* nums, int n, double mean, double sigma_mu_ratio) {
	int i;
	gaussian_approx(nums, n, 1.0, sigma_mu_ratio);
	for (i = 0; i < n; i += 1) {
		 // for 3sigma/mu = 30%, sigma/mu = 10%:
		 //   e^(28.5 * 10%) = 17.3X
		 //   e^(28.5 * 15%) = 71.8X hmm
		 // for 3sigma/mu = 24, sigma/mu = 8%:
		 //   e^(28.5 *  8%) = 9.8X
		 //   e^(28.5 * 12%) = 30.6X
		nums[i] = mean * exp(k5/k2*(nums[i] - 1.0));
	}
}
#endif // POWER_MODE && TEMPERATURE && VARIATION

#endif // VARIATION

// macro used in benchmark_name() to detect .fF.gz
#define DETECT_EXTENSION(name, ext) \
	if ((suffix = strstr(name, ext)) != NULL && strcmp(suffix, ext) == 0) \
		len = suffix - name;

void benchmark_name(char* dest, const char* fname) {
	// possible destination needs at most 10 bytes free (output string max length 9)
	const char *str, *pos;
	const int maxlen = 9; // small column width
	char *suffix, *onethread_case;
	int len;
	
	if (fname == NULL) {
		strcpy(dest, "(stdin)");
		return;
	}
	str = (char*) fname;
	onethread_case = strstr(fname, "/1t/thread_001.tt6e");
	do {
		pos = strchr(str, '/');
		if (pos == NULL) pos = strchr(str, '\\'); // in case of Windows filenames
		if (onethread_case != NULL && pos == onethread_case) {
			// example of this special case: bioperf/blast/1t/thread_001.tt6e.bz2 -> blast
			len = pos - str;
			if (len > maxlen)
				len = maxlen;
			strncpy(dest, str, len);
			dest[len] = '\0';
			return;
		}
		if (pos != NULL)
			str = pos+1; // just after the slash
	}
	while (pos != NULL);
	if (str[0] == '\0') { // strange case, slash-terminated filename
		strcpy(dest, "(null)");
		return;
	}
	else {
		len = strlen(str);
		DETECT_EXTENSION(str, ".ff.gz");
		DETECT_EXTENSION(str, ".fF.gz");
		DETECT_EXTENSION(str, ".tt6e.gz");
		DETECT_EXTENSION(str, ".ff.bz2");
		DETECT_EXTENSION(str, ".fF.bz2");
		DETECT_EXTENSION(str, ".tt6e.bz2");
		if (strncmp(str, "thread", 6) == 0 && str[0] != '\0') {
			// truncate typical nondescriptive output names generated by amber
			// e.g. 'thread_001.tt6e.bz2' generates 't_001'
			dest[0] = 't';
			len -= 5; // remove 'hread'
			if (len > maxlen)
				len = maxlen;
			strncpy(dest+1, str+6, len - 1);
			dest[len] = '\0';
		}
		else {
			if (len > maxlen)
				len = maxlen;
			strncpy(dest, str, len);
			dest[len] = '\0';
		}
	}


}

struct string_node {
	char* str;
	struct string_node* next;
} *strlist_head = NULL;

char* stralloc(int n) {
// non-reentrant: do not call this function inside run_processor()
	char* pstr;
	struct string_node* pnode;
	
	assert(n > 0);
	pstr = malloc(n * sizeof(char));
	pnode = malloc(sizeof(struct string_node));

	pnode->str = pstr;
	pnode->next = strlist_head;
	strlist_head = pnode;
	
	return pstr;
}

void strfree(char* pstr) {
// non-reentrant: do not call this function inside run_processor()
	struct string_node *back, *pnode;
	
	assert(pstr != NULL);
	back = NULL;
	pnode = strlist_head;
	while (pnode != NULL) {
		if (pstr == pnode->str) {
			if (back == NULL)
				strlist_head = pnode->next;
			else
				back->next = pnode->next;
			free(pstr);
			free(pnode);
			pstr = NULL;
			break;
		}
	}
	if (pstr != NULL)
		fatal("tried to free a string not in the allocated list\n");

	return;
}

void free_strings() {
// also non-reentrant
	while (strlist_head != NULL) {
		struct string_node* back;
		back = strlist_head;
		strlist_head = strlist_head->next;
		free(back->str);
		free(back);
	}
	assert(strlist_head == NULL);
}
